/*
  X86 EMULATOR by Kamil Korolkiewicz 2007
  Plik: x86.h
  Opis: Opis procesora x86 (v. 0.0)
*/


//rejestry procesora

dword   cpuregs[8];                  //podstawowe rejestry (32)
word    cpusregs[8];                 //rejestry segmentowe (16)
ftword  fpuregs[8];                  //rejestry fpu/mmx    (80)
dword   crregs[8];                   //rejestry cr
dword   drregs[8];                   //rejestry dr
dword   eflags;                      //rejestr eflags
dword   eip;                         //wskaznik rozkazow
qword   tsc;                         //rejestr licznika instrukcji tsc

//nazwy rejestrow

//podstawowe rejestry (32) (b/w/d)
char *cpuregsnames[8][3]=
{
 {"al","ax","eax"},                  //0
 {"cl","cx","ecx"},                  //1
 {"dl","dx","edx"},                  //2
 {"bl","bx","ebx"},                  //3
 {"ah","sp","esp"},                  //4
 {"ch","bp","ebp"},                  //5
 {"dh","si","esi"},                  //6
 {"bh","di","edi"},                  //7
};

char *cpusregsnames[8][3]=
{
 {"es","es","es"},
 {"cs","cs","cs"},
 {"ss","ss","ss"},
 {"ds","ds","ds"},
 {"fs","fs","fs"},
 {"gs","gs","gs"},
 {"","",""},
 {"","",""}
};

char *fpuregsnames[8][3]=
{
 {"st(0)","st(0)","st(0)"},
 {"st(1)","st(1)","st(1)"},
 {"st(2)","st(2)","st(2)"},
 {"st(3)","st(3)","st(3)"},
 {"st(4)","st(4)","st(4)"},
 {"st(5)","st(5)","st(5)"},
 {"st(6)","st(6)","st(6)"},
 {"st(7)","st(7)","st(7)"},
};

char *crregsnames[8][3]=
{
 {"cr0","cr0","cr0"},
 {"cr1","cr1","cr1"},
 {"cr2","cr2","cr2"},
 {"cr3","cr3","cr3"},
 {"cr4","cr4","cr4"},
 {"cr5","cr5","cr5"},
 {"cr6","cr6","cr6"},
 {"cr7","cr7","cr7"},
};

char *drregsnames[8][3]=
{
 {"dr0","dr0","dr0"},
 {"dr1","dr1","dr1"},
 {"dr2","dr2","dr2"},
 {"dr3","dr3","dr3"},
 {"dr4","dr4","dr4"},
 {"dr5","dr5","dr5"},
 {"dr6","dr6","dr6"},
 {"dr7","dr7","dr7"},
};

//tablica zamiany typu rejestru na tablice rejestrow
ptr regnamestab[REG_COUNT]=
{
 cpuregsnames,
 cpusregsnames,
 fpuregsnames,
 crregsnames,
 drregsnames
};

//nazwy adresowania (w/d(67h))
char *adnames[8][2]=
{
 {"bx+si","eax"},
 {"bx+di","ecx"},
 {"bp+si","edx"},
 {"bp+di","ebx"},
 {"si",AD_SIB},
 {"di",AD_IMMBP},
 {AD_IMMBP,"esi"},
 {"bx","edi"}
};

//nazwy rozmiarow pamieci
//kolejno 1,2,4,6,8,10,16 bitow
char *adsizenames[7]=
{
 "byte ptr ","word ptr ","dword ptr ","sword ptr ","qword ptr ","tword ptr ","dqword ptr "
};

//nazwy prefiksow
char *prefixnames[9]=
{
 "es:","cs:","ss:","ds:","fs:","gs","lock","repne","rep"
};

//lista opcode

extern op_t extoptab[256];
extern op_t modrm[256];
extern op_t modsm[256];
extern op_t modcr[256];
extern op_t moddr[256];
extern op_t g8081[256];
extern op_t g8283[256];
extern op_t gF6F7[256];
extern op_t gFE[256];
extern op_t gFF[256];
extern op_t gD0D3C0C1[256];
extern op_t g0F00[256];
extern op_t g0F01[256];
extern op_t gD8[256];
extern op_t gD9[256];
extern op_t gDA[256];
extern op_t gDB[256];
extern op_t gDC[256];
extern op_t gDD[256];
extern op_t gDE[256];
extern op_t gDF[256];
extern op_t g0FBA[256];


//glowne kody "opcode"
op_t mainoptab[256]=
{
 /*0(0x00)*/ {"add",F_CP,0,ADD,modrm},                 //addb m,r
 /*1(0x01)*/ {"add",F_CP+F_D,0,ADD,modrm},               //addw m,r
 /*2(0x02)*/ {"add",F_CP+F_DIR,0,ADD,modrm},             //addb r,m
 /*3(0x03)*/ {"add",F_CP+F_D+F_DIR,0,ADD,modrm},         //addw r,m
 /*4(0x04)*/ {"add",F_PAR+F_IMM,AL<<8,ADD,0},       //add al,imm8
 /*5(0x05)*/ {"add",F_PAR+F_IMM+F_D,EAX<<8,ADD,0},  //add (e)ax,imm16/32
 /*6(0x06)*/ {"push",F_PAR+F_D,SEGREGS+(ES<<8),PUSH,0},   //push es
 /*7(0x07)*/ {"pop",F_PAR+F_D,SEGREGS+(ES<<8),POP,0},     //pop es
 /*8(0x08)*/ {"or",F_CP,0,OR,modrm},                   //orb m,r
 /*9(0x09)*/ {"or",F_CP+F_D,0,OR,modrm},                 //orw m,r
 /*10(0x0A)*/ {"or",F_CP+F_DIR,0,OR,modrm},              //orb r,m
 /*11(0x0B)*/ {"or",F_CP+F_D+F_DIR,0,OR,modrm},          //orw r,m
 /*12(0x0C)*/ {"or",F_PAR+F_IMM,AL<<8,OR,0},        //or al,imm8
 /*13(0x0D)*/ {"or",F_PAR+F_IMM+F_D,EAX<<8,OR,0},   //or (e)ax,imm16/32
 /*14(0x0E)*/ {"push",F_PAR+F_D,SEGREGS+(CS<<8),PUSH,0},  //push cs
 /*15(0x0F)*/ {0,0,0,0,extoptab},                         //instrukcje rozszerzone
 /*16(0x10)*/ {"adc",F_CP,0,ADC,modrm},                //adcb m,r
 /*17(0x11)*/ {"adc",F_CP+F_D,0,ADC,modrm},              //adcw m,r
 /*18(0x12)*/ {"adc",F_CP+F_DIR,0,ADC,modrm},            //adcb r,m
 /*19(0x13)*/ {"adc",F_CP+F_D+F_DIR,0,ADC,modrm},        //adcw r,m
 /*20(0x14)*/ {"adc",F_PAR+F_IMM,AL<<8,ADC,0},      //adc al,imm8
 /*21(0x15)*/ {"adc",F_PAR+F_IMM+F_D,EAX<<8,ADC,0}, //adc (e)ax,imm16/32
 /*22(0x16)*/ {"push",F_PAR+F_D,SEGREGS+(SS<<8),PUSH,0},  //push ss
 /*23(0x17)*/ {"pop",F_PAR+F_D,SEGREGS+(SS<<8),POP,0},    //pop ss
 /*24(0x18)*/ {"sbb",F_CP,0,SBB,modrm},                //sbbb m,r
 /*25(0x19)*/ {"sbb",F_CP+F_D,0,SBB,modrm},              //sbbw m,r
 /*26(0x1A)*/ {"sbb",F_CP+F_DIR,0,SBB,modrm},            //sbbb r,m
 /*27(0x1B)*/ {"sbb",F_CP+F_D+F_DIR,0,SBB,modrm},        //sbbw r,m
 /*28(0x1C)*/ {"sbb",F_PAR+F_IMM,AL<<8,SBB,0},      //sbb al,imm8
 /*29(0x1D)*/ {"sbb",F_PAR+F_IMM+F_D,EAX<<8,SBB,0}, //sbb (e)ax,imm16/32
 /*30(0x1E)*/ {"push",F_PAR+F_D,SEGREGS+(DS<<8),PUSH,0},  //push ds
 /*31(0x1F)*/ {"pop",F_PAR+F_D,SEGREGS+(DS<<8),POP,0},    //pop ds
 /*32(0x20)*/ {"and",F_CP,0,AND,modrm},                //andb m,r
 /*33(0x21)*/ {"and",F_CP+F_D,0,AND,modrm},              //andw m,r
 /*34(0x22)*/ {"and",F_CP+F_DIR,0,AND,modrm},            //andb r,m
 /*35(0x23)*/ {"and",F_CP+F_D+F_DIR,0,AND,modrm},        //andw r,m
 /*36(0x24)*/ {"and",F_PAR+F_IMM,AL<<8,AND,0},      //and al,imm8
 /*37(0x25)*/ {"and",F_PAR+F_IMM+F_D,EAX<<8,AND,0}, //and (e)ax,imm16/32
 /*38(0x26)*/ {0,F_ES,0,0,0},                       //es
 /*39(0x27)*/ {"daa",0,0,DAA,0},                    //daa
 /*40(0x28)*/ {"sub",F_CP,0,SUB,modrm},                //subb m,r
 /*41(0x29)*/ {"sub",F_CP+F_D,0,SUB,modrm},              //subw m,r
 /*42(0x2A)*/ {"sub",F_CP+F_DIR,0,SUB,modrm},            //subb r,m
 /*43(0x2B)*/ {"sub",F_CP+F_D+F_DIR,0,SUB,modrm},        //subw r,m
 /*44(0x2C)*/ {"sub",F_PAR+F_IMM,AL<<8,SUB,0},      //sub al,imm8
 /*45(0x2D)*/ {"sub",F_PAR+F_IMM+F_D,EAX<<8,SUB,0}, //sub (e)ax,imm16/32
 /*46(0x2E)*/ {0,F_CS,0,0,0},                       //cs
 /*47(0x2F)*/ {"das",0,0,DAS,0},                    //das
 /*48(0x30)*/ {"xor",F_CP,0,XOR,modrm},                //xorb m,r
 /*49(0x31)*/ {"xor",F_CP+F_D,0,XOR,modrm},              //xorw m,r
 /*50(0x32)*/ {"xor",F_CP+F_DIR,0,XOR,modrm},            //xorb r,m
 /*51(0x33)*/ {"xor",F_CP+F_D+F_DIR,0,XOR,modrm},        //xorw r,m
 /*52(0x34)*/ {"xor",F_PAR+F_IMM,AL<<8,XOR,0},      //xor al,imm8
 /*53(0x35)*/ {"xor",F_PAR+F_IMM+F_D,EAX<<8,XOR,0}, //xor (e)ax,imm16/32
 /*54(0x36)*/ {0,F_SS,0,0,0},                       //ss
 /*55(0x37)*/ {"aaa",0,0,AAA,0},                    //aaa
 /*56(0x38)*/ {"cmp",F_CP,0,CMP,modrm},                //cmpb m,r
 /*57(0x39)*/ {"cmp",F_CP+F_D,0,CMP,modrm},              //cmpw m,r
 /*58(0x3A)*/ {"cmp",F_CP+F_DIR,0,CMP,modrm},            //cmpb r,m
 /*59(0x3B)*/ {"cmp",F_CP+F_D+F_DIR,0,CMP,modrm},        //cmpw r,m
 /*60(0x3C)*/ {"cmp",F_PAR+F_IMM,AL<<8,CMP,0},      //cmp al,imm8
 /*61(0x3D)*/ {"cmp",F_PAR+F_IMM+F_D,EAX<<8,CMP,0}, //cmp (e)ax,imm16/32
 /*62(0x3E)*/ {0,F_DS,0,0,0},                       //ds
 /*63(0x3F)*/ {"aas",0,0,AAS,0},                    //aas
 /*64(0x40)*/ {"inc",F_PAR+F_D,EAX<<8,INC,0},             //inc (e)ax
 /*65(0x41)*/ {"inc",F_PAR+F_D,ECX<<8,INC,0},             //inc (e)cx
 /*66(0x42)*/ {"inc",F_PAR+F_D,EDX<<8,INC,0},             //inc (e)dx
 /*67(0x43)*/ {"inc",F_PAR+F_D,EBX<<8,INC,0},             //inc (e)bx
 /*68(0x44)*/ {"inc",F_PAR+F_D,ESP<<8,INC,0},             //inc (e)sp
 /*69(0x45)*/ {"inc",F_PAR+F_D,EBP<<8,INC,0},             //inc (e)bp
 /*70(0x46)*/ {"inc",F_PAR+F_D,ESI<<8,INC,0},             //inc (e)si
 /*71(0x47)*/ {"inc",F_PAR+F_D,EDI<<8,INC,0},             //inc (e)di
 /*72(0x48)*/ {"dec",F_PAR+F_D,EAX<<8,DEC,0},             //dec (e)ax
 /*73(0x49)*/ {"dec",F_PAR+F_D,ECX<<8,DEC,0},             //dec (e)cx
 /*74(0x4A)*/ {"dec",F_PAR+F_D,EDX<<8,DEC,0},             //dec (e)dx
 /*75(0x4B)*/ {"dec",F_PAR+F_D,EBX<<8,DEC,0},             //dec (e)bx
 /*76(0x4C)*/ {"dec",F_PAR+F_D,ESP<<8,DEC,0},             //dec (e)sp
 /*77(0x4D)*/ {"dec",F_PAR+F_D,EBP<<8,DEC,0},             //dec (e)bp
 /*78(0x4E)*/ {"dec",F_PAR+F_D,ESI<<8,DEC,0},             //dec (e)si
 /*79(0x4F)*/ {"dec",F_PAR+F_D,EDI<<8,DEC,0},             //dec (e)di
 /*80(0x50)*/ {"push",F_PAR+F_D,EAX<<8,PUSH,0},           //push (e)ax
 /*81(0x51)*/ {"push",F_PAR+F_D,ECX<<8,PUSH,0},           //push (e)cx
 /*82(0x52)*/ {"push",F_PAR+F_D,EDX<<8,PUSH,0},           //push (e)dx
 /*83(0x53)*/ {"push",F_PAR+F_D,EBX<<8,PUSH,0},           //push (e)bx
 /*84(0x54)*/ {"push",F_PAR+F_D,ESP<<8,PUSH,0},           //push (e)sp
 /*85(0x55)*/ {"push",F_PAR+F_D,EBP<<8,PUSH,0},           //push (e)bp
 /*86(0x56)*/ {"push",F_PAR+F_D,ESI<<8,PUSH,0},           //push (e)si
 /*87(0x57)*/ {"push",F_PAR+F_D,EDI<<8,PUSH,0},           //push (e)di
 /*88(0x58)*/ {"pop",F_PAR+F_D,EAX<<8,POP,0},             //pop (e)ax
 /*89(0x59)*/ {"pop",F_PAR+F_D,ECX<<8,POP,0},             //pop (e)cx
 /*90(0x5A)*/ {"pop",F_PAR+F_D,EDX<<8,POP,0},             //pop (e)dx
 /*91(0x5B)*/ {"pop",F_PAR+F_D,EBX<<8,POP,0},             //pop (e)bx
 /*92(0x5C)*/ {"pop",F_PAR+F_D,ESP<<8,POP,0},             //pop (e)sp
 /*93(0x5D)*/ {"pop",F_PAR+F_D,EBP<<8,POP,0},             //pop (e)bp
 /*94(0x5E)*/ {"pop",F_PAR+F_D,ESI<<8,POP,0},             //pop (e)si
 /*95(0x5F)*/ {"pop",F_PAR+F_D,EDI<<8,POP,0},             //pop (e)di
 /*96(0x60)*/ {"pusha#",F_D+F_FORMS,0,PUSHA,0},             //pusha(d)
 /*97(0x61)*/ {"popa#",F_D+F_FORMS,0,POPA,0},               //popa(d)
 /*98(0x62)*/ {"bound",F_CP+F_D+F_DIR,0,BOUND,modrm},    //bound r,m
 /*99(0x63)*/ {"arpl",F_CP+F_D+F_DIR,0,ARPL,modrm},      //arpl r,m 
 /*100(0x64)*/ {0,F_FS,0,0,0},                      //fs
 /*101(0x65)*/ {0,F_GS,0,0,0},                      //gs
 /*102(0x66)*/ {0,F_DP,0,0,0},                      //66h
 /*103(0x67)*/ {0,F_MP,0,0,0},                      //67h
 /*104(0x68)*/ {"push",F_D+F_IMM,0,PUSH,0},         //pushw/d immed
 /*105(0x69)*/ {"imul",F_CP+F_D+F_IMM+F_DIR,0,IMUL,modrm},         //imul r,m,imm16/32
 /*106(0x6A)*/ {"push",F_IMM,0,PUSH,0},             //push immed
 /*107(0x6B)*/ {"imul",F_CP+F_D+F_IMM+F_IMMB+F_DIR,0,IMUL,modrm},        //imul r,m,imm8
 /*108(0x6C)*/ {"insb",0,0,INS,0},                  //insb
 /*109(0x6D)*/ {"ins*",F_D+F_FORMS,0,INS,0},                  //insw/d
 /*110(0x6E)*/ {"outsb",0,0,OUTS,0},                //outsb
 /*111(0x6F)*/ {"outs*",F_D+F_FORMS,0,OUTS,0},                //outsw/d
 /*112(0x70)*/ {"jo",F_IMM+F_REL,0,JO,0},                 //jo imm8 (OF=1)
 /*113(0x71)*/ {"jno",F_IMM+F_REL,0,JNO,0},               //jno imm8 (OF=0)
 /*114(0x72)*/ {"jb",F_IMM+F_REL,0,JC,0},                 //jb imm8 (CF=1)
 /*115(0x73)*/ {"jae",F_IMM+F_REL,0,JNC,0},               //jae imm8 (CF=0)
 /*116(0x74)*/ {"je",F_IMM+F_REL,0,JZ,0},                 //je imm8 (ZF=1)
 /*117(0x75)*/ {"jne",F_IMM+F_REL,0,JNZ,0},               //jne imm8 (ZF=0)
 /*118(0x76)*/ {"jbe",F_IMM+F_REL,0,JCZ,0},               //jbe imm8 (CF=1 lub ZF=1)
 /*119(0x77)*/ {"ja",F_IMM+F_REL,0,JNCZ,0},               //ja imm8 (CF=0 i ZF=0)
 /*120(0x78)*/ {"js",F_IMM+F_REL,0,JS,0},                 //js imm8 (SF=1)
 /*121(0x79)*/ {"jns",F_IMM+F_REL,0,JNS,0},               //jns imm8 (SF=0)
 /*122(0x7A)*/ {"jp",F_IMM+F_REL,0,JP,0},                 //jp imm8 (PF=1)
 /*123(0x7B)*/ {"jnp",F_IMM+F_REL,0,JNP,0},               //jnp imm8 (PF=0)
 /*124(0x7C)*/ {"jl",F_IMM+F_REL,0,JL,0},                 //jl imm8 (SF<>OF)
 /*125(0x7D)*/ {"jge",F_IMM+F_REL,0,JNL,0},               //jge imm8 (SF=OF)
 /*126(0x7E)*/ {"jle",F_IMM+F_REL,0,JLE,0},               //jle imm8 (ZF=1 lub SF<>OF)
 /*127(0x7F)*/ {"jg",F_IMM+F_REL,0,JNLE,0},               //jg imm8 (ZF=0 i SF=OF)
 /*128(0x80)*/ {0,F_PAR+F_IMM,0,0,g8081},           //add/or/adc/sbb/and/sub/xor/cmp rmb,imm8
 /*129(0x81)*/ {0,F_PAR+F_IMM+F_D,0,0,g8081},       //add/or/adc/sbb/and/sub/xor/cmp rmw/d,imm16/32
 /*130(0x82)*/ {0,F_PAR+F_IMM,0,0,g8283},           //add/or/adc/sbb/and/sub/xor/cmp rmb,imm8
 /*131(0x83)*/ {0,F_PAR+F_IMM+F_IMMB+F_D,0,0,g8283},       //add/or/adc/sbb/and/sub/xor/cmp rmw/d,imm8
 /*132(0x84)*/ {"test",F_CP,0,TEST,modrm},          //testb m,r
 /*133(0x85)*/ {"test",F_CP+F_D,0,TEST,modrm},      //testw m,r
 /*134(0x86)*/ {"xchg",F_CP,0,XCHG,modrm},          //xchgb m,r
 /*135(0x87)*/ {"xchg",F_CP+F_D,0,XCHG,modrm},      //xchgw m,r
 /*136(0x88)*/ {"mov",F_CP,0,MOV,modrm},            //movb m,r
 /*137(0x89)*/ {"mov",F_CP+F_D,0,MOV,modrm},        //movw/d m,r
 /*138(0x8A)*/ {"mov",F_CP+F_DIR,0,MOV,modrm},      //movb r,m
 /*139(0x8B)*/ {"mov",F_CP+F_D+F_DIR,0,MOV,modrm},  //movw/d r,m
 /*140(0x8C)*/ {"mov",F_D+F_CP,0,MOVSG,modsm},      //mov m,seg
 /*141(0x8D)*/ {"lea",F_CP+F_D,0,LEA,modrm},        //lea r,m
 /*142(0x8E)*/ {"mov",F_D+F_CP+F_DIR,0,MOVSG,modsm},    //mov seg,m
 /*143(0x8F)*/ {"pop",F_D,0,POP,modrm},             //pop rm
 /*144(0x90)*/ {"nop",0,0,NOP,0},                        //nop
 /*145(0x91)*/ {"xchg",0,(EAX<<24)+(ECX<<8),XCHG,0},     //xchg (e)cx,(e)ax
 /*146(0x92)*/ {"xchg",0,(EAX<<24)+(EDX<<8),XCHG,0},     //xchg (e)dx,(e)ax
 /*147(0x93)*/ {"xchg",0,(EAX<<24)+(EBX<<8),XCHG,0},     //xchg (e)bx,(e)ax
 /*148(0x94)*/ {"xchg",0,(EAX<<24)+(ESP<<8),XCHG,0},     //xchg (e)sp,(e)ax
 /*149(0x95)*/ {"xchg",0,(EAX<<24)+(EBP<<8),XCHG,0},     //xchg (e)bp,(e)ax
 /*150(0x96)*/ {"xchg",0,(EAX<<24)+(ESI<<8),XCHG,0},     //xchg (e)si,(e)ax
 /*151(0x97)*/ {"xchg",0,(EAX<<24)+(EDI<<8),XCHG,0},     //xchg (e)di,(e)ax
 /*152(0x98)*/ {"cbw|cwde",F_D+F_FORMS,0,CBW,0},         //cbw/(66h)cwde
 /*153(0x99)*/ {"cwd|cdq",F_D+F_FORMS,0,CWD,0},          //cwd/(66h)cdq
 /*154(0x9A)*/ {"call",F_D+F_FAR+F_PTR,0,CALLF,0},       //call far seg:ofs
 /*155(0x9B)*/ {"wait",0,0,WAIT,0},                      //wait(fwait)
 /*156(0x9C)*/ {"pushf#",F_D+F_FORMS,0,PUSHF,0},                 //pushf(d)
 /*157(0x9D)*/ {"popf#",F_D+F_FORMS,0,POPF,0},                   //popf(d)
 /*158(0x9E)*/ {"sahf",0,0,SAHF,0},                      //sahf
 /*159(0x9F)*/ {"lahf",0,0,LAHF,0},                      //lahf
 /*160(0xA0)*/ {"mov",F_PAR+F_MIMM,AL<<8,MOV,0},               //mov al,[imm16/32]
 /*161(0xA1)*/ {"mov",F_PAR+F_D+F_MIMM,EAX<<8,MOV,0},          //mov (e)ax,[imm16/32]
 /*162(0xA2)*/ {"mov",F_DIR+F_PAR+F_MIMM,AL<<8,MOV,0},         //mov [imm16/32],al
 /*163(0xA3)*/ {"mov",F_DIR+F_PAR+F_D+F_MIMM,EAX<<8,MOV,0},    //mov [imm16/32],(e)ax
 /*164(0xA4)*/ {"movsb",0,0,MOVS,0},                     //movsb
 /*165(0xA5)*/ {"movs*",F_D+F_FORMS,0,MOVS,0},                   //movsw/d
 /*166(0xA6)*/ {"cmpsb",0,0,CMPS,0},                     //cmpsb
 /*167(0xA7)*/ {"cmps*",F_D+F_FORMS,0,CMPS,0},                   //cmpsw/d
 /*168(0xA8)*/ {"test",F_PAR+F_IMM,0,TEST,0},                  //test al,imm8
 /*169(0xA9)*/ {"test",F_PAR+F_D+F_IMM,0,TEST,0},              //test (e)ax,imm16/32
 /*170(0xAA)*/ {"stosb",0,0,STOS,0},                     //stosb
 /*171(0xAB)*/ {"stos*",F_D+F_FORMS,0,STOS,0},                   //stosw/d
 /*172(0xAC)*/ {"lodsb",0,0,LODS,0},                     //lodsb
 /*173(0xAD)*/ {"lods*",F_D+F_FORMS,0,LODS,0},                   //lodsw/d
 /*174(0xAE)*/ {"scasb",0,0,SCAS,0},                     //scasb
 /*175(0xAF)*/ {"scas*",F_D+F_FORMS,0,SCAS,0},                     //scasw/d
 /*176(0xB0)*/ {"mov",F_IMM+F_PAR,AL<<8,MOV,0},       //movb al,imm
 /*177(0xB1)*/ {"mov",F_IMM+F_PAR,CL<<8,MOV,0},       //movb cl,imm
 /*178(0xB2)*/ {"mov",F_IMM+F_PAR,DL<<8,MOV,0},       //movb dl,imm
 /*179(0xB3)*/ {"mov",F_IMM+F_PAR,BL<<8,MOV,0},       //movb bl,imm
 /*180(0xB4)*/ {"mov",F_IMM+F_PAR,AH<<8,MOV,0},       //movb ah,imm
 /*181(0xB5)*/ {"mov",F_IMM+F_PAR,CH<<8,MOV,0},       //movb ch,imm
 /*182(0xB6)*/ {"mov",F_IMM+F_PAR,DH<<8,MOV,0},       //movb dh,imm
 /*183(0xB7)*/ {"mov",F_IMM+F_PAR,BH<<8,MOV,0},       //movb bh,imm 
 /*184(0xB8)*/ {"mov",F_D+F_IMM+F_PAR,EAX<<8,MOV,0},      //movw (e)ax,imm
 /*185(0xB9)*/ {"mov",F_D+F_IMM+F_PAR,ECX<<8,MOV,0},      //movw (e)cx,imm
 /*186(0xBA)*/ {"mov",F_D+F_IMM+F_PAR,EDX<<8,MOV,0},      //movw (e)dx,imm
 /*187(0xBB)*/ {"mov",F_D+F_IMM+F_PAR,EBX<<8,MOV,0},      //movw (e)bx,imm
 /*188(0xBC)*/ {"mov",F_D+F_IMM+F_PAR,ESP<<8,MOV,0},      //movw (e)sp,imm
 /*189(0xBD)*/ {"mov",F_D+F_IMM+F_PAR,EBP<<8,MOV,0},      //movw (e)bp,imm
 /*190(0xBE)*/ {"mov",F_D+F_IMM+F_PAR,ESI<<8,MOV,0},      //movw (e)si,imm
 /*191(0xBF)*/ {"mov",F_D+F_IMM+F_PAR,EDI<<8,MOV,0},      //movw (e)di,imm 
 /*192(0xC0)*/ {0,F_PAR+F_IMM,0,0,gD0D3C0C1},               //rol/ror/rcl/rcr/shl/shr/sar rmb,imm8
 /*193(0xC1)*/ {0,F_PAR+F_D+F_IMM+F_IMMB,0,0,gD0D3C0C1},           //rol/ror/rcl/rcr/shl/shr/sar rmw/d,imm8
 /*194(0xC2)*/ {"retn",F_D+F_IMM,0,RETN,0},               //retn imm16
 /*195(0xC3)*/ {"retn",0,0,RETN,0},                       //retn
 /*196(0xC4)*/ {"les",F_CP+F_D+F_DIR,0,LES,modrm},               //les r,m
 /*197(0xC5)*/ {"lds",F_CP+F_D+F_DIR,0,LDS,modrm},               //lds r,m
 /*198(0xC6)*/ {"mov",F_IMM,0,MOV,modrm},             //movb m,immed
 /*199(0xC7)*/ {"mov",F_D+F_IMM,0,MOV,modrm},             //movw/d m,immed
 /*200(0xC8)*/ {"enter",F_DIMM,0,ENTER,0},            //enter imm16,imm8
 /*201(0xC9)*/ {"leave",0,0,LEAVE,0},                 //leave
 /*202(0xCA)*/ {"retf",F_IMM,0,RETF,0},               //retf imm16
 /*203(0xCB)*/ {"retf",0,0,RETF,0},                   //retf
 /*204(0xCC)*/ {"int3",0,0,INT3,0},                   //int 3
 /*205(0xCD)*/ {"int",F_IMM,0,INTN,0},                 //int imm8
 /*206(0xCE)*/ {"into",0,0,INTO,0},                   //int 4 (if overflow)
 /*207(0xCF)*/ {"iret",0,0,IRET,0},                   //iret(d)
 /*208(0xD0)*/ {0,F_PAR+F_IMM+F_IMM1,0,0,gD0D3C0C1},              //rol/ror/rcl/rcr/shl/shr/sar rmb,1
 /*209(0xD1)*/ {0,F_PAR+F_D+F_IMM+F_IMM1,0,0,gD0D3C0C1},            //rol/ror/rcl/rcr/shl/shr/sar rmw/d,1
 /*210(0xD2)*/ {0,F_PAR+F_XCL,0,0,gD0D3C0C1},          //rol/ror/rcl/rcr/shl/shr/sar rmb,cl
 /*211(0xD3)*/ {0,F_PAR+F_D+F_XCL,0,0,gD0D3C0C1},        //rol/ror/rcl/rcr/shl/shr/sar rmw/d,cl
 /*212(0xD4)*/ {"aam",F_IMM,0,AAM,0},            //aam imm8
 /*213(0xD5)*/ {"aad",F_IMM,0,AAD,0},            //aad imm8
 /*214(0xD6)*/ {0,0,0,0,0},                      //setalc doit!!!
 /*215(0xD7)*/ {"xlat",0,0,XLAT,0},              //xlat
 /*216(0xD8)*/ {0,F_PAR+F_DW,0,0,gD8},           //fadd/fmul/fcom/fcomp/fsub/fsubr/fdiv/fdivr
 /*217(0xD9)*/ {0,0,0,0,gD9},                    //patrz w 24319101.pdf A.5.2
 /*218(0xDA)*/ {0,0,0,0,gDA},                    //fiadd/fimul/ficom/ficomp/fisub/fisubr/fidiv/fidivr i fcmov...
 /*219(0xDB)*/ {0,0,0,0,gDB},                    //fild/fist/fistp/fld ex/fstp ex/fclex/finit oraz fcmov...
 /*220(0xDC)*/ {0,F_PAR+F_D+F_QW,0,0,gDC},       //fadd/fmul/fcom/fcomp/fsub/fsubr/fdiv/fdivr
 /*221(0xDD)*/ {0,0,0,0,gDD},                    //patrz w 24319101.pdf A.5.6
 /*222(0xDE)*/ {0,0,0,0,gDE},                    //patrz w 24319101.pdf A.5.7
 /*223(0xDF)*/ {0,F_PAR,0,0,gDF},                //patrz w 24319101.pdf A.5.8
 /*224(0xE0)*/ {"loopne",F_IMM+F_REL,0,LOOPNE,0},           //loopne(z) imm8 (rel)
 /*225(0xE1)*/ {"loope",F_IMM+F_REL,0,LOOPE,0},             //loope(z) imm8 (rel)
 /*226(0xE2)*/ {"loop",F_IMM+F_REL,0,LOOP,0},               //loop imm8 (rel)
 /*227(0xE3)*/ {"j$cxz",F_D+F_IMMB+F_IMM+F_FORMS+F_REL,0,JCXZ,0},      //j(e)cxz imm8 (rel)
 /*228(0xE4)*/ {"in",F_PAR+F_IMM,AL<<8,INP,0},         //in al,imm8
 /*229(0xE5)*/ {"in",F_PAR+F_IMM+F_D+F_IMMB,EAX<<8,INP,0},    //in (e)ax,imm8
 /*230(0xE6)*/ {"out",F_PAR+F_IMM,AL<<24,OUTP,0},       //out imm8,al
 /*231(0xE7)*/ {"out",F_PAR+F_IMM+F_D+F_IMMB,EAX<<24,OUTP,0},  //out imm8,(e)ax
 /*232(0xE8)*/ {"call",F_D+F_IMM+F_REL,0,CALL,0},           //call imm16/32 (rel)
 /*233(0xE9)*/ {"jmp",F_D+F_IMM+F_REL,0,JMP,0},             //jmp imm16/32 (rel)
 /*234(0xEA)*/ {"jmp",F_D+F_FAR+F_PTR,0,JMPF,0},          //jmp far seg:ofs
 /*235(0xEB)*/ {"jmp",F_IMM+F_REL,0,JMP,0},                 //jmp imm8 (rel)
 /*236(0xEC)*/ {"in",F_PAR+F_CP+F_PAR2D,(AL<<8)+(EDX<<24),INP,0},         //in al,dx
 /*237(0xED)*/ {"in",F_PAR+F_CP+F_D+F_PAR2D,(EAX<<8)+(EDX<<24),INP,0},            //in (e)ax,dx
 /*238(0xEE)*/ {"out",F_PAR+F_CP+F_PAR1D,(AL<<24)+(EDX<<8),OUTP,0},       //out dx,al
 /*239(0xEF)*/ {"out",F_PAR+F_CP+F_D+F_PAR1D,(AL<<24)+(EDX<<8),OUTP,0},           //out dx,(e)ax
 /*240(0xF0)*/ {0,F_LOCK,0,0,0},                    //lock
 /*241(0xF1)*/ {0,0,0,0,0},                         //?
 /*242(0xF2)*/ {0,F_REPNE,0,0,0},                   //repne
 /*243(0xF3)*/ {0,F_REP,0,0,0},                     //rep
 /*244(0xF4)*/ {"hlt",0,0,HLT,0},                   //hlt
 /*245(0xF5)*/ {"cmc",0,0,CMC,0},                   //cmc
 /*246(0xF6)*/ {0,F_PAR,0,0,gF6F7},           //test/not/neg/mul/imul/div/idiv rmb,imm8
 /*247(0xF7)*/ {0,F_PAR+F_D,0,0,gF6F7},           //test/not/neg/mul/imul/div/idiv rmw,imm16/32
 /*248(0xF8)*/ {"clc",0,0,CLC,0},                   //clc
 /*249(0xF9)*/ {"stc",0,0,STC,0},                   //stc
 /*250(0xFA)*/ {"cli",0,0,CLI,0},                   //cli
 /*251(0xFB)*/ {"sti",0,0,STI,0},                   //sti
 /*252(0xFC)*/ {"cld",0,0,CLD,0},                   //cld
 /*253(0xFD)*/ {"std",0,0,STD,0},                   //std
 /*254(0xFE)*/ {0,F_PAR,0,0,gFE},                   //inc/dec rmb
 /*255(0xFF)*/ {0,F_PAR+F_D,0,0,gFF}                //inc/dec/calln/callf/jmpn/jmpf/push rmw/d
};

//instrukcje rozszerzone
op_t extoptab[256]=
{
 /*0(0x00)*/ {0,F_PAR+F_D+F_PAR1D,0,0,g0F00},
 /*1(0x01)*/ {0,F_PAR+F_PAR1D,0,0,g0F01},
 /*2(0x02)*/ {"lar",F_D+F_CP+F_DIR,0,LAR,modrm},     //lar r/rm
 /*3(0x03)*/ {"lsl",F_D+F_CP+F_DIR,0,LSL,modrm},     //lsl r/rm
 /*4(0x04)*/ {0,0,0,0,0},
 /*5(0x05)*/ {0,0,0,0,0},
 /*6(0x06)*/ {"clts",0,0,CLTS,0},                    //clts
 /*7(0x07)*/ {0,0,0,0,0},
 /*8(0x08)*/ {"invd",0,0,INVD,0},                    //invd
 /*9(0x09)*/ {"wbinvd",0,0,WBINVD,0},                //wbinvd
 /*10(0x0A)*/ {0,0,0,0,0},
 /*11(0x0B)*/ {"ud2",0,0,0,0},                       //ud2
 /*12(0x0C)*/ {0,0,0,0,0},
 /*13(0x0D)*/ {0,0,0,0,0},
 /*14(0x0E)*/ {0,0,0,0,0},
 /*15(0x0F)*/ {0,0,0,0,0},
 /*16(0x10)*/ {0,0,0,0,0},
 /*17(0x11)*/ {0,0,0,0,0},
 /*18(0x12)*/ {0,0,0,0,0},
 /*19(0x13)*/ {0,0,0,0,0},
 /*20(0x14)*/ {0,0,0,0,0},
 /*21(0x15)*/ {0,0,0,0,0},
 /*22(0x16)*/ {0,0,0,0,0},
 /*23(0x17)*/ {0,0,0,0,0},
 /*24(0x18)*/ {0,0,0,0,0},
 /*25(0x19)*/ {0,0,0,0,0},
 /*26(0x1A)*/ {0,0,0,0,0},
 /*27(0x1B)*/ {0,0,0,0,0},
 /*28(0x1C)*/ {0,0,0,0,0},
 /*29(0x1D)*/ {0,0,0,0,0},
 /*30(0x1E)*/ {0,0,0,0,0},
 /*31(0x1F)*/ {0,0,0,0,0},
 /*32(0x20)*/ {"mov",F_PAR+F_CP+F_D,0,MOVC,modcr},        //mov r,cr
 /*33(0x21)*/ {"mov",F_PAR+F_CP+F_D,0,MOVD,moddr},        //mov r,dr
 /*34(0x22)*/ {"mov",F_PAR+F_CP+F_DIR+F_D,0,MOVC,modcr},  //mov cr,r
 /*35(0x23)*/ {"mov",F_PAR+F_CP+F_DIR+F_D,0,MOVD,moddr},  //mov dr,r
 /*36(0x24)*/ {0,0,0,0,0},
 /*37(0x25)*/ {0,0,0,0,0},
 /*38(0x26)*/ {0,0,0,0,0},
 /*39(0x27)*/ {0,0,0,0,0},
 /*40(0x28)*/ {0,0,0,0,0},
 /*41(0x29)*/ {0,0,0,0,0},
 /*42(0x2A)*/ {0,0,0,0,0},
 /*43(0x2B)*/ {0,0,0,0,0},
 /*44(0x2C)*/ {0,0,0,0,0},
 /*45(0x2D)*/ {0,0,0,0,0},
 /*46(0x2E)*/ {0,0,0,0,0},
 /*47(0x2F)*/ {0,0,0,0,0},
 /*48(0x30)*/ {"wrmsr",0,0,WRMSR,0},                      //wrmsr
 /*49(0x31)*/ {"rdtsc",0,0,RDTSC,0},                      //rdtsc
 /*50(0x32)*/ {"rdmsr",0,0,RDMSR,0},                      //rdmsr
 /*51(0x33)*/ {"rdpmc",0,0,RDPMC,0},                      //rdpmc
 /*52(0x34)*/ {0,0,0,0,0},
 /*53(0x35)*/ {0,0,0,0,0},
 /*54(0x36)*/ {0,0,0,0,0},
 /*55(0x37)*/ {0,0,0,0,0},
 /*56(0x38)*/ {0,0,0,0,0},
 /*57(0x39)*/ {0,0,0,0,0},
 /*58(0x3A)*/ {0,0,0,0,0},
 /*59(0x3B)*/ {0,0,0,0,0},
 /*60(0x3C)*/ {0,0,0,0,0},
 /*61(0x3D)*/ {0,0,0,0,0},
 /*62(0x3E)*/ {0,0,0,0,0},
 /*63(0x3F)*/ {0,0,0,0,0},
 /*64(0x40)*/ {"cmovo",F_CP+F_DIR+F_D,0,CMOVO,modrm},                  //cmovo r/m8 (OF=1)
 /*65(0x41)*/ {"cmovno",F_CP+F_DIR+F_D,0,CMOVNO,modrm},                //cmovno r/m8 (OF=0)
 /*66(0x42)*/ {"cmovb",F_CP+F_DIR+F_D,0,CMOVC,modrm},                  //cmovb r/m8 (CF=1)
 /*67(0x43)*/ {"cmovae",F_CP+F_DIR+F_D,0,CMOVNC,modrm},                //cmovae r/m8 (CF=0)
 /*68(0x44)*/ {"cmove",F_CP+F_DIR+F_D,0,CMOVZ,modrm},                  //cmovz r/m8 (ZF=1)
 /*69(0x45)*/ {"cmovne",F_CP+F_DIR+F_D,0,CMOVNZ,modrm},                //cmovne r/m8 (ZF=0)
 /*70(0x46)*/ {"cmovbe",F_CP+F_DIR+F_D,0,CMOVCZ,modrm},                //cmovbe r/m8 (CF=1 lub ZF=1)
 /*71(0x47)*/ {"cmova",F_CP+F_DIR+F_D,0,CMOVNCZ,modrm},                //cmova r/m8 (CF=0 i ZF=0)
 /*72(0x48)*/ {"cmovs",F_CP+F_DIR+F_D,0,CMOVS,modrm},                  //cmovs r/m8 (SF=1)
 /*73(0x49)*/ {"cmovns",F_CP+F_DIR+F_D,0,CMOVNS,modrm},                //cmovns r/m8 (SF=0)
 /*74(0x4A)*/ {"cmovp",F_CP+F_DIR+F_D,0,CMOVP,modrm},                  //cmovp r/m8 (PF=1)
 /*75(0x4B)*/ {"cmovnp",F_CP+F_DIR+F_D,0,CMOVNP,modrm},                //cmovnp r/m8 (PF=0)
 /*76(0x4C)*/ {"cmovl",F_CP+F_DIR+F_D,0,CMOVL,modrm},                  //cmovl r/m8 (SF<>OF)
 /*77(0x4D)*/ {"cmovnl",F_CP+F_DIR+F_D,0,CMOVNL,modrm},                //cmovnl r/m8 (SF=OF)
 /*78(0x4E)*/ {"cmovle",F_CP+F_DIR+F_D,0,CMOVLE,modrm},                //cmovle r/m8 (ZF=1 lub SF<>OF)
 /*79(0x4F)*/ {"cmovg",F_CP+F_DIR+F_D,0,CMOVNLE,modrm},                //cmovg r/m8 (ZF=0 i SF=OF)
 /*80(0x50)*/ {0,0,0,0,0},
 /*81(0x51)*/ {0,0,0,0,0},
 /*82(0x52)*/ {0,0,0,0,0},
 /*83(0x53)*/ {0,0,0,0,0},
 /*84(0x54)*/ {0,0,0,0,0},
 /*85(0x55)*/ {0,0,0,0,0},
 /*86(0x56)*/ {0,0,0,0,0},
 /*87(0x57)*/ {0,0,0,0,0},
 /*88(0x58)*/ {0,0,0,0,0},
 /*89(0x59)*/ {0,0,0,0,0},
 /*90(0x5A)*/ {0,0,0,0,0},
 /*91(0x5B)*/ {0,0,0,0,0},
 /*92(0x5C)*/ {0,0,0,0,0},
 /*93(0x5D)*/ {0,0,0,0,0},
 /*94(0x5E)*/ {0,0,0,0,0},
 /*95(0x5F)*/ {0,0,0,0,0},
 /*96(0x60)*/ {0,0,0,0,0},
 /*97(0x61)*/ {0,0,0,0,0},
 /*98(0x62)*/ {0,0,0,0,0},
 /*99(0x63)*/ {0,0,0,0,0},
 /*100(0x64)*/ {0,0,0,0,0},
 /*101(0x65)*/ {0,0,0,0,0},
 /*102(0x66)*/ {0,0,0,0,0},
 /*103(0x67)*/ {0,0,0,0,0},
 /*104(0x68)*/ {0,0,0,0,0},
 /*105(0x69)*/ {0,0,0,0,0},
 /*106(0x6A)*/ {0,0,0,0,0},
 /*107(0x6B)*/ {0,0,0,0,0},
 /*108(0x6C)*/ {0,0,0,0,0},
 /*109(0x6D)*/ {0,0,0,0,0},
 /*110(0x6E)*/ {0,0,0,0,0},
 /*111(0x6F)*/ {0,0,0,0,0},
 /*112(0x70)*/ {0,0,0,0,0},
 /*113(0x71)*/ {0,0,0,0,0},
 /*114(0x72)*/ {0,0,0,0,0},
 /*115(0x73)*/ {0,0,0,0,0},
 /*116(0x74)*/ {0,0,0,0,0},
 /*117(0x75)*/ {0,0,0,0,0},
 /*118(0x76)*/ {0,0,0,0,0},
 /*119(0x77)*/ {0,0,0,0,0},
 /*120(0x78)*/ {0,0,0,0,0},
 /*121(0x79)*/ {0,0,0,0,0},
 /*122(0x7A)*/ {0,0,0,0,0},
 /*123(0x7B)*/ {0,0,0,0,0},
 /*124(0x7C)*/ {0,0,0,0,0},
 /*125(0x7D)*/ {0,0,0,0,0},
 /*126(0x7E)*/ {0,0,0,0,0},
 /*127(0x7F)*/ {0,0,0,0,0},
 /*128(0x80)*/ {"jo",F_IMM+F_D+F_REL,0,JO,0},                 //jo imm16/32 (OF=1)
 /*129(0x81)*/ {"jno",F_IMM+F_D+F_REL,0,JNO,0},               //jno imm16/32 (OF=0)
 /*130(0x82)*/ {"jb",F_IMM+F_D+F_REL,0,JC,0},                 //jb imm16/32 (CF=1)
 /*131(0x83)*/ {"jae",F_IMM+F_D+F_REL,0,JNC,0},               //jae imm16/32 (CF=0)
 /*132(0x84)*/ {"je",F_IMM+F_D+F_REL,0,JZ,0},                 //je imm16/32 (ZF=1)
 /*133(0x85)*/ {"jne",F_IMM+F_D+F_REL,0,JNZ,0},               //jne imm16/32 (ZF=0)
 /*134(0x86)*/ {"jbe",F_IMM+F_D+F_REL,0,JCZ,0},               //jbe imm16/32 (CF=1 lub ZF=1)
 /*135(0x87)*/ {"ja",F_IMM+F_D+F_REL,0,JNCZ,0},               //ja imm16/32 (CF=0 i ZF=0)
 /*136(0x88)*/ {"js",F_IMM+F_D+F_REL,0,JS,0},                 //js imm16/32 (SF=1)
 /*137(0x89)*/ {"jns",F_IMM+F_D+F_REL,0,JNS,0},               //jns imm16/32 (SF=0)
 /*138(0x8A)*/ {"jp",F_IMM+F_D+F_REL,0,JP,0},                 //jp imm16/32 (PF=1)
 /*139(0x8B)*/ {"jnp",F_IMM+F_D+F_REL,0,JNP,0},               //jnp imm16/32 (PF=0)
 /*140(0x8C)*/ {"jl",F_IMM+F_D+F_REL,0,JL,0},                 //jl imm16/32 (SF<>OF)
 /*141(0x8D)*/ {"jge",F_IMM+F_D+F_REL,0,JNL,0},               //jge imm16/32 (SF=OF)
 /*142(0x8E)*/ {"jle",F_IMM+F_D+F_REL,0,JLE,0},               //jle imm16/32 (ZF=1 lub SF<>OF)
 /*143(0x8F)*/ {"jg",F_IMM+F_D+F_REL,0,JNLE,0},               //jg imm16/32 (ZF=0 i SF=OF)
 /*144(0x90)*/ {"seto",0,0,SETO,modrm},                       //seto r/m8 (OF=1)
 /*145(0x91)*/ {"setno",0,0,SETNO,modrm},                     //setno r/m8 (OF=0)
 /*146(0x92)*/ {"setb",0,0,SETC,modrm},                       //setb r/m8 (CF=1)
 /*147(0x93)*/ {"setae",0,0,SETNC,modrm},                     //setae r/m8 (CF=0)
 /*148(0x94)*/ {"sete",0,0,SETZ,modrm},                       //setz r/m8 (ZF=1)
 /*149(0x95)*/ {"setne",0,0,SETNZ,modrm},                     //setne r/m8 (ZF=0)
 /*150(0x96)*/ {"setbe",0,0,SETCZ,modrm},                     //setbe r/m8 (CF=1 lub ZF=1)
 /*151(0x97)*/ {"seta",0,0,SETNCZ,modrm},                     //seta r/m8 (CF=0 i ZF=0)
 /*152(0x98)*/ {"sets",0,0,SETS,modrm},                       //sets r/m8 (SF=1)
 /*153(0x99)*/ {"setns",0,0,SETNS,modrm},                     //setns r/m8 (SF=0)
 /*154(0x9A)*/ {"setp",0,0,SETP,modrm},                       //setp r/m8 (PF=1)
 /*155(0x9B)*/ {"setnp",0,0,SETNP,modrm},                     //setnp r/m8 (PF=0)
 /*156(0x9C)*/ {"setl",0,0,SETL,modrm},                       //setl r/m8 (SF<>OF)
 /*157(0x9D)*/ {"setnl",0,0,SETNL,modrm},                     //setnl r/m8 (SF=OF)
 /*158(0x9E)*/ {"setle",0,0,SETLE,modrm},                     //setle r/m8 (ZF=1 lub SF<>OF)
 /*159(0x9F)*/ {"setg",0,0,SETNLE,modrm},                     //setg r/m8 (ZF=0 i SF=OF)
 /*160(0xA0)*/ {"push",F_PAR+F_D,SEGREGS+(FS<<8),PUSH,0},     //push fs
 /*161(0xA1)*/ {"pop",F_PAR+F_D,SEGREGS+(FS<<8),POP,0},       //pop fs
 /*162(0xA2)*/ {"cpuid",0,0,CPUID,0},                         //cpuid
 /*163(0xA3)*/ {"bt",F_CP+F_D,0,BT,modrm},                    //bt rm/r
 /*164(0xA4)*/ {"shld",F_D+F_CP+F_IMM+F_IMMB,0,SHLD,modrm},   //shld m,r,imm8
 /*165(0xA5)*/ {"shld",F_D+F_CP+F_XCL,0,SHLD,modrm},     //shld m,r,cl
 /*166(0xA6)*/ {0,0,0,0,0},
 /*167(0xA7)*/ {0,0,0,0,0},
 /*168(0xA8)*/ {"push",F_PAR+F_D,SEGREGS+(GS<<8),PUSH,0},     //push gs
 /*169(0xA9)*/ {"pop",F_PAR+F_D,SEGREGS+(GS<<8),POP,0},       //pop gs
 /*170(0xAA)*/ {"rsm",0,0,RSM,0},                             //rsm
 /*171(0xAB)*/ {"bts",F_D+F_CP,0,BTS,modrm},                  //bts rm/r
 /*172(0xAC)*/ {"shrd",F_D+F_CP+F_IMM+F_IMMB,0,SHRD,modrm},   //shrd m,r,imm8
 /*173(0xAD)*/ {"shrd",F_D+F_CP+F_XCL,0,SHRD,modrm},     //shrd m,r,cl
 /*174(0xAE)*/ {0,0,0,0,0},
 /*175(0xAF)*/ {"imul",F_D+F_CP+F_DIR,0,IMUL,modrm},                //imul r,m
 /*176(0xB0)*/ {"cmpxchg",F_CP,0,CMPXCHG,modrm},             //cmpxchg rm/r
 /*177(0xB1)*/ {"cmpxchg",F_D+F_CP,0,CMPXCHG,modrm},          //cmpxchg rm/r
 /*178(0xB2)*/ {"lss",F_D+F_CP+F_DIR,0,LSS,modrm},            //lss r,rm
 /*179(0xB3)*/ {"btr",F_D+F_CP,0,BTR,modrm},                  //btr rm/r
 /*180(0xB4)*/ {"lfs",F_D+F_CP+F_DIR,0,LFS,modrm},            //lfs r,rm
 /*181(0xB5)*/ {"lgs",F_D+F_CP+F_DIR,0,LGS,modrm},            //lgs r,rm
 /*182(0xB6)*/ {"movzx",F_CP+F_DIR+F_D+F_SD+F_PAR2B,0,MOVZX8,modrm}, //movzx r/rm8
 /*183(0xB7)*/ {"movzx",F_CP+F_DIR+F_D+F_W+F_PAR2D,0,MOVZX16,modrm}, //movzx r/rm16
 /*184(0xB8)*/ {0,0,0,0,0},
 /*185(0xB9)*/ {0,0,0,0,0},
 /*186(0xBA)*/ {0,F_PAR+F_CP+F_D+F_IMM+F_IMMB,0,0,g0FBA},            //bt/bts/btr/btc r/m,imm8
 /*187(0xBB)*/ {"btc",F_D+F_CP,0,BTC,0},                             //btc rm/r
 /*188(0xBC)*/ {"bsf",F_D+F_CP+F_DIR,0,BSF,0},                       //bsf r/rm
 /*189(0xBD)*/ {"bsr",F_D+F_CP+F_DIR,0,BSR,0},                       //bsr r/rm
 /*190(0xBE)*/ {"movsx",F_CP+F_DIR+F_D+F_SD+F_PAR2B,0,MOVSX8,modrm}, //movsx r/rm8
 /*191(0xBF)*/ {"movsx",F_CP+F_DIR+F_D+F_W+F_PAR2D,0,MOVSX16,modrm}, //movsx r/rm16
 /*192(0xC0)*/ {"xadd",F_CP,0,XADD,modrm},                           //xadd rm/r
 /*193(0xC1)*/ {"xadd",F_CP+F_D,0,XADD,modrm},                       //xadd rm/r
 /*194(0xC2)*/ {0,0,0,0,0},
 /*195(0xC3)*/ {0,0,0,0,0},
 /*196(0xC4)*/ {0,0,0,0,0},
 /*197(0xC5)*/ {0,0,0,0,0},
 /*198(0xC6)*/ {0,0,0,0,0},
 /*199(0xC7)*/ {0,0,0,0,0},                   //cmpxchg8b - doit
 /*200(0xC8)*/ {"bswap",F_PAR+F_D+F_DP,EAX<<8,BSWAP,0},              //bswap eax
 /*201(0xC9)*/ {"bswap",F_PAR+F_D+F_DP,ECX<<8,BSWAP,0},              //bswap ecx
 /*202(0xCA)*/ {"bswap",F_PAR+F_D+F_DP,EDX<<8,BSWAP,0},              //bswap edx
 /*203(0xCB)*/ {"bswap",F_PAR+F_D+F_DP,EBX<<8,BSWAP,0},              //bswap ebx
 /*204(0xCC)*/ {"bswap",F_PAR+F_D+F_DP,ESP<<8,BSWAP,0},              //bswap esp
 /*205(0xCD)*/ {"bswap",F_PAR+F_D+F_DP,EBP<<8,BSWAP,0},              //bswap ebp
 /*206(0xCE)*/ {"bswap",F_PAR+F_D+F_DP,ESI<<8,BSWAP,0},              //bswap esi
 /*207(0xCF)*/ {"bswap",F_PAR+F_D+F_DP,EDI<<8,BSWAP,0},              //bswap edi
 /*208(0xD0)*/ {0,0,0,0,0},
 /*209(0xD1)*/ {0,0,0,0,0},
 /*210(0xD2)*/ {0,0,0,0,0},
 /*211(0xD3)*/ {0,0,0,0,0},
 /*212(0xD4)*/ {0,0,0,0,0},
 /*213(0xD5)*/ {0,0,0,0,0},
 /*214(0xD6)*/ {0,0,0,0,0},
 /*215(0xD7)*/ {0,0,0,0,0},
 /*216(0xD8)*/ {0,0,0,0,0},
 /*217(0xD9)*/ {0,0,0,0,0},
 /*218(0xDA)*/ {0,0,0,0,0},
 /*219(0xDB)*/ {0,0,0,0,0},
 /*220(0xDC)*/ {0,0,0,0,0},
 /*221(0xDD)*/ {0,0,0,0,0},
 /*222(0xDE)*/ {0,0,0,0,0},
 /*223(0xDF)*/ {0,0,0,0,0},
 /*224(0xE0)*/ {0,0,0,0,0},
 /*225(0xE1)*/ {0,0,0,0,0},
 /*226(0xE2)*/ {0,0,0,0,0},
 /*227(0xE3)*/ {0,0,0,0,0},
 /*228(0xE4)*/ {0,0,0,0,0},
 /*229(0xE5)*/ {0,0,0,0,0},
 /*230(0xE6)*/ {0,0,0,0,0},
 /*231(0xE7)*/ {0,0,0,0,0},
 /*232(0xE8)*/ {0,0,0,0,0},
 /*233(0xE9)*/ {0,0,0,0,0},
 /*234(0xEA)*/ {0,0,0,0,0},
 /*235(0xEB)*/ {0,0,0,0,0},
 /*236(0xEC)*/ {0,0,0,0,0},
 /*237(0xED)*/ {0,0,0,0,0},
 /*238(0xEE)*/ {0,0,0,0,0},
 /*239(0xEF)*/ {0,0,0,0,0},
 /*240(0xF0)*/ {0,0,0,0,0},
 /*241(0xF1)*/ {0,0,0,0,0},
 /*242(0xF2)*/ {0,0,0,0,0},
 /*243(0xF3)*/ {0,0,0,0,0},
 /*244(0xF4)*/ {0,0,0,0,0},
 /*245(0xF5)*/ {0,0,0,0,0},
 /*246(0xF6)*/ {0,0,0,0,0},
 /*247(0xF7)*/ {0,0,0,0,0},
 /*248(0xF8)*/ {0,0,0,0,0},
 /*249(0xF9)*/ {0,0,0,0,0},
 /*250(0xFA)*/ {0,0,0,0,0},
 /*251(0xFB)*/ {0,0,0,0,0},
 /*252(0xFC)*/ {0,0,0,0,0},
 /*253(0xFD)*/ {0,0,0,0,0},
 /*254(0xFE)*/ {0,0,0,0,0},
 /*255(0xFF)*/ {0,0,0,0,0}
};

//MnRegMem
op_t modrm[256]=
{
 /*0(0x00)*/ {0,F_PAR,16,0,0},
 /*1(0x01)*/ {0,F_PAR,272,0,0},
 /*2(0x02)*/ {0,F_PAR,528,0,0},
 /*3(0x03)*/ {0,F_PAR,784,0,0},
 /*4(0x04)*/ {0,F_PAR,1040,0,0},
 /*5(0x05)*/ {0,F_PAR,1296,0,0},
 /*6(0x06)*/ {0,F_PAR,1552,0,0},
 /*7(0x07)*/ {0,F_PAR,1808,0,0},
 /*8(0x08)*/ {0,F_PAR,16777232,0,0},
 /*9(0x09)*/ {0,F_PAR,16777488,0,0},
 /*10(0x0A)*/ {0,F_PAR,16777744,0,0},
 /*11(0x0B)*/ {0,F_PAR,16778000,0,0},
 /*12(0x0C)*/ {0,F_PAR,16778256,0,0},
 /*13(0x0D)*/ {0,F_PAR,16778512,0,0},
 /*14(0x0E)*/ {0,F_PAR,16778768,0,0},
 /*15(0x0F)*/ {0,F_PAR,16779024,0,0},
 /*16(0x10)*/ {0,F_PAR,33554448,0,0},
 /*17(0x11)*/ {0,F_PAR,33554704,0,0},
 /*18(0x12)*/ {0,F_PAR,33554960,0,0},
 /*19(0x13)*/ {0,F_PAR,33555216,0,0},
 /*20(0x14)*/ {0,F_PAR,33555472,0,0},
 /*21(0x15)*/ {0,F_PAR,33555728,0,0},
 /*22(0x16)*/ {0,F_PAR,33555984,0,0},
 /*23(0x17)*/ {0,F_PAR,33556240,0,0},
 /*24(0x18)*/ {0,F_PAR,50331664,0,0},
 /*25(0x19)*/ {0,F_PAR,50331920,0,0},
 /*26(0x1A)*/ {0,F_PAR,50332176,0,0},
 /*27(0x1B)*/ {0,F_PAR,50332432,0,0},
 /*28(0x1C)*/ {0,F_PAR,50332688,0,0},
 /*29(0x1D)*/ {0,F_PAR,50332944,0,0},
 /*30(0x1E)*/ {0,F_PAR,50333200,0,0},
 /*31(0x1F)*/ {0,F_PAR,50333456,0,0},
 /*32(0x20)*/ {0,F_PAR,67108880,0,0},
 /*33(0x21)*/ {0,F_PAR,67109136,0,0},
 /*34(0x22)*/ {0,F_PAR,67109392,0,0},
 /*35(0x23)*/ {0,F_PAR,67109648,0,0},
 /*36(0x24)*/ {0,F_PAR,67109904,0,0},
 /*37(0x25)*/ {0,F_PAR,67110160,0,0},
 /*38(0x26)*/ {0,F_PAR,67110416,0,0},
 /*39(0x27)*/ {0,F_PAR,67110672,0,0},
 /*40(0x28)*/ {0,F_PAR,83886096,0,0},
 /*41(0x29)*/ {0,F_PAR,83886352,0,0},
 /*42(0x2A)*/ {0,F_PAR,83886608,0,0},
 /*43(0x2B)*/ {0,F_PAR,83886864,0,0},
 /*44(0x2C)*/ {0,F_PAR,83887120,0,0},
 /*45(0x2D)*/ {0,F_PAR,83887376,0,0},
 /*46(0x2E)*/ {0,F_PAR,83887632,0,0},
 /*47(0x2F)*/ {0,F_PAR,83887888,0,0},
 /*48(0x30)*/ {0,F_PAR,100663312,0,0},
 /*49(0x31)*/ {0,F_PAR,100663568,0,0},
 /*50(0x32)*/ {0,F_PAR,100663824,0,0},
 /*51(0x33)*/ {0,F_PAR,100664080,0,0},
 /*52(0x34)*/ {0,F_PAR,100664336,0,0},
 /*53(0x35)*/ {0,F_PAR,100664592,0,0},
 /*54(0x36)*/ {0,F_PAR,100664848,0,0},
 /*55(0x37)*/ {0,F_PAR,100665104,0,0},
 /*56(0x38)*/ {0,F_PAR,117440528,0,0},
 /*57(0x39)*/ {0,F_PAR,117440784,0,0},
 /*58(0x3A)*/ {0,F_PAR,117441040,0,0},
 /*59(0x3B)*/ {0,F_PAR,117441296,0,0},
 /*60(0x3C)*/ {0,F_PAR,117441552,0,0},
 /*61(0x3D)*/ {0,F_PAR,117441808,0,0},
 /*62(0x3E)*/ {0,F_PAR,117442064,0,0},
 /*63(0x3F)*/ {0,F_PAR,117442320,0,0},
 /*64(0x40)*/ {0,F_PAR,17,0,0},
 /*65(0x41)*/ {0,F_PAR,273,0,0},
 /*66(0x42)*/ {0,F_PAR,529,0,0},
 /*67(0x43)*/ {0,F_PAR,785,0,0},
 /*68(0x44)*/ {0,F_PAR,1041,0,0},
 /*69(0x45)*/ {0,F_PAR,1297,0,0},
 /*70(0x46)*/ {0,F_PAR,1553,0,0},
 /*71(0x47)*/ {0,F_PAR,1809,0,0},
 /*72(0x48)*/ {0,F_PAR,16777233,0,0},
 /*73(0x49)*/ {0,F_PAR,16777489,0,0},
 /*74(0x4A)*/ {0,F_PAR,16777745,0,0},
 /*75(0x4B)*/ {0,F_PAR,16778001,0,0},
 /*76(0x4C)*/ {0,F_PAR,16778257,0,0},
 /*77(0x4D)*/ {0,F_PAR,16778513,0,0},
 /*78(0x4E)*/ {0,F_PAR,16778769,0,0},
 /*79(0x4F)*/ {0,F_PAR,16779025,0,0},
 /*80(0x50)*/ {0,F_PAR,33554449,0,0},
 /*81(0x51)*/ {0,F_PAR,33554705,0,0},
 /*82(0x52)*/ {0,F_PAR,33554961,0,0},
 /*83(0x53)*/ {0,F_PAR,33555217,0,0},
 /*84(0x54)*/ {0,F_PAR,33555473,0,0},
 /*85(0x55)*/ {0,F_PAR,33555729,0,0},
 /*86(0x56)*/ {0,F_PAR,33555985,0,0},
 /*87(0x57)*/ {0,F_PAR,33556241,0,0},
 /*88(0x58)*/ {0,F_PAR,50331665,0,0},
 /*89(0x59)*/ {0,F_PAR,50331921,0,0},
 /*90(0x5A)*/ {0,F_PAR,50332177,0,0},
 /*91(0x5B)*/ {0,F_PAR,50332433,0,0},
 /*92(0x5C)*/ {0,F_PAR,50332689,0,0},
 /*93(0x5D)*/ {0,F_PAR,50332945,0,0},
 /*94(0x5E)*/ {0,F_PAR,50333201,0,0},
 /*95(0x5F)*/ {0,F_PAR,50333457,0,0},
 /*96(0x60)*/ {0,F_PAR,67108881,0,0},
 /*97(0x61)*/ {0,F_PAR,67109137,0,0},
 /*98(0x62)*/ {0,F_PAR,67109393,0,0},
 /*99(0x63)*/ {0,F_PAR,67109649,0,0},
 /*100(0x64)*/ {0,F_PAR,67109905,0,0},
 /*101(0x65)*/ {0,F_PAR,67110161,0,0},
 /*102(0x66)*/ {0,F_PAR,67110417,0,0},
 /*103(0x67)*/ {0,F_PAR,67110673,0,0},
 /*104(0x68)*/ {0,F_PAR,83886097,0,0},
 /*105(0x69)*/ {0,F_PAR,83886353,0,0},
 /*106(0x6A)*/ {0,F_PAR,83886609,0,0},
 /*107(0x6B)*/ {0,F_PAR,83886865,0,0},
 /*108(0x6C)*/ {0,F_PAR,83887121,0,0},
 /*109(0x6D)*/ {0,F_PAR,83887377,0,0},
 /*110(0x6E)*/ {0,F_PAR,83887633,0,0},
 /*111(0x6F)*/ {0,F_PAR,83887889,0,0},
 /*112(0x70)*/ {0,F_PAR,100663313,0,0},
 /*113(0x71)*/ {0,F_PAR,100663569,0,0},
 /*114(0x72)*/ {0,F_PAR,100663825,0,0},
 /*115(0x73)*/ {0,F_PAR,100664081,0,0},
 /*116(0x74)*/ {0,F_PAR,100664337,0,0},
 /*117(0x75)*/ {0,F_PAR,100664593,0,0},
 /*118(0x76)*/ {0,F_PAR,100664849,0,0},
 /*119(0x77)*/ {0,F_PAR,100665105,0,0},
 /*120(0x78)*/ {0,F_PAR,117440529,0,0},
 /*121(0x79)*/ {0,F_PAR,117440785,0,0},
 /*122(0x7A)*/ {0,F_PAR,117441041,0,0},
 /*123(0x7B)*/ {0,F_PAR,117441297,0,0},
 /*124(0x7C)*/ {0,F_PAR,117441553,0,0},
 /*125(0x7D)*/ {0,F_PAR,117441809,0,0},
 /*126(0x7E)*/ {0,F_PAR,117442065,0,0},
 /*127(0x7F)*/ {0,F_PAR,117442321,0,0},
 /*128(0x80)*/ {0,F_PAR,18,0,0},
 /*129(0x81)*/ {0,F_PAR,274,0,0},
 /*130(0x82)*/ {0,F_PAR,530,0,0},
 /*131(0x83)*/ {0,F_PAR,786,0,0},
 /*132(0x84)*/ {0,F_PAR,1042,0,0},
 /*133(0x85)*/ {0,F_PAR,1298,0,0},
 /*134(0x86)*/ {0,F_PAR,1554,0,0},
 /*135(0x87)*/ {0,F_PAR,1810,0,0},
 /*136(0x88)*/ {0,F_PAR,16777234,0,0},
 /*137(0x89)*/ {0,F_PAR,16777490,0,0},
 /*138(0x8A)*/ {0,F_PAR,16777746,0,0},
 /*139(0x8B)*/ {0,F_PAR,16778002,0,0},
 /*140(0x8C)*/ {0,F_PAR,16778258,0,0},
 /*141(0x8D)*/ {0,F_PAR,16778514,0,0},
 /*142(0x8E)*/ {0,F_PAR,16778770,0,0},
 /*143(0x8F)*/ {0,F_PAR,16779026,0,0},
 /*144(0x90)*/ {0,F_PAR,33554450,0,0},
 /*145(0x91)*/ {0,F_PAR,33554706,0,0},
 /*146(0x92)*/ {0,F_PAR,33554962,0,0},
 /*147(0x93)*/ {0,F_PAR,33555218,0,0},
 /*148(0x94)*/ {0,F_PAR,33555474,0,0},
 /*149(0x95)*/ {0,F_PAR,33555730,0,0},
 /*150(0x96)*/ {0,F_PAR,33555986,0,0},
 /*151(0x97)*/ {0,F_PAR,33556242,0,0},
 /*152(0x98)*/ {0,F_PAR,50331666,0,0},
 /*153(0x99)*/ {0,F_PAR,50331922,0,0},
 /*154(0x9A)*/ {0,F_PAR,50332178,0,0},
 /*155(0x9B)*/ {0,F_PAR,50332434,0,0},
 /*156(0x9C)*/ {0,F_PAR,50332690,0,0},
 /*157(0x9D)*/ {0,F_PAR,50332946,0,0},
 /*158(0x9E)*/ {0,F_PAR,50333202,0,0},
 /*159(0x9F)*/ {0,F_PAR,50333458,0,0},
 /*160(0xA0)*/ {0,F_PAR,67108882,0,0},
 /*161(0xA1)*/ {0,F_PAR,67109138,0,0},
 /*162(0xA2)*/ {0,F_PAR,67109394,0,0},
 /*163(0xA3)*/ {0,F_PAR,67109650,0,0},
 /*164(0xA4)*/ {0,F_PAR,67109906,0,0},
 /*165(0xA5)*/ {0,F_PAR,67110162,0,0},
 /*166(0xA6)*/ {0,F_PAR,67110418,0,0},
 /*167(0xA7)*/ {0,F_PAR,67110674,0,0},
 /*168(0xA8)*/ {0,F_PAR,83886098,0,0},
 /*169(0xA9)*/ {0,F_PAR,83886354,0,0},
 /*170(0xAA)*/ {0,F_PAR,83886610,0,0},
 /*171(0xAB)*/ {0,F_PAR,83886866,0,0},
 /*172(0xAC)*/ {0,F_PAR,83887122,0,0},
 /*173(0xAD)*/ {0,F_PAR,83887378,0,0},
 /*174(0xAE)*/ {0,F_PAR,83887634,0,0},
 /*175(0xAF)*/ {0,F_PAR,83887890,0,0},
 /*176(0xB0)*/ {0,F_PAR,100663314,0,0},
 /*177(0xB1)*/ {0,F_PAR,100663570,0,0},
 /*178(0xB2)*/ {0,F_PAR,100663826,0,0},
 /*179(0xB3)*/ {0,F_PAR,100664082,0,0},
 /*180(0xB4)*/ {0,F_PAR,100664338,0,0},
 /*181(0xB5)*/ {0,F_PAR,100664594,0,0},
 /*182(0xB6)*/ {0,F_PAR,100664850,0,0},
 /*183(0xB7)*/ {0,F_PAR,100665106,0,0},
 /*184(0xB8)*/ {0,F_PAR,117440530,0,0},
 /*185(0xB9)*/ {0,F_PAR,117440786,0,0},
 /*186(0xBA)*/ {0,F_PAR,117441042,0,0},
 /*187(0xBB)*/ {0,F_PAR,117441298,0,0},
 /*188(0xBC)*/ {0,F_PAR,117441554,0,0},
 /*189(0xBD)*/ {0,F_PAR,117441810,0,0},
 /*190(0xBE)*/ {0,F_PAR,117442066,0,0},
 /*191(0xBF)*/ {0,F_PAR,117442322,0,0},
 /*192(0xC0)*/ {0,F_PAR,0,0,0},
 /*193(0xC1)*/ {0,F_PAR,256,0,0},
 /*194(0xC2)*/ {0,F_PAR,512,0,0},
 /*195(0xC3)*/ {0,F_PAR,768,0,0},
 /*196(0xC4)*/ {0,F_PAR,1024,0,0},
 /*197(0xC5)*/ {0,F_PAR,1280,0,0},
 /*198(0xC6)*/ {0,F_PAR,1536,0,0},
 /*199(0xC7)*/ {0,F_PAR,1792,0,0},
 /*200(0xC8)*/ {0,F_PAR,16777216,0,0},
 /*201(0xC9)*/ {0,F_PAR,16777472,0,0},
 /*202(0xCA)*/ {0,F_PAR,16777728,0,0},
 /*203(0xCB)*/ {0,F_PAR,16777984,0,0},
 /*204(0xCC)*/ {0,F_PAR,16778240,0,0},
 /*205(0xCD)*/ {0,F_PAR,16778496,0,0},
 /*206(0xCE)*/ {0,F_PAR,16778752,0,0},
 /*207(0xCF)*/ {0,F_PAR,16779008,0,0},
 /*208(0xD0)*/ {0,F_PAR,33554432,0,0},
 /*209(0xD1)*/ {0,F_PAR,33554688,0,0},
 /*210(0xD2)*/ {0,F_PAR,33554944,0,0},
 /*211(0xD3)*/ {0,F_PAR,33555200,0,0},
 /*212(0xD4)*/ {0,F_PAR,33555456,0,0},
 /*213(0xD5)*/ {0,F_PAR,33555712,0,0},
 /*214(0xD6)*/ {0,F_PAR,33555968,0,0},
 /*215(0xD7)*/ {0,F_PAR,33556224,0,0},
 /*216(0xD8)*/ {0,F_PAR,50331648,0,0},
 /*217(0xD9)*/ {0,F_PAR,50331904,0,0},
 /*218(0xDA)*/ {0,F_PAR,50332160,0,0},
 /*219(0xDB)*/ {0,F_PAR,50332416,0,0},
 /*220(0xDC)*/ {0,F_PAR,50332672,0,0},
 /*221(0xDD)*/ {0,F_PAR,50332928,0,0},
 /*222(0xDE)*/ {0,F_PAR,50333184,0,0},
 /*223(0xDF)*/ {0,F_PAR,50333440,0,0},
 /*224(0xE0)*/ {0,F_PAR,67108864,0,0},
 /*225(0xE1)*/ {0,F_PAR,67109120,0,0},
 /*226(0xE2)*/ {0,F_PAR,67109376,0,0},
 /*227(0xE3)*/ {0,F_PAR,67109632,0,0},
 /*228(0xE4)*/ {0,F_PAR,67109888,0,0},
 /*229(0xE5)*/ {0,F_PAR,67110144,0,0},
 /*230(0xE6)*/ {0,F_PAR,67110400,0,0},
 /*231(0xE7)*/ {0,F_PAR,67110656,0,0},
 /*232(0xE8)*/ {0,F_PAR,83886080,0,0},
 /*233(0xE9)*/ {0,F_PAR,83886336,0,0},
 /*234(0xEA)*/ {0,F_PAR,83886592,0,0},
 /*235(0xEB)*/ {0,F_PAR,83886848,0,0},
 /*236(0xEC)*/ {0,F_PAR,83887104,0,0},
 /*237(0xED)*/ {0,F_PAR,83887360,0,0},
 /*238(0xEE)*/ {0,F_PAR,83887616,0,0},
 /*239(0xEF)*/ {0,F_PAR,83887872,0,0},
 /*240(0xF0)*/ {0,F_PAR,100663296,0,0},
 /*241(0xF1)*/ {0,F_PAR,100663552,0,0},
 /*242(0xF2)*/ {0,F_PAR,100663808,0,0},
 /*243(0xF3)*/ {0,F_PAR,100664064,0,0},
 /*244(0xF4)*/ {0,F_PAR,100664320,0,0},
 /*245(0xF5)*/ {0,F_PAR,100664576,0,0},
 /*246(0xF6)*/ {0,F_PAR,100664832,0,0},
 /*247(0xF7)*/ {0,F_PAR,100665088,0,0},
 /*248(0xF8)*/ {0,F_PAR,117440512,0,0},
 /*249(0xF9)*/ {0,F_PAR,117440768,0,0},
 /*250(0xFA)*/ {0,F_PAR,117441024,0,0},
 /*251(0xFB)*/ {0,F_PAR,117441280,0,0},
 /*252(0xFC)*/ {0,F_PAR,117441536,0,0},
 /*253(0xFD)*/ {0,F_PAR,117441792,0,0},
 /*254(0xFE)*/ {0,F_PAR,117442048,0,0},
 /*255(0xFF)*/ {0,F_PAR,117442304,0,0}
};

//MnSegMem
op_t modsm[256]=
{
 /*0(0x00)*/ {0,F_PAR,65552,0,0},
 /*1(0x01)*/ {0,F_PAR,65808,0,0},
 /*2(0x02)*/ {0,F_PAR,66064,0,0},
 /*3(0x03)*/ {0,F_PAR,66320,0,0},
 /*4(0x04)*/ {0,F_PAR,66576,0,0},
 /*5(0x05)*/ {0,F_PAR,66832,0,0},
 /*6(0x06)*/ {0,F_PAR,67088,0,0},
 /*7(0x07)*/ {0,F_PAR,67344,0,0},
 /*8(0x08)*/ {0,F_PAR,16842768,0,0},
 /*9(0x09)*/ {0,F_PAR,16843024,0,0},
 /*10(0x0A)*/ {0,F_PAR,16843280,0,0},
 /*11(0x0B)*/ {0,F_PAR,16843536,0,0},
 /*12(0x0C)*/ {0,F_PAR,16843792,0,0},
 /*13(0x0D)*/ {0,F_PAR,16844048,0,0},
 /*14(0x0E)*/ {0,F_PAR,16844304,0,0},
 /*15(0x0F)*/ {0,F_PAR,16844560,0,0},
 /*16(0x10)*/ {0,F_PAR,33619984,0,0},
 /*17(0x11)*/ {0,F_PAR,33620240,0,0},
 /*18(0x12)*/ {0,F_PAR,33620496,0,0},
 /*19(0x13)*/ {0,F_PAR,33620752,0,0},
 /*20(0x14)*/ {0,F_PAR,33621008,0,0},
 /*21(0x15)*/ {0,F_PAR,33621264,0,0},
 /*22(0x16)*/ {0,F_PAR,33621520,0,0},
 /*23(0x17)*/ {0,F_PAR,33621776,0,0},
 /*24(0x18)*/ {0,F_PAR,50397200,0,0},
 /*25(0x19)*/ {0,F_PAR,50397456,0,0},
 /*26(0x1A)*/ {0,F_PAR,50397712,0,0},
 /*27(0x1B)*/ {0,F_PAR,50397968,0,0},
 /*28(0x1C)*/ {0,F_PAR,50398224,0,0},
 /*29(0x1D)*/ {0,F_PAR,50398480,0,0},
 /*30(0x1E)*/ {0,F_PAR,50398736,0,0},
 /*31(0x1F)*/ {0,F_PAR,50398992,0,0},
 /*32(0x20)*/ {0,F_PAR,67174416,0,0},
 /*33(0x21)*/ {0,F_PAR,67174672,0,0},
 /*34(0x22)*/ {0,F_PAR,67174928,0,0},
 /*35(0x23)*/ {0,F_PAR,67175184,0,0},
 /*36(0x24)*/ {0,F_PAR,67175440,0,0},
 /*37(0x25)*/ {0,F_PAR,67175696,0,0},
 /*38(0x26)*/ {0,F_PAR,67175952,0,0},
 /*39(0x27)*/ {0,F_PAR,67176208,0,0},
 /*40(0x28)*/ {0,F_PAR,83951632,0,0},
 /*41(0x29)*/ {0,F_PAR,83951888,0,0},
 /*42(0x2A)*/ {0,F_PAR,83952144,0,0},
 /*43(0x2B)*/ {0,F_PAR,83952400,0,0},
 /*44(0x2C)*/ {0,F_PAR,83952656,0,0},
 /*45(0x2D)*/ {0,F_PAR,83952912,0,0},
 /*46(0x2E)*/ {0,F_PAR,83953168,0,0},
 /*47(0x2F)*/ {0,F_PAR,83953424,0,0},
 /*48(0x30)*/ {0,F_PAR,100728848,0,0},
 /*49(0x31)*/ {0,F_PAR,100729104,0,0},
 /*50(0x32)*/ {0,F_PAR,100729360,0,0},
 /*51(0x33)*/ {0,F_PAR,100729616,0,0},
 /*52(0x34)*/ {0,F_PAR,100729872,0,0},
 /*53(0x35)*/ {0,F_PAR,100730128,0,0},
 /*54(0x36)*/ {0,F_PAR,100730384,0,0},
 /*55(0x37)*/ {0,F_PAR,100730640,0,0},
 /*56(0x38)*/ {0,F_PAR,117506064,0,0},
 /*57(0x39)*/ {0,F_PAR,117506320,0,0},
 /*58(0x3A)*/ {0,F_PAR,117506576,0,0},
 /*59(0x3B)*/ {0,F_PAR,117506832,0,0},
 /*60(0x3C)*/ {0,F_PAR,117507088,0,0},
 /*61(0x3D)*/ {0,F_PAR,117507344,0,0},
 /*62(0x3E)*/ {0,F_PAR,117507600,0,0},
 /*63(0x3F)*/ {0,F_PAR,117507856,0,0},
 /*64(0x40)*/ {0,F_PAR,65553,0,0},
 /*65(0x41)*/ {0,F_PAR,65809,0,0},
 /*66(0x42)*/ {0,F_PAR,66065,0,0},
 /*67(0x43)*/ {0,F_PAR,66321,0,0},
 /*68(0x44)*/ {0,F_PAR,66577,0,0},
 /*69(0x45)*/ {0,F_PAR,66833,0,0},
 /*70(0x46)*/ {0,F_PAR,67089,0,0},
 /*71(0x47)*/ {0,F_PAR,67345,0,0},
 /*72(0x48)*/ {0,F_PAR,16842769,0,0},
 /*73(0x49)*/ {0,F_PAR,16843025,0,0},
 /*74(0x4A)*/ {0,F_PAR,16843281,0,0},
 /*75(0x4B)*/ {0,F_PAR,16843537,0,0},
 /*76(0x4C)*/ {0,F_PAR,16843793,0,0},
 /*77(0x4D)*/ {0,F_PAR,16844049,0,0},
 /*78(0x4E)*/ {0,F_PAR,16844305,0,0},
 /*79(0x4F)*/ {0,F_PAR,16844561,0,0},
 /*80(0x50)*/ {0,F_PAR,33619985,0,0},
 /*81(0x51)*/ {0,F_PAR,33620241,0,0},
 /*82(0x52)*/ {0,F_PAR,33620497,0,0},
 /*83(0x53)*/ {0,F_PAR,33620753,0,0},
 /*84(0x54)*/ {0,F_PAR,33621009,0,0},
 /*85(0x55)*/ {0,F_PAR,33621265,0,0},
 /*86(0x56)*/ {0,F_PAR,33621521,0,0},
 /*87(0x57)*/ {0,F_PAR,33621777,0,0},
 /*88(0x58)*/ {0,F_PAR,50397201,0,0},
 /*89(0x59)*/ {0,F_PAR,50397457,0,0},
 /*90(0x5A)*/ {0,F_PAR,50397713,0,0},
 /*91(0x5B)*/ {0,F_PAR,50397969,0,0},
 /*92(0x5C)*/ {0,F_PAR,50398225,0,0},
 /*93(0x5D)*/ {0,F_PAR,50398481,0,0},
 /*94(0x5E)*/ {0,F_PAR,50398737,0,0},
 /*95(0x5F)*/ {0,F_PAR,50398993,0,0},
 /*96(0x60)*/ {0,F_PAR,67174417,0,0},
 /*97(0x61)*/ {0,F_PAR,67174673,0,0},
 /*98(0x62)*/ {0,F_PAR,67174929,0,0},
 /*99(0x63)*/ {0,F_PAR,67175185,0,0},
 /*100(0x64)*/ {0,F_PAR,67175441,0,0},
 /*101(0x65)*/ {0,F_PAR,67175697,0,0},
 /*102(0x66)*/ {0,F_PAR,67175953,0,0},
 /*103(0x67)*/ {0,F_PAR,67176209,0,0},
 /*104(0x68)*/ {0,F_PAR,83951633,0,0},
 /*105(0x69)*/ {0,F_PAR,83951889,0,0},
 /*106(0x6A)*/ {0,F_PAR,83952145,0,0},
 /*107(0x6B)*/ {0,F_PAR,83952401,0,0},
 /*108(0x6C)*/ {0,F_PAR,83952657,0,0},
 /*109(0x6D)*/ {0,F_PAR,83952913,0,0},
 /*110(0x6E)*/ {0,F_PAR,83953169,0,0},
 /*111(0x6F)*/ {0,F_PAR,83953425,0,0},
 /*112(0x70)*/ {0,F_PAR,100728849,0,0},
 /*113(0x71)*/ {0,F_PAR,100729105,0,0},
 /*114(0x72)*/ {0,F_PAR,100729361,0,0},
 /*115(0x73)*/ {0,F_PAR,100729617,0,0},
 /*116(0x74)*/ {0,F_PAR,100729873,0,0},
 /*117(0x75)*/ {0,F_PAR,100730129,0,0},
 /*118(0x76)*/ {0,F_PAR,100730385,0,0},
 /*119(0x77)*/ {0,F_PAR,100730641,0,0},
 /*120(0x78)*/ {0,F_PAR,117506065,0,0},
 /*121(0x79)*/ {0,F_PAR,117506321,0,0},
 /*122(0x7A)*/ {0,F_PAR,117506577,0,0},
 /*123(0x7B)*/ {0,F_PAR,117506833,0,0},
 /*124(0x7C)*/ {0,F_PAR,117507089,0,0},
 /*125(0x7D)*/ {0,F_PAR,117507345,0,0},
 /*126(0x7E)*/ {0,F_PAR,117507601,0,0},
 /*127(0x7F)*/ {0,F_PAR,117507857,0,0},
 /*128(0x80)*/ {0,F_PAR,65554,0,0},
 /*129(0x81)*/ {0,F_PAR,65810,0,0},
 /*130(0x82)*/ {0,F_PAR,66066,0,0},
 /*131(0x83)*/ {0,F_PAR,66322,0,0},
 /*132(0x84)*/ {0,F_PAR,66578,0,0},
 /*133(0x85)*/ {0,F_PAR,66834,0,0},
 /*134(0x86)*/ {0,F_PAR,67090,0,0},
 /*135(0x87)*/ {0,F_PAR,67346,0,0},
 /*136(0x88)*/ {0,F_PAR,16842770,0,0},
 /*137(0x89)*/ {0,F_PAR,16843026,0,0},
 /*138(0x8A)*/ {0,F_PAR,16843282,0,0},
 /*139(0x8B)*/ {0,F_PAR,16843538,0,0},
 /*140(0x8C)*/ {0,F_PAR,16843794,0,0},
 /*141(0x8D)*/ {0,F_PAR,16844050,0,0},
 /*142(0x8E)*/ {0,F_PAR,16844306,0,0},
 /*143(0x8F)*/ {0,F_PAR,16844562,0,0},
 /*144(0x90)*/ {0,F_PAR,33619986,0,0},
 /*145(0x91)*/ {0,F_PAR,33620242,0,0},
 /*146(0x92)*/ {0,F_PAR,33620498,0,0},
 /*147(0x93)*/ {0,F_PAR,33620754,0,0},
 /*148(0x94)*/ {0,F_PAR,33621010,0,0},
 /*149(0x95)*/ {0,F_PAR,33621266,0,0},
 /*150(0x96)*/ {0,F_PAR,33621522,0,0},
 /*151(0x97)*/ {0,F_PAR,33621778,0,0},
 /*152(0x98)*/ {0,F_PAR,50397202,0,0},
 /*153(0x99)*/ {0,F_PAR,50397458,0,0},
 /*154(0x9A)*/ {0,F_PAR,50397714,0,0},
 /*155(0x9B)*/ {0,F_PAR,50397970,0,0},
 /*156(0x9C)*/ {0,F_PAR,50398226,0,0},
 /*157(0x9D)*/ {0,F_PAR,50398482,0,0},
 /*158(0x9E)*/ {0,F_PAR,50398738,0,0},
 /*159(0x9F)*/ {0,F_PAR,50398994,0,0},
 /*160(0xA0)*/ {0,F_PAR,67174418,0,0},
 /*161(0xA1)*/ {0,F_PAR,67174674,0,0},
 /*162(0xA2)*/ {0,F_PAR,67174930,0,0},
 /*163(0xA3)*/ {0,F_PAR,67175186,0,0},
 /*164(0xA4)*/ {0,F_PAR,67175442,0,0},
 /*165(0xA5)*/ {0,F_PAR,67175698,0,0},
 /*166(0xA6)*/ {0,F_PAR,67175954,0,0},
 /*167(0xA7)*/ {0,F_PAR,67176210,0,0},
 /*168(0xA8)*/ {0,F_PAR,83951634,0,0},
 /*169(0xA9)*/ {0,F_PAR,83951890,0,0},
 /*170(0xAA)*/ {0,F_PAR,83952146,0,0},
 /*171(0xAB)*/ {0,F_PAR,83952402,0,0},
 /*172(0xAC)*/ {0,F_PAR,83952658,0,0},
 /*173(0xAD)*/ {0,F_PAR,83952914,0,0},
 /*174(0xAE)*/ {0,F_PAR,83953170,0,0},
 /*175(0xAF)*/ {0,F_PAR,83953426,0,0},
 /*176(0xB0)*/ {0,F_PAR,100728850,0,0},
 /*177(0xB1)*/ {0,F_PAR,100729106,0,0},
 /*178(0xB2)*/ {0,F_PAR,100729362,0,0},
 /*179(0xB3)*/ {0,F_PAR,100729618,0,0},
 /*180(0xB4)*/ {0,F_PAR,100729874,0,0},
 /*181(0xB5)*/ {0,F_PAR,100730130,0,0},
 /*182(0xB6)*/ {0,F_PAR,100730386,0,0},
 /*183(0xB7)*/ {0,F_PAR,100730642,0,0},
 /*184(0xB8)*/ {0,F_PAR,117506066,0,0},
 /*185(0xB9)*/ {0,F_PAR,117506322,0,0},
 /*186(0xBA)*/ {0,F_PAR,117506578,0,0},
 /*187(0xBB)*/ {0,F_PAR,117506834,0,0},
 /*188(0xBC)*/ {0,F_PAR,117507090,0,0},
 /*189(0xBD)*/ {0,F_PAR,117507346,0,0},
 /*190(0xBE)*/ {0,F_PAR,117507602,0,0},
 /*191(0xBF)*/ {0,F_PAR,117507858,0,0},
 /*192(0xC0)*/ {0,F_PAR,65536,0,0},
 /*193(0xC1)*/ {0,F_PAR,65792,0,0},
 /*194(0xC2)*/ {0,F_PAR,66048,0,0},
 /*195(0xC3)*/ {0,F_PAR,66304,0,0},
 /*196(0xC4)*/ {0,F_PAR,66560,0,0},
 /*197(0xC5)*/ {0,F_PAR,66816,0,0},
 /*198(0xC6)*/ {0,F_PAR,67072,0,0},
 /*199(0xC7)*/ {0,F_PAR,67328,0,0},
 /*200(0xC8)*/ {0,F_PAR,16842752,0,0},
 /*201(0xC9)*/ {0,F_PAR,16843008,0,0},
 /*202(0xCA)*/ {0,F_PAR,16843264,0,0},
 /*203(0xCB)*/ {0,F_PAR,16843520,0,0},
 /*204(0xCC)*/ {0,F_PAR,16843776,0,0},
 /*205(0xCD)*/ {0,F_PAR,16844032,0,0},
 /*206(0xCE)*/ {0,F_PAR,16844288,0,0},
 /*207(0xCF)*/ {0,F_PAR,16844544,0,0},
 /*208(0xD0)*/ {0,F_PAR,33619968,0,0},
 /*209(0xD1)*/ {0,F_PAR,33620224,0,0},
 /*210(0xD2)*/ {0,F_PAR,33620480,0,0},
 /*211(0xD3)*/ {0,F_PAR,33620736,0,0},
 /*212(0xD4)*/ {0,F_PAR,33620992,0,0},
 /*213(0xD5)*/ {0,F_PAR,33621248,0,0},
 /*214(0xD6)*/ {0,F_PAR,33621504,0,0},
 /*215(0xD7)*/ {0,F_PAR,33621760,0,0},
 /*216(0xD8)*/ {0,F_PAR,50397184,0,0},
 /*217(0xD9)*/ {0,F_PAR,50397440,0,0},
 /*218(0xDA)*/ {0,F_PAR,50397696,0,0},
 /*219(0xDB)*/ {0,F_PAR,50397952,0,0},
 /*220(0xDC)*/ {0,F_PAR,50398208,0,0},
 /*221(0xDD)*/ {0,F_PAR,50398464,0,0},
 /*222(0xDE)*/ {0,F_PAR,50398720,0,0},
 /*223(0xDF)*/ {0,F_PAR,50398976,0,0},
 /*224(0xE0)*/ {0,F_PAR,67174400,0,0},
 /*225(0xE1)*/ {0,F_PAR,67174656,0,0},
 /*226(0xE2)*/ {0,F_PAR,67174912,0,0},
 /*227(0xE3)*/ {0,F_PAR,67175168,0,0},
 /*228(0xE4)*/ {0,F_PAR,67175424,0,0},
 /*229(0xE5)*/ {0,F_PAR,67175680,0,0},
 /*230(0xE6)*/ {0,F_PAR,67175936,0,0},
 /*231(0xE7)*/ {0,F_PAR,67176192,0,0},
 /*232(0xE8)*/ {0,F_PAR,83951616,0,0},
 /*233(0xE9)*/ {0,F_PAR,83951872,0,0},
 /*234(0xEA)*/ {0,F_PAR,83952128,0,0},
 /*235(0xEB)*/ {0,F_PAR,83952384,0,0},
 /*236(0xEC)*/ {0,F_PAR,83952640,0,0},
 /*237(0xED)*/ {0,F_PAR,83952896,0,0},
 /*238(0xEE)*/ {0,F_PAR,83953152,0,0},
 /*239(0xEF)*/ {0,F_PAR,83953408,0,0},
 /*240(0xF0)*/ {0,F_PAR,100728832,0,0},
 /*241(0xF1)*/ {0,F_PAR,100729088,0,0},
 /*242(0xF2)*/ {0,F_PAR,100729344,0,0},
 /*243(0xF3)*/ {0,F_PAR,100729600,0,0},
 /*244(0xF4)*/ {0,F_PAR,100729856,0,0},
 /*245(0xF5)*/ {0,F_PAR,100730112,0,0},
 /*246(0xF6)*/ {0,F_PAR,100730368,0,0},
 /*247(0xF7)*/ {0,F_PAR,100730624,0,0},
 /*248(0xF8)*/ {0,F_PAR,117506048,0,0},
 /*249(0xF9)*/ {0,F_PAR,117506304,0,0},
 /*250(0xFA)*/ {0,F_PAR,117506560,0,0},
 /*251(0xFB)*/ {0,F_PAR,117506816,0,0},
 /*252(0xFC)*/ {0,F_PAR,117507072,0,0},
 /*253(0xFD)*/ {0,F_PAR,117507328,0,0},
 /*254(0xFE)*/ {0,F_PAR,117507584,0,0},
 /*255(0xFF)*/ {0,F_PAR,117507840,0,0}
};

//rejestry cr
op_t modcr[256]=
{
 /*0(0x00)*/ {0,0,196608,0,0},
 /*1(0x01)*/ {0,0,16973824,0,0},
 /*2(0x02)*/ {0,0,33751040,0,0},
 /*3(0x03)*/ {0,0,50528256,0,0},
 /*4(0x04)*/ {0,0,67305472,0,0},
 /*5(0x05)*/ {0,0,84082688,0,0},
 /*6(0x06)*/ {0,0,100859904,0,0},
 /*7(0x07)*/ {0,0,117637120,0,0},
 /*8(0x08)*/ {0,0,196864,0,0},
 /*9(0x09)*/ {0,0,16974080,0,0},
 /*10(0x0A)*/ {0,0,33751296,0,0},
 /*11(0x0B)*/ {0,0,50528512,0,0},
 /*12(0x0C)*/ {0,0,67305728,0,0},
 /*13(0x0D)*/ {0,0,84082944,0,0},
 /*14(0x0E)*/ {0,0,100860160,0,0},
 /*15(0x0F)*/ {0,0,117637376,0,0},
 /*16(0x10)*/ {0,0,197120,0,0},
 /*17(0x11)*/ {0,0,16974336,0,0},
 /*18(0x12)*/ {0,0,33751552,0,0},
 /*19(0x13)*/ {0,0,50528768,0,0},
 /*20(0x14)*/ {0,0,67305984,0,0},
 /*21(0x15)*/ {0,0,84083200,0,0},
 /*22(0x16)*/ {0,0,100860416,0,0},
 /*23(0x17)*/ {0,0,117637632,0,0},
 /*24(0x18)*/ {0,0,197376,0,0},
 /*25(0x19)*/ {0,0,16974592,0,0},
 /*26(0x1A)*/ {0,0,33751808,0,0},
 /*27(0x1B)*/ {0,0,50529024,0,0},
 /*28(0x1C)*/ {0,0,67306240,0,0},
 /*29(0x1D)*/ {0,0,84083456,0,0},
 /*30(0x1E)*/ {0,0,100860672,0,0},
 /*31(0x1F)*/ {0,0,117637888,0,0},
 /*32(0x20)*/ {0,0,197632,0,0},
 /*33(0x21)*/ {0,0,16974848,0,0},
 /*34(0x22)*/ {0,0,33752064,0,0},
 /*35(0x23)*/ {0,0,50529280,0,0},
 /*36(0x24)*/ {0,0,67306496,0,0},
 /*37(0x25)*/ {0,0,84083712,0,0},
 /*38(0x26)*/ {0,0,100860928,0,0},
 /*39(0x27)*/ {0,0,117638144,0,0},
 /*40(0x28)*/ {0,0,197888,0,0},
 /*41(0x29)*/ {0,0,16975104,0,0},
 /*42(0x2A)*/ {0,0,33752320,0,0},
 /*43(0x2B)*/ {0,0,50529536,0,0},
 /*44(0x2C)*/ {0,0,67306752,0,0},
 /*45(0x2D)*/ {0,0,84083968,0,0},
 /*46(0x2E)*/ {0,0,100861184,0,0},
 /*47(0x2F)*/ {0,0,117638400,0,0},
 /*48(0x30)*/ {0,0,198144,0,0},
 /*49(0x31)*/ {0,0,16975360,0,0},
 /*50(0x32)*/ {0,0,33752576,0,0},
 /*51(0x33)*/ {0,0,50529792,0,0},
 /*52(0x34)*/ {0,0,67307008,0,0},
 /*53(0x35)*/ {0,0,84084224,0,0},
 /*54(0x36)*/ {0,0,100861440,0,0},
 /*55(0x37)*/ {0,0,117638656,0,0},
 /*56(0x38)*/ {0,0,198400,0,0},
 /*57(0x39)*/ {0,0,16975616,0,0},
 /*58(0x3A)*/ {0,0,33752832,0,0},
 /*59(0x3B)*/ {0,0,50530048,0,0},
 /*60(0x3C)*/ {0,0,67307264,0,0},
 /*61(0x3D)*/ {0,0,84084480,0,0},
 /*62(0x3E)*/ {0,0,100861696,0,0},
 /*63(0x3F)*/ {0,0,117638912,0,0},
 /*64(0x40)*/ {0,0,196608,0,0},
 /*65(0x41)*/ {0,0,16973824,0,0},
 /*66(0x42)*/ {0,0,33751040,0,0},
 /*67(0x43)*/ {0,0,50528256,0,0},
 /*68(0x44)*/ {0,0,67305472,0,0},
 /*69(0x45)*/ {0,0,84082688,0,0},
 /*70(0x46)*/ {0,0,100859904,0,0},
 /*71(0x47)*/ {0,0,117637120,0,0},
 /*72(0x48)*/ {0,0,196864,0,0},
 /*73(0x49)*/ {0,0,16974080,0,0},
 /*74(0x4A)*/ {0,0,33751296,0,0},
 /*75(0x4B)*/ {0,0,50528512,0,0},
 /*76(0x4C)*/ {0,0,67305728,0,0},
 /*77(0x4D)*/ {0,0,84082944,0,0},
 /*78(0x4E)*/ {0,0,100860160,0,0},
 /*79(0x4F)*/ {0,0,117637376,0,0},
 /*80(0x50)*/ {0,0,197120,0,0},
 /*81(0x51)*/ {0,0,16974336,0,0},
 /*82(0x52)*/ {0,0,33751552,0,0},
 /*83(0x53)*/ {0,0,50528768,0,0},
 /*84(0x54)*/ {0,0,67305984,0,0},
 /*85(0x55)*/ {0,0,84083200,0,0},
 /*86(0x56)*/ {0,0,100860416,0,0},
 /*87(0x57)*/ {0,0,117637632,0,0},
 /*88(0x58)*/ {0,0,197376,0,0},
 /*89(0x59)*/ {0,0,16974592,0,0},
 /*90(0x5A)*/ {0,0,33751808,0,0},
 /*91(0x5B)*/ {0,0,50529024,0,0},
 /*92(0x5C)*/ {0,0,67306240,0,0},
 /*93(0x5D)*/ {0,0,84083456,0,0},
 /*94(0x5E)*/ {0,0,100860672,0,0},
 /*95(0x5F)*/ {0,0,117637888,0,0},
 /*96(0x60)*/ {0,0,197632,0,0},
 /*97(0x61)*/ {0,0,16974848,0,0},
 /*98(0x62)*/ {0,0,33752064,0,0},
 /*99(0x63)*/ {0,0,50529280,0,0},
 /*100(0x64)*/ {0,0,67306496,0,0},
 /*101(0x65)*/ {0,0,84083712,0,0},
 /*102(0x66)*/ {0,0,100860928,0,0},
 /*103(0x67)*/ {0,0,117638144,0,0},
 /*104(0x68)*/ {0,0,197888,0,0},
 /*105(0x69)*/ {0,0,16975104,0,0},
 /*106(0x6A)*/ {0,0,33752320,0,0},
 /*107(0x6B)*/ {0,0,50529536,0,0},
 /*108(0x6C)*/ {0,0,67306752,0,0},
 /*109(0x6D)*/ {0,0,84083968,0,0},
 /*110(0x6E)*/ {0,0,100861184,0,0},
 /*111(0x6F)*/ {0,0,117638400,0,0},
 /*112(0x70)*/ {0,0,198144,0,0},
 /*113(0x71)*/ {0,0,16975360,0,0},
 /*114(0x72)*/ {0,0,33752576,0,0},
 /*115(0x73)*/ {0,0,50529792,0,0},
 /*116(0x74)*/ {0,0,67307008,0,0},
 /*117(0x75)*/ {0,0,84084224,0,0},
 /*118(0x76)*/ {0,0,100861440,0,0},
 /*119(0x77)*/ {0,0,117638656,0,0},
 /*120(0x78)*/ {0,0,198400,0,0},
 /*121(0x79)*/ {0,0,16975616,0,0},
 /*122(0x7A)*/ {0,0,33752832,0,0},
 /*123(0x7B)*/ {0,0,50530048,0,0},
 /*124(0x7C)*/ {0,0,67307264,0,0},
 /*125(0x7D)*/ {0,0,84084480,0,0},
 /*126(0x7E)*/ {0,0,100861696,0,0},
 /*127(0x7F)*/ {0,0,117638912,0,0},
 /*128(0x80)*/ {0,0,196608,0,0},
 /*129(0x81)*/ {0,0,16973824,0,0},
 /*130(0x82)*/ {0,0,33751040,0,0},
 /*131(0x83)*/ {0,0,50528256,0,0},
 /*132(0x84)*/ {0,0,67305472,0,0},
 /*133(0x85)*/ {0,0,84082688,0,0},
 /*134(0x86)*/ {0,0,100859904,0,0},
 /*135(0x87)*/ {0,0,117637120,0,0},
 /*136(0x88)*/ {0,0,196864,0,0},
 /*137(0x89)*/ {0,0,16974080,0,0},
 /*138(0x8A)*/ {0,0,33751296,0,0},
 /*139(0x8B)*/ {0,0,50528512,0,0},
 /*140(0x8C)*/ {0,0,67305728,0,0},
 /*141(0x8D)*/ {0,0,84082944,0,0},
 /*142(0x8E)*/ {0,0,100860160,0,0},
 /*143(0x8F)*/ {0,0,117637376,0,0},
 /*144(0x90)*/ {0,0,197120,0,0},
 /*145(0x91)*/ {0,0,16974336,0,0},
 /*146(0x92)*/ {0,0,33751552,0,0},
 /*147(0x93)*/ {0,0,50528768,0,0},
 /*148(0x94)*/ {0,0,67305984,0,0},
 /*149(0x95)*/ {0,0,84083200,0,0},
 /*150(0x96)*/ {0,0,100860416,0,0},
 /*151(0x97)*/ {0,0,117637632,0,0},
 /*152(0x98)*/ {0,0,197376,0,0},
 /*153(0x99)*/ {0,0,16974592,0,0},
 /*154(0x9A)*/ {0,0,33751808,0,0},
 /*155(0x9B)*/ {0,0,50529024,0,0},
 /*156(0x9C)*/ {0,0,67306240,0,0},
 /*157(0x9D)*/ {0,0,84083456,0,0},
 /*158(0x9E)*/ {0,0,100860672,0,0},
 /*159(0x9F)*/ {0,0,117637888,0,0},
 /*160(0xA0)*/ {0,0,197632,0,0},
 /*161(0xA1)*/ {0,0,16974848,0,0},
 /*162(0xA2)*/ {0,0,33752064,0,0},
 /*163(0xA3)*/ {0,0,50529280,0,0},
 /*164(0xA4)*/ {0,0,67306496,0,0},
 /*165(0xA5)*/ {0,0,84083712,0,0},
 /*166(0xA6)*/ {0,0,100860928,0,0},
 /*167(0xA7)*/ {0,0,117638144,0,0},
 /*168(0xA8)*/ {0,0,197888,0,0},
 /*169(0xA9)*/ {0,0,16975104,0,0},
 /*170(0xAA)*/ {0,0,33752320,0,0},
 /*171(0xAB)*/ {0,0,50529536,0,0},
 /*172(0xAC)*/ {0,0,67306752,0,0},
 /*173(0xAD)*/ {0,0,84083968,0,0},
 /*174(0xAE)*/ {0,0,100861184,0,0},
 /*175(0xAF)*/ {0,0,117638400,0,0},
 /*176(0xB0)*/ {0,0,198144,0,0},
 /*177(0xB1)*/ {0,0,16975360,0,0},
 /*178(0xB2)*/ {0,0,33752576,0,0},
 /*179(0xB3)*/ {0,0,50529792,0,0},
 /*180(0xB4)*/ {0,0,67307008,0,0},
 /*181(0xB5)*/ {0,0,84084224,0,0},
 /*182(0xB6)*/ {0,0,100861440,0,0},
 /*183(0xB7)*/ {0,0,117638656,0,0},
 /*184(0xB8)*/ {0,0,198400,0,0},
 /*185(0xB9)*/ {0,0,16975616,0,0},
 /*186(0xBA)*/ {0,0,33752832,0,0},
 /*187(0xBB)*/ {0,0,50530048,0,0},
 /*188(0xBC)*/ {0,0,67307264,0,0},
 /*189(0xBD)*/ {0,0,84084480,0,0},
 /*190(0xBE)*/ {0,0,100861696,0,0},
 /*191(0xBF)*/ {0,0,117638912,0,0},
 /*192(0xC0)*/ {0,0,196608,0,0},
 /*193(0xC1)*/ {0,0,16973824,0,0},
 /*194(0xC2)*/ {0,0,33751040,0,0},
 /*195(0xC3)*/ {0,0,50528256,0,0},
 /*196(0xC4)*/ {0,0,67305472,0,0},
 /*197(0xC5)*/ {0,0,84082688,0,0},
 /*198(0xC6)*/ {0,0,100859904,0,0},
 /*199(0xC7)*/ {0,0,117637120,0,0},
 /*200(0xC8)*/ {0,0,196864,0,0},
 /*201(0xC9)*/ {0,0,16974080,0,0},
 /*202(0xCA)*/ {0,0,33751296,0,0},
 /*203(0xCB)*/ {0,0,50528512,0,0},
 /*204(0xCC)*/ {0,0,67305728,0,0},
 /*205(0xCD)*/ {0,0,84082944,0,0},
 /*206(0xCE)*/ {0,0,100860160,0,0},
 /*207(0xCF)*/ {0,0,117637376,0,0},
 /*208(0xD0)*/ {0,0,197120,0,0},
 /*209(0xD1)*/ {0,0,16974336,0,0},
 /*210(0xD2)*/ {0,0,33751552,0,0},
 /*211(0xD3)*/ {0,0,50528768,0,0},
 /*212(0xD4)*/ {0,0,67305984,0,0},
 /*213(0xD5)*/ {0,0,84083200,0,0},
 /*214(0xD6)*/ {0,0,100860416,0,0},
 /*215(0xD7)*/ {0,0,117637632,0,0},
 /*216(0xD8)*/ {0,0,197376,0,0},
 /*217(0xD9)*/ {0,0,16974592,0,0},
 /*218(0xDA)*/ {0,0,33751808,0,0},
 /*219(0xDB)*/ {0,0,50529024,0,0},
 /*220(0xDC)*/ {0,0,67306240,0,0},
 /*221(0xDD)*/ {0,0,84083456,0,0},
 /*222(0xDE)*/ {0,0,100860672,0,0},
 /*223(0xDF)*/ {0,0,117637888,0,0},
 /*224(0xE0)*/ {0,0,197632,0,0},
 /*225(0xE1)*/ {0,0,16974848,0,0},
 /*226(0xE2)*/ {0,0,33752064,0,0},
 /*227(0xE3)*/ {0,0,50529280,0,0},
 /*228(0xE4)*/ {0,0,67306496,0,0},
 /*229(0xE5)*/ {0,0,84083712,0,0},
 /*230(0xE6)*/ {0,0,100860928,0,0},
 /*231(0xE7)*/ {0,0,117638144,0,0},
 /*232(0xE8)*/ {0,0,197888,0,0},
 /*233(0xE9)*/ {0,0,16975104,0,0},
 /*234(0xEA)*/ {0,0,33752320,0,0},
 /*235(0xEB)*/ {0,0,50529536,0,0},
 /*236(0xEC)*/ {0,0,67306752,0,0},
 /*237(0xED)*/ {0,0,84083968,0,0},
 /*238(0xEE)*/ {0,0,100861184,0,0},
 /*239(0xEF)*/ {0,0,117638400,0,0},
 /*240(0xF0)*/ {0,0,198144,0,0},
 /*241(0xF1)*/ {0,0,16975360,0,0},
 /*242(0xF2)*/ {0,0,33752576,0,0},
 /*243(0xF3)*/ {0,0,50529792,0,0},
 /*244(0xF4)*/ {0,0,67307008,0,0},
 /*245(0xF5)*/ {0,0,84084224,0,0},
 /*246(0xF6)*/ {0,0,100861440,0,0},
 /*247(0xF7)*/ {0,0,117638656,0,0},
 /*248(0xF8)*/ {0,0,198400,0,0},
 /*249(0xF9)*/ {0,0,16975616,0,0},
 /*250(0xFA)*/ {0,0,33752832,0,0},
 /*251(0xFB)*/ {0,0,50530048,0,0},
 /*252(0xFC)*/ {0,0,67307264,0,0},
 /*253(0xFD)*/ {0,0,84084480,0,0},
 /*254(0xFE)*/ {0,0,100861696,0,0},
 /*255(0xFF)*/ {0,0,117638912,0,0}
};

//rejestry dr
op_t moddr[256]=
{
 /*0(0x00)*/ {0,0,262144,0,0},
 /*1(0x01)*/ {0,0,17039360,0,0},
 /*2(0x02)*/ {0,0,33816576,0,0},
 /*3(0x03)*/ {0,0,50593792,0,0},
 /*4(0x04)*/ {0,0,67371008,0,0},
 /*5(0x05)*/ {0,0,84148224,0,0},
 /*6(0x06)*/ {0,0,100925440,0,0},
 /*7(0x07)*/ {0,0,117702656,0,0},
 /*8(0x08)*/ {0,0,262400,0,0},
 /*9(0x09)*/ {0,0,17039616,0,0},
 /*10(0x0A)*/ {0,0,33816832,0,0},
 /*11(0x0B)*/ {0,0,50594048,0,0},
 /*12(0x0C)*/ {0,0,67371264,0,0},
 /*13(0x0D)*/ {0,0,84148480,0,0},
 /*14(0x0E)*/ {0,0,100925696,0,0},
 /*15(0x0F)*/ {0,0,117702912,0,0},
 /*16(0x10)*/ {0,0,262656,0,0},
 /*17(0x11)*/ {0,0,17039872,0,0},
 /*18(0x12)*/ {0,0,33817088,0,0},
 /*19(0x13)*/ {0,0,50594304,0,0},
 /*20(0x14)*/ {0,0,67371520,0,0},
 /*21(0x15)*/ {0,0,84148736,0,0},
 /*22(0x16)*/ {0,0,100925952,0,0},
 /*23(0x17)*/ {0,0,117703168,0,0},
 /*24(0x18)*/ {0,0,262912,0,0},
 /*25(0x19)*/ {0,0,17040128,0,0},
 /*26(0x1A)*/ {0,0,33817344,0,0},
 /*27(0x1B)*/ {0,0,50594560,0,0},
 /*28(0x1C)*/ {0,0,67371776,0,0},
 /*29(0x1D)*/ {0,0,84148992,0,0},
 /*30(0x1E)*/ {0,0,100926208,0,0},
 /*31(0x1F)*/ {0,0,117703424,0,0},
 /*32(0x20)*/ {0,0,263168,0,0},
 /*33(0x21)*/ {0,0,17040384,0,0},
 /*34(0x22)*/ {0,0,33817600,0,0},
 /*35(0x23)*/ {0,0,50594816,0,0},
 /*36(0x24)*/ {0,0,67372032,0,0},
 /*37(0x25)*/ {0,0,84149248,0,0},
 /*38(0x26)*/ {0,0,100926464,0,0},
 /*39(0x27)*/ {0,0,117703680,0,0},
 /*40(0x28)*/ {0,0,263424,0,0},
 /*41(0x29)*/ {0,0,17040640,0,0},
 /*42(0x2A)*/ {0,0,33817856,0,0},
 /*43(0x2B)*/ {0,0,50595072,0,0},
 /*44(0x2C)*/ {0,0,67372288,0,0},
 /*45(0x2D)*/ {0,0,84149504,0,0},
 /*46(0x2E)*/ {0,0,100926720,0,0},
 /*47(0x2F)*/ {0,0,117703936,0,0},
 /*48(0x30)*/ {0,0,263680,0,0},
 /*49(0x31)*/ {0,0,17040896,0,0},
 /*50(0x32)*/ {0,0,33818112,0,0},
 /*51(0x33)*/ {0,0,50595328,0,0},
 /*52(0x34)*/ {0,0,67372544,0,0},
 /*53(0x35)*/ {0,0,84149760,0,0},
 /*54(0x36)*/ {0,0,100926976,0,0},
 /*55(0x37)*/ {0,0,117704192,0,0},
 /*56(0x38)*/ {0,0,263936,0,0},
 /*57(0x39)*/ {0,0,17041152,0,0},
 /*58(0x3A)*/ {0,0,33818368,0,0},
 /*59(0x3B)*/ {0,0,50595584,0,0},
 /*60(0x3C)*/ {0,0,67372800,0,0},
 /*61(0x3D)*/ {0,0,84150016,0,0},
 /*62(0x3E)*/ {0,0,100927232,0,0},
 /*63(0x3F)*/ {0,0,117704448,0,0},
 /*64(0x40)*/ {0,0,262144,0,0},
 /*65(0x41)*/ {0,0,17039360,0,0},
 /*66(0x42)*/ {0,0,33816576,0,0},
 /*67(0x43)*/ {0,0,50593792,0,0},
 /*68(0x44)*/ {0,0,67371008,0,0},
 /*69(0x45)*/ {0,0,84148224,0,0},
 /*70(0x46)*/ {0,0,100925440,0,0},
 /*71(0x47)*/ {0,0,117702656,0,0},
 /*72(0x48)*/ {0,0,262400,0,0},
 /*73(0x49)*/ {0,0,17039616,0,0},
 /*74(0x4A)*/ {0,0,33816832,0,0},
 /*75(0x4B)*/ {0,0,50594048,0,0},
 /*76(0x4C)*/ {0,0,67371264,0,0},
 /*77(0x4D)*/ {0,0,84148480,0,0},
 /*78(0x4E)*/ {0,0,100925696,0,0},
 /*79(0x4F)*/ {0,0,117702912,0,0},
 /*80(0x50)*/ {0,0,262656,0,0},
 /*81(0x51)*/ {0,0,17039872,0,0},
 /*82(0x52)*/ {0,0,33817088,0,0},
 /*83(0x53)*/ {0,0,50594304,0,0},
 /*84(0x54)*/ {0,0,67371520,0,0},
 /*85(0x55)*/ {0,0,84148736,0,0},
 /*86(0x56)*/ {0,0,100925952,0,0},
 /*87(0x57)*/ {0,0,117703168,0,0},
 /*88(0x58)*/ {0,0,262912,0,0},
 /*89(0x59)*/ {0,0,17040128,0,0},
 /*90(0x5A)*/ {0,0,33817344,0,0},
 /*91(0x5B)*/ {0,0,50594560,0,0},
 /*92(0x5C)*/ {0,0,67371776,0,0},
 /*93(0x5D)*/ {0,0,84148992,0,0},
 /*94(0x5E)*/ {0,0,100926208,0,0},
 /*95(0x5F)*/ {0,0,117703424,0,0},
 /*96(0x60)*/ {0,0,263168,0,0},
 /*97(0x61)*/ {0,0,17040384,0,0},
 /*98(0x62)*/ {0,0,33817600,0,0},
 /*99(0x63)*/ {0,0,50594816,0,0},
 /*100(0x64)*/ {0,0,67372032,0,0},
 /*101(0x65)*/ {0,0,84149248,0,0},
 /*102(0x66)*/ {0,0,100926464,0,0},
 /*103(0x67)*/ {0,0,117703680,0,0},
 /*104(0x68)*/ {0,0,263424,0,0},
 /*105(0x69)*/ {0,0,17040640,0,0},
 /*106(0x6A)*/ {0,0,33817856,0,0},
 /*107(0x6B)*/ {0,0,50595072,0,0},
 /*108(0x6C)*/ {0,0,67372288,0,0},
 /*109(0x6D)*/ {0,0,84149504,0,0},
 /*110(0x6E)*/ {0,0,100926720,0,0},
 /*111(0x6F)*/ {0,0,117703936,0,0},
 /*112(0x70)*/ {0,0,263680,0,0},
 /*113(0x71)*/ {0,0,17040896,0,0},
 /*114(0x72)*/ {0,0,33818112,0,0},
 /*115(0x73)*/ {0,0,50595328,0,0},
 /*116(0x74)*/ {0,0,67372544,0,0},
 /*117(0x75)*/ {0,0,84149760,0,0},
 /*118(0x76)*/ {0,0,100926976,0,0},
 /*119(0x77)*/ {0,0,117704192,0,0},
 /*120(0x78)*/ {0,0,263936,0,0},
 /*121(0x79)*/ {0,0,17041152,0,0},
 /*122(0x7A)*/ {0,0,33818368,0,0},
 /*123(0x7B)*/ {0,0,50595584,0,0},
 /*124(0x7C)*/ {0,0,67372800,0,0},
 /*125(0x7D)*/ {0,0,84150016,0,0},
 /*126(0x7E)*/ {0,0,100927232,0,0},
 /*127(0x7F)*/ {0,0,117704448,0,0},
 /*128(0x80)*/ {0,0,262144,0,0},
 /*129(0x81)*/ {0,0,17039360,0,0},
 /*130(0x82)*/ {0,0,33816576,0,0},
 /*131(0x83)*/ {0,0,50593792,0,0},
 /*132(0x84)*/ {0,0,67371008,0,0},
 /*133(0x85)*/ {0,0,84148224,0,0},
 /*134(0x86)*/ {0,0,100925440,0,0},
 /*135(0x87)*/ {0,0,117702656,0,0},
 /*136(0x88)*/ {0,0,262400,0,0},
 /*137(0x89)*/ {0,0,17039616,0,0},
 /*138(0x8A)*/ {0,0,33816832,0,0},
 /*139(0x8B)*/ {0,0,50594048,0,0},
 /*140(0x8C)*/ {0,0,67371264,0,0},
 /*141(0x8D)*/ {0,0,84148480,0,0},
 /*142(0x8E)*/ {0,0,100925696,0,0},
 /*143(0x8F)*/ {0,0,117702912,0,0},
 /*144(0x90)*/ {0,0,262656,0,0},
 /*145(0x91)*/ {0,0,17039872,0,0},
 /*146(0x92)*/ {0,0,33817088,0,0},
 /*147(0x93)*/ {0,0,50594304,0,0},
 /*148(0x94)*/ {0,0,67371520,0,0},
 /*149(0x95)*/ {0,0,84148736,0,0},
 /*150(0x96)*/ {0,0,100925952,0,0},
 /*151(0x97)*/ {0,0,117703168,0,0},
 /*152(0x98)*/ {0,0,262912,0,0},
 /*153(0x99)*/ {0,0,17040128,0,0},
 /*154(0x9A)*/ {0,0,33817344,0,0},
 /*155(0x9B)*/ {0,0,50594560,0,0},
 /*156(0x9C)*/ {0,0,67371776,0,0},
 /*157(0x9D)*/ {0,0,84148992,0,0},
 /*158(0x9E)*/ {0,0,100926208,0,0},
 /*159(0x9F)*/ {0,0,117703424,0,0},
 /*160(0xA0)*/ {0,0,263168,0,0},
 /*161(0xA1)*/ {0,0,17040384,0,0},
 /*162(0xA2)*/ {0,0,33817600,0,0},
 /*163(0xA3)*/ {0,0,50594816,0,0},
 /*164(0xA4)*/ {0,0,67372032,0,0},
 /*165(0xA5)*/ {0,0,84149248,0,0},
 /*166(0xA6)*/ {0,0,100926464,0,0},
 /*167(0xA7)*/ {0,0,117703680,0,0},
 /*168(0xA8)*/ {0,0,263424,0,0},
 /*169(0xA9)*/ {0,0,17040640,0,0},
 /*170(0xAA)*/ {0,0,33817856,0,0},
 /*171(0xAB)*/ {0,0,50595072,0,0},
 /*172(0xAC)*/ {0,0,67372288,0,0},
 /*173(0xAD)*/ {0,0,84149504,0,0},
 /*174(0xAE)*/ {0,0,100926720,0,0},
 /*175(0xAF)*/ {0,0,117703936,0,0},
 /*176(0xB0)*/ {0,0,263680,0,0},
 /*177(0xB1)*/ {0,0,17040896,0,0},
 /*178(0xB2)*/ {0,0,33818112,0,0},
 /*179(0xB3)*/ {0,0,50595328,0,0},
 /*180(0xB4)*/ {0,0,67372544,0,0},
 /*181(0xB5)*/ {0,0,84149760,0,0},
 /*182(0xB6)*/ {0,0,100926976,0,0},
 /*183(0xB7)*/ {0,0,117704192,0,0},
 /*184(0xB8)*/ {0,0,263936,0,0},
 /*185(0xB9)*/ {0,0,17041152,0,0},
 /*186(0xBA)*/ {0,0,33818368,0,0},
 /*187(0xBB)*/ {0,0,50595584,0,0},
 /*188(0xBC)*/ {0,0,67372800,0,0},
 /*189(0xBD)*/ {0,0,84150016,0,0},
 /*190(0xBE)*/ {0,0,100927232,0,0},
 /*191(0xBF)*/ {0,0,117704448,0,0},
 /*192(0xC0)*/ {0,0,262144,0,0},
 /*193(0xC1)*/ {0,0,17039360,0,0},
 /*194(0xC2)*/ {0,0,33816576,0,0},
 /*195(0xC3)*/ {0,0,50593792,0,0},
 /*196(0xC4)*/ {0,0,67371008,0,0},
 /*197(0xC5)*/ {0,0,84148224,0,0},
 /*198(0xC6)*/ {0,0,100925440,0,0},
 /*199(0xC7)*/ {0,0,117702656,0,0},
 /*200(0xC8)*/ {0,0,262400,0,0},
 /*201(0xC9)*/ {0,0,17039616,0,0},
 /*202(0xCA)*/ {0,0,33816832,0,0},
 /*203(0xCB)*/ {0,0,50594048,0,0},
 /*204(0xCC)*/ {0,0,67371264,0,0},
 /*205(0xCD)*/ {0,0,84148480,0,0},
 /*206(0xCE)*/ {0,0,100925696,0,0},
 /*207(0xCF)*/ {0,0,117702912,0,0},
 /*208(0xD0)*/ {0,0,262656,0,0},
 /*209(0xD1)*/ {0,0,17039872,0,0},
 /*210(0xD2)*/ {0,0,33817088,0,0},
 /*211(0xD3)*/ {0,0,50594304,0,0},
 /*212(0xD4)*/ {0,0,67371520,0,0},
 /*213(0xD5)*/ {0,0,84148736,0,0},
 /*214(0xD6)*/ {0,0,100925952,0,0},
 /*215(0xD7)*/ {0,0,117703168,0,0},
 /*216(0xD8)*/ {0,0,262912,0,0},
 /*217(0xD9)*/ {0,0,17040128,0,0},
 /*218(0xDA)*/ {0,0,33817344,0,0},
 /*219(0xDB)*/ {0,0,50594560,0,0},
 /*220(0xDC)*/ {0,0,67371776,0,0},
 /*221(0xDD)*/ {0,0,84148992,0,0},
 /*222(0xDE)*/ {0,0,100926208,0,0},
 /*223(0xDF)*/ {0,0,117703424,0,0},
 /*224(0xE0)*/ {0,0,263168,0,0},
 /*225(0xE1)*/ {0,0,17040384,0,0},
 /*226(0xE2)*/ {0,0,33817600,0,0},
 /*227(0xE3)*/ {0,0,50594816,0,0},
 /*228(0xE4)*/ {0,0,67372032,0,0},
 /*229(0xE5)*/ {0,0,84149248,0,0},
 /*230(0xE6)*/ {0,0,100926464,0,0},
 /*231(0xE7)*/ {0,0,117703680,0,0},
 /*232(0xE8)*/ {0,0,263424,0,0},
 /*233(0xE9)*/ {0,0,17040640,0,0},
 /*234(0xEA)*/ {0,0,33817856,0,0},
 /*235(0xEB)*/ {0,0,50595072,0,0},
 /*236(0xEC)*/ {0,0,67372288,0,0},
 /*237(0xED)*/ {0,0,84149504,0,0},
 /*238(0xEE)*/ {0,0,100926720,0,0},
 /*239(0xEF)*/ {0,0,117703936,0,0},
 /*240(0xF0)*/ {0,0,263680,0,0},
 /*241(0xF1)*/ {0,0,17040896,0,0},
 /*242(0xF2)*/ {0,0,33818112,0,0},
 /*243(0xF3)*/ {0,0,50595328,0,0},
 /*244(0xF4)*/ {0,0,67372544,0,0},
 /*245(0xF5)*/ {0,0,84149760,0,0},
 /*246(0xF6)*/ {0,0,100926976,0,0},
 /*247(0xF7)*/ {0,0,117704192,0,0},
 /*248(0xF8)*/ {0,0,263936,0,0},
 /*249(0xF9)*/ {0,0,17041152,0,0},
 /*250(0xFA)*/ {0,0,33818368,0,0},
 /*251(0xFB)*/ {0,0,50595584,0,0},
 /*252(0xFC)*/ {0,0,67372800,0,0},
 /*253(0xFD)*/ {0,0,84150016,0,0},
 /*254(0xFE)*/ {0,0,100927232,0,0},
 /*255(0xFF)*/ {0,0,117704448,0,0}
};

//grupa 80h-81h
op_t g8081[256]=
{
 /*0(0x00)*/ {"add",0,16,ADD,0},
 /*1(0x01)*/ {"add",0,272,ADD,0},
 /*2(0x02)*/ {"add",0,528,ADD,0},
 /*3(0x03)*/ {"add",0,784,ADD,0},
 /*4(0x04)*/ {"add",0,1040,ADD,0},
 /*5(0x05)*/ {"add",0,1296,ADD,0},
 /*6(0x06)*/ {"add",0,1552,ADD,0},
 /*7(0x07)*/ {"add",0,1808,ADD,0},
 /*8(0x08)*/ {"or",0,16,OR,0},
 /*9(0x09)*/ {"or",0,272,OR,0},
 /*10(0x0A)*/ {"or",0,528,OR,0},
 /*11(0x0B)*/ {"or",0,784,OR,0},
 /*12(0x0C)*/ {"or",0,1040,OR,0},
 /*13(0x0D)*/ {"or",0,1296,OR,0},
 /*14(0x0E)*/ {"or",0,1552,OR,0},
 /*15(0x0F)*/ {"or",0,1808,OR,0},
 /*16(0x10)*/ {"adc",0,16,ADC,0},
 /*17(0x11)*/ {"adc",0,272,ADC,0},
 /*18(0x12)*/ {"adc",0,528,ADC,0},
 /*19(0x13)*/ {"adc",0,784,ADC,0},
 /*20(0x14)*/ {"adc",0,1040,ADC,0},
 /*21(0x15)*/ {"adc",0,1296,ADC,0},
 /*22(0x16)*/ {"adc",0,1552,ADC,0},
 /*23(0x17)*/ {"adc",0,1808,ADC,0},
 /*24(0x18)*/ {"sbb",0,16,SBB,0},
 /*25(0x19)*/ {"sbb",0,272,SBB,0},
 /*26(0x1A)*/ {"sbb",0,528,SBB,0},
 /*27(0x1B)*/ {"sbb",0,784,SBB,0},
 /*28(0x1C)*/ {"sbb",0,1040,SBB,0},
 /*29(0x1D)*/ {"sbb",0,1296,SBB,0},
 /*30(0x1E)*/ {"sbb",0,1552,SBB,0},
 /*31(0x1F)*/ {"sbb",0,1808,SBB,0},
 /*32(0x20)*/ {"and",0,16,AND,0},
 /*33(0x21)*/ {"and",0,272,AND,0},
 /*34(0x22)*/ {"and",0,528,AND,0},
 /*35(0x23)*/ {"and",0,784,AND,0},
 /*36(0x24)*/ {"and",0,1040,AND,0},
 /*37(0x25)*/ {"and",0,1296,AND,0},
 /*38(0x26)*/ {"and",0,1552,AND,0},
 /*39(0x27)*/ {"and",0,1808,AND,0},
 /*40(0x28)*/ {"sub",0,16,SUB,0},
 /*41(0x29)*/ {"sub",0,272,SUB,0},
 /*42(0x2A)*/ {"sub",0,528,SUB,0},
 /*43(0x2B)*/ {"sub",0,784,SUB,0},
 /*44(0x2C)*/ {"sub",0,1040,SUB,0},
 /*45(0x2D)*/ {"sub",0,1296,SUB,0},
 /*46(0x2E)*/ {"sub",0,1552,SUB,0},
 /*47(0x2F)*/ {"sub",0,1808,SUB,0},
 /*48(0x30)*/ {"xor",0,16,XOR,0},
 /*49(0x31)*/ {"xor",0,272,XOR,0},
 /*50(0x32)*/ {"xor",0,528,XOR,0},
 /*51(0x33)*/ {"xor",0,784,XOR,0},
 /*52(0x34)*/ {"xor",0,1040,XOR,0},
 /*53(0x35)*/ {"xor",0,1296,XOR,0},
 /*54(0x36)*/ {"xor",0,1552,XOR,0},
 /*55(0x37)*/ {"xor",0,1808,XOR,0},
 /*56(0x38)*/ {"cmp",0,16,CMP,0},
 /*57(0x39)*/ {"cmp",0,272,CMP,0},
 /*58(0x3A)*/ {"cmp",0,528,CMP,0},
 /*59(0x3B)*/ {"cmp",0,784,CMP,0},
 /*60(0x3C)*/ {"cmp",0,1040,CMP,0},
 /*61(0x3D)*/ {"cmp",0,1296,CMP,0},
 /*62(0x3E)*/ {"cmp",0,1552,CMP,0},
 /*63(0x3F)*/ {"cmp",0,1808,CMP,0},
 /*64(0x40)*/ {"add",0,17,ADD,0},
 /*65(0x41)*/ {"add",0,273,ADD,0},
 /*66(0x42)*/ {"add",0,529,ADD,0},
 /*67(0x43)*/ {"add",0,785,ADD,0},
 /*68(0x44)*/ {"add",0,1041,ADD,0},
 /*69(0x45)*/ {"add",0,1297,ADD,0},
 /*70(0x46)*/ {"add",0,1553,ADD,0},
 /*71(0x47)*/ {"add",0,1809,ADD,0},
 /*72(0x48)*/ {"or",0,17,OR,0},
 /*73(0x49)*/ {"or",0,273,OR,0},
 /*74(0x4A)*/ {"or",0,529,OR,0},
 /*75(0x4B)*/ {"or",0,785,OR,0},
 /*76(0x4C)*/ {"or",0,1041,OR,0},
 /*77(0x4D)*/ {"or",0,1297,OR,0},
 /*78(0x4E)*/ {"or",0,1553,OR,0},
 /*79(0x4F)*/ {"or",0,1809,OR,0},
 /*80(0x50)*/ {"adc",0,17,ADC,0},
 /*81(0x51)*/ {"adc",0,273,ADC,0},
 /*82(0x52)*/ {"adc",0,529,ADC,0},
 /*83(0x53)*/ {"adc",0,785,ADC,0},
 /*84(0x54)*/ {"adc",0,1041,ADC,0},
 /*85(0x55)*/ {"adc",0,1297,ADC,0},
 /*86(0x56)*/ {"adc",0,1553,ADC,0},
 /*87(0x57)*/ {"adc",0,1809,ADC,0},
 /*88(0x58)*/ {"sbb",0,17,SBB,0},
 /*89(0x59)*/ {"sbb",0,273,SBB,0},
 /*90(0x5A)*/ {"sbb",0,529,SBB,0},
 /*91(0x5B)*/ {"sbb",0,785,SBB,0},
 /*92(0x5C)*/ {"sbb",0,1041,SBB,0},
 /*93(0x5D)*/ {"sbb",0,1297,SBB,0},
 /*94(0x5E)*/ {"sbb",0,1553,SBB,0},
 /*95(0x5F)*/ {"sbb",0,1809,SBB,0},
 /*96(0x60)*/ {"and",0,17,AND,0},
 /*97(0x61)*/ {"and",0,273,AND,0},
 /*98(0x62)*/ {"and",0,529,AND,0},
 /*99(0x63)*/ {"and",0,785,AND,0},
 /*100(0x64)*/ {"and",0,1041,AND,0},
 /*101(0x65)*/ {"and",0,1297,AND,0},
 /*102(0x66)*/ {"and",0,1553,AND,0},
 /*103(0x67)*/ {"and",0,1809,AND,0},
 /*104(0x68)*/ {"sub",0,17,SUB,0},
 /*105(0x69)*/ {"sub",0,273,SUB,0},
 /*106(0x6A)*/ {"sub",0,529,SUB,0},
 /*107(0x6B)*/ {"sub",0,785,SUB,0},
 /*108(0x6C)*/ {"sub",0,1041,SUB,0},
 /*109(0x6D)*/ {"sub",0,1297,SUB,0},
 /*110(0x6E)*/ {"sub",0,1553,SUB,0},
 /*111(0x6F)*/ {"sub",0,1809,SUB,0},
 /*112(0x70)*/ {"xor",0,17,XOR,0},
 /*113(0x71)*/ {"xor",0,273,XOR,0},
 /*114(0x72)*/ {"xor",0,529,XOR,0},
 /*115(0x73)*/ {"xor",0,785,XOR,0},
 /*116(0x74)*/ {"xor",0,1041,XOR,0},
 /*117(0x75)*/ {"xor",0,1297,XOR,0},
 /*118(0x76)*/ {"xor",0,1553,XOR,0},
 /*119(0x77)*/ {"xor",0,1809,XOR,0},
 /*120(0x78)*/ {"cmp",0,17,CMP,0},
 /*121(0x79)*/ {"cmp",0,273,CMP,0},
 /*122(0x7A)*/ {"cmp",0,529,CMP,0},
 /*123(0x7B)*/ {"cmp",0,785,CMP,0},
 /*124(0x7C)*/ {"cmp",0,1041,CMP,0},
 /*125(0x7D)*/ {"cmp",0,1297,CMP,0},
 /*126(0x7E)*/ {"cmp",0,1553,CMP,0},
 /*127(0x7F)*/ {"cmp",0,1809,CMP,0},
 /*128(0x80)*/ {"add",0,18,ADD,0},
 /*129(0x81)*/ {"add",0,274,ADD,0},
 /*130(0x82)*/ {"add",0,530,ADD,0},
 /*131(0x83)*/ {"add",0,786,ADD,0},
 /*132(0x84)*/ {"add",0,1042,ADD,0},
 /*133(0x85)*/ {"add",0,1298,ADD,0},
 /*134(0x86)*/ {"add",0,1554,ADD,0},
 /*135(0x87)*/ {"add",0,1810,ADD,0},
 /*136(0x88)*/ {"or",0,18,OR,0},
 /*137(0x89)*/ {"or",0,274,OR,0},
 /*138(0x8A)*/ {"or",0,530,OR,0},
 /*139(0x8B)*/ {"or",0,786,OR,0},
 /*140(0x8C)*/ {"or",0,1042,OR,0},
 /*141(0x8D)*/ {"or",0,1298,OR,0},
 /*142(0x8E)*/ {"or",0,1554,OR,0},
 /*143(0x8F)*/ {"or",0,1810,OR,0},
 /*144(0x90)*/ {"adc",0,18,ADC,0},
 /*145(0x91)*/ {"adc",0,274,ADC,0},
 /*146(0x92)*/ {"adc",0,530,ADC,0},
 /*147(0x93)*/ {"adc",0,786,ADC,0},
 /*148(0x94)*/ {"adc",0,1042,ADC,0},
 /*149(0x95)*/ {"adc",0,1298,ADC,0},
 /*150(0x96)*/ {"adc",0,1554,ADC,0},
 /*151(0x97)*/ {"adc",0,1810,ADC,0},
 /*152(0x98)*/ {"sbb",0,18,SBB,0},
 /*153(0x99)*/ {"sbb",0,274,SBB,0},
 /*154(0x9A)*/ {"sbb",0,530,SBB,0},
 /*155(0x9B)*/ {"sbb",0,786,SBB,0},
 /*156(0x9C)*/ {"sbb",0,1042,SBB,0},
 /*157(0x9D)*/ {"sbb",0,1298,SBB,0},
 /*158(0x9E)*/ {"sbb",0,1554,SBB,0},
 /*159(0x9F)*/ {"sbb",0,1810,SBB,0},
 /*160(0xA0)*/ {"and",0,18,AND,0},
 /*161(0xA1)*/ {"and",0,274,AND,0},
 /*162(0xA2)*/ {"and",0,530,AND,0},
 /*163(0xA3)*/ {"and",0,786,AND,0},
 /*164(0xA4)*/ {"and",0,1042,AND,0},
 /*165(0xA5)*/ {"and",0,1298,AND,0},
 /*166(0xA6)*/ {"and",0,1554,AND,0},
 /*167(0xA7)*/ {"and",0,1810,AND,0},
 /*168(0xA8)*/ {"sub",0,18,SUB,0},
 /*169(0xA9)*/ {"sub",0,274,SUB,0},
 /*170(0xAA)*/ {"sub",0,530,SUB,0},
 /*171(0xAB)*/ {"sub",0,786,SUB,0},
 /*172(0xAC)*/ {"sub",0,1042,SUB,0},
 /*173(0xAD)*/ {"sub",0,1298,SUB,0},
 /*174(0xAE)*/ {"sub",0,1554,SUB,0},
 /*175(0xAF)*/ {"sub",0,1810,SUB,0},
 /*176(0xB0)*/ {"xor",0,18,XOR,0},
 /*177(0xB1)*/ {"xor",0,274,XOR,0},
 /*178(0xB2)*/ {"xor",0,530,XOR,0},
 /*179(0xB3)*/ {"xor",0,786,XOR,0},
 /*180(0xB4)*/ {"xor",0,1042,XOR,0},
 /*181(0xB5)*/ {"xor",0,1298,XOR,0},
 /*182(0xB6)*/ {"xor",0,1554,XOR,0},
 /*183(0xB7)*/ {"xor",0,1810,XOR,0},
 /*184(0xB8)*/ {"cmp",0,18,CMP,0},
 /*185(0xB9)*/ {"cmp",0,274,CMP,0},
 /*186(0xBA)*/ {"cmp",0,530,CMP,0},
 /*187(0xBB)*/ {"cmp",0,786,CMP,0},
 /*188(0xBC)*/ {"cmp",0,1042,CMP,0},
 /*189(0xBD)*/ {"cmp",0,1298,CMP,0},
 /*190(0xBE)*/ {"cmp",0,1554,CMP,0},
 /*191(0xBF)*/ {"cmp",0,1810,CMP,0},
 /*192(0xC0)*/ {"add",0,0,ADD,0},
 /*193(0xC1)*/ {"add",0,256,ADD,0},
 /*194(0xC2)*/ {"add",0,512,ADD,0},
 /*195(0xC3)*/ {"add",0,768,ADD,0},
 /*196(0xC4)*/ {"add",0,1024,ADD,0},
 /*197(0xC5)*/ {"add",0,1280,ADD,0},
 /*198(0xC6)*/ {"add",0,1536,ADD,0},
 /*199(0xC7)*/ {"add",0,1792,ADD,0},
 /*200(0xC8)*/ {"or",0,0,OR,0},
 /*201(0xC9)*/ {"or",0,256,OR,0},
 /*202(0xCA)*/ {"or",0,512,OR,0},
 /*203(0xCB)*/ {"or",0,768,OR,0},
 /*204(0xCC)*/ {"or",0,1024,OR,0},
 /*205(0xCD)*/ {"or",0,1280,OR,0},
 /*206(0xCE)*/ {"or",0,1536,OR,0},
 /*207(0xCF)*/ {"or",0,1792,OR,0},
 /*208(0xD0)*/ {"adc",0,0,ADC,0},
 /*209(0xD1)*/ {"adc",0,256,ADC,0},
 /*210(0xD2)*/ {"adc",0,512,ADC,0},
 /*211(0xD3)*/ {"adc",0,768,ADC,0},
 /*212(0xD4)*/ {"adc",0,1024,ADC,0},
 /*213(0xD5)*/ {"adc",0,1280,ADC,0},
 /*214(0xD6)*/ {"adc",0,1536,ADC,0},
 /*215(0xD7)*/ {"adc",0,1792,ADC,0},
 /*216(0xD8)*/ {"sbb",0,0,SBB,0},
 /*217(0xD9)*/ {"sbb",0,256,SBB,0},
 /*218(0xDA)*/ {"sbb",0,512,SBB,0},
 /*219(0xDB)*/ {"sbb",0,768,SBB,0},
 /*220(0xDC)*/ {"sbb",0,1024,SBB,0},
 /*221(0xDD)*/ {"sbb",0,1280,SBB,0},
 /*222(0xDE)*/ {"sbb",0,1536,SBB,0},
 /*223(0xDF)*/ {"sbb",0,1792,SBB,0},
 /*224(0xE0)*/ {"and",0,0,AND,0},
 /*225(0xE1)*/ {"and",0,256,AND,0},
 /*226(0xE2)*/ {"and",0,512,AND,0},
 /*227(0xE3)*/ {"and",0,768,AND,0},
 /*228(0xE4)*/ {"and",0,1024,AND,0},
 /*229(0xE5)*/ {"and",0,1280,AND,0},
 /*230(0xE6)*/ {"and",0,1536,AND,0},
 /*231(0xE7)*/ {"and",0,1792,AND,0},
 /*232(0xE8)*/ {"sub",0,0,SUB,0},
 /*233(0xE9)*/ {"sub",0,256,SUB,0},
 /*234(0xEA)*/ {"sub",0,512,SUB,0},
 /*235(0xEB)*/ {"sub",0,768,SUB,0},
 /*236(0xEC)*/ {"sub",0,1024,SUB,0},
 /*237(0xED)*/ {"sub",0,1280,SUB,0},
 /*238(0xEE)*/ {"sub",0,1536,SUB,0},
 /*239(0xEF)*/ {"sub",0,1792,SUB,0},
 /*240(0xF0)*/ {"xor",0,0,XOR,0},
 /*241(0xF1)*/ {"xor",0,256,XOR,0},
 /*242(0xF2)*/ {"xor",0,512,XOR,0},
 /*243(0xF3)*/ {"xor",0,768,XOR,0},
 /*244(0xF4)*/ {"xor",0,1024,XOR,0},
 /*245(0xF5)*/ {"xor",0,1280,XOR,0},
 /*246(0xF6)*/ {"xor",0,1536,XOR,0},
 /*247(0xF7)*/ {"xor",0,1792,XOR,0},
 /*248(0xF8)*/ {"cmp",0,0,CMP,0},
 /*249(0xF9)*/ {"cmp",0,256,CMP,0},
 /*250(0xFA)*/ {"cmp",0,512,CMP,0},
 /*251(0xFB)*/ {"cmp",0,768,CMP,0},
 /*252(0xFC)*/ {"cmp",0,1024,CMP,0},
 /*253(0xFD)*/ {"cmp",0,1280,CMP,0},
 /*254(0xFE)*/ {"cmp",0,1536,CMP,0},
 /*255(0xFF)*/ {"cmp",0,1792,CMP,0}
};

//grupa 82h-83h
//funkcje typu add word,byte
op_t g8283[256]=
{
 /*0(0x00)*/ {"add",0,16,ADDX,0},
 /*1(0x01)*/ {"add",0,272,ADDX,0},
 /*2(0x02)*/ {"add",0,528,ADDX,0},
 /*3(0x03)*/ {"add",0,784,ADDX,0},
 /*4(0x04)*/ {"add",0,1040,ADDX,0},
 /*5(0x05)*/ {"add",0,1296,ADDX,0},
 /*6(0x06)*/ {"add",0,1552,ADDX,0},
 /*7(0x07)*/ {"add",0,1808,ADDX,0},
 /*8(0x08)*/ {"or",0,16,ORX,0},
 /*9(0x09)*/ {"or",0,272,ORX,0},
 /*10(0x0A)*/ {"or",0,528,ORX,0},
 /*11(0x0B)*/ {"or",0,784,ORX,0},
 /*12(0x0C)*/ {"or",0,1040,ORX,0},
 /*13(0x0D)*/ {"or",0,1296,ORX,0},
 /*14(0x0E)*/ {"or",0,1552,ORX,0},
 /*15(0x0F)*/ {"or",0,1808,ORX,0},
 /*16(0x10)*/ {"adc",0,16,ADCX,0},
 /*17(0x11)*/ {"adc",0,272,ADCX,0},
 /*18(0x12)*/ {"adc",0,528,ADCX,0},
 /*19(0x13)*/ {"adc",0,784,ADCX,0},
 /*20(0x14)*/ {"adc",0,1040,ADCX,0},
 /*21(0x15)*/ {"adc",0,1296,ADCX,0},
 /*22(0x16)*/ {"adc",0,1552,ADCX,0},
 /*23(0x17)*/ {"adc",0,1808,ADCX,0},
 /*24(0x18)*/ {"sbb",0,16,SBBX,0},
 /*25(0x19)*/ {"sbb",0,272,SBBX,0},
 /*26(0x1A)*/ {"sbb",0,528,SBBX,0},
 /*27(0x1B)*/ {"sbb",0,784,SBBX,0},
 /*28(0x1C)*/ {"sbb",0,1040,SBBX,0},
 /*29(0x1D)*/ {"sbb",0,1296,SBBX,0},
 /*30(0x1E)*/ {"sbb",0,1552,SBBX,0},
 /*31(0x1F)*/ {"sbb",0,1808,SBBX,0},
 /*32(0x20)*/ {"and",0,16,ANDX,0},
 /*33(0x21)*/ {"and",0,272,ANDX,0},
 /*34(0x22)*/ {"and",0,528,ANDX,0},
 /*35(0x23)*/ {"and",0,784,ANDX,0},
 /*36(0x24)*/ {"and",0,1040,ANDX,0},
 /*37(0x25)*/ {"and",0,1296,ANDX,0},
 /*38(0x26)*/ {"and",0,1552,ANDX,0},
 /*39(0x27)*/ {"and",0,1808,ANDX,0},
 /*40(0x28)*/ {"sub",0,16,SUBX,0},
 /*41(0x29)*/ {"sub",0,272,SUBX,0},
 /*42(0x2A)*/ {"sub",0,528,SUBX,0},
 /*43(0x2B)*/ {"sub",0,784,SUBX,0},
 /*44(0x2C)*/ {"sub",0,1040,SUBX,0},
 /*45(0x2D)*/ {"sub",0,1296,SUBX,0},
 /*46(0x2E)*/ {"sub",0,1552,SUBX,0},
 /*47(0x2F)*/ {"sub",0,1808,SUBX,0},
 /*48(0x30)*/ {"xor",0,16,XORX,0},
 /*49(0x31)*/ {"xor",0,272,XORX,0},
 /*50(0x32)*/ {"xor",0,528,XORX,0},
 /*51(0x33)*/ {"xor",0,784,XORX,0},
 /*52(0x34)*/ {"xor",0,1040,XORX,0},
 /*53(0x35)*/ {"xor",0,1296,XORX,0},
 /*54(0x36)*/ {"xor",0,1552,XORX,0},
 /*55(0x37)*/ {"xor",0,1808,XORX,0},
 /*56(0x38)*/ {"cmp",0,16,CMPX,0},
 /*57(0x39)*/ {"cmp",0,272,CMPX,0},
 /*58(0x3A)*/ {"cmp",0,528,CMPX,0},
 /*59(0x3B)*/ {"cmp",0,784,CMPX,0},
 /*60(0x3C)*/ {"cmp",0,1040,CMPX,0},
 /*61(0x3D)*/ {"cmp",0,1296,CMPX,0},
 /*62(0x3E)*/ {"cmp",0,1552,CMPX,0},
 /*63(0x3F)*/ {"cmp",0,1808,CMPX,0},
 /*64(0x40)*/ {"add",0,17,ADDX,0},
 /*65(0x41)*/ {"add",0,273,ADDX,0},
 /*66(0x42)*/ {"add",0,529,ADDX,0},
 /*67(0x43)*/ {"add",0,785,ADDX,0},
 /*68(0x44)*/ {"add",0,1041,ADDX,0},
 /*69(0x45)*/ {"add",0,1297,ADDX,0},
 /*70(0x46)*/ {"add",0,1553,ADDX,0},
 /*71(0x47)*/ {"add",0,1809,ADDX,0},
 /*72(0x48)*/ {"or",0,17,ORX,0},
 /*73(0x49)*/ {"or",0,273,ORX,0},
 /*74(0x4A)*/ {"or",0,529,ORX,0},
 /*75(0x4B)*/ {"or",0,785,ORX,0},
 /*76(0x4C)*/ {"or",0,1041,ORX,0},
 /*77(0x4D)*/ {"or",0,1297,ORX,0},
 /*78(0x4E)*/ {"or",0,1553,ORX,0},
 /*79(0x4F)*/ {"or",0,1809,ORX,0},
 /*80(0x50)*/ {"adc",0,17,ADCX,0},
 /*81(0x51)*/ {"adc",0,273,ADCX,0},
 /*82(0x52)*/ {"adc",0,529,ADCX,0},
 /*83(0x53)*/ {"adc",0,785,ADCX,0},
 /*84(0x54)*/ {"adc",0,1041,ADCX,0},
 /*85(0x55)*/ {"adc",0,1297,ADCX,0},
 /*86(0x56)*/ {"adc",0,1553,ADCX,0},
 /*87(0x57)*/ {"adc",0,1809,ADCX,0},
 /*88(0x58)*/ {"sbb",0,17,SBBX,0},
 /*89(0x59)*/ {"sbb",0,273,SBBX,0},
 /*90(0x5A)*/ {"sbb",0,529,SBBX,0},
 /*91(0x5B)*/ {"sbb",0,785,SBBX,0},
 /*92(0x5C)*/ {"sbb",0,1041,SBBX,0},
 /*93(0x5D)*/ {"sbb",0,1297,SBBX,0},
 /*94(0x5E)*/ {"sbb",0,1553,SBBX,0},
 /*95(0x5F)*/ {"sbb",0,1809,SBBX,0},
 /*96(0x60)*/ {"and",0,17,ANDX,0},
 /*97(0x61)*/ {"and",0,273,ANDX,0},
 /*98(0x62)*/ {"and",0,529,ANDX,0},
 /*99(0x63)*/ {"and",0,785,ANDX,0},
 /*100(0x64)*/ {"and",0,1041,ANDX,0},
 /*101(0x65)*/ {"and",0,1297,ANDX,0},
 /*102(0x66)*/ {"and",0,1553,ANDX,0},
 /*103(0x67)*/ {"and",0,1809,ANDX,0},
 /*104(0x68)*/ {"sub",0,17,SUBX,0},
 /*105(0x69)*/ {"sub",0,273,SUBX,0},
 /*106(0x6A)*/ {"sub",0,529,SUBX,0},
 /*107(0x6B)*/ {"sub",0,785,SUBX,0},
 /*108(0x6C)*/ {"sub",0,1041,SUBX,0},
 /*109(0x6D)*/ {"sub",0,1297,SUBX,0},
 /*110(0x6E)*/ {"sub",0,1553,SUBX,0},
 /*111(0x6F)*/ {"sub",0,1809,SUBX,0},
 /*112(0x70)*/ {"xor",0,17,XORX,0},
 /*113(0x71)*/ {"xor",0,273,XORX,0},
 /*114(0x72)*/ {"xor",0,529,XORX,0},
 /*115(0x73)*/ {"xor",0,785,XORX,0},
 /*116(0x74)*/ {"xor",0,1041,XORX,0},
 /*117(0x75)*/ {"xor",0,1297,XORX,0},
 /*118(0x76)*/ {"xor",0,1553,XORX,0},
 /*119(0x77)*/ {"xor",0,1809,XORX,0},
 /*120(0x78)*/ {"cmp",0,17,CMPX,0},
 /*121(0x79)*/ {"cmp",0,273,CMPX,0},
 /*122(0x7A)*/ {"cmp",0,529,CMPX,0},
 /*123(0x7B)*/ {"cmp",0,785,CMPX,0},
 /*124(0x7C)*/ {"cmp",0,1041,CMPX,0},
 /*125(0x7D)*/ {"cmp",0,1297,CMPX,0},
 /*126(0x7E)*/ {"cmp",0,1553,CMPX,0},
 /*127(0x7F)*/ {"cmp",0,1809,CMPX,0},
 /*128(0x80)*/ {"add",0,18,ADDX,0},
 /*129(0x81)*/ {"add",0,274,ADDX,0},
 /*130(0x82)*/ {"add",0,530,ADDX,0},
 /*131(0x83)*/ {"add",0,786,ADDX,0},
 /*132(0x84)*/ {"add",0,1042,ADDX,0},
 /*133(0x85)*/ {"add",0,1298,ADDX,0},
 /*134(0x86)*/ {"add",0,1554,ADDX,0},
 /*135(0x87)*/ {"add",0,1810,ADDX,0},
 /*136(0x88)*/ {"or",0,18,ORX,0},
 /*137(0x89)*/ {"or",0,274,ORX,0},
 /*138(0x8A)*/ {"or",0,530,ORX,0},
 /*139(0x8B)*/ {"or",0,786,ORX,0},
 /*140(0x8C)*/ {"or",0,1042,ORX,0},
 /*141(0x8D)*/ {"or",0,1298,ORX,0},
 /*142(0x8E)*/ {"or",0,1554,ORX,0},
 /*143(0x8F)*/ {"or",0,1810,ORX,0},
 /*144(0x90)*/ {"adc",0,18,ADCX,0},
 /*145(0x91)*/ {"adc",0,274,ADCX,0},
 /*146(0x92)*/ {"adc",0,530,ADCX,0},
 /*147(0x93)*/ {"adc",0,786,ADCX,0},
 /*148(0x94)*/ {"adc",0,1042,ADCX,0},
 /*149(0x95)*/ {"adc",0,1298,ADCX,0},
 /*150(0x96)*/ {"adc",0,1554,ADCX,0},
 /*151(0x97)*/ {"adc",0,1810,ADCX,0},
 /*152(0x98)*/ {"sbb",0,18,SBBX,0},
 /*153(0x99)*/ {"sbb",0,274,SBBX,0},
 /*154(0x9A)*/ {"sbb",0,530,SBBX,0},
 /*155(0x9B)*/ {"sbb",0,786,SBBX,0},
 /*156(0x9C)*/ {"sbb",0,1042,SBBX,0},
 /*157(0x9D)*/ {"sbb",0,1298,SBBX,0},
 /*158(0x9E)*/ {"sbb",0,1554,SBBX,0},
 /*159(0x9F)*/ {"sbb",0,1810,SBBX,0},
 /*160(0xA0)*/ {"and",0,18,ANDX,0},
 /*161(0xA1)*/ {"and",0,274,ANDX,0},
 /*162(0xA2)*/ {"and",0,530,ANDX,0},
 /*163(0xA3)*/ {"and",0,786,ANDX,0},
 /*164(0xA4)*/ {"and",0,1042,ANDX,0},
 /*165(0xA5)*/ {"and",0,1298,ANDX,0},
 /*166(0xA6)*/ {"and",0,1554,ANDX,0},
 /*167(0xA7)*/ {"and",0,1810,ANDX,0},
 /*168(0xA8)*/ {"sub",0,18,SUBX,0},
 /*169(0xA9)*/ {"sub",0,274,SUBX,0},
 /*170(0xAA)*/ {"sub",0,530,SUBX,0},
 /*171(0xAB)*/ {"sub",0,786,SUBX,0},
 /*172(0xAC)*/ {"sub",0,1042,SUBX,0},
 /*173(0xAD)*/ {"sub",0,1298,SUBX,0},
 /*174(0xAE)*/ {"sub",0,1554,SUBX,0},
 /*175(0xAF)*/ {"sub",0,1810,SUBX,0},
 /*176(0xB0)*/ {"xor",0,18,XORX,0},
 /*177(0xB1)*/ {"xor",0,274,XORX,0},
 /*178(0xB2)*/ {"xor",0,530,XORX,0},
 /*179(0xB3)*/ {"xor",0,786,XORX,0},
 /*180(0xB4)*/ {"xor",0,1042,XORX,0},
 /*181(0xB5)*/ {"xor",0,1298,XORX,0},
 /*182(0xB6)*/ {"xor",0,1554,XORX,0},
 /*183(0xB7)*/ {"xor",0,1810,XORX,0},
 /*184(0xB8)*/ {"cmp",0,18,CMPX,0},
 /*185(0xB9)*/ {"cmp",0,274,CMPX,0},
 /*186(0xBA)*/ {"cmp",0,530,CMPX,0},
 /*187(0xBB)*/ {"cmp",0,786,CMPX,0},
 /*188(0xBC)*/ {"cmp",0,1042,CMPX,0},
 /*189(0xBD)*/ {"cmp",0,1298,CMPX,0},
 /*190(0xBE)*/ {"cmp",0,1554,CMPX,0},
 /*191(0xBF)*/ {"cmp",0,1810,CMPX,0},
 /*192(0xC0)*/ {"add",0,0,ADDX,0},
 /*193(0xC1)*/ {"add",0,256,ADDX,0},
 /*194(0xC2)*/ {"add",0,512,ADDX,0},
 /*195(0xC3)*/ {"add",0,768,ADDX,0},
 /*196(0xC4)*/ {"add",0,1024,ADDX,0},
 /*197(0xC5)*/ {"add",0,1280,ADDX,0},
 /*198(0xC6)*/ {"add",0,1536,ADDX,0},
 /*199(0xC7)*/ {"add",0,1792,ADDX,0},
 /*200(0xC8)*/ {"or",0,0,ORX,0},
 /*201(0xC9)*/ {"or",0,256,ORX,0},
 /*202(0xCA)*/ {"or",0,512,ORX,0},
 /*203(0xCB)*/ {"or",0,768,ORX,0},
 /*204(0xCC)*/ {"or",0,1024,ORX,0},
 /*205(0xCD)*/ {"or",0,1280,ORX,0},
 /*206(0xCE)*/ {"or",0,1536,ORX,0},
 /*207(0xCF)*/ {"or",0,1792,ORX,0},
 /*208(0xD0)*/ {"adc",0,0,ADCX,0},
 /*209(0xD1)*/ {"adc",0,256,ADCX,0},
 /*210(0xD2)*/ {"adc",0,512,ADCX,0},
 /*211(0xD3)*/ {"adc",0,768,ADCX,0},
 /*212(0xD4)*/ {"adc",0,1024,ADCX,0},
 /*213(0xD5)*/ {"adc",0,1280,ADCX,0},
 /*214(0xD6)*/ {"adc",0,1536,ADCX,0},
 /*215(0xD7)*/ {"adc",0,1792,ADCX,0},
 /*216(0xD8)*/ {"sbb",0,0,SBBX,0},
 /*217(0xD9)*/ {"sbb",0,256,SBBX,0},
 /*218(0xDA)*/ {"sbb",0,512,SBBX,0},
 /*219(0xDB)*/ {"sbb",0,768,SBBX,0},
 /*220(0xDC)*/ {"sbb",0,1024,SBBX,0},
 /*221(0xDD)*/ {"sbb",0,1280,SBBX,0},
 /*222(0xDE)*/ {"sbb",0,1536,SBBX,0},
 /*223(0xDF)*/ {"sbb",0,1792,SBBX,0},
 /*224(0xE0)*/ {"and",0,0,ANDX,0},
 /*225(0xE1)*/ {"and",0,256,ANDX,0},
 /*226(0xE2)*/ {"and",0,512,ANDX,0},
 /*227(0xE3)*/ {"and",0,768,ANDX,0},
 /*228(0xE4)*/ {"and",0,1024,ANDX,0},
 /*229(0xE5)*/ {"and",0,1280,ANDX,0},
 /*230(0xE6)*/ {"and",0,1536,ANDX,0},
 /*231(0xE7)*/ {"and",0,1792,ANDX,0},
 /*232(0xE8)*/ {"sub",0,0,SUBX,0},
 /*233(0xE9)*/ {"sub",0,256,SUBX,0},
 /*234(0xEA)*/ {"sub",0,512,SUBX,0},
 /*235(0xEB)*/ {"sub",0,768,SUBX,0},
 /*236(0xEC)*/ {"sub",0,1024,SUBX,0},
 /*237(0xED)*/ {"sub",0,1280,SUBX,0},
 /*238(0xEE)*/ {"sub",0,1536,SUBX,0},
 /*239(0xEF)*/ {"sub",0,1792,SUBX,0},
 /*240(0xF0)*/ {"xor",0,0,XORX,0},
 /*241(0xF1)*/ {"xor",0,256,XORX,0},
 /*242(0xF2)*/ {"xor",0,512,XORX,0},
 /*243(0xF3)*/ {"xor",0,768,XORX,0},
 /*244(0xF4)*/ {"xor",0,1024,XORX,0},
 /*245(0xF5)*/ {"xor",0,1280,XORX,0},
 /*246(0xF6)*/ {"xor",0,1536,XORX,0},
 /*247(0xF7)*/ {"xor",0,1792,XORX,0},
 /*248(0xF8)*/ {"cmp",0,0,CMPX,0},
 /*249(0xF9)*/ {"cmp",0,256,CMPX,0},
 /*250(0xFA)*/ {"cmp",0,512,CMPX,0},
 /*251(0xFB)*/ {"cmp",0,768,CMPX,0},
 /*252(0xFC)*/ {"cmp",0,1024,CMPX,0},
 /*253(0xFD)*/ {"cmp",0,1280,CMPX,0},
 /*254(0xFE)*/ {"cmp",0,1536,CMPX,0},
 /*255(0xFF)*/ {"cmp",0,1792,CMPX,0}
};

//grupa F6h-F7h
//grupa instrukcji jedno-argumentowych (oprocz innych wersji test,imul,idiv)
op_t gF6F7[256]=
{
 /*0(0x00)*/ {"test",F_IMM,16,TEST,0},
 /*1(0x01)*/ {"test",F_IMM,272,TEST,0},
 /*2(0x02)*/ {"test",F_IMM,528,TEST,0},
 /*3(0x03)*/ {"test",F_IMM,784,TEST,0},
 /*4(0x04)*/ {"test",F_IMM,1040,TEST,0},
 /*5(0x05)*/ {"test",F_IMM,1296,TEST,0},
 /*6(0x06)*/ {"test",F_IMM,1552,TEST,0},
 /*7(0x07)*/ {"test",F_IMM,1808,TEST,0},
 /*8(0x08)*/ {0,0,16,0,0},
 /*9(0x09)*/ {0,0,272,0,0},
 /*10(0x0A)*/ {0,0,528,0,0},
 /*11(0x0B)*/ {0,0,784,0,0},
 /*12(0x0C)*/ {0,0,1040,0,0},
 /*13(0x0D)*/ {0,0,1296,0,0},
 /*14(0x0E)*/ {0,0,1552,0,0},
 /*15(0x0F)*/ {0,0,1808,0,0},
 /*16(0x10)*/ {"not",0,16,NOT,0},
 /*17(0x11)*/ {"not",0,272,NOT,0},
 /*18(0x12)*/ {"not",0,528,NOT,0},
 /*19(0x13)*/ {"not",0,784,NOT,0},
 /*20(0x14)*/ {"not",0,1040,NOT,0},
 /*21(0x15)*/ {"not",0,1296,NOT,0},
 /*22(0x16)*/ {"not",0,1552,NOT,0},
 /*23(0x17)*/ {"not",0,1808,NOT,0},
 /*24(0x18)*/ {"neg",0,16,NEG,0},
 /*25(0x19)*/ {"neg",0,272,NEG,0},
 /*26(0x1A)*/ {"neg",0,528,NEG,0},
 /*27(0x1B)*/ {"neg",0,784,NEG,0},
 /*28(0x1C)*/ {"neg",0,1040,NEG,0},
 /*29(0x1D)*/ {"neg",0,1296,NEG,0},
 /*30(0x1E)*/ {"neg",0,1552,NEG,0},
 /*31(0x1F)*/ {"neg",0,1808,NEG,0},
 /*32(0x20)*/ {"mul",0,16,MUL,0},
 /*33(0x21)*/ {"mul",0,272,MUL,0},
 /*34(0x22)*/ {"mul",0,528,MUL,0},
 /*35(0x23)*/ {"mul",0,784,MUL,0},
 /*36(0x24)*/ {"mul",0,1040,MUL,0},
 /*37(0x25)*/ {"mul",0,1296,MUL,0},
 /*38(0x26)*/ {"mul",0,1552,MUL,0},
 /*39(0x27)*/ {"mul",0,1808,MUL,0},
 /*40(0x28)*/ {"imul",0,16,IMUL,0},
 /*41(0x29)*/ {"imul",0,272,IMUL,0},
 /*42(0x2A)*/ {"imul",0,528,IMUL,0},
 /*43(0x2B)*/ {"imul",0,784,IMUL,0},
 /*44(0x2C)*/ {"imul",0,1040,IMUL,0},
 /*45(0x2D)*/ {"imul",0,1296,IMUL,0},
 /*46(0x2E)*/ {"imul",0,1552,IMUL,0},
 /*47(0x2F)*/ {"imul",0,1808,IMUL,0},
 /*48(0x30)*/ {"div",0,16,DIV,0},
 /*49(0x31)*/ {"div",0,272,DIV,0},
 /*50(0x32)*/ {"div",0,528,DIV,0},
 /*51(0x33)*/ {"div",0,784,DIV,0},
 /*52(0x34)*/ {"div",0,1040,DIV,0},
 /*53(0x35)*/ {"div",0,1296,DIV,0},
 /*54(0x36)*/ {"div",0,1552,DIV,0},
 /*55(0x37)*/ {"div",0,1808,DIV,0},
 /*56(0x38)*/ {"idiv",0,16,IDIV,0},
 /*57(0x39)*/ {"idiv",0,272,IDIV,0},
 /*58(0x3A)*/ {"idiv",0,528,IDIV,0},
 /*59(0x3B)*/ {"idiv",0,784,IDIV,0},
 /*60(0x3C)*/ {"idiv",0,1040,IDIV,0},
 /*61(0x3D)*/ {"idiv",0,1296,IDIV,0},
 /*62(0x3E)*/ {"idiv",0,1552,IDIV,0},
 /*63(0x3F)*/ {"idiv",0,1808,IDIV,0},
 /*64(0x40)*/ {"test",F_IMM,17,TEST,0},
 /*65(0x41)*/ {"test",F_IMM,273,TEST,0},
 /*66(0x42)*/ {"test",F_IMM,529,TEST,0},
 /*67(0x43)*/ {"test",F_IMM,785,TEST,0},
 /*68(0x44)*/ {"test",F_IMM,1041,TEST,0},
 /*69(0x45)*/ {"test",F_IMM,1297,TEST,0},
 /*70(0x46)*/ {"test",F_IMM,1553,TEST,0},
 /*71(0x47)*/ {"test",F_IMM,1809,TEST,0},
 /*72(0x48)*/ {0,0,17,0,0},
 /*73(0x49)*/ {0,0,273,0,0},
 /*74(0x4A)*/ {0,0,529,0,0},
 /*75(0x4B)*/ {0,0,785,0,0},
 /*76(0x4C)*/ {0,0,1041,0,0},
 /*77(0x4D)*/ {0,0,1297,0,0},
 /*78(0x4E)*/ {0,0,1553,0,0},
 /*79(0x4F)*/ {0,0,1809,0,0},
 /*80(0x50)*/ {"not",0,17,NOT,0},
 /*81(0x51)*/ {"not",0,273,NOT,0},
 /*82(0x52)*/ {"not",0,529,NOT,0},
 /*83(0x53)*/ {"not",0,785,NOT,0},
 /*84(0x54)*/ {"not",0,1041,NOT,0},
 /*85(0x55)*/ {"not",0,1297,NOT,0},
 /*86(0x56)*/ {"not",0,1553,NOT,0},
 /*87(0x57)*/ {"not",0,1809,NOT,0},
 /*88(0x58)*/ {"neg",0,17,NEG,0},
 /*89(0x59)*/ {"neg",0,273,NEG,0},
 /*90(0x5A)*/ {"neg",0,529,NEG,0},
 /*91(0x5B)*/ {"neg",0,785,NEG,0},
 /*92(0x5C)*/ {"neg",0,1041,NEG,0},
 /*93(0x5D)*/ {"neg",0,1297,NEG,0},
 /*94(0x5E)*/ {"neg",0,1553,NEG,0},
 /*95(0x5F)*/ {"neg",0,1809,NEG,0},
 /*96(0x60)*/ {"mul",0,17,MUL,0},
 /*97(0x61)*/ {"mul",0,273,MUL,0},
 /*98(0x62)*/ {"mul",0,529,MUL,0},
 /*99(0x63)*/ {"mul",0,785,MUL,0},
 /*100(0x64)*/ {"mul",0,1041,MUL,0},
 /*101(0x65)*/ {"mul",0,1297,MUL,0},
 /*102(0x66)*/ {"mul",0,1553,MUL,0},
 /*103(0x67)*/ {"mul",0,1809,MUL,0},
 /*104(0x68)*/ {"imul",0,17,IMUL,0},
 /*105(0x69)*/ {"imul",0,273,IMUL,0},
 /*106(0x6A)*/ {"imul",0,529,IMUL,0},
 /*107(0x6B)*/ {"imul",0,785,IMUL,0},
 /*108(0x6C)*/ {"imul",0,1041,IMUL,0},
 /*109(0x6D)*/ {"imul",0,1297,IMUL,0},
 /*110(0x6E)*/ {"imul",0,1553,IMUL,0},
 /*111(0x6F)*/ {"imul",0,1809,IMUL,0},
 /*112(0x70)*/ {"div",0,17,DIV,0},
 /*113(0x71)*/ {"div",0,273,DIV,0},
 /*114(0x72)*/ {"div",0,529,DIV,0},
 /*115(0x73)*/ {"div",0,785,DIV,0},
 /*116(0x74)*/ {"div",0,1041,DIV,0},
 /*117(0x75)*/ {"div",0,1297,DIV,0},
 /*118(0x76)*/ {"div",0,1553,DIV,0},
 /*119(0x77)*/ {"div",0,1809,DIV,0},
 /*120(0x78)*/ {"idiv",0,17,IDIV,0},
 /*121(0x79)*/ {"idiv",0,273,IDIV,0},
 /*122(0x7A)*/ {"idiv",0,529,IDIV,0},
 /*123(0x7B)*/ {"idiv",0,785,IDIV,0},
 /*124(0x7C)*/ {"idiv",0,1041,IDIV,0},
 /*125(0x7D)*/ {"idiv",0,1297,IDIV,0},
 /*126(0x7E)*/ {"idiv",0,1553,IDIV,0},
 /*127(0x7F)*/ {"idiv",0,1809,IDIV,0},
 /*128(0x80)*/ {"test",F_IMM,18,TEST,0},
 /*129(0x81)*/ {"test",F_IMM,274,TEST,0},
 /*130(0x82)*/ {"test",F_IMM,530,TEST,0},
 /*131(0x83)*/ {"test",F_IMM,786,TEST,0},
 /*132(0x84)*/ {"test",F_IMM,1042,TEST,0},
 /*133(0x85)*/ {"test",F_IMM,1298,TEST,0},
 /*134(0x86)*/ {"test",F_IMM,1554,TEST,0},
 /*135(0x87)*/ {"test",F_IMM,1810,TEST,0},
 /*136(0x88)*/ {0,0,18,0,0},
 /*137(0x89)*/ {0,0,274,0,0},
 /*138(0x8A)*/ {0,0,530,0,0},
 /*139(0x8B)*/ {0,0,786,0,0},
 /*140(0x8C)*/ {0,0,1042,0,0},
 /*141(0x8D)*/ {0,0,1298,0,0},
 /*142(0x8E)*/ {0,0,1554,0,0},
 /*143(0x8F)*/ {0,0,1810,0,0},
 /*144(0x90)*/ {"not",0,18,NOT,0},
 /*145(0x91)*/ {"not",0,274,NOT,0},
 /*146(0x92)*/ {"not",0,530,NOT,0},
 /*147(0x93)*/ {"not",0,786,NOT,0},
 /*148(0x94)*/ {"not",0,1042,NOT,0},
 /*149(0x95)*/ {"not",0,1298,NOT,0},
 /*150(0x96)*/ {"not",0,1554,NOT,0},
 /*151(0x97)*/ {"not",0,1810,NOT,0},
 /*152(0x98)*/ {"neg",0,18,NEG,0},
 /*153(0x99)*/ {"neg",0,274,NEG,0},
 /*154(0x9A)*/ {"neg",0,530,NEG,0},
 /*155(0x9B)*/ {"neg",0,786,NEG,0},
 /*156(0x9C)*/ {"neg",0,1042,NEG,0},
 /*157(0x9D)*/ {"neg",0,1298,NEG,0},
 /*158(0x9E)*/ {"neg",0,1554,NEG,0},
 /*159(0x9F)*/ {"neg",0,1810,NEG,0},
 /*160(0xA0)*/ {"mul",0,18,MUL,0},
 /*161(0xA1)*/ {"mul",0,274,MUL,0},
 /*162(0xA2)*/ {"mul",0,530,MUL,0},
 /*163(0xA3)*/ {"mul",0,786,MUL,0},
 /*164(0xA4)*/ {"mul",0,1042,MUL,0},
 /*165(0xA5)*/ {"mul",0,1298,MUL,0},
 /*166(0xA6)*/ {"mul",0,1554,MUL,0},
 /*167(0xA7)*/ {"mul",0,1810,MUL,0},
 /*168(0xA8)*/ {"imul",0,18,IMUL,0},
 /*169(0xA9)*/ {"imul",0,274,IMUL,0},
 /*170(0xAA)*/ {"imul",0,530,IMUL,0},
 /*171(0xAB)*/ {"imul",0,786,IMUL,0},
 /*172(0xAC)*/ {"imul",0,1042,IMUL,0},
 /*173(0xAD)*/ {"imul",0,1298,IMUL,0},
 /*174(0xAE)*/ {"imul",0,1554,IMUL,0},
 /*175(0xAF)*/ {"imul",0,1810,IMUL,0},
 /*176(0xB0)*/ {"div",0,18,DIV,0},
 /*177(0xB1)*/ {"div",0,274,DIV,0},
 /*178(0xB2)*/ {"div",0,530,DIV,0},
 /*179(0xB3)*/ {"div",0,786,DIV,0},
 /*180(0xB4)*/ {"div",0,1042,DIV,0},
 /*181(0xB5)*/ {"div",0,1298,DIV,0},
 /*182(0xB6)*/ {"div",0,1554,DIV,0},
 /*183(0xB7)*/ {"div",0,1810,DIV,0},
 /*184(0xB8)*/ {"idiv",0,18,IDIV,0},
 /*185(0xB9)*/ {"idiv",0,274,IDIV,0},
 /*186(0xBA)*/ {"idiv",0,530,IDIV,0},
 /*187(0xBB)*/ {"idiv",0,786,IDIV,0},
 /*188(0xBC)*/ {"idiv",0,1042,IDIV,0},
 /*189(0xBD)*/ {"idiv",0,1298,IDIV,0},
 /*190(0xBE)*/ {"idiv",0,1554,IDIV,0},
 /*191(0xBF)*/ {"idiv",0,1810,IDIV,0},
 /*192(0xC0)*/ {"test",F_IMM,0,TEST,0},
 /*193(0xC1)*/ {"test",F_IMM,256,TEST,0},
 /*194(0xC2)*/ {"test",F_IMM,512,TEST,0},
 /*195(0xC3)*/ {"test",F_IMM,768,TEST,0},
 /*196(0xC4)*/ {"test",F_IMM,1024,TEST,0},
 /*197(0xC5)*/ {"test",F_IMM,1280,TEST,0},
 /*198(0xC6)*/ {"test",F_IMM,1536,TEST,0},
 /*199(0xC7)*/ {"test",F_IMM,1792,TEST,0},
 /*200(0xC8)*/ {0,0,0,0,0},
 /*201(0xC9)*/ {0,0,256,0,0},
 /*202(0xCA)*/ {0,0,512,0,0},
 /*203(0xCB)*/ {0,0,768,0,0},
 /*204(0xCC)*/ {0,0,1024,0,0},
 /*205(0xCD)*/ {0,0,1280,0,0},
 /*206(0xCE)*/ {0,0,1536,0,0},
 /*207(0xCF)*/ {0,0,1792,0,0},
 /*208(0xD0)*/ {"not",0,0,NOT,0},
 /*209(0xD1)*/ {"not",0,256,NOT,0},
 /*210(0xD2)*/ {"not",0,512,NOT,0},
 /*211(0xD3)*/ {"not",0,768,NOT,0},
 /*212(0xD4)*/ {"not",0,1024,NOT,0},
 /*213(0xD5)*/ {"not",0,1280,NOT,0},
 /*214(0xD6)*/ {"not",0,1536,NOT,0},
 /*215(0xD7)*/ {"not",0,1792,NOT,0},
 /*216(0xD8)*/ {"neg",0,0,NEG,0},
 /*217(0xD9)*/ {"neg",0,256,NEG,0},
 /*218(0xDA)*/ {"neg",0,512,NEG,0},
 /*219(0xDB)*/ {"neg",0,768,NEG,0},
 /*220(0xDC)*/ {"neg",0,1024,NEG,0},
 /*221(0xDD)*/ {"neg",0,1280,NEG,0},
 /*222(0xDE)*/ {"neg",0,1536,NEG,0},
 /*223(0xDF)*/ {"neg",0,1792,NEG,0},
 /*224(0xE0)*/ {"mul",0,0,MUL,0},
 /*225(0xE1)*/ {"mul",0,256,MUL,0},
 /*226(0xE2)*/ {"mul",0,512,MUL,0},
 /*227(0xE3)*/ {"mul",0,768,MUL,0},
 /*228(0xE4)*/ {"mul",0,1024,MUL,0},
 /*229(0xE5)*/ {"mul",0,1280,MUL,0},
 /*230(0xE6)*/ {"mul",0,1536,MUL,0},
 /*231(0xE7)*/ {"mul",0,1792,MUL,0},
 /*232(0xE8)*/ {"imul",0,0,IMUL,0},
 /*233(0xE9)*/ {"imul",0,256,IMUL,0},
 /*234(0xEA)*/ {"imul",0,512,IMUL,0},
 /*235(0xEB)*/ {"imul",0,768,IMUL,0},
 /*236(0xEC)*/ {"imul",0,1024,IMUL,0},
 /*237(0xED)*/ {"imul",0,1280,IMUL,0},
 /*238(0xEE)*/ {"imul",0,1536,IMUL,0},
 /*239(0xEF)*/ {"imul",0,1792,IMUL,0},
 /*240(0xF0)*/ {"div",0,0,DIV,0},
 /*241(0xF1)*/ {"div",0,256,DIV,0},
 /*242(0xF2)*/ {"div",0,512,DIV,0},
 /*243(0xF3)*/ {"div",0,768,DIV,0},
 /*244(0xF4)*/ {"div",0,1024,DIV,0},
 /*245(0xF5)*/ {"div",0,1280,DIV,0},
 /*246(0xF6)*/ {"div",0,1536,DIV,0},
 /*247(0xF7)*/ {"div",0,1792,DIV,0},
 /*248(0xF8)*/ {"idiv",0,0,IDIV,0},
 /*249(0xF9)*/ {"idiv",0,256,IDIV,0},
 /*250(0xFA)*/ {"idiv",0,512,IDIV,0},
 /*251(0xFB)*/ {"idiv",0,768,IDIV,0},
 /*252(0xFC)*/ {"idiv",0,1024,IDIV,0},
 /*253(0xFD)*/ {"idiv",0,1280,IDIV,0},
 /*254(0xFE)*/ {"idiv",0,1536,IDIV,0},
 /*255(0xFF)*/ {"idiv",0,1792,IDIV,0}
};

//grupa FEh
//tylko inc rmb i dec rmb
op_t gFE[256]=
{
 /*0(0x00)*/ {"inc",0,16,INC,0},
 /*1(0x01)*/ {"inc",0,272,INC,0},
 /*2(0x02)*/ {"inc",0,528,INC,0},
 /*3(0x03)*/ {"inc",0,784,INC,0},
 /*4(0x04)*/ {"inc",0,1040,INC,0},
 /*5(0x05)*/ {"inc",0,1296,INC,0},
 /*6(0x06)*/ {"inc",0,1552,INC,0},
 /*7(0x07)*/ {"inc",0,1808,INC,0},
 /*8(0x08)*/ {"dec",0,16,DEC,0},
 /*9(0x09)*/ {"dec",0,272,DEC,0},
 /*10(0x0A)*/ {"dec",0,528,DEC,0},
 /*11(0x0B)*/ {"dec",0,784,DEC,0},
 /*12(0x0C)*/ {"dec",0,1040,DEC,0},
 /*13(0x0D)*/ {"dec",0,1296,DEC,0},
 /*14(0x0E)*/ {"dec",0,1552,DEC,0},
 /*15(0x0F)*/ {"dec",0,1808,DEC,0},
 /*16(0x10)*/ {0,0,16,0,0},
 /*17(0x11)*/ {0,0,272,0,0},
 /*18(0x12)*/ {0,0,528,0,0},
 /*19(0x13)*/ {0,0,784,0,0},
 /*20(0x14)*/ {0,0,1040,0,0},
 /*21(0x15)*/ {0,0,1296,0,0},
 /*22(0x16)*/ {0,0,1552,0,0},
 /*23(0x17)*/ {0,0,1808,0,0},
 /*24(0x18)*/ {0,0,16,0,0},
 /*25(0x19)*/ {0,0,272,0,0},
 /*26(0x1A)*/ {0,0,528,0,0},
 /*27(0x1B)*/ {0,0,784,0,0},
 /*28(0x1C)*/ {0,0,1040,0,0},
 /*29(0x1D)*/ {0,0,1296,0,0},
 /*30(0x1E)*/ {0,0,1552,0,0},
 /*31(0x1F)*/ {0,0,1808,0,0},
 /*32(0x20)*/ {0,0,16,0,0},
 /*33(0x21)*/ {0,0,272,0,0},
 /*34(0x22)*/ {0,0,528,0,0},
 /*35(0x23)*/ {0,0,784,0,0},
 /*36(0x24)*/ {0,0,1040,0,0},
 /*37(0x25)*/ {0,0,1296,0,0},
 /*38(0x26)*/ {0,0,1552,0,0},
 /*39(0x27)*/ {0,0,1808,0,0},
 /*40(0x28)*/ {0,0,16,0,0},
 /*41(0x29)*/ {0,0,272,0,0},
 /*42(0x2A)*/ {0,0,528,0,0},
 /*43(0x2B)*/ {0,0,784,0,0},
 /*44(0x2C)*/ {0,0,1040,0,0},
 /*45(0x2D)*/ {0,0,1296,0,0},
 /*46(0x2E)*/ {0,0,1552,0,0},
 /*47(0x2F)*/ {0,0,1808,0,0},
 /*48(0x30)*/ {0,0,16,0,0},
 /*49(0x31)*/ {0,0,272,0,0},
 /*50(0x32)*/ {0,0,528,0,0},
 /*51(0x33)*/ {0,0,784,0,0},
 /*52(0x34)*/ {0,0,1040,0,0},
 /*53(0x35)*/ {0,0,1296,0,0},
 /*54(0x36)*/ {0,0,1552,0,0},
 /*55(0x37)*/ {0,0,1808,0,0},
 /*56(0x38)*/ {0,0,16,0,0},
 /*57(0x39)*/ {0,0,272,0,0},
 /*58(0x3A)*/ {0,0,528,0,0},
 /*59(0x3B)*/ {0,0,784,0,0},
 /*60(0x3C)*/ {0,0,1040,0,0},
 /*61(0x3D)*/ {0,0,1296,0,0},
 /*62(0x3E)*/ {0,0,1552,0,0},
 /*63(0x3F)*/ {0,0,1808,0,0},
 /*64(0x40)*/ {"inc",0,17,INC,0},
 /*65(0x41)*/ {"inc",0,273,INC,0},
 /*66(0x42)*/ {"inc",0,529,INC,0},
 /*67(0x43)*/ {"inc",0,785,INC,0},
 /*68(0x44)*/ {"inc",0,1041,INC,0},
 /*69(0x45)*/ {"inc",0,1297,INC,0},
 /*70(0x46)*/ {"inc",0,1553,INC,0},
 /*71(0x47)*/ {"inc",0,1809,INC,0},
 /*72(0x48)*/ {"dec",0,17,DEC,0},
 /*73(0x49)*/ {"dec",0,273,DEC,0},
 /*74(0x4A)*/ {"dec",0,529,DEC,0},
 /*75(0x4B)*/ {"dec",0,785,DEC,0},
 /*76(0x4C)*/ {"dec",0,1041,DEC,0},
 /*77(0x4D)*/ {"dec",0,1297,DEC,0},
 /*78(0x4E)*/ {"dec",0,1553,DEC,0},
 /*79(0x4F)*/ {"dec",0,1809,DEC,0},
 /*80(0x50)*/ {0,0,17,0,0},
 /*81(0x51)*/ {0,0,273,0,0},
 /*82(0x52)*/ {0,0,529,0,0},
 /*83(0x53)*/ {0,0,785,0,0},
 /*84(0x54)*/ {0,0,1041,0,0},
 /*85(0x55)*/ {0,0,1297,0,0},
 /*86(0x56)*/ {0,0,1553,0,0},
 /*87(0x57)*/ {0,0,1809,0,0},
 /*88(0x58)*/ {0,0,17,0,0},
 /*89(0x59)*/ {0,0,273,0,0},
 /*90(0x5A)*/ {0,0,529,0,0},
 /*91(0x5B)*/ {0,0,785,0,0},
 /*92(0x5C)*/ {0,0,1041,0,0},
 /*93(0x5D)*/ {0,0,1297,0,0},
 /*94(0x5E)*/ {0,0,1553,0,0},
 /*95(0x5F)*/ {0,0,1809,0,0},
 /*96(0x60)*/ {0,0,17,0,0},
 /*97(0x61)*/ {0,0,273,0,0},
 /*98(0x62)*/ {0,0,529,0,0},
 /*99(0x63)*/ {0,0,785,0,0},
 /*100(0x64)*/ {0,0,1041,0,0},
 /*101(0x65)*/ {0,0,1297,0,0},
 /*102(0x66)*/ {0,0,1553,0,0},
 /*103(0x67)*/ {0,0,1809,0,0},
 /*104(0x68)*/ {0,0,17,0,0},
 /*105(0x69)*/ {0,0,273,0,0},
 /*106(0x6A)*/ {0,0,529,0,0},
 /*107(0x6B)*/ {0,0,785,0,0},
 /*108(0x6C)*/ {0,0,1041,0,0},
 /*109(0x6D)*/ {0,0,1297,0,0},
 /*110(0x6E)*/ {0,0,1553,0,0},
 /*111(0x6F)*/ {0,0,1809,0,0},
 /*112(0x70)*/ {0,0,17,0,0},
 /*113(0x71)*/ {0,0,273,0,0},
 /*114(0x72)*/ {0,0,529,0,0},
 /*115(0x73)*/ {0,0,785,0,0},
 /*116(0x74)*/ {0,0,1041,0,0},
 /*117(0x75)*/ {0,0,1297,0,0},
 /*118(0x76)*/ {0,0,1553,0,0},
 /*119(0x77)*/ {0,0,1809,0,0},
 /*120(0x78)*/ {0,0,17,0,0},
 /*121(0x79)*/ {0,0,273,0,0},
 /*122(0x7A)*/ {0,0,529,0,0},
 /*123(0x7B)*/ {0,0,785,0,0},
 /*124(0x7C)*/ {0,0,1041,0,0},
 /*125(0x7D)*/ {0,0,1297,0,0},
 /*126(0x7E)*/ {0,0,1553,0,0},
 /*127(0x7F)*/ {0,0,1809,0,0},
 /*128(0x80)*/ {"inc",0,18,INC,0},
 /*129(0x81)*/ {"inc",0,274,INC,0},
 /*130(0x82)*/ {"inc",0,530,INC,0},
 /*131(0x83)*/ {"inc",0,786,INC,0},
 /*132(0x84)*/ {"inc",0,1042,INC,0},
 /*133(0x85)*/ {"inc",0,1298,INC,0},
 /*134(0x86)*/ {"inc",0,1554,INC,0},
 /*135(0x87)*/ {"inc",0,1810,INC,0},
 /*136(0x88)*/ {"dec",0,18,DEC,0},
 /*137(0x89)*/ {"dec",0,274,DEC,0},
 /*138(0x8A)*/ {"dec",0,530,DEC,0},
 /*139(0x8B)*/ {"dec",0,786,DEC,0},
 /*140(0x8C)*/ {"dec",0,1042,DEC,0},
 /*141(0x8D)*/ {"dec",0,1298,DEC,0},
 /*142(0x8E)*/ {"dec",0,1554,DEC,0},
 /*143(0x8F)*/ {"dec",0,1810,DEC,0},
 /*144(0x90)*/ {0,0,18,0,0},
 /*145(0x91)*/ {0,0,274,0,0},
 /*146(0x92)*/ {0,0,530,0,0},
 /*147(0x93)*/ {0,0,786,0,0},
 /*148(0x94)*/ {0,0,1042,0,0},
 /*149(0x95)*/ {0,0,1298,0,0},
 /*150(0x96)*/ {0,0,1554,0,0},
 /*151(0x97)*/ {0,0,1810,0,0},
 /*152(0x98)*/ {0,0,18,0,0},
 /*153(0x99)*/ {0,0,274,0,0},
 /*154(0x9A)*/ {0,0,530,0,0},
 /*155(0x9B)*/ {0,0,786,0,0},
 /*156(0x9C)*/ {0,0,1042,0,0},
 /*157(0x9D)*/ {0,0,1298,0,0},
 /*158(0x9E)*/ {0,0,1554,0,0},
 /*159(0x9F)*/ {0,0,1810,0,0},
 /*160(0xA0)*/ {0,0,18,0,0},
 /*161(0xA1)*/ {0,0,274,0,0},
 /*162(0xA2)*/ {0,0,530,0,0},
 /*163(0xA3)*/ {0,0,786,0,0},
 /*164(0xA4)*/ {0,0,1042,0,0},
 /*165(0xA5)*/ {0,0,1298,0,0},
 /*166(0xA6)*/ {0,0,1554,0,0},
 /*167(0xA7)*/ {0,0,1810,0,0},
 /*168(0xA8)*/ {0,0,18,0,0},
 /*169(0xA9)*/ {0,0,274,0,0},
 /*170(0xAA)*/ {0,0,530,0,0},
 /*171(0xAB)*/ {0,0,786,0,0},
 /*172(0xAC)*/ {0,0,1042,0,0},
 /*173(0xAD)*/ {0,0,1298,0,0},
 /*174(0xAE)*/ {0,0,1554,0,0},
 /*175(0xAF)*/ {0,0,1810,0,0},
 /*176(0xB0)*/ {0,0,18,0,0},
 /*177(0xB1)*/ {0,0,274,0,0},
 /*178(0xB2)*/ {0,0,530,0,0},
 /*179(0xB3)*/ {0,0,786,0,0},
 /*180(0xB4)*/ {0,0,1042,0,0},
 /*181(0xB5)*/ {0,0,1298,0,0},
 /*182(0xB6)*/ {0,0,1554,0,0},
 /*183(0xB7)*/ {0,0,1810,0,0},
 /*184(0xB8)*/ {0,0,18,0,0},
 /*185(0xB9)*/ {0,0,274,0,0},
 /*186(0xBA)*/ {0,0,530,0,0},
 /*187(0xBB)*/ {0,0,786,0,0},
 /*188(0xBC)*/ {0,0,1042,0,0},
 /*189(0xBD)*/ {0,0,1298,0,0},
 /*190(0xBE)*/ {0,0,1554,0,0},
 /*191(0xBF)*/ {0,0,1810,0,0},
 /*192(0xC0)*/ {"inc",0,0,INC,0},
 /*193(0xC1)*/ {"inc",0,256,INC,0},
 /*194(0xC2)*/ {"inc",0,512,INC,0},
 /*195(0xC3)*/ {"inc",0,768,INC,0},
 /*196(0xC4)*/ {"inc",0,1024,INC,0},
 /*197(0xC5)*/ {"inc",0,1280,INC,0},
 /*198(0xC6)*/ {"inc",0,1536,INC,0},
 /*199(0xC7)*/ {"inc",0,1792,INC,0},
 /*200(0xC8)*/ {"dec",0,0,DEC,0},
 /*201(0xC9)*/ {"dec",0,256,DEC,0},
 /*202(0xCA)*/ {"dec",0,512,DEC,0},
 /*203(0xCB)*/ {"dec",0,768,DEC,0},
 /*204(0xCC)*/ {"dec",0,1024,DEC,0},
 /*205(0xCD)*/ {"dec",0,1280,DEC,0},
 /*206(0xCE)*/ {"dec",0,1536,DEC,0},
 /*207(0xCF)*/ {"dec",0,1792,DEC,0},
 /*208(0xD0)*/ {0,0,0,0,0},
 /*209(0xD1)*/ {0,0,256,0,0},
 /*210(0xD2)*/ {0,0,512,0,0},
 /*211(0xD3)*/ {0,0,768,0,0},
 /*212(0xD4)*/ {0,0,1024,0,0},
 /*213(0xD5)*/ {0,0,1280,0,0},
 /*214(0xD6)*/ {0,0,1536,0,0},
 /*215(0xD7)*/ {0,0,1792,0,0},
 /*216(0xD8)*/ {0,0,0,0,0},
 /*217(0xD9)*/ {0,0,256,0,0},
 /*218(0xDA)*/ {0,0,512,0,0},
 /*219(0xDB)*/ {0,0,768,0,0},
 /*220(0xDC)*/ {0,0,1024,0,0},
 /*221(0xDD)*/ {0,0,1280,0,0},
 /*222(0xDE)*/ {0,0,1536,0,0},
 /*223(0xDF)*/ {0,0,1792,0,0},
 /*224(0xE0)*/ {0,0,0,0,0},
 /*225(0xE1)*/ {0,0,256,0,0},
 /*226(0xE2)*/ {0,0,512,0,0},
 /*227(0xE3)*/ {0,0,768,0,0},
 /*228(0xE4)*/ {0,0,1024,0,0},
 /*229(0xE5)*/ {0,0,1280,0,0},
 /*230(0xE6)*/ {0,0,1536,0,0},
 /*231(0xE7)*/ {0,0,1792,0,0},
 /*232(0xE8)*/ {0,0,0,0,0},
 /*233(0xE9)*/ {0,0,256,0,0},
 /*234(0xEA)*/ {0,0,512,0,0},
 /*235(0xEB)*/ {0,0,768,0,0},
 /*236(0xEC)*/ {0,0,1024,0,0},
 /*237(0xED)*/ {0,0,1280,0,0},
 /*238(0xEE)*/ {0,0,1536,0,0},
 /*239(0xEF)*/ {0,0,1792,0,0},
 /*240(0xF0)*/ {0,0,0,0,0},
 /*241(0xF1)*/ {0,0,256,0,0},
 /*242(0xF2)*/ {0,0,512,0,0},
 /*243(0xF3)*/ {0,0,768,0,0},
 /*244(0xF4)*/ {0,0,1024,0,0},
 /*245(0xF5)*/ {0,0,1280,0,0},
 /*246(0xF6)*/ {0,0,1536,0,0},
 /*247(0xF7)*/ {0,0,1792,0,0},
 /*248(0xF8)*/ {0,0,0,0,0},
 /*249(0xF9)*/ {0,0,256,0,0},
 /*250(0xFA)*/ {0,0,512,0,0},
 /*251(0xFB)*/ {0,0,768,0,0},
 /*252(0xFC)*/ {0,0,1024,0,0},
 /*253(0xFD)*/ {0,0,1280,0,0},
 /*254(0xFE)*/ {0,0,1536,0,0},
 /*255(0xFF)*/ {0,0,1792,0,0}
};

//grupa FFh
//instrukcje inc/dec/push oraz skokow
op_t gFF[256]=
{
 /*0(0x00)*/ {"inc",0,16,INC,0},
 /*1(0x01)*/ {"inc",0,272,INC,0},
 /*2(0x02)*/ {"inc",0,528,INC,0},
 /*3(0x03)*/ {"inc",0,784,INC,0},
 /*4(0x04)*/ {"inc",0,1040,INC,0},
 /*5(0x05)*/ {"inc",0,1296,INC,0},
 /*6(0x06)*/ {"inc",0,1552,INC,0},
 /*7(0x07)*/ {"inc",0,1808,INC,0},
 /*8(0x08)*/ {"dec",0,16,DEC,0},
 /*9(0x09)*/ {"dec",0,272,DEC,0},
 /*10(0x0A)*/ {"dec",0,528,DEC,0},
 /*11(0x0B)*/ {"dec",0,784,DEC,0},
 /*12(0x0C)*/ {"dec",0,1040,DEC,0},
 /*13(0x0D)*/ {"dec",0,1296,DEC,0},
 /*14(0x0E)*/ {"dec",0,1552,DEC,0},
 /*15(0x0F)*/ {"dec",0,1808,DEC,0},
 /*16(0x10)*/ {"call",0,16,CALLN,0},
 /*17(0x11)*/ {"call",0,272,CALLN,0},
 /*18(0x12)*/ {"call",0,528,CALLN,0},
 /*19(0x13)*/ {"call",0,784,CALLN,0},
 /*20(0x14)*/ {"call",0,1040,CALLN,0},
 /*21(0x15)*/ {"call",0,1296,CALLN,0},
 /*22(0x16)*/ {"call",0,1552,CALLN,0},
 /*23(0x17)*/ {"call",0,1808,CALLN,0},
 /*24(0x18)*/ {"call",F_FAR,16,CALLF,0},
 /*25(0x19)*/ {"call",F_FAR,272,CALLF,0},
 /*26(0x1A)*/ {"call",F_FAR,528,CALLF,0},
 /*27(0x1B)*/ {"call",F_FAR,784,CALLF,0},
 /*28(0x1C)*/ {"call",F_FAR,1040,CALLF,0},
 /*29(0x1D)*/ {"call",F_FAR,1296,CALLF,0},
 /*30(0x1E)*/ {"call",F_FAR,1552,CALLF,0},
 /*31(0x1F)*/ {"call",F_FAR,1808,CALLF,0},
 /*32(0x20)*/ {"jmp",0,16,JMPN,0},
 /*33(0x21)*/ {"jmp",0,272,JMPN,0},
 /*34(0x22)*/ {"jmp",0,528,JMPN,0},
 /*35(0x23)*/ {"jmp",0,784,JMPN,0},
 /*36(0x24)*/ {"jmp",0,1040,JMPN,0},
 /*37(0x25)*/ {"jmp",0,1296,JMPN,0},
 /*38(0x26)*/ {"jmp",0,1552,JMPN,0},
 /*39(0x27)*/ {"jmp",0,1808,JMPN,0},
 /*40(0x28)*/ {"jmp",F_FAR,16,JMPF,0},
 /*41(0x29)*/ {"jmp",F_FAR,272,JMPF,0},
 /*42(0x2A)*/ {"jmp",F_FAR,528,JMPF,0},
 /*43(0x2B)*/ {"jmp",F_FAR,784,JMPF,0},
 /*44(0x2C)*/ {"jmp",F_FAR,1040,JMPF,0},
 /*45(0x2D)*/ {"jmp",F_FAR,1296,JMPF,0},
 /*46(0x2E)*/ {"jmp",F_FAR,1552,JMPF,0},
 /*47(0x2F)*/ {"jmp",F_FAR,1808,JMPF,0},
 /*48(0x30)*/ {"push",0,16,PUSH,0},
 /*49(0x31)*/ {"push",0,272,PUSH,0},
 /*50(0x32)*/ {"push",0,528,PUSH,0},
 /*51(0x33)*/ {"push",0,784,PUSH,0},
 /*52(0x34)*/ {"push",0,1040,PUSH,0},
 /*53(0x35)*/ {"push",0,1296,PUSH,0},
 /*54(0x36)*/ {"push",0,1552,PUSH,0},
 /*55(0x37)*/ {"push",0,1808,PUSH,0},
 /*56(0x38)*/ {0,0,16,0,0},
 /*57(0x39)*/ {0,0,272,0,0},
 /*58(0x3A)*/ {0,0,528,0,0},
 /*59(0x3B)*/ {0,0,784,0,0},
 /*60(0x3C)*/ {0,0,1040,0,0},
 /*61(0x3D)*/ {0,0,1296,0,0},
 /*62(0x3E)*/ {0,0,1552,0,0},
 /*63(0x3F)*/ {0,0,1808,0,0},
 /*64(0x40)*/ {"inc",0,17,INC,0},
 /*65(0x41)*/ {"inc",0,273,INC,0},
 /*66(0x42)*/ {"inc",0,529,INC,0},
 /*67(0x43)*/ {"inc",0,785,INC,0},
 /*68(0x44)*/ {"inc",0,1041,INC,0},
 /*69(0x45)*/ {"inc",0,1297,INC,0},
 /*70(0x46)*/ {"inc",0,1553,INC,0},
 /*71(0x47)*/ {"inc",0,1809,INC,0},
 /*72(0x48)*/ {"dec",0,17,DEC,0},
 /*73(0x49)*/ {"dec",0,273,DEC,0},
 /*74(0x4A)*/ {"dec",0,529,DEC,0},
 /*75(0x4B)*/ {"dec",0,785,DEC,0},
 /*76(0x4C)*/ {"dec",0,1041,DEC,0},
 /*77(0x4D)*/ {"dec",0,1297,DEC,0},
 /*78(0x4E)*/ {"dec",0,1553,DEC,0},
 /*79(0x4F)*/ {"dec",0,1809,DEC,0},
 /*80(0x50)*/ {"call",0,17,CALLN,0},
 /*81(0x51)*/ {"call",0,273,CALLN,0},
 /*82(0x52)*/ {"call",0,529,CALLN,0},
 /*83(0x53)*/ {"call",0,785,CALLN,0},
 /*84(0x54)*/ {"call",0,1041,CALLN,0},
 /*85(0x55)*/ {"call",0,1297,CALLN,0},
 /*86(0x56)*/ {"call",0,1553,CALLN,0},
 /*87(0x57)*/ {"call",0,1809,CALLN,0},
 /*88(0x58)*/ {"call",F_FAR,17,CALLF,0},
 /*89(0x59)*/ {"call",F_FAR,273,CALLF,0},
 /*90(0x5A)*/ {"call",F_FAR,529,CALLF,0},
 /*91(0x5B)*/ {"call",F_FAR,785,CALLF,0},
 /*92(0x5C)*/ {"call",F_FAR,1041,CALLF,0},
 /*93(0x5D)*/ {"call",F_FAR,1297,CALLF,0},
 /*94(0x5E)*/ {"call",F_FAR,1553,CALLF,0},
 /*95(0x5F)*/ {"call",F_FAR,1809,CALLF,0},
 /*96(0x60)*/ {"jmp",0,17,JMPN,0},
 /*97(0x61)*/ {"jmp",0,273,JMPN,0},
 /*98(0x62)*/ {"jmp",0,529,JMPN,0},
 /*99(0x63)*/ {"jmp",0,785,JMPN,0},
 /*100(0x64)*/ {"jmp",0,1041,JMPN,0},
 /*101(0x65)*/ {"jmp",0,1297,JMPN,0},
 /*102(0x66)*/ {"jmp",0,1553,JMPN,0},
 /*103(0x67)*/ {"jmp",0,1809,JMPN,0},
 /*104(0x68)*/ {"jmp",F_FAR,17,JMPF,0},
 /*105(0x69)*/ {"jmp",F_FAR,273,JMPF,0},
 /*106(0x6A)*/ {"jmp",F_FAR,529,JMPF,0},
 /*107(0x6B)*/ {"jmp",F_FAR,785,JMPF,0},
 /*108(0x6C)*/ {"jmp",F_FAR,1041,JMPF,0},
 /*109(0x6D)*/ {"jmp",F_FAR,1297,JMPF,0},
 /*110(0x6E)*/ {"jmp",F_FAR,1553,JMPF,0},
 /*111(0x6F)*/ {"jmp",F_FAR,1809,JMPF,0},
 /*112(0x70)*/ {"push",0,17,PUSH,0},
 /*113(0x71)*/ {"push",0,273,PUSH,0},
 /*114(0x72)*/ {"push",0,529,PUSH,0},
 /*115(0x73)*/ {"push",0,785,PUSH,0},
 /*116(0x74)*/ {"push",0,1041,PUSH,0},
 /*117(0x75)*/ {"push",0,1297,PUSH,0},
 /*118(0x76)*/ {"push",0,1553,PUSH,0},
 /*119(0x77)*/ {"push",0,1809,PUSH,0},
 /*120(0x78)*/ {0,0,17,0,0},
 /*121(0x79)*/ {0,0,273,0,0},
 /*122(0x7A)*/ {0,0,529,0,0},
 /*123(0x7B)*/ {0,0,785,0,0},
 /*124(0x7C)*/ {0,0,1041,0,0},
 /*125(0x7D)*/ {0,0,1297,0,0},
 /*126(0x7E)*/ {0,0,1553,0,0},
 /*127(0x7F)*/ {0,0,1809,0,0},
 /*128(0x80)*/ {"inc",0,18,INC,0},
 /*129(0x81)*/ {"inc",0,274,INC,0},
 /*130(0x82)*/ {"inc",0,530,INC,0},
 /*131(0x83)*/ {"inc",0,786,INC,0},
 /*132(0x84)*/ {"inc",0,1042,INC,0},
 /*133(0x85)*/ {"inc",0,1298,INC,0},
 /*134(0x86)*/ {"inc",0,1554,INC,0},
 /*135(0x87)*/ {"inc",0,1810,INC,0},
 /*136(0x88)*/ {"dec",0,18,DEC,0},
 /*137(0x89)*/ {"dec",0,274,DEC,0},
 /*138(0x8A)*/ {"dec",0,530,DEC,0},
 /*139(0x8B)*/ {"dec",0,786,DEC,0},
 /*140(0x8C)*/ {"dec",0,1042,DEC,0},
 /*141(0x8D)*/ {"dec",0,1298,DEC,0},
 /*142(0x8E)*/ {"dec",0,1554,DEC,0},
 /*143(0x8F)*/ {"dec",0,1810,DEC,0},
 /*144(0x90)*/ {"call",0,18,CALLN,0},
 /*145(0x91)*/ {"call",0,274,CALLN,0},
 /*146(0x92)*/ {"call",0,530,CALLN,0},
 /*147(0x93)*/ {"call",0,786,CALLN,0},
 /*148(0x94)*/ {"call",0,1042,CALLN,0},
 /*149(0x95)*/ {"call",0,1298,CALLN,0},
 /*150(0x96)*/ {"call",0,1554,CALLN,0},
 /*151(0x97)*/ {"call",0,1810,CALLN,0},
 /*152(0x98)*/ {"call",F_FAR,18,CALLF,0},
 /*153(0x99)*/ {"call",F_FAR,274,CALLF,0},
 /*154(0x9A)*/ {"call",F_FAR,530,CALLF,0},
 /*155(0x9B)*/ {"call",F_FAR,786,CALLF,0},
 /*156(0x9C)*/ {"call",F_FAR,1042,CALLF,0},
 /*157(0x9D)*/ {"call",F_FAR,1298,CALLF,0},
 /*158(0x9E)*/ {"call",F_FAR,1554,CALLF,0},
 /*159(0x9F)*/ {"call",F_FAR,1810,CALLF,0},
 /*160(0xA0)*/ {"jmp",0,18,JMPN,0},
 /*161(0xA1)*/ {"jmp",0,274,JMPN,0},
 /*162(0xA2)*/ {"jmp",0,530,JMPN,0},
 /*163(0xA3)*/ {"jmp",0,786,JMPN,0},
 /*164(0xA4)*/ {"jmp",0,1042,JMPN,0},
 /*165(0xA5)*/ {"jmp",0,1298,JMPN,0},
 /*166(0xA6)*/ {"jmp",0,1554,JMPN,0},
 /*167(0xA7)*/ {"jmp",0,1810,JMPN,0},
 /*168(0xA8)*/ {"jmp",F_FAR,18,JMPF,0},
 /*169(0xA9)*/ {"jmp",F_FAR,274,JMPF,0},
 /*170(0xAA)*/ {"jmp",F_FAR,530,JMPF,0},
 /*171(0xAB)*/ {"jmp",F_FAR,786,JMPF,0},
 /*172(0xAC)*/ {"jmp",F_FAR,1042,JMPF,0},
 /*173(0xAD)*/ {"jmp",F_FAR,1298,JMPF,0},
 /*174(0xAE)*/ {"jmp",F_FAR,1554,JMPF,0},
 /*175(0xAF)*/ {"jmp",F_FAR,1810,JMPF,0},
 /*176(0xB0)*/ {"push",0,18,PUSH,0},
 /*177(0xB1)*/ {"push",0,274,PUSH,0},
 /*178(0xB2)*/ {"push",0,530,PUSH,0},
 /*179(0xB3)*/ {"push",0,786,PUSH,0},
 /*180(0xB4)*/ {"push",0,1042,PUSH,0},
 /*181(0xB5)*/ {"push",0,1298,PUSH,0},
 /*182(0xB6)*/ {"push",0,1554,PUSH,0},
 /*183(0xB7)*/ {"push",0,1810,PUSH,0},
 /*184(0xB8)*/ {0,0,18,0,0},
 /*185(0xB9)*/ {0,0,274,0,0},
 /*186(0xBA)*/ {0,0,530,0,0},
 /*187(0xBB)*/ {0,0,786,0,0},
 /*188(0xBC)*/ {0,0,1042,0,0},
 /*189(0xBD)*/ {0,0,1298,0,0},
 /*190(0xBE)*/ {0,0,1554,0,0},
 /*191(0xBF)*/ {0,0,1810,0,0},
 /*192(0xC0)*/ {"inc",0,0,INC,0},
 /*193(0xC1)*/ {"inc",0,256,INC,0},
 /*194(0xC2)*/ {"inc",0,512,INC,0},
 /*195(0xC3)*/ {"inc",0,768,INC,0},
 /*196(0xC4)*/ {"inc",0,1024,INC,0},
 /*197(0xC5)*/ {"inc",0,1280,INC,0},
 /*198(0xC6)*/ {"inc",0,1536,INC,0},
 /*199(0xC7)*/ {"inc",0,1792,INC,0},
 /*200(0xC8)*/ {"dec",0,0,DEC,0},
 /*201(0xC9)*/ {"dec",0,256,DEC,0},
 /*202(0xCA)*/ {"dec",0,512,DEC,0},
 /*203(0xCB)*/ {"dec",0,768,DEC,0},
 /*204(0xCC)*/ {"dec",0,1024,DEC,0},
 /*205(0xCD)*/ {"dec",0,1280,DEC,0},
 /*206(0xCE)*/ {"dec",0,1536,DEC,0},
 /*207(0xCF)*/ {"dec",0,1792,DEC,0},
 /*208(0xD0)*/ {"call",0,0,CALLN,0},
 /*209(0xD1)*/ {"call",0,256,CALLN,0},
 /*210(0xD2)*/ {"call",0,512,CALLN,0},
 /*211(0xD3)*/ {"call",0,768,CALLN,0},
 /*212(0xD4)*/ {"call",0,1024,CALLN,0},
 /*213(0xD5)*/ {"call",0,1280,CALLN,0},
 /*214(0xD6)*/ {"call",0,1536,CALLN,0},
 /*215(0xD7)*/ {"call",0,1792,CALLN,0},
 /*216(0xD8)*/ {"call",F_FAR,0,CALLF,0},
 /*217(0xD9)*/ {"call",F_FAR,256,CALLF,0},
 /*218(0xDA)*/ {"call",F_FAR,512,CALLF,0},
 /*219(0xDB)*/ {"call",F_FAR,768,CALLF,0},
 /*220(0xDC)*/ {"call",F_FAR,1024,CALLF,0},
 /*221(0xDD)*/ {"call",F_FAR,1280,CALLF,0},
 /*222(0xDE)*/ {"call",F_FAR,1536,CALLF,0},
 /*223(0xDF)*/ {"call",F_FAR,1792,CALLF,0},
 /*224(0xE0)*/ {"jmp",0,0,JMPN,0},
 /*225(0xE1)*/ {"jmp",0,256,JMPN,0},
 /*226(0xE2)*/ {"jmp",0,512,JMPN,0},
 /*227(0xE3)*/ {"jmp",0,768,JMPN,0},
 /*228(0xE4)*/ {"jmp",0,1024,JMPN,0},
 /*229(0xE5)*/ {"jmp",0,1280,JMPN,0},
 /*230(0xE6)*/ {"jmp",0,1536,JMPN,0},
 /*231(0xE7)*/ {"jmp",0,1792,JMPN,0},
 /*232(0xE8)*/ {"jmp",F_FAR,0,JMPF,0},
 /*233(0xE9)*/ {"jmp",F_FAR,256,JMPF,0},
 /*234(0xEA)*/ {"jmp",F_FAR,512,JMPF,0},
 /*235(0xEB)*/ {"jmp",F_FAR,768,JMPF,0},
 /*236(0xEC)*/ {"jmp",F_FAR,1024,JMPF,0},
 /*237(0xED)*/ {"jmp",F_FAR,1280,JMPF,0},
 /*238(0xEE)*/ {"jmp",F_FAR,1536,JMPF,0},
 /*239(0xEF)*/ {"jmp",F_FAR,1792,JMPF,0},
 /*240(0xF0)*/ {"push",0,0,PUSH,0},
 /*241(0xF1)*/ {"push",0,256,PUSH,0},
 /*242(0xF2)*/ {"push",0,512,PUSH,0},
 /*243(0xF3)*/ {"push",0,768,PUSH,0},
 /*244(0xF4)*/ {"push",0,1024,PUSH,0},
 /*245(0xF5)*/ {"push",0,1280,PUSH,0},
 /*246(0xF6)*/ {"push",0,1536,PUSH,0},
 /*247(0xF7)*/ {"push",0,1792,PUSH,0},
 /*248(0xF8)*/ {0,0,0,0,0},
 /*249(0xF9)*/ {0,0,256,0,0},
 /*250(0xFA)*/ {0,0,512,0,0},
 /*251(0xFB)*/ {0,0,768,0,0},
 /*252(0xFC)*/ {0,0,1024,0,0},
 /*253(0xFD)*/ {0,0,1280,0,0},
 /*254(0xFE)*/ {0,0,1536,0,0},
 /*255(0xFF)*/ {0,0,1792,0,0}
};


//grupa D0h-D3h oraz C0h-C1h
//grupa instrukcji przesunec
op_t gD0D3C0C1[256]=
{
 /*0(0x00)*/ {"rol",0,16,ROL,0},
 /*1(0x01)*/ {"rol",0,272,ROL,0},
 /*2(0x02)*/ {"rol",0,528,ROL,0},
 /*3(0x03)*/ {"rol",0,784,ROL,0},
 /*4(0x04)*/ {"rol",0,1040,ROL,0},
 /*5(0x05)*/ {"rol",0,1296,ROL,0},
 /*6(0x06)*/ {"rol",0,1552,ROL,0},
 /*7(0x07)*/ {"rol",0,1808,ROL,0},
 /*8(0x08)*/ {"ror",0,16,ROR,0},
 /*9(0x09)*/ {"ror",0,272,ROR,0},
 /*10(0x0A)*/ {"ror",0,528,ROR,0},
 /*11(0x0B)*/ {"ror",0,784,ROR,0},
 /*12(0x0C)*/ {"ror",0,1040,ROR,0},
 /*13(0x0D)*/ {"ror",0,1296,ROR,0},
 /*14(0x0E)*/ {"ror",0,1552,ROR,0},
 /*15(0x0F)*/ {"ror",0,1808,ROR,0},
 /*16(0x10)*/ {"rcl",0,16,RCL,0},
 /*17(0x11)*/ {"rcl",0,272,RCL,0},
 /*18(0x12)*/ {"rcl",0,528,RCL,0},
 /*19(0x13)*/ {"rcl",0,784,RCL,0},
 /*20(0x14)*/ {"rcl",0,1040,RCL,0},
 /*21(0x15)*/ {"rcl",0,1296,RCL,0},
 /*22(0x16)*/ {"rcl",0,1552,RCL,0},
 /*23(0x17)*/ {"rcl",0,1808,RCL,0},
 /*24(0x18)*/ {"rcr",0,16,RCR,0},
 /*25(0x19)*/ {"rcr",0,272,RCR,0},
 /*26(0x1A)*/ {"rcr",0,528,RCR,0},
 /*27(0x1B)*/ {"rcr",0,784,RCR,0},
 /*28(0x1C)*/ {"rcr",0,1040,RCR,0},
 /*29(0x1D)*/ {"rcr",0,1296,RCR,0},
 /*30(0x1E)*/ {"rcr",0,1552,RCR,0},
 /*31(0x1F)*/ {"rcr",0,1808,RCR,0},
 /*32(0x20)*/ {"shl",0,16,SHL,0},
 /*33(0x21)*/ {"shl",0,272,SHL,0},
 /*34(0x22)*/ {"shl",0,528,SHL,0},
 /*35(0x23)*/ {"shl",0,784,SHL,0},
 /*36(0x24)*/ {"shl",0,1040,SHL,0},
 /*37(0x25)*/ {"shl",0,1296,SHL,0},
 /*38(0x26)*/ {"shl",0,1552,SHL,0},
 /*39(0x27)*/ {"shl",0,1808,SHL,0},
 /*40(0x28)*/ {"shr",0,16,SHR,0},
 /*41(0x29)*/ {"shr",0,272,SHR,0},
 /*42(0x2A)*/ {"shr",0,528,SHR,0},
 /*43(0x2B)*/ {"shr",0,784,SHR,0},
 /*44(0x2C)*/ {"shr",0,1040,SHR,0},
 /*45(0x2D)*/ {"shr",0,1296,SHR,0},
 /*46(0x2E)*/ {"shr",0,1552,SHR,0},
 /*47(0x2F)*/ {"shr",0,1808,SHR,0},
 /*48(0x30)*/ {0,0,16,0,0},
 /*49(0x31)*/ {0,0,272,0,0},
 /*50(0x32)*/ {0,0,528,0,0},
 /*51(0x33)*/ {0,0,784,0,0},
 /*52(0x34)*/ {0,0,1040,0,0},
 /*53(0x35)*/ {0,0,1296,0,0},
 /*54(0x36)*/ {0,0,1552,0,0},
 /*55(0x37)*/ {0,0,1808,0,0},
 /*56(0x38)*/ {"sar",0,16,SAR,0},
 /*57(0x39)*/ {"sar",0,272,SAR,0},
 /*58(0x3A)*/ {"sar",0,528,SAR,0},
 /*59(0x3B)*/ {"sar",0,784,SAR,0},
 /*60(0x3C)*/ {"sar",0,1040,SAR,0},
 /*61(0x3D)*/ {"sar",0,1296,SAR,0},
 /*62(0x3E)*/ {"sar",0,1552,SAR,0},
 /*63(0x3F)*/ {"sar",0,1808,SAR,0},
 /*64(0x40)*/ {"rol",0,17,ROL,0},
 /*65(0x41)*/ {"rol",0,273,ROL,0},
 /*66(0x42)*/ {"rol",0,529,ROL,0},
 /*67(0x43)*/ {"rol",0,785,ROL,0},
 /*68(0x44)*/ {"rol",0,1041,ROL,0},
 /*69(0x45)*/ {"rol",0,1297,ROL,0},
 /*70(0x46)*/ {"rol",0,1553,ROL,0},
 /*71(0x47)*/ {"rol",0,1809,ROL,0},
 /*72(0x48)*/ {"ror",0,17,ROR,0},
 /*73(0x49)*/ {"ror",0,273,ROR,0},
 /*74(0x4A)*/ {"ror",0,529,ROR,0},
 /*75(0x4B)*/ {"ror",0,785,ROR,0},
 /*76(0x4C)*/ {"ror",0,1041,ROR,0},
 /*77(0x4D)*/ {"ror",0,1297,ROR,0},
 /*78(0x4E)*/ {"ror",0,1553,ROR,0},
 /*79(0x4F)*/ {"ror",0,1809,ROR,0},
 /*80(0x50)*/ {"rcl",0,17,RCL,0},
 /*81(0x51)*/ {"rcl",0,273,RCL,0},
 /*82(0x52)*/ {"rcl",0,529,RCL,0},
 /*83(0x53)*/ {"rcl",0,785,RCL,0},
 /*84(0x54)*/ {"rcl",0,1041,RCL,0},
 /*85(0x55)*/ {"rcl",0,1297,RCL,0},
 /*86(0x56)*/ {"rcl",0,1553,RCL,0},
 /*87(0x57)*/ {"rcl",0,1809,RCL,0},
 /*88(0x58)*/ {"rcr",0,17,RCR,0},
 /*89(0x59)*/ {"rcr",0,273,RCR,0},
 /*90(0x5A)*/ {"rcr",0,529,RCR,0},
 /*91(0x5B)*/ {"rcr",0,785,RCR,0},
 /*92(0x5C)*/ {"rcr",0,1041,RCR,0},
 /*93(0x5D)*/ {"rcr",0,1297,RCR,0},
 /*94(0x5E)*/ {"rcr",0,1553,RCR,0},
 /*95(0x5F)*/ {"rcr",0,1809,RCR,0},
 /*96(0x60)*/ {"shl",0,17,SHL,0},
 /*97(0x61)*/ {"shl",0,273,SHL,0},
 /*98(0x62)*/ {"shl",0,529,SHL,0},
 /*99(0x63)*/ {"shl",0,785,SHL,0},
 /*100(0x64)*/ {"shl",0,1041,SHL,0},
 /*101(0x65)*/ {"shl",0,1297,SHL,0},
 /*102(0x66)*/ {"shl",0,1553,SHL,0},
 /*103(0x67)*/ {"shl",0,1809,SHL,0},
 /*104(0x68)*/ {"shr",0,17,SHR,0},
 /*105(0x69)*/ {"shr",0,273,SHR,0},
 /*106(0x6A)*/ {"shr",0,529,SHR,0},
 /*107(0x6B)*/ {"shr",0,785,SHR,0},
 /*108(0x6C)*/ {"shr",0,1041,SHR,0},
 /*109(0x6D)*/ {"shr",0,1297,SHR,0},
 /*110(0x6E)*/ {"shr",0,1553,SHR,0},
 /*111(0x6F)*/ {"shr",0,1809,SHR,0},
 /*112(0x70)*/ {0,0,17,0,0},
 /*113(0x71)*/ {0,0,273,0,0},
 /*114(0x72)*/ {0,0,529,0,0},
 /*115(0x73)*/ {0,0,785,0,0},
 /*116(0x74)*/ {0,0,1041,0,0},
 /*117(0x75)*/ {0,0,1297,0,0},
 /*118(0x76)*/ {0,0,1553,0,0},
 /*119(0x77)*/ {0,0,1809,0,0},
 /*120(0x78)*/ {"sar",0,17,SAR,0},
 /*121(0x79)*/ {"sar",0,273,SAR,0},
 /*122(0x7A)*/ {"sar",0,529,SAR,0},
 /*123(0x7B)*/ {"sar",0,785,SAR,0},
 /*124(0x7C)*/ {"sar",0,1041,SAR,0},
 /*125(0x7D)*/ {"sar",0,1297,SAR,0},
 /*126(0x7E)*/ {"sar",0,1553,SAR,0},
 /*127(0x7F)*/ {"sar",0,1809,SAR,0},
 /*128(0x80)*/ {"rol",0,18,ROL,0},
 /*129(0x81)*/ {"rol",0,274,ROL,0},
 /*130(0x82)*/ {"rol",0,530,ROL,0},
 /*131(0x83)*/ {"rol",0,786,ROL,0},
 /*132(0x84)*/ {"rol",0,1042,ROL,0},
 /*133(0x85)*/ {"rol",0,1298,ROL,0},
 /*134(0x86)*/ {"rol",0,1554,ROL,0},
 /*135(0x87)*/ {"rol",0,1810,ROL,0},
 /*136(0x88)*/ {"ror",0,18,ROR,0},
 /*137(0x89)*/ {"ror",0,274,ROR,0},
 /*138(0x8A)*/ {"ror",0,530,ROR,0},
 /*139(0x8B)*/ {"ror",0,786,ROR,0},
 /*140(0x8C)*/ {"ror",0,1042,ROR,0},
 /*141(0x8D)*/ {"ror",0,1298,ROR,0},
 /*142(0x8E)*/ {"ror",0,1554,ROR,0},
 /*143(0x8F)*/ {"ror",0,1810,ROR,0},
 /*144(0x90)*/ {"rcl",0,18,RCL,0},
 /*145(0x91)*/ {"rcl",0,274,RCL,0},
 /*146(0x92)*/ {"rcl",0,530,RCL,0},
 /*147(0x93)*/ {"rcl",0,786,RCL,0},
 /*148(0x94)*/ {"rcl",0,1042,RCL,0},
 /*149(0x95)*/ {"rcl",0,1298,RCL,0},
 /*150(0x96)*/ {"rcl",0,1554,RCL,0},
 /*151(0x97)*/ {"rcl",0,1810,RCL,0},
 /*152(0x98)*/ {"rcr",0,18,RCR,0},
 /*153(0x99)*/ {"rcr",0,274,RCR,0},
 /*154(0x9A)*/ {"rcr",0,530,RCR,0},
 /*155(0x9B)*/ {"rcr",0,786,RCR,0},
 /*156(0x9C)*/ {"rcr",0,1042,RCR,0},
 /*157(0x9D)*/ {"rcr",0,1298,RCR,0},
 /*158(0x9E)*/ {"rcr",0,1554,RCR,0},
 /*159(0x9F)*/ {"rcr",0,1810,RCR,0},
 /*160(0xA0)*/ {"shl",0,18,SHL,0},
 /*161(0xA1)*/ {"shl",0,274,SHL,0},
 /*162(0xA2)*/ {"shl",0,530,SHL,0},
 /*163(0xA3)*/ {"shl",0,786,SHL,0},
 /*164(0xA4)*/ {"shl",0,1042,SHL,0},
 /*165(0xA5)*/ {"shl",0,1298,SHL,0},
 /*166(0xA6)*/ {"shl",0,1554,SHL,0},
 /*167(0xA7)*/ {"shl",0,1810,SHL,0},
 /*168(0xA8)*/ {"shr",0,18,SHR,0},
 /*169(0xA9)*/ {"shr",0,274,SHR,0},
 /*170(0xAA)*/ {"shr",0,530,SHR,0},
 /*171(0xAB)*/ {"shr",0,786,SHR,0},
 /*172(0xAC)*/ {"shr",0,1042,SHR,0},
 /*173(0xAD)*/ {"shr",0,1298,SHR,0},
 /*174(0xAE)*/ {"shr",0,1554,SHR,0},
 /*175(0xAF)*/ {"shr",0,1810,SHR,0},
 /*176(0xB0)*/ {0,0,18,0,0},
 /*177(0xB1)*/ {0,0,274,0,0},
 /*178(0xB2)*/ {0,0,530,0,0},
 /*179(0xB3)*/ {0,0,786,0,0},
 /*180(0xB4)*/ {0,0,1042,0,0},
 /*181(0xB5)*/ {0,0,1298,0,0},
 /*182(0xB6)*/ {0,0,1554,0,0},
 /*183(0xB7)*/ {0,0,1810,0,0},
 /*184(0xB8)*/ {"sar",0,18,SAR,0},
 /*185(0xB9)*/ {"sar",0,274,SAR,0},
 /*186(0xBA)*/ {"sar",0,530,SAR,0},
 /*187(0xBB)*/ {"sar",0,786,SAR,0},
 /*188(0xBC)*/ {"sar",0,1042,SAR,0},
 /*189(0xBD)*/ {"sar",0,1298,SAR,0},
 /*190(0xBE)*/ {"sar",0,1554,SAR,0},
 /*191(0xBF)*/ {"sar",0,1810,SAR,0},
 /*192(0xC0)*/ {"rol",0,0,ROL,0},
 /*193(0xC1)*/ {"rol",0,256,ROL,0},
 /*194(0xC2)*/ {"rol",0,512,ROL,0},
 /*195(0xC3)*/ {"rol",0,768,ROL,0},
 /*196(0xC4)*/ {"rol",0,1024,ROL,0},
 /*197(0xC5)*/ {"rol",0,1280,ROL,0},
 /*198(0xC6)*/ {"rol",0,1536,ROL,0},
 /*199(0xC7)*/ {"rol",0,1792,ROL,0},
 /*200(0xC8)*/ {"ror",0,0,ROR,0},
 /*201(0xC9)*/ {"ror",0,256,ROR,0},
 /*202(0xCA)*/ {"ror",0,512,ROR,0},
 /*203(0xCB)*/ {"ror",0,768,ROR,0},
 /*204(0xCC)*/ {"ror",0,1024,ROR,0},
 /*205(0xCD)*/ {"ror",0,1280,ROR,0},
 /*206(0xCE)*/ {"ror",0,1536,ROR,0},
 /*207(0xCF)*/ {"ror",0,1792,ROR,0},
 /*208(0xD0)*/ {"rcl",0,0,RCL,0},
 /*209(0xD1)*/ {"rcl",0,256,RCL,0},
 /*210(0xD2)*/ {"rcl",0,512,RCL,0},
 /*211(0xD3)*/ {"rcl",0,768,RCL,0},
 /*212(0xD4)*/ {"rcl",0,1024,RCL,0},
 /*213(0xD5)*/ {"rcl",0,1280,RCL,0},
 /*214(0xD6)*/ {"rcl",0,1536,RCL,0},
 /*215(0xD7)*/ {"rcl",0,1792,RCL,0},
 /*216(0xD8)*/ {"rcr",0,0,RCR,0},
 /*217(0xD9)*/ {"rcr",0,256,RCR,0},
 /*218(0xDA)*/ {"rcr",0,512,RCR,0},
 /*219(0xDB)*/ {"rcr",0,768,RCR,0},
 /*220(0xDC)*/ {"rcr",0,1024,RCR,0},
 /*221(0xDD)*/ {"rcr",0,1280,RCR,0},
 /*222(0xDE)*/ {"rcr",0,1536,RCR,0},
 /*223(0xDF)*/ {"rcr",0,1792,RCR,0},
 /*224(0xE0)*/ {"shl",0,0,SHL,0},
 /*225(0xE1)*/ {"shl",0,256,SHL,0},
 /*226(0xE2)*/ {"shl",0,512,SHL,0},
 /*227(0xE3)*/ {"shl",0,768,SHL,0},
 /*228(0xE4)*/ {"shl",0,1024,SHL,0},
 /*229(0xE5)*/ {"shl",0,1280,SHL,0},
 /*230(0xE6)*/ {"shl",0,1536,SHL,0},
 /*231(0xE7)*/ {"shl",0,1792,SHL,0},
 /*232(0xE8)*/ {"shr",0,0,SHR,0},
 /*233(0xE9)*/ {"shr",0,256,SHR,0},
 /*234(0xEA)*/ {"shr",0,512,SHR,0},
 /*235(0xEB)*/ {"shr",0,768,SHR,0},
 /*236(0xEC)*/ {"shr",0,1024,SHR,0},
 /*237(0xED)*/ {"shr",0,1280,SHR,0},
 /*238(0xEE)*/ {"shr",0,1536,SHR,0},
 /*239(0xEF)*/ {"shr",0,1792,SHR,0},
 /*240(0xF0)*/ {0,0,0,0,0},
 /*241(0xF1)*/ {0,0,256,0,0},
 /*242(0xF2)*/ {0,0,512,0,0},
 /*243(0xF3)*/ {0,0,768,0,0},
 /*244(0xF4)*/ {0,0,1024,0,0},
 /*245(0xF5)*/ {0,0,1280,0,0},
 /*246(0xF6)*/ {0,0,1536,0,0},
 /*247(0xF7)*/ {0,0,1792,0,0},
 /*248(0xF8)*/ {"sar",0,0,SAR,0},
 /*249(0xF9)*/ {"sar",0,256,SAR,0},
 /*250(0xFA)*/ {"sar",0,512,SAR,0},
 /*251(0xFB)*/ {"sar",0,768,SAR,0},
 /*252(0xFC)*/ {"sar",0,1024,SAR,0},
 /*253(0xFD)*/ {"sar",0,1280,SAR,0},
 /*254(0xFE)*/ {"sar",0,1536,SAR,0},
 /*255(0xFF)*/ {"sar",0,1792,SAR,0}
};

//grupa 0F00 (instrukcje systemowe)
op_t g0F00[256]=
{
 /*0(0x00)*/ {"sldt",F_W,16,SLDT,0},
 /*1(0x01)*/ {"sldt",F_W,272,SLDT,0},
 /*2(0x02)*/ {"sldt",F_W,528,SLDT,0},
 /*3(0x03)*/ {"sldt",F_W,784,SLDT,0},
 /*4(0x04)*/ {"sldt",F_W,1040,SLDT,0},
 /*5(0x05)*/ {"sldt",F_W,1296,SLDT,0},
 /*6(0x06)*/ {"sldt",F_W,1552,SLDT,0},
 /*7(0x07)*/ {"sldt",F_W,1808,SLDT,0},
 /*8(0x08)*/ {"str",F_W,16,STR,0},
 /*9(0x09)*/ {"str",F_W,272,STR,0},
 /*10(0x0A)*/ {"str",F_W,528,STR,0},
 /*11(0x0B)*/ {"str",F_W,784,STR,0},
 /*12(0x0C)*/ {"str",F_W,1040,STR,0},
 /*13(0x0D)*/ {"str",F_W,1296,STR,0},
 /*14(0x0E)*/ {"str",F_W,1552,STR,0},
 /*15(0x0F)*/ {"str",F_W,1808,STR,0},
 /*16(0x10)*/ {"lldt",F_W,16,LLDT,0},
 /*17(0x11)*/ {"lldt",F_W,272,LLDT,0},
 /*18(0x12)*/ {"lldt",F_W,528,LLDT,0},
 /*19(0x13)*/ {"lldt",F_W,784,LLDT,0},
 /*20(0x14)*/ {"lldt",F_W,1040,LLDT,0},
 /*21(0x15)*/ {"lldt",F_W,1296,LLDT,0},
 /*22(0x16)*/ {"lldt",F_W,1552,LLDT,0},
 /*23(0x17)*/ {"lldt",F_W,1808,LLDT,0},
 /*24(0x18)*/ {"ltr",F_W,16,LTR,0},
 /*25(0x19)*/ {"ltr",F_W,272,LTR,0},
 /*26(0x1A)*/ {"ltr",F_W,528,LTR,0},
 /*27(0x1B)*/ {"ltr",F_W,784,LTR,0},
 /*28(0x1C)*/ {"ltr",F_W,1040,LTR,0},
 /*29(0x1D)*/ {"ltr",F_W,1296,LTR,0},
 /*30(0x1E)*/ {"ltr",F_W,1552,LTR,0},
 /*31(0x1F)*/ {"ltr",F_W,1808,LTR,0},
 /*32(0x20)*/ {"verr",F_W,16,VERR,0},
 /*33(0x21)*/ {"verr",F_W,272,VERR,0},
 /*34(0x22)*/ {"verr",F_W,528,VERR,0},
 /*35(0x23)*/ {"verr",F_W,784,VERR,0},
 /*36(0x24)*/ {"verr",F_W,1040,VERR,0},
 /*37(0x25)*/ {"verr",F_W,1296,VERR,0},
 /*38(0x26)*/ {"verr",F_W,1552,VERR,0},
 /*39(0x27)*/ {"verr",F_W,1808,VERR,0},
 /*40(0x28)*/ {"verw",F_W,16,VERW,0},
 /*41(0x29)*/ {"verw",F_W,272,VERW,0},
 /*42(0x2A)*/ {"verw",F_W,528,VERW,0},
 /*43(0x2B)*/ {"verw",F_W,784,VERW,0},
 /*44(0x2C)*/ {"verw",F_W,1040,VERW,0},
 /*45(0x2D)*/ {"verw",F_W,1296,VERW,0},
 /*46(0x2E)*/ {"verw",F_W,1552,VERW,0},
 /*47(0x2F)*/ {"verw",F_W,1808,VERW,0},
 /*48(0x30)*/ {0,F_W,16,0,0},
 /*49(0x31)*/ {0,F_W,272,0,0},
 /*50(0x32)*/ {0,F_W,528,0,0},
 /*51(0x33)*/ {0,F_W,784,0,0},
 /*52(0x34)*/ {0,F_W,1040,0,0},
 /*53(0x35)*/ {0,F_W,1296,0,0},
 /*54(0x36)*/ {0,F_W,1552,0,0},
 /*55(0x37)*/ {0,F_W,1808,0,0},
 /*56(0x38)*/ {0,F_W,16,0,0},
 /*57(0x39)*/ {0,F_W,272,0,0},
 /*58(0x3A)*/ {0,F_W,528,0,0},
 /*59(0x3B)*/ {0,F_W,784,0,0},
 /*60(0x3C)*/ {0,F_W,1040,0,0},
 /*61(0x3D)*/ {0,F_W,1296,0,0},
 /*62(0x3E)*/ {0,F_W,1552,0,0},
 /*63(0x3F)*/ {0,F_W,1808,0,0},
 /*64(0x40)*/ {"sldt",F_W,17,SLDT,0},
 /*65(0x41)*/ {"sldt",F_W,273,SLDT,0},
 /*66(0x42)*/ {"sldt",F_W,529,SLDT,0},
 /*67(0x43)*/ {"sldt",F_W,785,SLDT,0},
 /*68(0x44)*/ {"sldt",F_W,1041,SLDT,0},
 /*69(0x45)*/ {"sldt",F_W,1297,SLDT,0},
 /*70(0x46)*/ {"sldt",F_W,1553,SLDT,0},
 /*71(0x47)*/ {"sldt",F_W,1809,SLDT,0},
 /*72(0x48)*/ {"str",F_W,17,STR,0},
 /*73(0x49)*/ {"str",F_W,273,STR,0},
 /*74(0x4A)*/ {"str",F_W,529,STR,0},
 /*75(0x4B)*/ {"str",F_W,785,STR,0},
 /*76(0x4C)*/ {"str",F_W,1041,STR,0},
 /*77(0x4D)*/ {"str",F_W,1297,STR,0},
 /*78(0x4E)*/ {"str",F_W,1553,STR,0},
 /*79(0x4F)*/ {"str",F_W,1809,STR,0},
 /*80(0x50)*/ {"lldt",F_W,17,LLDT,0},
 /*81(0x51)*/ {"lldt",F_W,273,LLDT,0},
 /*82(0x52)*/ {"lldt",F_W,529,LLDT,0},
 /*83(0x53)*/ {"lldt",F_W,785,LLDT,0},
 /*84(0x54)*/ {"lldt",F_W,1041,LLDT,0},
 /*85(0x55)*/ {"lldt",F_W,1297,LLDT,0},
 /*86(0x56)*/ {"lldt",F_W,1553,LLDT,0},
 /*87(0x57)*/ {"lldt",F_W,1809,LLDT,0},
 /*88(0x58)*/ {"ltr",F_W,17,LTR,0},
 /*89(0x59)*/ {"ltr",F_W,273,LTR,0},
 /*90(0x5A)*/ {"ltr",F_W,529,LTR,0},
 /*91(0x5B)*/ {"ltr",F_W,785,LTR,0},
 /*92(0x5C)*/ {"ltr",F_W,1041,LTR,0},
 /*93(0x5D)*/ {"ltr",F_W,1297,LTR,0},
 /*94(0x5E)*/ {"ltr",F_W,1553,LTR,0},
 /*95(0x5F)*/ {"ltr",F_W,1809,LTR,0},
 /*96(0x60)*/ {"verr",F_W,17,VERR,0},
 /*97(0x61)*/ {"verr",F_W,273,VERR,0},
 /*98(0x62)*/ {"verr",F_W,529,VERR,0},
 /*99(0x63)*/ {"verr",F_W,785,VERR,0},
 /*100(0x64)*/ {"verr",F_W,1041,VERR,0},
 /*101(0x65)*/ {"verr",F_W,1297,VERR,0},
 /*102(0x66)*/ {"verr",F_W,1553,VERR,0},
 /*103(0x67)*/ {"verr",F_W,1809,VERR,0},
 /*104(0x68)*/ {"verw",F_W,17,VERW,0},
 /*105(0x69)*/ {"verw",F_W,273,VERW,0},
 /*106(0x6A)*/ {"verw",F_W,529,VERW,0},
 /*107(0x6B)*/ {"verw",F_W,785,VERW,0},
 /*108(0x6C)*/ {"verw",F_W,1041,VERW,0},
 /*109(0x6D)*/ {"verw",F_W,1297,VERW,0},
 /*110(0x6E)*/ {"verw",F_W,1553,VERW,0},
 /*111(0x6F)*/ {"verw",F_W,1809,VERW,0},
 /*112(0x70)*/ {0,F_W,17,0,0},
 /*113(0x71)*/ {0,F_W,273,0,0},
 /*114(0x72)*/ {0,F_W,529,0,0},
 /*115(0x73)*/ {0,F_W,785,0,0},
 /*116(0x74)*/ {0,F_W,1041,0,0},
 /*117(0x75)*/ {0,F_W,1297,0,0},
 /*118(0x76)*/ {0,F_W,1553,0,0},
 /*119(0x77)*/ {0,F_W,1809,0,0},
 /*120(0x78)*/ {0,F_W,17,0,0},
 /*121(0x79)*/ {0,F_W,273,0,0},
 /*122(0x7A)*/ {0,F_W,529,0,0},
 /*123(0x7B)*/ {0,F_W,785,0,0},
 /*124(0x7C)*/ {0,F_W,1041,0,0},
 /*125(0x7D)*/ {0,F_W,1297,0,0},
 /*126(0x7E)*/ {0,F_W,1553,0,0},
 /*127(0x7F)*/ {0,F_W,1809,0,0},
 /*128(0x80)*/ {"sldt",F_W,18,SLDT,0},
 /*129(0x81)*/ {"sldt",F_W,274,SLDT,0},
 /*130(0x82)*/ {"sldt",F_W,530,SLDT,0},
 /*131(0x83)*/ {"sldt",F_W,786,SLDT,0},
 /*132(0x84)*/ {"sldt",F_W,1042,SLDT,0},
 /*133(0x85)*/ {"sldt",F_W,1298,SLDT,0},
 /*134(0x86)*/ {"sldt",F_W,1554,SLDT,0},
 /*135(0x87)*/ {"sldt",F_W,1810,SLDT,0},
 /*136(0x88)*/ {"str",F_W,18,STR,0},
 /*137(0x89)*/ {"str",F_W,274,STR,0},
 /*138(0x8A)*/ {"str",F_W,530,STR,0},
 /*139(0x8B)*/ {"str",F_W,786,STR,0},
 /*140(0x8C)*/ {"str",F_W,1042,STR,0},
 /*141(0x8D)*/ {"str",F_W,1298,STR,0},
 /*142(0x8E)*/ {"str",F_W,1554,STR,0},
 /*143(0x8F)*/ {"str",F_W,1810,STR,0},
 /*144(0x90)*/ {"lldt",F_W,18,LLDT,0},
 /*145(0x91)*/ {"lldt",F_W,274,LLDT,0},
 /*146(0x92)*/ {"lldt",F_W,530,LLDT,0},
 /*147(0x93)*/ {"lldt",F_W,786,LLDT,0},
 /*148(0x94)*/ {"lldt",F_W,1042,LLDT,0},
 /*149(0x95)*/ {"lldt",F_W,1298,LLDT,0},
 /*150(0x96)*/ {"lldt",F_W,1554,LLDT,0},
 /*151(0x97)*/ {"lldt",F_W,1810,LLDT,0},
 /*152(0x98)*/ {"ltr",F_W,18,LTR,0},
 /*153(0x99)*/ {"ltr",F_W,274,LTR,0},
 /*154(0x9A)*/ {"ltr",F_W,530,LTR,0},
 /*155(0x9B)*/ {"ltr",F_W,786,LTR,0},
 /*156(0x9C)*/ {"ltr",F_W,1042,LTR,0},
 /*157(0x9D)*/ {"ltr",F_W,1298,LTR,0},
 /*158(0x9E)*/ {"ltr",F_W,1554,LTR,0},
 /*159(0x9F)*/ {"ltr",F_W,1810,LTR,0},
 /*160(0xA0)*/ {"verr",F_W,18,VERR,0},
 /*161(0xA1)*/ {"verr",F_W,274,VERR,0},
 /*162(0xA2)*/ {"verr",F_W,530,VERR,0},
 /*163(0xA3)*/ {"verr",F_W,786,VERR,0},
 /*164(0xA4)*/ {"verr",F_W,1042,VERR,0},
 /*165(0xA5)*/ {"verr",F_W,1298,VERR,0},
 /*166(0xA6)*/ {"verr",F_W,1554,VERR,0},
 /*167(0xA7)*/ {"verr",F_W,1810,VERR,0},
 /*168(0xA8)*/ {"verw",F_W,18,VERW,0},
 /*169(0xA9)*/ {"verw",F_W,274,VERW,0},
 /*170(0xAA)*/ {"verw",F_W,530,VERW,0},
 /*171(0xAB)*/ {"verw",F_W,786,VERW,0},
 /*172(0xAC)*/ {"verw",F_W,1042,VERW,0},
 /*173(0xAD)*/ {"verw",F_W,1298,VERW,0},
 /*174(0xAE)*/ {"verw",F_W,1554,VERW,0},
 /*175(0xAF)*/ {"verw",F_W,1810,VERW,0},
 /*176(0xB0)*/ {0,F_W,18,0,0},
 /*177(0xB1)*/ {0,F_W,274,0,0},
 /*178(0xB2)*/ {0,F_W,530,0,0},
 /*179(0xB3)*/ {0,F_W,786,0,0},
 /*180(0xB4)*/ {0,F_W,1042,0,0},
 /*181(0xB5)*/ {0,F_W,1298,0,0},
 /*182(0xB6)*/ {0,F_W,1554,0,0},
 /*183(0xB7)*/ {0,F_W,1810,0,0},
 /*184(0xB8)*/ {0,F_W,18,0,0},
 /*185(0xB9)*/ {0,F_W,274,0,0},
 /*186(0xBA)*/ {0,F_W,530,0,0},
 /*187(0xBB)*/ {0,F_W,786,0,0},
 /*188(0xBC)*/ {0,F_W,1042,0,0},
 /*189(0xBD)*/ {0,F_W,1298,0,0},
 /*190(0xBE)*/ {0,F_W,1554,0,0},
 /*191(0xBF)*/ {0,F_W,1810,0,0},
 /*192(0xC0)*/ {"sldt",F_W,0,SLDT,0},
 /*193(0xC1)*/ {"sldt",F_W,256,SLDT,0},
 /*194(0xC2)*/ {"sldt",F_W,512,SLDT,0},
 /*195(0xC3)*/ {"sldt",F_W,768,SLDT,0},
 /*196(0xC4)*/ {"sldt",F_W,1024,SLDT,0},
 /*197(0xC5)*/ {"sldt",F_W,1280,SLDT,0},
 /*198(0xC6)*/ {"sldt",F_W,1536,SLDT,0},
 /*199(0xC7)*/ {"sldt",F_W,1792,SLDT,0},
 /*200(0xC8)*/ {"str",F_W,0,STR,0},
 /*201(0xC9)*/ {"str",F_W,256,STR,0},
 /*202(0xCA)*/ {"str",F_W,512,STR,0},
 /*203(0xCB)*/ {"str",F_W,768,STR,0},
 /*204(0xCC)*/ {"str",F_W,1024,STR,0},
 /*205(0xCD)*/ {"str",F_W,1280,STR,0},
 /*206(0xCE)*/ {"str",F_W,1536,STR,0},
 /*207(0xCF)*/ {"str",F_W,1792,STR,0},
 /*208(0xD0)*/ {"lldt",F_W,0,LLDT,0},
 /*209(0xD1)*/ {"lldt",F_W,256,LLDT,0},
 /*210(0xD2)*/ {"lldt",F_W,512,LLDT,0},
 /*211(0xD3)*/ {"lldt",F_W,768,LLDT,0},
 /*212(0xD4)*/ {"lldt",F_W,1024,LLDT,0},
 /*213(0xD5)*/ {"lldt",F_W,1280,LLDT,0},
 /*214(0xD6)*/ {"lldt",F_W,1536,LLDT,0},
 /*215(0xD7)*/ {"lldt",F_W,1792,LLDT,0},
 /*216(0xD8)*/ {"ltr",F_W,0,LTR,0},
 /*217(0xD9)*/ {"ltr",F_W,256,LTR,0},
 /*218(0xDA)*/ {"ltr",F_W,512,LTR,0},
 /*219(0xDB)*/ {"ltr",F_W,768,LTR,0},
 /*220(0xDC)*/ {"ltr",F_W,1024,LTR,0},
 /*221(0xDD)*/ {"ltr",F_W,1280,LTR,0},
 /*222(0xDE)*/ {"ltr",F_W,1536,LTR,0},
 /*223(0xDF)*/ {"ltr",F_W,1792,LTR,0},
 /*224(0xE0)*/ {"verr",F_W,0,VERR,0},
 /*225(0xE1)*/ {"verr",F_W,256,VERR,0},
 /*226(0xE2)*/ {"verr",F_W,512,VERR,0},
 /*227(0xE3)*/ {"verr",F_W,768,VERR,0},
 /*228(0xE4)*/ {"verr",F_W,1024,VERR,0},
 /*229(0xE5)*/ {"verr",F_W,1280,VERR,0},
 /*230(0xE6)*/ {"verr",F_W,1536,VERR,0},
 /*231(0xE7)*/ {"verr",F_W,1792,VERR,0},
 /*232(0xE8)*/ {"verw",F_W,0,VERW,0},
 /*233(0xE9)*/ {"verw",F_W,256,VERW,0},
 /*234(0xEA)*/ {"verw",F_W,512,VERW,0},
 /*235(0xEB)*/ {"verw",F_W,768,VERW,0},
 /*236(0xEC)*/ {"verw",F_W,1024,VERW,0},
 /*237(0xED)*/ {"verw",F_W,1280,VERW,0},
 /*238(0xEE)*/ {"verw",F_W,1536,VERW,0},
 /*239(0xEF)*/ {"verw",F_W,1792,VERW,0},
 /*240(0xF0)*/ {0,F_W,0,0,0},
 /*241(0xF1)*/ {0,F_W,256,0,0},
 /*242(0xF2)*/ {0,F_W,512,0,0},
 /*243(0xF3)*/ {0,F_W,768,0,0},
 /*244(0xF4)*/ {0,F_W,1024,0,0},
 /*245(0xF5)*/ {0,F_W,1280,0,0},
 /*246(0xF6)*/ {0,F_W,1536,0,0},
 /*247(0xF7)*/ {0,F_W,1792,0,0},
 /*248(0xF8)*/ {0,F_W,0,0,0},
 /*249(0xF9)*/ {0,F_W,256,0,0},
 /*250(0xFA)*/ {0,F_W,512,0,0},
 /*251(0xFB)*/ {0,F_W,768,0,0},
 /*252(0xFC)*/ {0,F_W,1024,0,0},
 /*253(0xFD)*/ {0,F_W,1280,0,0},
 /*254(0xFE)*/ {0,F_W,1536,0,0},
 /*255(0xFF)*/ {0,F_W,1792,0,0}
};

//grupa 0F01 (instrukcje systemowe)
op_t g0F01[256]=
{
 /*0(0x00)*/ {"sgdt",F_SW,16,SGDT,0},
 /*1(0x01)*/ {"sgdt",F_SW,272,SGDT,0},
 /*2(0x02)*/ {"sgdt",F_SW,528,SGDT,0},
 /*3(0x03)*/ {"sgdt",F_SW,784,SGDT,0},
 /*4(0x04)*/ {"sgdt",F_SW,1040,SGDT,0},
 /*5(0x05)*/ {"sgdt",F_SW,1296,SGDT,0},
 /*6(0x06)*/ {"sgdt",F_SW,1552,SGDT,0},
 /*7(0x07)*/ {"sgdt",F_SW,1808,SGDT,0},
 /*8(0x08)*/ {"sidt",F_SW,16,SIDT,0},
 /*9(0x09)*/ {"sidt",F_SW,272,SIDT,0},
 /*10(0x0A)*/ {"sidt",F_SW,528,SIDT,0},
 /*11(0x0B)*/ {"sidt",F_SW,784,SIDT,0},
 /*12(0x0C)*/ {"sidt",F_SW,1040,SIDT,0},
 /*13(0x0D)*/ {"sidt",F_SW,1296,SIDT,0},
 /*14(0x0E)*/ {"sidt",F_SW,1552,SIDT,0},
 /*15(0x0F)*/ {"sidt",F_SW,1808,SIDT,0},
 /*16(0x10)*/ {"lgdt",F_SW,16,LGDT,0},
 /*17(0x11)*/ {"lgdt",F_SW,272,LGDT,0},
 /*18(0x12)*/ {"lgdt",F_SW,528,LGDT,0},
 /*19(0x13)*/ {"lgdt",F_SW,784,LGDT,0},
 /*20(0x14)*/ {"lgdt",F_SW,1040,LGDT,0},
 /*21(0x15)*/ {"lgdt",F_SW,1296,LGDT,0},
 /*22(0x16)*/ {"lgdt",F_SW,1552,LGDT,0},
 /*23(0x17)*/ {"lgdt",F_SW,1808,LGDT,0},
 /*24(0x18)*/ {"lidt",F_SW,16,LIDT,0},
 /*25(0x19)*/ {"lidt",F_SW,272,LIDT,0},
 /*26(0x1A)*/ {"lidt",F_SW,528,LIDT,0},
 /*27(0x1B)*/ {"lidt",F_SW,784,LIDT,0},
 /*28(0x1C)*/ {"lidt",F_SW,1040,LIDT,0},
 /*29(0x1D)*/ {"lidt",F_SW,1296,LIDT,0},
 /*30(0x1E)*/ {"lidt",F_SW,1552,LIDT,0},
 /*31(0x1F)*/ {"lidt",F_SW,1808,LIDT,0},
 /*32(0x20)*/ {"smsw",F_W,16,SMSW,0},
 /*33(0x21)*/ {"smsw",F_W,272,SMSW,0},
 /*34(0x22)*/ {"smsw",F_W,528,SMSW,0},
 /*35(0x23)*/ {"smsw",F_W,784,SMSW,0},
 /*36(0x24)*/ {"smsw",F_W,1040,SMSW,0},
 /*37(0x25)*/ {"smsw",F_W,1296,SMSW,0},
 /*38(0x26)*/ {"smsw",F_W,1552,SMSW,0},
 /*39(0x27)*/ {"smsw",F_W,1808,SMSW,0},
 /*40(0x28)*/ {0,F_W,16,0,0},
 /*41(0x29)*/ {0,F_W,272,0,0},
 /*42(0x2A)*/ {0,F_W,528,0,0},
 /*43(0x2B)*/ {0,F_W,784,0,0},
 /*44(0x2C)*/ {0,F_W,1040,0,0},
 /*45(0x2D)*/ {0,F_W,1296,0,0},
 /*46(0x2E)*/ {0,F_W,1552,0,0},
 /*47(0x2F)*/ {0,F_W,1808,0,0},
 /*48(0x30)*/ {"lmsw",F_W,16,LMSW,0},
 /*49(0x31)*/ {"lmsw",F_W,272,LMSW,0},
 /*50(0x32)*/ {"lmsw",F_W,528,LMSW,0},
 /*51(0x33)*/ {"lmsw",F_W,784,LMSW,0},
 /*52(0x34)*/ {"lmsw",F_W,1040,LMSW,0},
 /*53(0x35)*/ {"lmsw",F_W,1296,LMSW,0},
 /*54(0x36)*/ {"lmsw",F_W,1552,LMSW,0},
 /*55(0x37)*/ {"lmsw",F_W,1808,LMSW,0},
 /*56(0x38)*/ {"invlpg",F_W,16,INVLPG,0},
 /*57(0x39)*/ {"invlpg",F_W,272,INVLPG,0},
 /*58(0x3A)*/ {"invlpg",F_W,528,INVLPG,0},
 /*59(0x3B)*/ {"invlpg",F_W,784,INVLPG,0},
 /*60(0x3C)*/ {"invlpg",F_W,1040,INVLPG,0},
 /*61(0x3D)*/ {"invlpg",F_W,1296,INVLPG,0},
 /*62(0x3E)*/ {"invlpg",F_W,1552,INVLPG,0},
 /*63(0x3F)*/ {"invlpg",F_W,1808,INVLPG,0},
 /*64(0x40)*/ {"sgdt",F_SW,17,SGDT,0},
 /*65(0x41)*/ {"sgdt",F_SW,273,SGDT,0},
 /*66(0x42)*/ {"sgdt",F_SW,529,SGDT,0},
 /*67(0x43)*/ {"sgdt",F_SW,785,SGDT,0},
 /*68(0x44)*/ {"sgdt",F_SW,1041,SGDT,0},
 /*69(0x45)*/ {"sgdt",F_SW,1297,SGDT,0},
 /*70(0x46)*/ {"sgdt",F_SW,1553,SGDT,0},
 /*71(0x47)*/ {"sgdt",F_SW,1809,SGDT,0},
 /*72(0x48)*/ {"sidt",F_SW,17,SIDT,0},
 /*73(0x49)*/ {"sidt",F_SW,273,SIDT,0},
 /*74(0x4A)*/ {"sidt",F_SW,529,SIDT,0},
 /*75(0x4B)*/ {"sidt",F_SW,785,SIDT,0},
 /*76(0x4C)*/ {"sidt",F_SW,1041,SIDT,0},
 /*77(0x4D)*/ {"sidt",F_SW,1297,SIDT,0},
 /*78(0x4E)*/ {"sidt",F_SW,1553,SIDT,0},
 /*79(0x4F)*/ {"sidt",F_SW,1809,SIDT,0},
 /*80(0x50)*/ {"lgdt",F_SW,17,LGDT,0},
 /*81(0x51)*/ {"lgdt",F_SW,273,LGDT,0},
 /*82(0x52)*/ {"lgdt",F_SW,529,LGDT,0},
 /*83(0x53)*/ {"lgdt",F_SW,785,LGDT,0},
 /*84(0x54)*/ {"lgdt",F_SW,1041,LGDT,0},
 /*85(0x55)*/ {"lgdt",F_SW,1297,LGDT,0},
 /*86(0x56)*/ {"lgdt",F_SW,1553,LGDT,0},
 /*87(0x57)*/ {"lgdt",F_SW,1809,LGDT,0},
 /*88(0x58)*/ {"lidt",F_SW,17,LIDT,0},
 /*89(0x59)*/ {"lidt",F_SW,273,LIDT,0},
 /*90(0x5A)*/ {"lidt",F_SW,529,LIDT,0},
 /*91(0x5B)*/ {"lidt",F_SW,785,LIDT,0},
 /*92(0x5C)*/ {"lidt",F_SW,1041,LIDT,0},
 /*93(0x5D)*/ {"lidt",F_SW,1297,LIDT,0},
 /*94(0x5E)*/ {"lidt",F_SW,1553,LIDT,0},
 /*95(0x5F)*/ {"lidt",F_SW,1809,LIDT,0},
 /*96(0x60)*/ {"smsw",F_W,17,SMSW,0},
 /*97(0x61)*/ {"smsw",F_W,273,SMSW,0},
 /*98(0x62)*/ {"smsw",F_W,529,SMSW,0},
 /*99(0x63)*/ {"smsw",F_W,785,SMSW,0},
 /*100(0x64)*/ {"smsw",F_W,1041,SMSW,0},
 /*101(0x65)*/ {"smsw",F_W,1297,SMSW,0},
 /*102(0x66)*/ {"smsw",F_W,1553,SMSW,0},
 /*103(0x67)*/ {"smsw",F_W,1809,SMSW,0},
 /*104(0x68)*/ {0,F_W,17,0,0},
 /*105(0x69)*/ {0,F_W,273,0,0},
 /*106(0x6A)*/ {0,F_W,529,0,0},
 /*107(0x6B)*/ {0,F_W,785,0,0},
 /*108(0x6C)*/ {0,F_W,1041,0,0},
 /*109(0x6D)*/ {0,F_W,1297,0,0},
 /*110(0x6E)*/ {0,F_W,1553,0,0},
 /*111(0x6F)*/ {0,F_W,1809,0,0},
 /*112(0x70)*/ {"lmsw",F_W,17,LMSW,0},
 /*113(0x71)*/ {"lmsw",F_W,273,LMSW,0},
 /*114(0x72)*/ {"lmsw",F_W,529,LMSW,0},
 /*115(0x73)*/ {"lmsw",F_W,785,LMSW,0},
 /*116(0x74)*/ {"lmsw",F_W,1041,LMSW,0},
 /*117(0x75)*/ {"lmsw",F_W,1297,LMSW,0},
 /*118(0x76)*/ {"lmsw",F_W,1553,LMSW,0},
 /*119(0x77)*/ {"lmsw",F_W,1809,LMSW,0},
 /*120(0x78)*/ {"invlpg",F_W,17,INVLPG,0},
 /*121(0x79)*/ {"invlpg",F_W,273,INVLPG,0},
 /*122(0x7A)*/ {"invlpg",F_W,529,INVLPG,0},
 /*123(0x7B)*/ {"invlpg",F_W,785,INVLPG,0},
 /*124(0x7C)*/ {"invlpg",F_W,1041,INVLPG,0},
 /*125(0x7D)*/ {"invlpg",F_W,1297,INVLPG,0},
 /*126(0x7E)*/ {"invlpg",F_W,1553,INVLPG,0},
 /*127(0x7F)*/ {"invlpg",F_W,1809,INVLPG,0},
 /*128(0x80)*/ {"sgdt",F_SW,18,SGDT,0},
 /*129(0x81)*/ {"sgdt",F_SW,274,SGDT,0},
 /*130(0x82)*/ {"sgdt",F_SW,530,SGDT,0},
 /*131(0x83)*/ {"sgdt",F_SW,786,SGDT,0},
 /*132(0x84)*/ {"sgdt",F_SW,1042,SGDT,0},
 /*133(0x85)*/ {"sgdt",F_SW,1298,SGDT,0},
 /*134(0x86)*/ {"sgdt",F_SW,1554,SGDT,0},
 /*135(0x87)*/ {"sgdt",F_SW,1810,SGDT,0},
 /*136(0x88)*/ {"sidt",F_SW,18,SIDT,0},
 /*137(0x89)*/ {"sidt",F_SW,274,SIDT,0},
 /*138(0x8A)*/ {"sidt",F_SW,530,SIDT,0},
 /*139(0x8B)*/ {"sidt",F_SW,786,SIDT,0},
 /*140(0x8C)*/ {"sidt",F_SW,1042,SIDT,0},
 /*141(0x8D)*/ {"sidt",F_SW,1298,SIDT,0},
 /*142(0x8E)*/ {"sidt",F_SW,1554,SIDT,0},
 /*143(0x8F)*/ {"sidt",F_SW,1810,SIDT,0},
 /*144(0x90)*/ {"lgdt",F_SW,18,LGDT,0},
 /*145(0x91)*/ {"lgdt",F_SW,274,LGDT,0},
 /*146(0x92)*/ {"lgdt",F_SW,530,LGDT,0},
 /*147(0x93)*/ {"lgdt",F_SW,786,LGDT,0},
 /*148(0x94)*/ {"lgdt",F_SW,1042,LGDT,0},
 /*149(0x95)*/ {"lgdt",F_SW,1298,LGDT,0},
 /*150(0x96)*/ {"lgdt",F_SW,1554,LGDT,0},
 /*151(0x97)*/ {"lgdt",F_SW,1810,LGDT,0},
 /*152(0x98)*/ {"lidt",F_SW,18,LIDT,0},
 /*153(0x99)*/ {"lidt",F_SW,274,LIDT,0},
 /*154(0x9A)*/ {"lidt",F_SW,530,LIDT,0},
 /*155(0x9B)*/ {"lidt",F_SW,786,LIDT,0},
 /*156(0x9C)*/ {"lidt",F_SW,1042,LIDT,0},
 /*157(0x9D)*/ {"lidt",F_SW,1298,LIDT,0},
 /*158(0x9E)*/ {"lidt",F_SW,1554,LIDT,0},
 /*159(0x9F)*/ {"lidt",F_SW,1810,LIDT,0},
 /*160(0xA0)*/ {"smsw",F_W,18,SMSW,0},
 /*161(0xA1)*/ {"smsw",F_W,274,SMSW,0},
 /*162(0xA2)*/ {"smsw",F_W,530,SMSW,0},
 /*163(0xA3)*/ {"smsw",F_W,786,SMSW,0},
 /*164(0xA4)*/ {"smsw",F_W,1042,SMSW,0},
 /*165(0xA5)*/ {"smsw",F_W,1298,SMSW,0},
 /*166(0xA6)*/ {"smsw",F_W,1554,SMSW,0},
 /*167(0xA7)*/ {"smsw",F_W,1810,SMSW,0},
 /*168(0xA8)*/ {0,F_W,18,0,0},
 /*169(0xA9)*/ {0,F_W,274,0,0},
 /*170(0xAA)*/ {0,F_W,530,0,0},
 /*171(0xAB)*/ {0,F_W,786,0,0},
 /*172(0xAC)*/ {0,F_W,1042,0,0},
 /*173(0xAD)*/ {0,F_W,1298,0,0},
 /*174(0xAE)*/ {0,F_W,1554,0,0},
 /*175(0xAF)*/ {0,F_W,1810,0,0},
 /*176(0xB0)*/ {"lmsw",F_W,18,LMSW,0},
 /*177(0xB1)*/ {"lmsw",F_W,274,LMSW,0},
 /*178(0xB2)*/ {"lmsw",F_W,530,LMSW,0},
 /*179(0xB3)*/ {"lmsw",F_W,786,LMSW,0},
 /*180(0xB4)*/ {"lmsw",F_W,1042,LMSW,0},
 /*181(0xB5)*/ {"lmsw",F_W,1298,LMSW,0},
 /*182(0xB6)*/ {"lmsw",F_W,1554,LMSW,0},
 /*183(0xB7)*/ {"lmsw",F_W,1810,LMSW,0},
 /*184(0xB8)*/ {"invlpg",F_W,18,INVLPG,0},
 /*185(0xB9)*/ {"invlpg",F_W,274,INVLPG,0},
 /*186(0xBA)*/ {"invlpg",F_W,530,INVLPG,0},
 /*187(0xBB)*/ {"invlpg",F_W,786,INVLPG,0},
 /*188(0xBC)*/ {"invlpg",F_W,1042,INVLPG,0},
 /*189(0xBD)*/ {"invlpg",F_W,1298,INVLPG,0},
 /*190(0xBE)*/ {"invlpg",F_W,1554,INVLPG,0},
 /*191(0xBF)*/ {"invlpg",F_W,1810,INVLPG,0},
 /*192(0xC0)*/ {"sgdt",F_SW,0,SGDT,0},
 /*193(0xC1)*/ {"sgdt",F_SW,256,SGDT,0},
 /*194(0xC2)*/ {"sgdt",F_SW,512,SGDT,0},
 /*195(0xC3)*/ {"sgdt",F_SW,768,SGDT,0},
 /*196(0xC4)*/ {"sgdt",F_SW,1024,SGDT,0},
 /*197(0xC5)*/ {"sgdt",F_SW,1280,SGDT,0},
 /*198(0xC6)*/ {"sgdt",F_SW,1536,SGDT,0},
 /*199(0xC7)*/ {"sgdt",F_SW,1792,SGDT,0},
 /*200(0xC8)*/ {"sidt",F_SW,0,SIDT,0},
 /*201(0xC9)*/ {"sidt",F_SW,256,SIDT,0},
 /*202(0xCA)*/ {"sidt",F_SW,512,SIDT,0},
 /*203(0xCB)*/ {"sidt",F_SW,768,SIDT,0},
 /*204(0xCC)*/ {"sidt",F_SW,1024,SIDT,0},
 /*205(0xCD)*/ {"sidt",F_SW,1280,SIDT,0},
 /*206(0xCE)*/ {"sidt",F_SW,1536,SIDT,0},
 /*207(0xCF)*/ {"sidt",F_SW,1792,SIDT,0},
 /*208(0xD0)*/ {"lgdt",F_SW,0,LGDT,0},
 /*209(0xD1)*/ {"lgdt",F_SW,256,LGDT,0},
 /*210(0xD2)*/ {"lgdt",F_SW,512,LGDT,0},
 /*211(0xD3)*/ {"lgdt",F_SW,768,LGDT,0},
 /*212(0xD4)*/ {"lgdt",F_SW,1024,LGDT,0},
 /*213(0xD5)*/ {"lgdt",F_SW,1280,LGDT,0},
 /*214(0xD6)*/ {"lgdt",F_SW,1536,LGDT,0},
 /*215(0xD7)*/ {"lgdt",F_SW,1792,LGDT,0},
 /*216(0xD8)*/ {"lidt",F_SW,0,LIDT,0},
 /*217(0xD9)*/ {"lidt",F_SW,256,LIDT,0},
 /*218(0xDA)*/ {"lidt",F_SW,512,LIDT,0},
 /*219(0xDB)*/ {"lidt",F_SW,768,LIDT,0},
 /*220(0xDC)*/ {"lidt",F_SW,1024,LIDT,0},
 /*221(0xDD)*/ {"lidt",F_SW,1280,LIDT,0},
 /*222(0xDE)*/ {"lidt",F_SW,1536,LIDT,0},
 /*223(0xDF)*/ {"lidt",F_SW,1792,LIDT,0},
 /*224(0xE0)*/ {"smsw",F_W,0,SMSW,0},
 /*225(0xE1)*/ {"smsw",F_W,256,SMSW,0},
 /*226(0xE2)*/ {"smsw",F_W,512,SMSW,0},
 /*227(0xE3)*/ {"smsw",F_W,768,SMSW,0},
 /*228(0xE4)*/ {"smsw",F_W,1024,SMSW,0},
 /*229(0xE5)*/ {"smsw",F_W,1280,SMSW,0},
 /*230(0xE6)*/ {"smsw",F_W,1536,SMSW,0},
 /*231(0xE7)*/ {"smsw",F_W,1792,SMSW,0},
 /*232(0xE8)*/ {0,F_W,0,0,0},
 /*233(0xE9)*/ {0,F_W,256,0,0},
 /*234(0xEA)*/ {0,F_W,512,0,0},
 /*235(0xEB)*/ {0,F_W,768,0,0},
 /*236(0xEC)*/ {0,F_W,1024,0,0},
 /*237(0xED)*/ {0,F_W,1280,0,0},
 /*238(0xEE)*/ {0,F_W,1536,0,0},
 /*239(0xEF)*/ {0,F_W,1792,0,0},
 /*240(0xF0)*/ {"lmsw",F_W,0,LMSW,0},
 /*241(0xF1)*/ {"lmsw",F_W,256,LMSW,0},
 /*242(0xF2)*/ {"lmsw",F_W,512,LMSW,0},
 /*243(0xF3)*/ {"lmsw",F_W,768,LMSW,0},
 /*244(0xF4)*/ {"lmsw",F_W,1024,LMSW,0},
 /*245(0xF5)*/ {"lmsw",F_W,1280,LMSW,0},
 /*246(0xF6)*/ {"lmsw",F_W,1536,LMSW,0},
 /*247(0xF7)*/ {"lmsw",F_W,1792,LMSW,0},
 /*248(0xF8)*/ {"invlpg",F_W,0,INVLPG,0},
 /*249(0xF9)*/ {"invlpg",F_W,256,INVLPG,0},
 /*250(0xFA)*/ {"invlpg",F_W,512,INVLPG,0},
 /*251(0xFB)*/ {"invlpg",F_W,768,INVLPG,0},
 /*252(0xFC)*/ {"invlpg",F_W,1024,INVLPG,0},
 /*253(0xFD)*/ {"invlpg",F_W,1280,INVLPG,0},
 /*254(0xFE)*/ {"invlpg",F_W,1536,INVLPG,0},
 /*255(0xFF)*/ {"invlpg",F_W,1792,INVLPG,0}
};

//grupa D8H - instrukcje koprocesora
op_t gD8[256]=
{
 /*0(0x00)*/ {"fadd",0,16,FADD,0},
 /*1(0x01)*/ {"fadd",0,272,FADD,0},
 /*2(0x02)*/ {"fadd",0,528,FADD,0},
 /*3(0x03)*/ {"fadd",0,784,FADD,0},
 /*4(0x04)*/ {"fadd",0,1040,FADD,0},
 /*5(0x05)*/ {"fadd",0,1296,FADD,0},
 /*6(0x06)*/ {"fadd",0,1552,FADD,0},
 /*7(0x07)*/ {"fadd",0,1808,FADD,0},
 /*8(0x08)*/ {"fmul",0,16,FMUL,0},
 /*9(0x09)*/ {"fmul",0,272,FMUL,0},
 /*10(0x0A)*/ {"fmul",0,528,FMUL,0},
 /*11(0x0B)*/ {"fmul",0,784,FMUL,0},
 /*12(0x0C)*/ {"fmul",0,1040,FMUL,0},
 /*13(0x0D)*/ {"fmul",0,1296,FMUL,0},
 /*14(0x0E)*/ {"fmul",0,1552,FMUL,0},
 /*15(0x0F)*/ {"fmul",0,1808,FMUL,0},
 /*16(0x10)*/ {"fcom",0,16,FCOM,0},
 /*17(0x11)*/ {"fcom",0,272,FCOM,0},
 /*18(0x12)*/ {"fcom",0,528,FCOM,0},
 /*19(0x13)*/ {"fcom",0,784,FCOM,0},
 /*20(0x14)*/ {"fcom",0,1040,FCOM,0},
 /*21(0x15)*/ {"fcom",0,1296,FCOM,0},
 /*22(0x16)*/ {"fcom",0,1552,FCOM,0},
 /*23(0x17)*/ {"fcom",0,1808,FCOM,0},
 /*24(0x18)*/ {"fcomp",0,16,FCOMP,0},
 /*25(0x19)*/ {"fcomp",0,272,FCOMP,0},
 /*26(0x1A)*/ {"fcomp",0,528,FCOMP,0},
 /*27(0x1B)*/ {"fcomp",0,784,FCOMP,0},
 /*28(0x1C)*/ {"fcomp",0,1040,FCOMP,0},
 /*29(0x1D)*/ {"fcomp",0,1296,FCOMP,0},
 /*30(0x1E)*/ {"fcomp",0,1552,FCOMP,0},
 /*31(0x1F)*/ {"fcomp",0,1808,FCOMP,0},
 /*32(0x20)*/ {"fsub",0,16,FSUB,0},
 /*33(0x21)*/ {"fsub",0,272,FSUB,0},
 /*34(0x22)*/ {"fsub",0,528,FSUB,0},
 /*35(0x23)*/ {"fsub",0,784,FSUB,0},
 /*36(0x24)*/ {"fsub",0,1040,FSUB,0},
 /*37(0x25)*/ {"fsub",0,1296,FSUB,0},
 /*38(0x26)*/ {"fsub",0,1552,FSUB,0},
 /*39(0x27)*/ {"fsub",0,1808,FSUB,0},
 /*40(0x28)*/ {"fsubr",0,16,FSUBR,0},
 /*41(0x29)*/ {"fsubr",0,272,FSUBR,0},
 /*42(0x2A)*/ {"fsubr",0,528,FSUBR,0},
 /*43(0x2B)*/ {"fsubr",0,784,FSUBR,0},
 /*44(0x2C)*/ {"fsubr",0,1040,FSUBR,0},
 /*45(0x2D)*/ {"fsubr",0,1296,FSUBR,0},
 /*46(0x2E)*/ {"fsubr",0,1552,FSUBR,0},
 /*47(0x2F)*/ {"fsubr",0,1808,FSUBR,0},
 /*48(0x30)*/ {"fdiv",0,16,FDIV,0},
 /*49(0x31)*/ {"fdiv",0,272,FDIV,0},
 /*50(0x32)*/ {"fdiv",0,528,FDIV,0},
 /*51(0x33)*/ {"fdiv",0,784,FDIV,0},
 /*52(0x34)*/ {"fdiv",0,1040,FDIV,0},
 /*53(0x35)*/ {"fdiv",0,1296,FDIV,0},
 /*54(0x36)*/ {"fdiv",0,1552,FDIV,0},
 /*55(0x37)*/ {"fdiv",0,1808,FDIV,0},
 /*56(0x38)*/ {"fdivr",0,16,FDIVR,0},
 /*57(0x39)*/ {"fdivr",0,272,FDIVR,0},
 /*58(0x3A)*/ {"fdivr",0,528,FDIVR,0},
 /*59(0x3B)*/ {"fdivr",0,784,FDIVR,0},
 /*60(0x3C)*/ {"fdivr",0,1040,FDIVR,0},
 /*61(0x3D)*/ {"fdivr",0,1296,FDIVR,0},
 /*62(0x3E)*/ {"fdivr",0,1552,FDIVR,0},
 /*63(0x3F)*/ {"fdivr",0,1808,FDIVR,0},
 /*64(0x40)*/ {"fadd",0,17,FADD,0},
 /*65(0x41)*/ {"fadd",0,273,FADD,0},
 /*66(0x42)*/ {"fadd",0,529,FADD,0},
 /*67(0x43)*/ {"fadd",0,785,FADD,0},
 /*68(0x44)*/ {"fadd",0,1041,FADD,0},
 /*69(0x45)*/ {"fadd",0,1297,FADD,0},
 /*70(0x46)*/ {"fadd",0,1553,FADD,0},
 /*71(0x47)*/ {"fadd",0,1809,FADD,0},
 /*72(0x48)*/ {"fmul",0,17,FMUL,0},
 /*73(0x49)*/ {"fmul",0,273,FMUL,0},
 /*74(0x4A)*/ {"fmul",0,529,FMUL,0},
 /*75(0x4B)*/ {"fmul",0,785,FMUL,0},
 /*76(0x4C)*/ {"fmul",0,1041,FMUL,0},
 /*77(0x4D)*/ {"fmul",0,1297,FMUL,0},
 /*78(0x4E)*/ {"fmul",0,1553,FMUL,0},
 /*79(0x4F)*/ {"fmul",0,1809,FMUL,0},
 /*80(0x50)*/ {"fcom",0,17,FCOM,0},
 /*81(0x51)*/ {"fcom",0,273,FCOM,0},
 /*82(0x52)*/ {"fcom",0,529,FCOM,0},
 /*83(0x53)*/ {"fcom",0,785,FCOM,0},
 /*84(0x54)*/ {"fcom",0,1041,FCOM,0},
 /*85(0x55)*/ {"fcom",0,1297,FCOM,0},
 /*86(0x56)*/ {"fcom",0,1553,FCOM,0},
 /*87(0x57)*/ {"fcom",0,1809,FCOM,0},
 /*88(0x58)*/ {"fcomp",0,17,FCOMP,0},
 /*89(0x59)*/ {"fcomp",0,273,FCOMP,0},
 /*90(0x5A)*/ {"fcomp",0,529,FCOMP,0},
 /*91(0x5B)*/ {"fcomp",0,785,FCOMP,0},
 /*92(0x5C)*/ {"fcomp",0,1041,FCOMP,0},
 /*93(0x5D)*/ {"fcomp",0,1297,FCOMP,0},
 /*94(0x5E)*/ {"fcomp",0,1553,FCOMP,0},
 /*95(0x5F)*/ {"fcomp",0,1809,FCOMP,0},
 /*96(0x60)*/ {"fsub",0,17,FSUB,0},
 /*97(0x61)*/ {"fsub",0,273,FSUB,0},
 /*98(0x62)*/ {"fsub",0,529,FSUB,0},
 /*99(0x63)*/ {"fsub",0,785,FSUB,0},
 /*100(0x64)*/ {"fsub",0,1041,FSUB,0},
 /*101(0x65)*/ {"fsub",0,1297,FSUB,0},
 /*102(0x66)*/ {"fsub",0,1553,FSUB,0},
 /*103(0x67)*/ {"fsub",0,1809,FSUB,0},
 /*104(0x68)*/ {"fsubr",0,17,FSUBR,0},
 /*105(0x69)*/ {"fsubr",0,273,FSUBR,0},
 /*106(0x6A)*/ {"fsubr",0,529,FSUBR,0},
 /*107(0x6B)*/ {"fsubr",0,785,FSUBR,0},
 /*108(0x6C)*/ {"fsubr",0,1041,FSUBR,0},
 /*109(0x6D)*/ {"fsubr",0,1297,FSUBR,0},
 /*110(0x6E)*/ {"fsubr",0,1553,FSUBR,0},
 /*111(0x6F)*/ {"fsubr",0,1809,FSUBR,0},
 /*112(0x70)*/ {"fdiv",0,17,FDIV,0},
 /*113(0x71)*/ {"fdiv",0,273,FDIV,0},
 /*114(0x72)*/ {"fdiv",0,529,FDIV,0},
 /*115(0x73)*/ {"fdiv",0,785,FDIV,0},
 /*116(0x74)*/ {"fdiv",0,1041,FDIV,0},
 /*117(0x75)*/ {"fdiv",0,1297,FDIV,0},
 /*118(0x76)*/ {"fdiv",0,1553,FDIV,0},
 /*119(0x77)*/ {"fdiv",0,1809,FDIV,0},
 /*120(0x78)*/ {"fdivr",0,17,FDIVR,0},
 /*121(0x79)*/ {"fdivr",0,273,FDIVR,0},
 /*122(0x7A)*/ {"fdivr",0,529,FDIVR,0},
 /*123(0x7B)*/ {"fdivr",0,785,FDIVR,0},
 /*124(0x7C)*/ {"fdivr",0,1041,FDIVR,0},
 /*125(0x7D)*/ {"fdivr",0,1297,FDIVR,0},
 /*126(0x7E)*/ {"fdivr",0,1553,FDIVR,0},
 /*127(0x7F)*/ {"fdivr",0,1809,FDIVR,0},
 /*128(0x80)*/ {"fadd",0,18,FADD,0},
 /*129(0x81)*/ {"fadd",0,274,FADD,0},
 /*130(0x82)*/ {"fadd",0,530,FADD,0},
 /*131(0x83)*/ {"fadd",0,786,FADD,0},
 /*132(0x84)*/ {"fadd",0,1042,FADD,0},
 /*133(0x85)*/ {"fadd",0,1298,FADD,0},
 /*134(0x86)*/ {"fadd",0,1554,FADD,0},
 /*135(0x87)*/ {"fadd",0,1810,FADD,0},
 /*136(0x88)*/ {"fmul",0,18,FMUL,0},
 /*137(0x89)*/ {"fmul",0,274,FMUL,0},
 /*138(0x8A)*/ {"fmul",0,530,FMUL,0},
 /*139(0x8B)*/ {"fmul",0,786,FMUL,0},
 /*140(0x8C)*/ {"fmul",0,1042,FMUL,0},
 /*141(0x8D)*/ {"fmul",0,1298,FMUL,0},
 /*142(0x8E)*/ {"fmul",0,1554,FMUL,0},
 /*143(0x8F)*/ {"fmul",0,1810,FMUL,0},
 /*144(0x90)*/ {"fcom",0,18,FCOM,0},
 /*145(0x91)*/ {"fcom",0,274,FCOM,0},
 /*146(0x92)*/ {"fcom",0,530,FCOM,0},
 /*147(0x93)*/ {"fcom",0,786,FCOM,0},
 /*148(0x94)*/ {"fcom",0,1042,FCOM,0},
 /*149(0x95)*/ {"fcom",0,1298,FCOM,0},
 /*150(0x96)*/ {"fcom",0,1554,FCOM,0},
 /*151(0x97)*/ {"fcom",0,1810,FCOM,0},
 /*152(0x98)*/ {"fcomp",0,18,FCOMP,0},
 /*153(0x99)*/ {"fcomp",0,274,FCOMP,0},
 /*154(0x9A)*/ {"fcomp",0,530,FCOMP,0},
 /*155(0x9B)*/ {"fcomp",0,786,FCOMP,0},
 /*156(0x9C)*/ {"fcomp",0,1042,FCOMP,0},
 /*157(0x9D)*/ {"fcomp",0,1298,FCOMP,0},
 /*158(0x9E)*/ {"fcomp",0,1554,FCOMP,0},
 /*159(0x9F)*/ {"fcomp",0,1810,FCOMP,0},
 /*160(0xA0)*/ {"fsub",0,18,FSUB,0},
 /*161(0xA1)*/ {"fsub",0,274,FSUB,0},
 /*162(0xA2)*/ {"fsub",0,530,FSUB,0},
 /*163(0xA3)*/ {"fsub",0,786,FSUB,0},
 /*164(0xA4)*/ {"fsub",0,1042,FSUB,0},
 /*165(0xA5)*/ {"fsub",0,1298,FSUB,0},
 /*166(0xA6)*/ {"fsub",0,1554,FSUB,0},
 /*167(0xA7)*/ {"fsub",0,1810,FSUB,0},
 /*168(0xA8)*/ {"fsubr",0,18,FSUBR,0},
 /*169(0xA9)*/ {"fsubr",0,274,FSUBR,0},
 /*170(0xAA)*/ {"fsubr",0,530,FSUBR,0},
 /*171(0xAB)*/ {"fsubr",0,786,FSUBR,0},
 /*172(0xAC)*/ {"fsubr",0,1042,FSUBR,0},
 /*173(0xAD)*/ {"fsubr",0,1298,FSUBR,0},
 /*174(0xAE)*/ {"fsubr",0,1554,FSUBR,0},
 /*175(0xAF)*/ {"fsubr",0,1810,FSUBR,0},
 /*176(0xB0)*/ {"fdiv",0,18,FDIV,0},
 /*177(0xB1)*/ {"fdiv",0,274,FDIV,0},
 /*178(0xB2)*/ {"fdiv",0,530,FDIV,0},
 /*179(0xB3)*/ {"fdiv",0,786,FDIV,0},
 /*180(0xB4)*/ {"fdiv",0,1042,FDIV,0},
 /*181(0xB5)*/ {"fdiv",0,1298,FDIV,0},
 /*182(0xB6)*/ {"fdiv",0,1554,FDIV,0},
 /*183(0xB7)*/ {"fdiv",0,1810,FDIV,0},
 /*184(0xB8)*/ {"fdivr",0,18,FDIVR,0},
 /*185(0xB9)*/ {"fdivr",0,274,FDIVR,0},
 /*186(0xBA)*/ {"fdivr",0,530,FDIVR,0},
 /*187(0xBB)*/ {"fdivr",0,786,FDIVR,0},
 /*188(0xBC)*/ {"fdivr",0,1042,FDIVR,0},
 /*189(0xBD)*/ {"fdivr",0,1298,FDIVR,0},
 /*190(0xBE)*/ {"fdivr",0,1554,FDIVR,0},
 /*191(0xBF)*/ {"fdivr",0,1810,FDIVR,0},
 /*192(0xC0)*/ {"fadd",F_CP,131074,FADD,0},
 /*193(0xC1)*/ {"fadd",F_CP,16908290,FADD,0},
 /*194(0xC2)*/ {"fadd",F_CP,33685506,FADD,0},
 /*195(0xC3)*/ {"fadd",F_CP,50462722,FADD,0},
 /*196(0xC4)*/ {"fadd",F_CP,67239938,FADD,0},
 /*197(0xC5)*/ {"fadd",F_CP,84017154,FADD,0},
 /*198(0xC6)*/ {"fadd",F_CP,100794370,FADD,0},
 /*199(0xC7)*/ {"fadd",F_CP,117571586,FADD,0},
 /*200(0xC8)*/ {"fmul",F_CP,131074,FMUL,0},
 /*201(0xC9)*/ {"fmul",F_CP,16908290,FMUL,0},
 /*202(0xCA)*/ {"fmul",F_CP,33685506,FMUL,0},
 /*203(0xCB)*/ {"fmul",F_CP,50462722,FMUL,0},
 /*204(0xCC)*/ {"fmul",F_CP,67239938,FMUL,0},
 /*205(0xCD)*/ {"fmul",F_CP,84017154,FMUL,0},
 /*206(0xCE)*/ {"fmul",F_CP,100794370,FMUL,0},
 /*207(0xCF)*/ {"fmul",F_CP,117571586,FMUL,0},
 /*208(0xD0)*/ {"fcom",F_CP,131074,FCOM,0},
 /*209(0xD1)*/ {"fcom",F_CP,16908290,FCOM,0},
 /*210(0xD2)*/ {"fcom",F_CP,33685506,FCOM,0},
 /*211(0xD3)*/ {"fcom",F_CP,50462722,FCOM,0},
 /*212(0xD4)*/ {"fcom",F_CP,67239938,FCOM,0},
 /*213(0xD5)*/ {"fcom",F_CP,84017154,FCOM,0},
 /*214(0xD6)*/ {"fcom",F_CP,100794370,FCOM,0},
 /*215(0xD7)*/ {"fcom",F_CP,117571586,FCOM,0},
 /*216(0xD8)*/ {"fcomp",F_CP,131074,FCOMP,0},
 /*217(0xD9)*/ {"fcomp",F_CP,16908290,FCOMP,0},
 /*218(0xDA)*/ {"fcomp",F_CP,33685506,FCOMP,0},
 /*219(0xDB)*/ {"fcomp",F_CP,50462722,FCOMP,0},
 /*220(0xDC)*/ {"fcomp",F_CP,67239938,FCOMP,0},
 /*221(0xDD)*/ {"fcomp",F_CP,84017154,FCOMP,0},
 /*222(0xDE)*/ {"fcomp",F_CP,100794370,FCOMP,0},
 /*223(0xDF)*/ {"fcomp",F_CP,117571586,FCOMP,0},
 /*224(0xE0)*/ {"fsub",F_CP,131074,FSUB,0},
 /*225(0xE1)*/ {"fsub",F_CP,16908290,FSUB,0},
 /*226(0xE2)*/ {"fsub",F_CP,33685506,FSUB,0},
 /*227(0xE3)*/ {"fsub",F_CP,50462722,FSUB,0},
 /*228(0xE4)*/ {"fsub",F_CP,67239938,FSUB,0},
 /*229(0xE5)*/ {"fsub",F_CP,84017154,FSUB,0},
 /*230(0xE6)*/ {"fsub",F_CP,100794370,FSUB,0},
 /*231(0xE7)*/ {"fsub",F_CP,117571586,FSUB,0},
 /*232(0xE8)*/ {"fsubr",F_CP,131074,FSUBR,0},
 /*233(0xE9)*/ {"fsubr",F_CP,16908290,FSUBR,0},
 /*234(0xEA)*/ {"fsubr",F_CP,33685506,FSUBR,0},
 /*235(0xEB)*/ {"fsubr",F_CP,50462722,FSUBR,0},
 /*236(0xEC)*/ {"fsubr",F_CP,67239938,FSUBR,0},
 /*237(0xED)*/ {"fsubr",F_CP,84017154,FSUBR,0},
 /*238(0xEE)*/ {"fsubr",F_CP,100794370,FSUBR,0},
 /*239(0xEF)*/ {"fsubr",F_CP,117571586,FSUBR,0},
 /*240(0xF0)*/ {"fdiv",F_CP,131074,FDIV,0},
 /*241(0xF1)*/ {"fdiv",F_CP,16908290,FDIV,0},
 /*242(0xF2)*/ {"fdiv",F_CP,33685506,FDIV,0},
 /*243(0xF3)*/ {"fdiv",F_CP,50462722,FDIV,0},
 /*244(0xF4)*/ {"fdiv",F_CP,67239938,FDIV,0},
 /*245(0xF5)*/ {"fdiv",F_CP,84017154,FDIV,0},
 /*246(0xF6)*/ {"fdiv",F_CP,100794370,FDIV,0},
 /*247(0xF7)*/ {"fdiv",F_CP,117571586,FDIV,0},
 /*248(0xF8)*/ {"fdivr",F_CP,131074,FDIVR,0},
 /*249(0xF9)*/ {"fdivr",F_CP,16908290,FDIVR,0},
 /*250(0xFA)*/ {"fdivr",F_CP,33685506,FDIVR,0},
 /*251(0xFB)*/ {"fdivr",F_CP,50462722,FDIVR,0},
 /*252(0xFC)*/ {"fdivr",F_CP,67239938,FDIVR,0},
 /*253(0xFD)*/ {"fdivr",F_CP,84017154,FDIVR,0},
 /*254(0xFE)*/ {"fdivr",F_CP,100794370,FDIVR,0},
 /*255(0xFF)*/ {"fdivr",F_CP,117571586,FDIVR,0}
};

//grupa D9H - instrukcje koprocesora
op_t gD9[256]=
{
 /*0(0x00)*/ {"fld",F_PAR+F_DW,16,FLD,0},
 /*1(0x01)*/ {"fld",F_PAR+F_DW,272,FLD,0},
 /*2(0x02)*/ {"fld",F_PAR+F_DW,528,FLD,0},
 /*3(0x03)*/ {"fld",F_PAR+F_DW,784,FLD,0},
 /*4(0x04)*/ {"fld",F_PAR+F_DW,1040,FLD,0},
 /*5(0x05)*/ {"fld",F_PAR+F_DW,1296,FLD,0},
 /*6(0x06)*/ {"fld",F_PAR+F_DW,1552,FLD,0},
 /*7(0x07)*/ {"fld",F_PAR+F_DW,1808,FLD,0},
 /*8(0x08)*/ {0,F_PAR+F_DW,16,0,0},
 /*9(0x09)*/ {0,F_PAR+F_DW,272,0,0},
 /*10(0x0A)*/ {0,F_PAR+F_DW,528,0,0},
 /*11(0x0B)*/ {0,F_PAR+F_DW,784,0,0},
 /*12(0x0C)*/ {0,F_PAR+F_DW,1040,0,0},
 /*13(0x0D)*/ {0,F_PAR+F_DW,1296,0,0},
 /*14(0x0E)*/ {0,F_PAR+F_DW,1552,0,0},
 /*15(0x0F)*/ {0,F_PAR+F_DW,1808,0,0},
 /*16(0x10)*/ {"fst",F_PAR+F_DW,16,FST,0},
 /*17(0x11)*/ {"fst",F_PAR+F_DW,272,FST,0},
 /*18(0x12)*/ {"fst",F_PAR+F_DW,528,FST,0},
 /*19(0x13)*/ {"fst",F_PAR+F_DW,784,FST,0},
 /*20(0x14)*/ {"fst",F_PAR+F_DW,1040,FST,0},
 /*21(0x15)*/ {"fst",F_PAR+F_DW,1296,FST,0},
 /*22(0x16)*/ {"fst",F_PAR+F_DW,1552,FST,0},
 /*23(0x17)*/ {"fst",F_PAR+F_DW,1808,FST,0},
 /*24(0x18)*/ {"fstp",F_PAR+F_DW,16,FSTP,0},
 /*25(0x19)*/ {"fstp",F_PAR+F_DW,272,FSTP,0},
 /*26(0x1A)*/ {"fstp",F_PAR+F_DW,528,FSTP,0},
 /*27(0x1B)*/ {"fstp",F_PAR+F_DW,784,FSTP,0},
 /*28(0x1C)*/ {"fstp",F_PAR+F_DW,1040,FSTP,0},
 /*29(0x1D)*/ {"fstp",F_PAR+F_DW,1296,FSTP,0},
 /*30(0x1E)*/ {"fstp",F_PAR+F_DW,1552,FSTP,0},
 /*31(0x1F)*/ {"fstp",F_PAR+F_DW,1808,FSTP,0},
 /*32(0x20)*/ {"fldenv",F_PAR,16,FLDENV,0},
 /*33(0x21)*/ {"fldenv",F_PAR,272,FLDENV,0},
 /*34(0x22)*/ {"fldenv",F_PAR,528,FLDENV,0},
 /*35(0x23)*/ {"fldenv",F_PAR,784,FLDENV,0},
 /*36(0x24)*/ {"fldenv",F_PAR,1040,FLDENV,0},
 /*37(0x25)*/ {"fldenv",F_PAR,1296,FLDENV,0},
 /*38(0x26)*/ {"fldenv",F_PAR,1552,FLDENV,0},
 /*39(0x27)*/ {"fldenv",F_PAR,1808,FLDENV,0},
 /*40(0x28)*/ {"fldcw",F_PAR,16,FLDCW,0},
 /*41(0x29)*/ {"fldcw",F_PAR,272,FLDCW,0},
 /*42(0x2A)*/ {"fldcw",F_PAR,528,FLDCW,0},
 /*43(0x2B)*/ {"fldcw",F_PAR,784,FLDCW,0},
 /*44(0x2C)*/ {"fldcw",F_PAR,1040,FLDCW,0},
 /*45(0x2D)*/ {"fldcw",F_PAR,1296,FLDCW,0},
 /*46(0x2E)*/ {"fldcw",F_PAR,1552,FLDCW,0},
 /*47(0x2F)*/ {"fldcw",F_PAR,1808,FLDCW,0},
 /*48(0x30)*/ {"fstenv",F_PAR,16,FSTENV,0},
 /*49(0x31)*/ {"fstenv",F_PAR,272,FSTENV,0},
 /*50(0x32)*/ {"fstenv",F_PAR,528,FSTENV,0},
 /*51(0x33)*/ {"fstenv",F_PAR,784,FSTENV,0},
 /*52(0x34)*/ {"fstenv",F_PAR,1040,FSTENV,0},
 /*53(0x35)*/ {"fstenv",F_PAR,1296,FSTENV,0},
 /*54(0x36)*/ {"fstenv",F_PAR,1552,FSTENV,0},
 /*55(0x37)*/ {"fstenv",F_PAR,1808,FSTENV,0},
 /*56(0x38)*/ {"fstcw",F_PAR,16,FSTCW,0},
 /*57(0x39)*/ {"fstcw",F_PAR,272,FSTCW,0},
 /*58(0x3A)*/ {"fstcw",F_PAR,528,FSTCW,0},
 /*59(0x3B)*/ {"fstcw",F_PAR,784,FSTCW,0},
 /*60(0x3C)*/ {"fstcw",F_PAR,1040,FSTCW,0},
 /*61(0x3D)*/ {"fstcw",F_PAR,1296,FSTCW,0},
 /*62(0x3E)*/ {"fstcw",F_PAR,1552,FSTCW,0},
 /*63(0x3F)*/ {"fstcw",F_PAR,1808,FSTCW,0},
 /*64(0x40)*/ {"fld",F_PAR+F_DW,17,FLD,0},
 /*65(0x41)*/ {"fld",F_PAR+F_DW,273,FLD,0},
 /*66(0x42)*/ {"fld",F_PAR+F_DW,529,FLD,0},
 /*67(0x43)*/ {"fld",F_PAR+F_DW,785,FLD,0},
 /*68(0x44)*/ {"fld",F_PAR+F_DW,1041,FLD,0},
 /*69(0x45)*/ {"fld",F_PAR+F_DW,1297,FLD,0},
 /*70(0x46)*/ {"fld",F_PAR+F_DW,1553,FLD,0},
 /*71(0x47)*/ {"fld",F_PAR+F_DW,1809,FLD,0},
 /*72(0x48)*/ {0,F_PAR+F_DW,17,0,0},
 /*73(0x49)*/ {0,F_PAR+F_DW,273,0,0},
 /*74(0x4A)*/ {0,F_PAR+F_DW,529,0,0},
 /*75(0x4B)*/ {0,F_PAR+F_DW,785,0,0},
 /*76(0x4C)*/ {0,F_PAR+F_DW,1041,0,0},
 /*77(0x4D)*/ {0,F_PAR+F_DW,1297,0,0},
 /*78(0x4E)*/ {0,F_PAR+F_DW,1553,0,0},
 /*79(0x4F)*/ {0,F_PAR+F_DW,1809,0,0},
 /*80(0x50)*/ {"fst",F_PAR+F_DW,17,FST,0},
 /*81(0x51)*/ {"fst",F_PAR+F_DW,273,FST,0},
 /*82(0x52)*/ {"fst",F_PAR+F_DW,529,FST,0},
 /*83(0x53)*/ {"fst",F_PAR+F_DW,785,FST,0},
 /*84(0x54)*/ {"fst",F_PAR+F_DW,1041,FST,0},
 /*85(0x55)*/ {"fst",F_PAR+F_DW,1297,FST,0},
 /*86(0x56)*/ {"fst",F_PAR+F_DW,1553,FST,0},
 /*87(0x57)*/ {"fst",F_PAR+F_DW,1809,FST,0},
 /*88(0x58)*/ {"fstp",F_PAR+F_DW,17,FSTP,0},
 /*89(0x59)*/ {"fstp",F_PAR+F_DW,273,FSTP,0},
 /*90(0x5A)*/ {"fstp",F_PAR+F_DW,529,FSTP,0},
 /*91(0x5B)*/ {"fstp",F_PAR+F_DW,785,FSTP,0},
 /*92(0x5C)*/ {"fstp",F_PAR+F_DW,1041,FSTP,0},
 /*93(0x5D)*/ {"fstp",F_PAR+F_DW,1297,FSTP,0},
 /*94(0x5E)*/ {"fstp",F_PAR+F_DW,1553,FSTP,0},
 /*95(0x5F)*/ {"fstp",F_PAR+F_DW,1809,FSTP,0},
 /*96(0x60)*/ {"fldenv",F_PAR,17,FLDENV,0},
 /*97(0x61)*/ {"fldenv",F_PAR,273,FLDENV,0},
 /*98(0x62)*/ {"fldenv",F_PAR,529,FLDENV,0},
 /*99(0x63)*/ {"fldenv",F_PAR,785,FLDENV,0},
 /*100(0x64)*/ {"fldenv",F_PAR,1041,FLDENV,0},
 /*101(0x65)*/ {"fldenv",F_PAR,1297,FLDENV,0},
 /*102(0x66)*/ {"fldenv",F_PAR,1553,FLDENV,0},
 /*103(0x67)*/ {"fldenv",F_PAR,1809,FLDENV,0},
 /*104(0x68)*/ {"fldcw",F_PAR,17,FLDCW,0},
 /*105(0x69)*/ {"fldcw",F_PAR,273,FLDCW,0},
 /*106(0x6A)*/ {"fldcw",F_PAR,529,FLDCW,0},
 /*107(0x6B)*/ {"fldcw",F_PAR,785,FLDCW,0},
 /*108(0x6C)*/ {"fldcw",F_PAR,1041,FLDCW,0},
 /*109(0x6D)*/ {"fldcw",F_PAR,1297,FLDCW,0},
 /*110(0x6E)*/ {"fldcw",F_PAR,1553,FLDCW,0},
 /*111(0x6F)*/ {"fldcw",F_PAR,1809,FLDCW,0},
 /*112(0x70)*/ {"fstenv",F_PAR,17,FSTENV,0},
 /*113(0x71)*/ {"fstenv",F_PAR,273,FSTENV,0},
 /*114(0x72)*/ {"fstenv",F_PAR,529,FSTENV,0},
 /*115(0x73)*/ {"fstenv",F_PAR,785,FSTENV,0},
 /*116(0x74)*/ {"fstenv",F_PAR,1041,FSTENV,0},
 /*117(0x75)*/ {"fstenv",F_PAR,1297,FSTENV,0},
 /*118(0x76)*/ {"fstenv",F_PAR,1553,FSTENV,0},
 /*119(0x77)*/ {"fstenv",F_PAR,1809,FSTENV,0},
 /*120(0x78)*/ {"fstcw",F_PAR,17,FSTCW,0},
 /*121(0x79)*/ {"fstcw",F_PAR,273,FSTCW,0},
 /*122(0x7A)*/ {"fstcw",F_PAR,529,FSTCW,0},
 /*123(0x7B)*/ {"fstcw",F_PAR,785,FSTCW,0},
 /*124(0x7C)*/ {"fstcw",F_PAR,1041,FSTCW,0},
 /*125(0x7D)*/ {"fstcw",F_PAR,1297,FSTCW,0},
 /*126(0x7E)*/ {"fstcw",F_PAR,1553,FSTCW,0},
 /*127(0x7F)*/ {"fstcw",F_PAR,1809,FSTCW,0},
 /*128(0x80)*/ {"fld",F_PAR+F_DW,18,FLD,0},
 /*129(0x81)*/ {"fld",F_PAR+F_DW,274,FLD,0},
 /*130(0x82)*/ {"fld",F_PAR+F_DW,530,FLD,0},
 /*131(0x83)*/ {"fld",F_PAR+F_DW,786,FLD,0},
 /*132(0x84)*/ {"fld",F_PAR+F_DW,1042,FLD,0},
 /*133(0x85)*/ {"fld",F_PAR+F_DW,1298,FLD,0},
 /*134(0x86)*/ {"fld",F_PAR+F_DW,1554,FLD,0},
 /*135(0x87)*/ {"fld",F_PAR+F_DW,1810,FLD,0},
 /*136(0x88)*/ {0,F_PAR+F_DW,18,0,0},
 /*137(0x89)*/ {0,F_PAR+F_DW,274,0,0},
 /*138(0x8A)*/ {0,F_PAR+F_DW,530,0,0},
 /*139(0x8B)*/ {0,F_PAR+F_DW,786,0,0},
 /*140(0x8C)*/ {0,F_PAR+F_DW,1042,0,0},
 /*141(0x8D)*/ {0,F_PAR+F_DW,1298,0,0},
 /*142(0x8E)*/ {0,F_PAR+F_DW,1554,0,0},
 /*143(0x8F)*/ {0,F_PAR+F_DW,1810,0,0},
 /*144(0x90)*/ {"fst",F_PAR+F_DW,18,FST,0},
 /*145(0x91)*/ {"fst",F_PAR+F_DW,274,FST,0},
 /*146(0x92)*/ {"fst",F_PAR+F_DW,530,FST,0},
 /*147(0x93)*/ {"fst",F_PAR+F_DW,786,FST,0},
 /*148(0x94)*/ {"fst",F_PAR+F_DW,1042,FST,0},
 /*149(0x95)*/ {"fst",F_PAR+F_DW,1298,FST,0},
 /*150(0x96)*/ {"fst",F_PAR+F_DW,1554,FST,0},
 /*151(0x97)*/ {"fst",F_PAR+F_DW,1810,FST,0},
 /*152(0x98)*/ {"fstp",F_PAR+F_DW,18,FSTP,0},
 /*153(0x99)*/ {"fstp",F_PAR+F_DW,274,FSTP,0},
 /*154(0x9A)*/ {"fstp",F_PAR+F_DW,530,FSTP,0},
 /*155(0x9B)*/ {"fstp",F_PAR+F_DW,786,FSTP,0},
 /*156(0x9C)*/ {"fstp",F_PAR+F_DW,1042,FSTP,0},
 /*157(0x9D)*/ {"fstp",F_PAR+F_DW,1298,FSTP,0},
 /*158(0x9E)*/ {"fstp",F_PAR+F_DW,1554,FSTP,0},
 /*159(0x9F)*/ {"fstp",F_PAR+F_DW,1810,FSTP,0},
 /*160(0xA0)*/ {"fldenv",F_PAR,18,FLDENV,0},
 /*161(0xA1)*/ {"fldenv",F_PAR,274,FLDENV,0},
 /*162(0xA2)*/ {"fldenv",F_PAR,530,FLDENV,0},
 /*163(0xA3)*/ {"fldenv",F_PAR,786,FLDENV,0},
 /*164(0xA4)*/ {"fldenv",F_PAR,1042,FLDENV,0},
 /*165(0xA5)*/ {"fldenv",F_PAR,1298,FLDENV,0},
 /*166(0xA6)*/ {"fldenv",F_PAR,1554,FLDENV,0},
 /*167(0xA7)*/ {"fldenv",F_PAR,1810,FLDENV,0},
 /*168(0xA8)*/ {"fldcw",F_PAR,18,FLDCW,0},
 /*169(0xA9)*/ {"fldcw",F_PAR,274,FLDCW,0},
 /*170(0xAA)*/ {"fldcw",F_PAR,530,FLDCW,0},
 /*171(0xAB)*/ {"fldcw",F_PAR,786,FLDCW,0},
 /*172(0xAC)*/ {"fldcw",F_PAR,1042,FLDCW,0},
 /*173(0xAD)*/ {"fldcw",F_PAR,1298,FLDCW,0},
 /*174(0xAE)*/ {"fldcw",F_PAR,1554,FLDCW,0},
 /*175(0xAF)*/ {"fldcw",F_PAR,1810,FLDCW,0},
 /*176(0xB0)*/ {"fstenv",F_PAR,18,FSTENV,0},
 /*177(0xB1)*/ {"fstenv",F_PAR,274,FSTENV,0},
 /*178(0xB2)*/ {"fstenv",F_PAR,530,FSTENV,0},
 /*179(0xB3)*/ {"fstenv",F_PAR,786,FSTENV,0},
 /*180(0xB4)*/ {"fstenv",F_PAR,1042,FSTENV,0},
 /*181(0xB5)*/ {"fstenv",F_PAR,1298,FSTENV,0},
 /*182(0xB6)*/ {"fstenv",F_PAR,1554,FSTENV,0},
 /*183(0xB7)*/ {"fstenv",F_PAR,1810,FSTENV,0},
 /*184(0xB8)*/ {"fstcw",F_PAR,18,FSTCW,0},
 /*185(0xB9)*/ {"fstcw",F_PAR,274,FSTCW,0},
 /*186(0xBA)*/ {"fstcw",F_PAR,530,FSTCW,0},
 /*187(0xBB)*/ {"fstcw",F_PAR,786,FSTCW,0},
 /*188(0xBC)*/ {"fstcw",F_PAR,1042,FSTCW,0},
 /*189(0xBD)*/ {"fstcw",F_PAR,1298,FSTCW,0},
 /*190(0xBE)*/ {"fstcw",F_PAR,1554,FSTCW,0},
 /*191(0xBF)*/ {"fstcw",F_PAR,1810,FSTCW,0},
 /*192(0xC0)*/ {"fld",F_PAR,2,FLD,0},
 /*193(0xC1)*/ {"fld",F_PAR,258,FLD,0},
 /*194(0xC2)*/ {"fld",F_PAR,514,FLD,0},
 /*195(0xC3)*/ {"fld",F_PAR,770,FLD,0},
 /*196(0xC4)*/ {"fld",F_PAR,1026,FLD,0},
 /*197(0xC5)*/ {"fld",F_PAR,1282,FLD,0},
 /*198(0xC6)*/ {"fld",F_PAR,1538,FLD,0},
 /*199(0xC7)*/ {"fld",F_PAR,1794,FLD,0},
 /*200(0xC8)*/ {"fxch",F_PAR,2,FXCH,0},
 /*201(0xC9)*/ {"fxch",F_PAR,258,FXCH,0},
 /*202(0xCA)*/ {"fxch",F_PAR,514,FXCH,0},
 /*203(0xCB)*/ {"fxch",F_PAR,770,FXCH,0},
 /*204(0xCC)*/ {"fxch",F_PAR,1026,FXCH,0},
 /*205(0xCD)*/ {"fxch",F_PAR,1282,FXCH,0},
 /*206(0xCE)*/ {"fxch",F_PAR,1538,FXCH,0},
 /*207(0xCF)*/ {"fxch",F_PAR,1794,FXCH,0},
 /*208(0xD0)*/ {"fnop",0,0,NOP,0},
 /*209(0xD1)*/ {0,0,0,0,0},
 /*210(0xD2)*/ {0,0,0,0,0},
 /*211(0xD3)*/ {0,0,0,0,0},
 /*212(0xD4)*/ {0,0,0,0,0},
 /*213(0xD5)*/ {0,0,0,0,0},
 /*214(0xD6)*/ {0,0,0,0,0},
 /*215(0xD7)*/ {0,0,0,0,0},
 /*216(0xD8)*/ {0,0,0,0,0},
 /*217(0xD9)*/ {0,0,0,0,0},
 /*218(0xDA)*/ {0,0,0,0,0},
 /*219(0xDB)*/ {0,0,0,0,0},
 /*220(0xDC)*/ {0,0,0,0,0},
 /*221(0xDD)*/ {0,0,0,0,0},
 /*222(0xDE)*/ {0,0,0,0,0},
 /*223(0xDF)*/ {0,0,0,0,0},
 /*224(0xE0)*/ {"fchs",0,0,FCHS,0},
 /*225(0xE1)*/ {"fabs",0,0,FABS,0},
 /*226(0xE2)*/ {0,0,0,0,0},
 /*227(0xE3)*/ {0,0,0,0,0},
 /*228(0xE4)*/ {"ftst",0,0,FTST,0},
 /*229(0xE5)*/ {"fxam",0,0,FXAM,0},
 /*230(0xE6)*/ {0,0,0,0,0},
 /*231(0xE7)*/ {0,0,0,0,0},
 /*232(0xE8)*/ {"fld1",0,0,FLD1,0},
 /*233(0xE9)*/ {"fldl2t",0,0,FLDL2T,0},
 /*234(0xEA)*/ {"fldl2e",0,0,FLDL2E,0},
 /*235(0xEB)*/ {"fldpi",0,0,FLDPI,0},
 /*236(0xEC)*/ {"fldlg2",0,0,FLDLG2,0},
 /*237(0xED)*/ {"fldln2",0,0,FLDLN2,0},
 /*238(0xEE)*/ {"fldz",0,0,FLDZ,0},
 /*239(0xEF)*/ {0,0,0,0,0},
 /*240(0xF0)*/ {0,0,0,0,0},
 /*241(0xF1)*/ {"fyl2x",0,0,FYL2X,0},
 /*242(0xF2)*/ {"fptan",0,0,FPTAN,0},
 /*243(0xF3)*/ {"fpatan",0,0,FPATAN,0},
 /*244(0xF4)*/ {"fxtract",0,0,FXTRACT,0},
 /*245(0xF5)*/ {"fprem1",0,0,FPREM1,0},
 /*246(0xF6)*/ {"fdecstp",0,0,FDECSTP,0},
 /*247(0xF7)*/ {"fincstp",0,0,FINCSTP,0},
 /*248(0xF8)*/ {"fprem",0,0,FPREM,0},
 /*249(0xF9)*/ {"fyl2xp1",0,0,FYL2XP1,0},
 /*250(0xFA)*/ {"fsqrt",0,0,FSQRT,0},
 /*251(0xFB)*/ {"fsincos",0,0,FSINCOS,0},
 /*252(0xFC)*/ {"frndint",0,0,FRNDINT,0},
 /*253(0xFD)*/ {"fscale",0,0,FSCALE,0},
 /*254(0xFE)*/ {"fsin",0,0,FSIN,0},
 /*255(0xFF)*/ {"fcos",0,0,FCOS,0}
};

//grupa DAH - instrukcje koprocesora
op_t gDA[256]=
{
 /*0(0x00)*/ {"fiadd",F_PAR+F_DW,16,FIADD,0},
 /*1(0x01)*/ {"fiadd",F_PAR+F_DW,272,FIADD,0},
 /*2(0x02)*/ {"fiadd",F_PAR+F_DW,528,FIADD,0},
 /*3(0x03)*/ {"fiadd",F_PAR+F_DW,784,FIADD,0},
 /*4(0x04)*/ {"fiadd",F_PAR+F_DW,1040,FIADD,0},
 /*5(0x05)*/ {"fiadd",F_PAR+F_DW,1296,FIADD,0},
 /*6(0x06)*/ {"fiadd",F_PAR+F_DW,1552,FIADD,0},
 /*7(0x07)*/ {"fiadd",F_PAR+F_DW,1808,FIADD,0},
 /*8(0x08)*/ {"fimul",F_PAR+F_DW,16,FIMUL,0},
 /*9(0x09)*/ {"fimul",F_PAR+F_DW,272,FIMUL,0},
 /*10(0x0A)*/ {"fimul",F_PAR+F_DW,528,FIMUL,0},
 /*11(0x0B)*/ {"fimul",F_PAR+F_DW,784,FIMUL,0},
 /*12(0x0C)*/ {"fimul",F_PAR+F_DW,1040,FIMUL,0},
 /*13(0x0D)*/ {"fimul",F_PAR+F_DW,1296,FIMUL,0},
 /*14(0x0E)*/ {"fimul",F_PAR+F_DW,1552,FIMUL,0},
 /*15(0x0F)*/ {"fimul",F_PAR+F_DW,1808,FIMUL,0},
 /*16(0x10)*/ {"ficom",F_PAR+F_DW,16,FICOM,0},
 /*17(0x11)*/ {"ficom",F_PAR+F_DW,272,FICOM,0},
 /*18(0x12)*/ {"ficom",F_PAR+F_DW,528,FICOM,0},
 /*19(0x13)*/ {"ficom",F_PAR+F_DW,784,FICOM,0},
 /*20(0x14)*/ {"ficom",F_PAR+F_DW,1040,FICOM,0},
 /*21(0x15)*/ {"ficom",F_PAR+F_DW,1296,FICOM,0},
 /*22(0x16)*/ {"ficom",F_PAR+F_DW,1552,FICOM,0},
 /*23(0x17)*/ {"ficom",F_PAR+F_DW,1808,FICOM,0},
 /*24(0x18)*/ {"ficomp",F_PAR+F_DW,16,FICOMP,0},
 /*25(0x19)*/ {"ficomp",F_PAR+F_DW,272,FICOMP,0},
 /*26(0x1A)*/ {"ficomp",F_PAR+F_DW,528,FICOMP,0},
 /*27(0x1B)*/ {"ficomp",F_PAR+F_DW,784,FICOMP,0},
 /*28(0x1C)*/ {"ficomp",F_PAR+F_DW,1040,FICOMP,0},
 /*29(0x1D)*/ {"ficomp",F_PAR+F_DW,1296,FICOMP,0},
 /*30(0x1E)*/ {"ficomp",F_PAR+F_DW,1552,FICOMP,0},
 /*31(0x1F)*/ {"ficomp",F_PAR+F_DW,1808,FICOMP,0},
 /*32(0x20)*/ {"fisub",F_PAR+F_DW,16,FISUB,0},
 /*33(0x21)*/ {"fisub",F_PAR+F_DW,272,FISUB,0},
 /*34(0x22)*/ {"fisub",F_PAR+F_DW,528,FISUB,0},
 /*35(0x23)*/ {"fisub",F_PAR+F_DW,784,FISUB,0},
 /*36(0x24)*/ {"fisub",F_PAR+F_DW,1040,FISUB,0},
 /*37(0x25)*/ {"fisub",F_PAR+F_DW,1296,FISUB,0},
 /*38(0x26)*/ {"fisub",F_PAR+F_DW,1552,FISUB,0},
 /*39(0x27)*/ {"fisub",F_PAR+F_DW,1808,FISUB,0},
 /*40(0x28)*/ {"fisubr",F_PAR+F_DW,16,FISUBR,0},
 /*41(0x29)*/ {"fisubr",F_PAR+F_DW,272,FISUBR,0},
 /*42(0x2A)*/ {"fisubr",F_PAR+F_DW,528,FISUBR,0},
 /*43(0x2B)*/ {"fisubr",F_PAR+F_DW,784,FISUBR,0},
 /*44(0x2C)*/ {"fisubr",F_PAR+F_DW,1040,FISUBR,0},
 /*45(0x2D)*/ {"fisubr",F_PAR+F_DW,1296,FISUBR,0},
 /*46(0x2E)*/ {"fisubr",F_PAR+F_DW,1552,FISUBR,0},
 /*47(0x2F)*/ {"fisubr",F_PAR+F_DW,1808,FISUBR,0},
 /*48(0x30)*/ {"fidiv",F_PAR+F_DW,16,FIDIV,0},
 /*49(0x31)*/ {"fidiv",F_PAR+F_DW,272,FIDIV,0},
 /*50(0x32)*/ {"fidiv",F_PAR+F_DW,528,FIDIV,0},
 /*51(0x33)*/ {"fidiv",F_PAR+F_DW,784,FIDIV,0},
 /*52(0x34)*/ {"fidiv",F_PAR+F_DW,1040,FIDIV,0},
 /*53(0x35)*/ {"fidiv",F_PAR+F_DW,1296,FIDIV,0},
 /*54(0x36)*/ {"fidiv",F_PAR+F_DW,1552,FIDIV,0},
 /*55(0x37)*/ {"fidiv",F_PAR+F_DW,1808,FIDIV,0},
 /*56(0x38)*/ {"fidivr",F_PAR+F_DW,16,FIDIVR,0},
 /*57(0x39)*/ {"fidivr",F_PAR+F_DW,272,FIDIVR,0},
 /*58(0x3A)*/ {"fidivr",F_PAR+F_DW,528,FIDIVR,0},
 /*59(0x3B)*/ {"fidivr",F_PAR+F_DW,784,FIDIVR,0},
 /*60(0x3C)*/ {"fidivr",F_PAR+F_DW,1040,FIDIVR,0},
 /*61(0x3D)*/ {"fidivr",F_PAR+F_DW,1296,FIDIVR,0},
 /*62(0x3E)*/ {"fidivr",F_PAR+F_DW,1552,FIDIVR,0},
 /*63(0x3F)*/ {"fidivr",F_PAR+F_DW,1808,FIDIVR,0},
 /*64(0x40)*/ {"fiadd",F_PAR+F_DW,17,FIADD,0},
 /*65(0x41)*/ {"fiadd",F_PAR+F_DW,273,FIADD,0},
 /*66(0x42)*/ {"fiadd",F_PAR+F_DW,529,FIADD,0},
 /*67(0x43)*/ {"fiadd",F_PAR+F_DW,785,FIADD,0},
 /*68(0x44)*/ {"fiadd",F_PAR+F_DW,1041,FIADD,0},
 /*69(0x45)*/ {"fiadd",F_PAR+F_DW,1297,FIADD,0},
 /*70(0x46)*/ {"fiadd",F_PAR+F_DW,1553,FIADD,0},
 /*71(0x47)*/ {"fiadd",F_PAR+F_DW,1809,FIADD,0},
 /*72(0x48)*/ {"fimul",F_PAR+F_DW,17,FIMUL,0},
 /*73(0x49)*/ {"fimul",F_PAR+F_DW,273,FIMUL,0},
 /*74(0x4A)*/ {"fimul",F_PAR+F_DW,529,FIMUL,0},
 /*75(0x4B)*/ {"fimul",F_PAR+F_DW,785,FIMUL,0},
 /*76(0x4C)*/ {"fimul",F_PAR+F_DW,1041,FIMUL,0},
 /*77(0x4D)*/ {"fimul",F_PAR+F_DW,1297,FIMUL,0},
 /*78(0x4E)*/ {"fimul",F_PAR+F_DW,1553,FIMUL,0},
 /*79(0x4F)*/ {"fimul",F_PAR+F_DW,1809,FIMUL,0},
 /*80(0x50)*/ {"ficom",F_PAR+F_DW,17,FICOM,0},
 /*81(0x51)*/ {"ficom",F_PAR+F_DW,273,FICOM,0},
 /*82(0x52)*/ {"ficom",F_PAR+F_DW,529,FICOM,0},
 /*83(0x53)*/ {"ficom",F_PAR+F_DW,785,FICOM,0},
 /*84(0x54)*/ {"ficom",F_PAR+F_DW,1041,FICOM,0},
 /*85(0x55)*/ {"ficom",F_PAR+F_DW,1297,FICOM,0},
 /*86(0x56)*/ {"ficom",F_PAR+F_DW,1553,FICOM,0},
 /*87(0x57)*/ {"ficom",F_PAR+F_DW,1809,FICOM,0},
 /*88(0x58)*/ {"ficomp",F_PAR+F_DW,17,FICOMP,0},
 /*89(0x59)*/ {"ficomp",F_PAR+F_DW,273,FICOMP,0},
 /*90(0x5A)*/ {"ficomp",F_PAR+F_DW,529,FICOMP,0},
 /*91(0x5B)*/ {"ficomp",F_PAR+F_DW,785,FICOMP,0},
 /*92(0x5C)*/ {"ficomp",F_PAR+F_DW,1041,FICOMP,0},
 /*93(0x5D)*/ {"ficomp",F_PAR+F_DW,1297,FICOMP,0},
 /*94(0x5E)*/ {"ficomp",F_PAR+F_DW,1553,FICOMP,0},
 /*95(0x5F)*/ {"ficomp",F_PAR+F_DW,1809,FICOMP,0},
 /*96(0x60)*/ {"fisub",F_PAR+F_DW,17,FISUB,0},
 /*97(0x61)*/ {"fisub",F_PAR+F_DW,273,FISUB,0},
 /*98(0x62)*/ {"fisub",F_PAR+F_DW,529,FISUB,0},
 /*99(0x63)*/ {"fisub",F_PAR+F_DW,785,FISUB,0},
 /*100(0x64)*/ {"fisub",F_PAR+F_DW,1041,FISUB,0},
 /*101(0x65)*/ {"fisub",F_PAR+F_DW,1297,FISUB,0},
 /*102(0x66)*/ {"fisub",F_PAR+F_DW,1553,FISUB,0},
 /*103(0x67)*/ {"fisub",F_PAR+F_DW,1809,FISUB,0},
 /*104(0x68)*/ {"fisubr",F_PAR+F_DW,17,FISUBR,0},
 /*105(0x69)*/ {"fisubr",F_PAR+F_DW,273,FISUBR,0},
 /*106(0x6A)*/ {"fisubr",F_PAR+F_DW,529,FISUBR,0},
 /*107(0x6B)*/ {"fisubr",F_PAR+F_DW,785,FISUBR,0},
 /*108(0x6C)*/ {"fisubr",F_PAR+F_DW,1041,FISUBR,0},
 /*109(0x6D)*/ {"fisubr",F_PAR+F_DW,1297,FISUBR,0},
 /*110(0x6E)*/ {"fisubr",F_PAR+F_DW,1553,FISUBR,0},
 /*111(0x6F)*/ {"fisubr",F_PAR+F_DW,1809,FISUBR,0},
 /*112(0x70)*/ {"fidiv",F_PAR+F_DW,17,FIDIV,0},
 /*113(0x71)*/ {"fidiv",F_PAR+F_DW,273,FIDIV,0},
 /*114(0x72)*/ {"fidiv",F_PAR+F_DW,529,FIDIV,0},
 /*115(0x73)*/ {"fidiv",F_PAR+F_DW,785,FIDIV,0},
 /*116(0x74)*/ {"fidiv",F_PAR+F_DW,1041,FIDIV,0},
 /*117(0x75)*/ {"fidiv",F_PAR+F_DW,1297,FIDIV,0},
 /*118(0x76)*/ {"fidiv",F_PAR+F_DW,1553,FIDIV,0},
 /*119(0x77)*/ {"fidiv",F_PAR+F_DW,1809,FIDIV,0},
 /*120(0x78)*/ {"fidivr",F_PAR+F_DW,17,FIDIVR,0},
 /*121(0x79)*/ {"fidivr",F_PAR+F_DW,273,FIDIVR,0},
 /*122(0x7A)*/ {"fidivr",F_PAR+F_DW,529,FIDIVR,0},
 /*123(0x7B)*/ {"fidivr",F_PAR+F_DW,785,FIDIVR,0},
 /*124(0x7C)*/ {"fidivr",F_PAR+F_DW,1041,FIDIVR,0},
 /*125(0x7D)*/ {"fidivr",F_PAR+F_DW,1297,FIDIVR,0},
 /*126(0x7E)*/ {"fidivr",F_PAR+F_DW,1553,FIDIVR,0},
 /*127(0x7F)*/ {"fidivr",F_PAR+F_DW,1809,FIDIVR,0},
 /*128(0x80)*/ {"fiadd",F_PAR+F_DW,18,FIADD,0},
 /*129(0x81)*/ {"fiadd",F_PAR+F_DW,274,FIADD,0},
 /*130(0x82)*/ {"fiadd",F_PAR+F_DW,530,FIADD,0},
 /*131(0x83)*/ {"fiadd",F_PAR+F_DW,786,FIADD,0},
 /*132(0x84)*/ {"fiadd",F_PAR+F_DW,1042,FIADD,0},
 /*133(0x85)*/ {"fiadd",F_PAR+F_DW,1298,FIADD,0},
 /*134(0x86)*/ {"fiadd",F_PAR+F_DW,1554,FIADD,0},
 /*135(0x87)*/ {"fiadd",F_PAR+F_DW,1810,FIADD,0},
 /*136(0x88)*/ {"fimul",F_PAR+F_DW,18,FIMUL,0},
 /*137(0x89)*/ {"fimul",F_PAR+F_DW,274,FIMUL,0},
 /*138(0x8A)*/ {"fimul",F_PAR+F_DW,530,FIMUL,0},
 /*139(0x8B)*/ {"fimul",F_PAR+F_DW,786,FIMUL,0},
 /*140(0x8C)*/ {"fimul",F_PAR+F_DW,1042,FIMUL,0},
 /*141(0x8D)*/ {"fimul",F_PAR+F_DW,1298,FIMUL,0},
 /*142(0x8E)*/ {"fimul",F_PAR+F_DW,1554,FIMUL,0},
 /*143(0x8F)*/ {"fimul",F_PAR+F_DW,1810,FIMUL,0},
 /*144(0x90)*/ {"ficom",F_PAR+F_DW,18,FICOM,0},
 /*145(0x91)*/ {"ficom",F_PAR+F_DW,274,FICOM,0},
 /*146(0x92)*/ {"ficom",F_PAR+F_DW,530,FICOM,0},
 /*147(0x93)*/ {"ficom",F_PAR+F_DW,786,FICOM,0},
 /*148(0x94)*/ {"ficom",F_PAR+F_DW,1042,FICOM,0},
 /*149(0x95)*/ {"ficom",F_PAR+F_DW,1298,FICOM,0},
 /*150(0x96)*/ {"ficom",F_PAR+F_DW,1554,FICOM,0},
 /*151(0x97)*/ {"ficom",F_PAR+F_DW,1810,FICOM,0},
 /*152(0x98)*/ {"ficomp",F_PAR+F_DW,18,FICOMP,0},
 /*153(0x99)*/ {"ficomp",F_PAR+F_DW,274,FICOMP,0},
 /*154(0x9A)*/ {"ficomp",F_PAR+F_DW,530,FICOMP,0},
 /*155(0x9B)*/ {"ficomp",F_PAR+F_DW,786,FICOMP,0},
 /*156(0x9C)*/ {"ficomp",F_PAR+F_DW,1042,FICOMP,0},
 /*157(0x9D)*/ {"ficomp",F_PAR+F_DW,1298,FICOMP,0},
 /*158(0x9E)*/ {"ficomp",F_PAR+F_DW,1554,FICOMP,0},
 /*159(0x9F)*/ {"ficomp",F_PAR+F_DW,1810,FICOMP,0},
 /*160(0xA0)*/ {"fisub",F_PAR+F_DW,18,FISUB,0},
 /*161(0xA1)*/ {"fisub",F_PAR+F_DW,274,FISUB,0},
 /*162(0xA2)*/ {"fisub",F_PAR+F_DW,530,FISUB,0},
 /*163(0xA3)*/ {"fisub",F_PAR+F_DW,786,FISUB,0},
 /*164(0xA4)*/ {"fisub",F_PAR+F_DW,1042,FISUB,0},
 /*165(0xA5)*/ {"fisub",F_PAR+F_DW,1298,FISUB,0},
 /*166(0xA6)*/ {"fisub",F_PAR+F_DW,1554,FISUB,0},
 /*167(0xA7)*/ {"fisub",F_PAR+F_DW,1810,FISUB,0},
 /*168(0xA8)*/ {"fisubr",F_PAR+F_DW,18,FISUBR,0},
 /*169(0xA9)*/ {"fisubr",F_PAR+F_DW,274,FISUBR,0},
 /*170(0xAA)*/ {"fisubr",F_PAR+F_DW,530,FISUBR,0},
 /*171(0xAB)*/ {"fisubr",F_PAR+F_DW,786,FISUBR,0},
 /*172(0xAC)*/ {"fisubr",F_PAR+F_DW,1042,FISUBR,0},
 /*173(0xAD)*/ {"fisubr",F_PAR+F_DW,1298,FISUBR,0},
 /*174(0xAE)*/ {"fisubr",F_PAR+F_DW,1554,FISUBR,0},
 /*175(0xAF)*/ {"fisubr",F_PAR+F_DW,1810,FISUBR,0},
 /*176(0xB0)*/ {"fidiv",F_PAR+F_DW,18,FIDIV,0},
 /*177(0xB1)*/ {"fidiv",F_PAR+F_DW,274,FIDIV,0},
 /*178(0xB2)*/ {"fidiv",F_PAR+F_DW,530,FIDIV,0},
 /*179(0xB3)*/ {"fidiv",F_PAR+F_DW,786,FIDIV,0},
 /*180(0xB4)*/ {"fidiv",F_PAR+F_DW,1042,FIDIV,0},
 /*181(0xB5)*/ {"fidiv",F_PAR+F_DW,1298,FIDIV,0},
 /*182(0xB6)*/ {"fidiv",F_PAR+F_DW,1554,FIDIV,0},
 /*183(0xB7)*/ {"fidiv",F_PAR+F_DW,1810,FIDIV,0},
 /*184(0xB8)*/ {"fidivr",F_PAR+F_DW,18,FIDIVR,0},
 /*185(0xB9)*/ {"fidivr",F_PAR+F_DW,274,FIDIVR,0},
 /*186(0xBA)*/ {"fidivr",F_PAR+F_DW,530,FIDIVR,0},
 /*187(0xBB)*/ {"fidivr",F_PAR+F_DW,786,FIDIVR,0},
 /*188(0xBC)*/ {"fidivr",F_PAR+F_DW,1042,FIDIVR,0},
 /*189(0xBD)*/ {"fidivr",F_PAR+F_DW,1298,FIDIVR,0},
 /*190(0xBE)*/ {"fidivr",F_PAR+F_DW,1554,FIDIVR,0},
 /*191(0xBF)*/ {"fidivr",F_PAR+F_DW,1810,FIDIVR,0},
 /*192(0xC0)*/ {"fcomvb",F_PAR,2,FCMOVB,0},
 /*193(0xC1)*/ {"fcomvb",F_PAR,258,FCMOVB,0},
 /*194(0xC2)*/ {"fcomvb",F_PAR,514,FCMOVB,0},
 /*195(0xC3)*/ {"fcomvb",F_PAR,770,FCMOVB,0},
 /*196(0xC4)*/ {"fcomvb",F_PAR,1026,FCMOVB,0},
 /*197(0xC5)*/ {"fcomvb",F_PAR,1282,FCMOVB,0},
 /*198(0xC6)*/ {"fcomvb",F_PAR,1538,FCMOVB,0},
 /*199(0xC7)*/ {"fcomvb",F_PAR,1794,FCMOVB,0},
 /*200(0xC8)*/ {"fcmove",F_PAR,2,FCMOVE,0},
 /*201(0xC9)*/ {"fcmove",F_PAR,258,FCMOVE,0},
 /*202(0xCA)*/ {"fcmove",F_PAR,514,FCMOVE,0},
 /*203(0xCB)*/ {"fcmove",F_PAR,770,FCMOVE,0},
 /*204(0xCC)*/ {"fcmove",F_PAR,1026,FCMOVE,0},
 /*205(0xCD)*/ {"fcmove",F_PAR,1282,FCMOVE,0},
 /*206(0xCE)*/ {"fcmove",F_PAR,1538,FCMOVE,0},
 /*207(0xCF)*/ {"fcmove",F_PAR,1794,FCMOVE,0},
 /*208(0xD0)*/ {"fcmovbe",F_PAR,2,FCMOVBE,0},
 /*209(0xD1)*/ {"fcmovbe",F_PAR,258,FCMOVBE,0},
 /*210(0xD2)*/ {"fcmovbe",F_PAR,514,FCMOVBE,0},
 /*211(0xD3)*/ {"fcmovbe",F_PAR,770,FCMOVBE,0},
 /*212(0xD4)*/ {"fcmovbe",F_PAR,1026,FCMOVBE,0},
 /*213(0xD5)*/ {"fcmovbe",F_PAR,1282,FCMOVBE,0},
 /*214(0xD6)*/ {"fcmovbe",F_PAR,1538,FCMOVBE,0},
 /*215(0xD7)*/ {"fcmovbe",F_PAR,1794,FCMOVBE,0},
 /*216(0xD8)*/ {"fcmovu",F_PAR,2,FCMOVU,0},
 /*217(0xD9)*/ {"fcmovu",F_PAR,258,FCMOVU,0},
 /*218(0xDA)*/ {"fcmovu",F_PAR,514,FCMOVU,0},
 /*219(0xDB)*/ {"fcmovu",F_PAR,770,FCMOVU,0},
 /*220(0xDC)*/ {"fcmovu",F_PAR,1026,FCMOVU,0},
 /*221(0xDD)*/ {"fcmovu",F_PAR,1282,FCMOVU,0},
 /*222(0xDE)*/ {"fcmovu",F_PAR,1538,FCMOVU,0},
 /*223(0xDF)*/ {"fcmovu",F_PAR,1794,FCMOVU,0},
 /*224(0xE0)*/ {0,0,0,0,0},
 /*225(0xE1)*/ {0,0,0,0,0},
 /*226(0xE2)*/ {0,0,0,0,0},
 /*227(0xE3)*/ {0,0,0,0,0},
 /*228(0xE4)*/ {0,0,0,0,0},
 /*229(0xE5)*/ {0,0,0,0,0},
 /*230(0xE6)*/ {0,0,0,0,0},
 /*231(0xE7)*/ {0,0,0,0,0},
 /*232(0xE8)*/ {0,0,0,0,0},
 /*233(0xE9)*/ {"fucompp",0,0,FUCOMPP,0},
 /*234(0xEA)*/ {0,0,0,0,0},
 /*235(0xEB)*/ {0,0,0,0,0},
 /*236(0xEC)*/ {0,0,0,0,0},
 /*237(0xED)*/ {0,0,0,0,0},
 /*238(0xEE)*/ {0,0,0,0,0},
 /*239(0xEF)*/ {0,0,0,0,0},
 /*240(0xF0)*/ {0,0,0,0,0},
 /*241(0xF1)*/ {0,0,0,0,0},
 /*242(0xF2)*/ {0,0,0,0,0},
 /*243(0xF3)*/ {0,0,0,0,0},
 /*244(0xF4)*/ {0,0,0,0,0},
 /*245(0xF5)*/ {0,0,0,0,0},
 /*246(0xF6)*/ {0,0,0,0,0},
 /*247(0xF7)*/ {0,0,0,0,0},
 /*248(0xF8)*/ {0,0,0,0,0},
 /*249(0xF9)*/ {0,0,0,0,0},
 /*250(0xFA)*/ {0,0,0,0,0},
 /*251(0xFB)*/ {0,0,0,0,0},
 /*252(0xFC)*/ {0,0,0,0,0},
 /*253(0xFD)*/ {0,0,0,0,0},
 /*254(0xFE)*/ {0,0,0,0,0},
 /*255(0xFF)*/ {0,0,0,0,0}
};

//grupa DBH - instrukcje koprocesora
op_t gDB[256]=
{
 /*0(0x00)*/ {"fild",F_PAR+F_DW+F_D,16,FILD,0},
 /*1(0x01)*/ {"fild",F_PAR+F_DW+F_D,272,FILD,0},
 /*2(0x02)*/ {"fild",F_PAR+F_DW+F_D,528,FILD,0},
 /*3(0x03)*/ {"fild",F_PAR+F_DW+F_D,784,FILD,0},
 /*4(0x04)*/ {"fild",F_PAR+F_DW+F_D,1040,FILD,0},
 /*5(0x05)*/ {"fild",F_PAR+F_DW+F_D,1296,FILD,0},
 /*6(0x06)*/ {"fild",F_PAR+F_DW+F_D,1552,FILD,0},
 /*7(0x07)*/ {"fild",F_PAR+F_DW+F_D,1808,FILD,0},
 /*8(0x08)*/ {0,F_PAR+F_DW+F_D,16,0,0},
 /*9(0x09)*/ {0,F_PAR+F_DW+F_D,272,0,0},
 /*10(0x0A)*/ {0,F_PAR+F_DW+F_D,528,0,0},
 /*11(0x0B)*/ {0,F_PAR+F_DW+F_D,784,0,0},
 /*12(0x0C)*/ {0,F_PAR+F_DW+F_D,1040,0,0},
 /*13(0x0D)*/ {0,F_PAR+F_DW+F_D,1296,0,0},
 /*14(0x0E)*/ {0,F_PAR+F_DW+F_D,1552,0,0},
 /*15(0x0F)*/ {0,F_PAR+F_DW+F_D,1808,0,0},
 /*16(0x10)*/ {"fist",F_PAR+F_DW+F_D,16,FIST,0},
 /*17(0x11)*/ {"fist",F_PAR+F_DW+F_D,272,FIST,0},
 /*18(0x12)*/ {"fist",F_PAR+F_DW+F_D,528,FIST,0},
 /*19(0x13)*/ {"fist",F_PAR+F_DW+F_D,784,FIST,0},
 /*20(0x14)*/ {"fist",F_PAR+F_DW+F_D,1040,FIST,0},
 /*21(0x15)*/ {"fist",F_PAR+F_DW+F_D,1296,FIST,0},
 /*22(0x16)*/ {"fist",F_PAR+F_DW+F_D,1552,FIST,0},
 /*23(0x17)*/ {"fist",F_PAR+F_DW+F_D,1808,FIST,0},
 /*24(0x18)*/ {"fistp",F_PAR+F_DW+F_D,16,FISTP,0},
 /*25(0x19)*/ {"fistp",F_PAR+F_DW+F_D,272,FISTP,0},
 /*26(0x1A)*/ {"fistp",F_PAR+F_DW+F_D,528,FISTP,0},
 /*27(0x1B)*/ {"fistp",F_PAR+F_DW+F_D,784,FISTP,0},
 /*28(0x1C)*/ {"fistp",F_PAR+F_DW+F_D,1040,FISTP,0},
 /*29(0x1D)*/ {"fistp",F_PAR+F_DW+F_D,1296,FISTP,0},
 /*30(0x1E)*/ {"fistp",F_PAR+F_DW+F_D,1552,FISTP,0},
 /*31(0x1F)*/ {"fistp",F_PAR+F_DW+F_D,1808,FISTP,0},
 /*32(0x20)*/ {0,F_PAR+F_DW+F_D,16,0,0},
 /*33(0x21)*/ {0,F_PAR+F_DW+F_D,272,0,0},
 /*34(0x22)*/ {0,F_PAR+F_DW+F_D,528,0,0},
 /*35(0x23)*/ {0,F_PAR+F_DW+F_D,784,0,0},
 /*36(0x24)*/ {0,F_PAR+F_DW+F_D,1040,0,0},
 /*37(0x25)*/ {0,F_PAR+F_DW+F_D,1296,0,0},
 /*38(0x26)*/ {0,F_PAR+F_DW+F_D,1552,0,0},
 /*39(0x27)*/ {0,F_PAR+F_DW+F_D,1808,0,0},
 /*40(0x28)*/ {"fld",F_PAR+F_TW+F_D+F_DP,16,FLD,0},
 /*41(0x29)*/ {"fld",F_PAR+F_TW+F_D+F_DP,272,FLD,0},
 /*42(0x2A)*/ {"fld",F_PAR+F_TW+F_D+F_DP,528,FLD,0},
 /*43(0x2B)*/ {"fld",F_PAR+F_TW+F_D+F_DP,784,FLD,0},
 /*44(0x2C)*/ {"fld",F_PAR+F_TW+F_D+F_DP,1040,FLD,0},
 /*45(0x2D)*/ {"fld",F_PAR+F_TW+F_D+F_DP,1296,FLD,0},
 /*46(0x2E)*/ {"fld",F_PAR+F_TW+F_D+F_DP,1552,FLD,0},
 /*47(0x2F)*/ {"fld",F_PAR+F_TW+F_D+F_DP,1808,FLD,0},
 /*48(0x30)*/ {0,F_PAR+F_TW+F_D+F_DP,16,0,0},
 /*49(0x31)*/ {0,F_PAR+F_TW+F_D+F_DP,272,0,0},
 /*50(0x32)*/ {0,F_PAR+F_TW+F_D+F_DP,528,0,0},
 /*51(0x33)*/ {0,F_PAR+F_TW+F_D+F_DP,784,0,0},
 /*52(0x34)*/ {0,F_PAR+F_TW+F_D+F_DP,1040,0,0},
 /*53(0x35)*/ {0,F_PAR+F_TW+F_D+F_DP,1296,0,0},
 /*54(0x36)*/ {0,F_PAR+F_TW+F_D+F_DP,1552,0,0},
 /*55(0x37)*/ {0,F_PAR+F_TW+F_D+F_DP,1808,0,0},
 /*56(0x38)*/ {"fstp",F_PAR+F_TW+F_D+F_DP,16,FSTP,0},
 /*57(0x39)*/ {"fstp",F_PAR+F_TW+F_D+F_DP,272,FSTP,0},
 /*58(0x3A)*/ {"fstp",F_PAR+F_TW+F_D+F_DP,528,FSTP,0},
 /*59(0x3B)*/ {"fstp",F_PAR+F_TW+F_D+F_DP,784,FSTP,0},
 /*60(0x3C)*/ {"fstp",F_PAR+F_TW+F_D+F_DP,1040,FSTP,0},
 /*61(0x3D)*/ {"fstp",F_PAR+F_TW+F_D+F_DP,1296,FSTP,0},
 /*62(0x3E)*/ {"fstp",F_PAR+F_TW+F_D+F_DP,1552,FSTP,0},
 /*63(0x3F)*/ {"fstp",F_PAR+F_TW+F_D+F_DP,1808,FSTP,0},
 /*64(0x40)*/ {"fild",F_PAR+F_DW+F_D,17,FILD,0},
 /*65(0x41)*/ {"fild",F_PAR+F_DW+F_D,273,FILD,0},
 /*66(0x42)*/ {"fild",F_PAR+F_DW+F_D,529,FILD,0},
 /*67(0x43)*/ {"fild",F_PAR+F_DW+F_D,785,FILD,0},
 /*68(0x44)*/ {"fild",F_PAR+F_DW+F_D,1041,FILD,0},
 /*69(0x45)*/ {"fild",F_PAR+F_DW+F_D,1297,FILD,0},
 /*70(0x46)*/ {"fild",F_PAR+F_DW+F_D,1553,FILD,0},
 /*71(0x47)*/ {"fild",F_PAR+F_DW+F_D,1809,FILD,0},
 /*72(0x48)*/ {0,F_PAR+F_DW+F_D,17,0,0},
 /*73(0x49)*/ {0,F_PAR+F_DW+F_D,273,0,0},
 /*74(0x4A)*/ {0,F_PAR+F_DW+F_D,529,0,0},
 /*75(0x4B)*/ {0,F_PAR+F_DW+F_D,785,0,0},
 /*76(0x4C)*/ {0,F_PAR+F_DW+F_D,1041,0,0},
 /*77(0x4D)*/ {0,F_PAR+F_DW+F_D,1297,0,0},
 /*78(0x4E)*/ {0,F_PAR+F_DW+F_D,1553,0,0},
 /*79(0x4F)*/ {0,F_PAR+F_DW+F_D,1809,0,0},
 /*80(0x50)*/ {"fist",F_PAR+F_DW+F_D,17,FIST,0},
 /*81(0x51)*/ {"fist",F_PAR+F_DW+F_D,273,FIST,0},
 /*82(0x52)*/ {"fist",F_PAR+F_DW+F_D,529,FIST,0},
 /*83(0x53)*/ {"fist",F_PAR+F_DW+F_D,785,FIST,0},
 /*84(0x54)*/ {"fist",F_PAR+F_DW+F_D,1041,FIST,0},
 /*85(0x55)*/ {"fist",F_PAR+F_DW+F_D,1297,FIST,0},
 /*86(0x56)*/ {"fist",F_PAR+F_DW+F_D,1553,FIST,0},
 /*87(0x57)*/ {"fist",F_PAR+F_DW+F_D,1809,FIST,0},
 /*88(0x58)*/ {"fistp",F_PAR+F_DW+F_D,17,FISTP,0},
 /*89(0x59)*/ {"fistp",F_PAR+F_DW+F_D,273,FISTP,0},
 /*90(0x5A)*/ {"fistp",F_PAR+F_DW+F_D,529,FISTP,0},
 /*91(0x5B)*/ {"fistp",F_PAR+F_DW+F_D,785,FISTP,0},
 /*92(0x5C)*/ {"fistp",F_PAR+F_DW+F_D,1041,FISTP,0},
 /*93(0x5D)*/ {"fistp",F_PAR+F_DW+F_D,1297,FISTP,0},
 /*94(0x5E)*/ {"fistp",F_PAR+F_DW+F_D,1553,FISTP,0},
 /*95(0x5F)*/ {"fistp",F_PAR+F_DW+F_D,1809,FISTP,0},
 /*96(0x60)*/ {0,F_PAR+F_DW+F_D,17,0,0},
 /*97(0x61)*/ {0,F_PAR+F_DW+F_D,273,0,0},
 /*98(0x62)*/ {0,F_PAR+F_DW+F_D,529,0,0},
 /*99(0x63)*/ {0,F_PAR+F_DW+F_D,785,0,0},
 /*100(0x64)*/ {0,F_PAR+F_DW+F_D,1041,0,0},
 /*101(0x65)*/ {0,F_PAR+F_DW+F_D,1297,0,0},
 /*102(0x66)*/ {0,F_PAR+F_DW+F_D,1553,0,0},
 /*103(0x67)*/ {0,F_PAR+F_DW+F_D,1809,0,0},
 /*104(0x68)*/ {"fld",F_PAR+F_TW+F_D+F_DP,17,FLD,0},
 /*105(0x69)*/ {"fld",F_PAR+F_TW+F_D+F_DP,273,FLD,0},
 /*106(0x6A)*/ {"fld",F_PAR+F_TW+F_D+F_DP,529,FLD,0},
 /*107(0x6B)*/ {"fld",F_PAR+F_TW+F_D+F_DP,785,FLD,0},
 /*108(0x6C)*/ {"fld",F_PAR+F_TW+F_D+F_DP,1041,FLD,0},
 /*109(0x6D)*/ {"fld",F_PAR+F_TW+F_D+F_DP,1297,FLD,0},
 /*110(0x6E)*/ {"fld",F_PAR+F_TW+F_D+F_DP,1553,FLD,0},
 /*111(0x6F)*/ {"fld",F_PAR+F_TW+F_D+F_DP,1809,FLD,0},
 /*112(0x70)*/ {0,F_PAR+F_TW+F_D+F_DP,17,0,0},
 /*113(0x71)*/ {0,F_PAR+F_TW+F_D+F_DP,273,0,0},
 /*114(0x72)*/ {0,F_PAR+F_TW+F_D+F_DP,529,0,0},
 /*115(0x73)*/ {0,F_PAR+F_TW+F_D+F_DP,785,0,0},
 /*116(0x74)*/ {0,F_PAR+F_TW+F_D+F_DP,1041,0,0},
 /*117(0x75)*/ {0,F_PAR+F_TW+F_D+F_DP,1297,0,0},
 /*118(0x76)*/ {0,F_PAR+F_TW+F_D+F_DP,1553,0,0},
 /*119(0x77)*/ {0,F_PAR+F_TW+F_D+F_DP,1809,0,0},
 /*120(0x78)*/ {"fstp",F_PAR+F_TW+F_D+F_DP,17,FSTP,0},
 /*121(0x79)*/ {"fstp",F_PAR+F_TW+F_D+F_DP,273,FSTP,0},
 /*122(0x7A)*/ {"fstp",F_PAR+F_TW+F_D+F_DP,529,FSTP,0},
 /*123(0x7B)*/ {"fstp",F_PAR+F_TW+F_D+F_DP,785,FSTP,0},
 /*124(0x7C)*/ {"fstp",F_PAR+F_TW+F_D+F_DP,1041,FSTP,0},
 /*125(0x7D)*/ {"fstp",F_PAR+F_TW+F_D+F_DP,1297,FSTP,0},
 /*126(0x7E)*/ {"fstp",F_PAR+F_TW+F_D+F_DP,1553,FSTP,0},
 /*127(0x7F)*/ {"fstp",F_PAR+F_TW+F_D+F_DP,1809,FSTP,0},
 /*128(0x80)*/ {"fild",F_PAR+F_DW+F_D,18,FILD,0},
 /*129(0x81)*/ {"fild",F_PAR+F_DW+F_D,274,FILD,0},
 /*130(0x82)*/ {"fild",F_PAR+F_DW+F_D,530,FILD,0},
 /*131(0x83)*/ {"fild",F_PAR+F_DW+F_D,786,FILD,0},
 /*132(0x84)*/ {"fild",F_PAR+F_DW+F_D,1042,FILD,0},
 /*133(0x85)*/ {"fild",F_PAR+F_DW+F_D,1298,FILD,0},
 /*134(0x86)*/ {"fild",F_PAR+F_DW+F_D,1554,FILD,0},
 /*135(0x87)*/ {"fild",F_PAR+F_DW+F_D,1810,FILD,0},
 /*136(0x88)*/ {0,F_PAR+F_DW+F_D,18,0,0},
 /*137(0x89)*/ {0,F_PAR+F_DW+F_D,274,0,0},
 /*138(0x8A)*/ {0,F_PAR+F_DW+F_D,530,0,0},
 /*139(0x8B)*/ {0,F_PAR+F_DW+F_D,786,0,0},
 /*140(0x8C)*/ {0,F_PAR+F_DW+F_D,1042,0,0},
 /*141(0x8D)*/ {0,F_PAR+F_DW+F_D,1298,0,0},
 /*142(0x8E)*/ {0,F_PAR+F_DW+F_D,1554,0,0},
 /*143(0x8F)*/ {0,F_PAR+F_DW+F_D,1810,0,0},
 /*144(0x90)*/ {"fist",F_PAR+F_DW+F_D,18,FIST,0},
 /*145(0x91)*/ {"fist",F_PAR+F_DW+F_D,274,FIST,0},
 /*146(0x92)*/ {"fist",F_PAR+F_DW+F_D,530,FIST,0},
 /*147(0x93)*/ {"fist",F_PAR+F_DW+F_D,786,FIST,0},
 /*148(0x94)*/ {"fist",F_PAR+F_DW+F_D,1042,FIST,0},
 /*149(0x95)*/ {"fist",F_PAR+F_DW+F_D,1298,FIST,0},
 /*150(0x96)*/ {"fist",F_PAR+F_DW+F_D,1554,FIST,0},
 /*151(0x97)*/ {"fist",F_PAR+F_DW+F_D,1810,FIST,0},
 /*152(0x98)*/ {"fistp",F_PAR+F_DW+F_D,18,FISTP,0},
 /*153(0x99)*/ {"fistp",F_PAR+F_DW+F_D,274,FISTP,0},
 /*154(0x9A)*/ {"fistp",F_PAR+F_DW+F_D,530,FISTP,0},
 /*155(0x9B)*/ {"fistp",F_PAR+F_DW+F_D,786,FISTP,0},
 /*156(0x9C)*/ {"fistp",F_PAR+F_DW+F_D,1042,FISTP,0},
 /*157(0x9D)*/ {"fistp",F_PAR+F_DW+F_D,1298,FISTP,0},
 /*158(0x9E)*/ {"fistp",F_PAR+F_DW+F_D,1554,FISTP,0},
 /*159(0x9F)*/ {"fistp",F_PAR+F_DW+F_D,1810,FISTP,0},
 /*160(0xA0)*/ {0,F_PAR+F_DW+F_D,18,0,0},
 /*161(0xA1)*/ {0,F_PAR+F_DW+F_D,274,0,0},
 /*162(0xA2)*/ {0,F_PAR+F_DW+F_D,530,0,0},
 /*163(0xA3)*/ {0,F_PAR+F_DW+F_D,786,0,0},
 /*164(0xA4)*/ {0,F_PAR+F_DW+F_D,1042,0,0},
 /*165(0xA5)*/ {0,F_PAR+F_DW+F_D,1298,0,0},
 /*166(0xA6)*/ {0,F_PAR+F_DW+F_D,1554,0,0},
 /*167(0xA7)*/ {0,F_PAR+F_DW+F_D,1810,0,0},
 /*168(0xA8)*/ {"fld",F_PAR+F_TW+F_D+F_DP,18,FLD,0},
 /*169(0xA9)*/ {"fld",F_PAR+F_TW+F_D+F_DP,274,FLD,0},
 /*170(0xAA)*/ {"fld",F_PAR+F_TW+F_D+F_DP,530,FLD,0},
 /*171(0xAB)*/ {"fld",F_PAR+F_TW+F_D+F_DP,786,FLD,0},
 /*172(0xAC)*/ {"fld",F_PAR+F_TW+F_D+F_DP,1042,FLD,0},
 /*173(0xAD)*/ {"fld",F_PAR+F_TW+F_D+F_DP,1298,FLD,0},
 /*174(0xAE)*/ {"fld",F_PAR+F_TW+F_D+F_DP,1554,FLD,0},
 /*175(0xAF)*/ {"fld",F_PAR+F_TW+F_D+F_DP,1810,FLD,0},
 /*176(0xB0)*/ {0,F_PAR+F_TW+F_D+F_DP,18,0,0},
 /*177(0xB1)*/ {0,F_PAR+F_TW+F_D+F_DP,274,0,0},
 /*178(0xB2)*/ {0,F_PAR+F_TW+F_D+F_DP,530,0,0},
 /*179(0xB3)*/ {0,F_PAR+F_TW+F_D+F_DP,786,0,0},
 /*180(0xB4)*/ {0,F_PAR+F_TW+F_D+F_DP,1042,0,0},
 /*181(0xB5)*/ {0,F_PAR+F_TW+F_D+F_DP,1298,0,0},
 /*182(0xB6)*/ {0,F_PAR+F_TW+F_D+F_DP,1554,0,0},
 /*183(0xB7)*/ {0,F_PAR+F_TW+F_D+F_DP,1810,0,0},
 /*184(0xB8)*/ {"fstp",F_PAR+F_TW+F_D+F_DP,18,FSTP,0},
 /*185(0xB9)*/ {"fstp",F_PAR+F_TW+F_D+F_DP,274,FSTP,0},
 /*186(0xBA)*/ {"fstp",F_PAR+F_TW+F_D+F_DP,530,FSTP,0},
 /*187(0xBB)*/ {"fstp",F_PAR+F_TW+F_D+F_DP,786,FSTP,0},
 /*188(0xBC)*/ {"fstp",F_PAR+F_TW+F_D+F_DP,1042,FSTP,0},
 /*189(0xBD)*/ {"fstp",F_PAR+F_TW+F_D+F_DP,1298,FSTP,0},
 /*190(0xBE)*/ {"fstp",F_PAR+F_TW+F_D+F_DP,1554,FSTP,0},
 /*191(0xBF)*/ {"fstp",F_PAR+F_TW+F_D+F_DP,1810,FSTP,0},
 /*192(0xC0)*/ {"fcomvnb",F_PAR,2,FCMOVNB,0},
 /*193(0xC1)*/ {"fcomvnb",F_PAR,258,FCMOVNB,0},
 /*194(0xC2)*/ {"fcomvnb",F_PAR,514,FCMOVNB,0},
 /*195(0xC3)*/ {"fcomvnb",F_PAR,770,FCMOVNB,0},
 /*196(0xC4)*/ {"fcomvnb",F_PAR,1026,FCMOVNB,0},
 /*197(0xC5)*/ {"fcomvnb",F_PAR,1282,FCMOVNB,0},
 /*198(0xC6)*/ {"fcomvnb",F_PAR,1538,FCMOVNB,0},
 /*199(0xC7)*/ {"fcomvnb",F_PAR,1794,FCMOVNB,0},
 /*200(0xC8)*/ {"fcmovne",F_PAR,2,FCMOVNE,0},
 /*201(0xC9)*/ {"fcmovne",F_PAR,258,FCMOVNE,0},
 /*202(0xCA)*/ {"fcmovne",F_PAR,514,FCMOVNE,0},
 /*203(0xCB)*/ {"fcmovne",F_PAR,770,FCMOVNE,0},
 /*204(0xCC)*/ {"fcmovne",F_PAR,1026,FCMOVNE,0},
 /*205(0xCD)*/ {"fcmovne",F_PAR,1282,FCMOVNE,0},
 /*206(0xCE)*/ {"fcmovne",F_PAR,1538,FCMOVNE,0},
 /*207(0xCF)*/ {"fcmovne",F_PAR,1794,FCMOVNE,0},
 /*208(0xD0)*/ {"fcmovnbe",F_PAR,2,FCMOVNBE,0},
 /*209(0xD1)*/ {"fcmovnbe",F_PAR,258,FCMOVNBE,0},
 /*210(0xD2)*/ {"fcmovnbe",F_PAR,514,FCMOVNBE,0},
 /*211(0xD3)*/ {"fcmovnbe",F_PAR,770,FCMOVNBE,0},
 /*212(0xD4)*/ {"fcmovnbe",F_PAR,1026,FCMOVNBE,0},
 /*213(0xD5)*/ {"fcmovnbe",F_PAR,1282,FCMOVNBE,0},
 /*214(0xD6)*/ {"fcmovnbe",F_PAR,1538,FCMOVNBE,0},
 /*215(0xD7)*/ {"fcmovnbe",F_PAR,1794,FCMOVNBE,0},
 /*216(0xD8)*/ {"fcmovnu",F_PAR,2,FCMOVNU,0},
 /*217(0xD9)*/ {"fcmovnu",F_PAR,258,FCMOVNU,0},
 /*218(0xDA)*/ {"fcmovnu",F_PAR,514,FCMOVNU,0},
 /*219(0xDB)*/ {"fcmovnu",F_PAR,770,FCMOVNU,0},
 /*220(0xDC)*/ {"fcmovnu",F_PAR,1026,FCMOVNU,0},
 /*221(0xDD)*/ {"fcmovnu",F_PAR,1282,FCMOVNU,0},
 /*222(0xDE)*/ {"fcmovnu",F_PAR,1538,FCMOVNU,0},
 /*223(0xDF)*/ {"fcmovnu",F_PAR,1794,FCMOVNU,0},
 /*224(0xE0)*/ {0,0,0,0,0},
 /*225(0xE1)*/ {0,0,0,0,0},
 /*226(0xE2)*/ {"fclex",0,0,FCLEX,0},
 /*227(0xE3)*/ {"finit",0,0,FINIT,0},
 /*228(0xE4)*/ {0,0,0,0,0},
 /*229(0xE5)*/ {0,0,0,0,0},
 /*230(0xE6)*/ {0,0,0,0,0},
 /*231(0xE7)*/ {0,0,0,0,0},
 /*232(0xE8)*/ {"fucomi",F_PAR,2,FUCOMI,0},
 /*233(0xE9)*/ {"fucomi",F_PAR,258,FUCOMI,0},
 /*234(0xEA)*/ {"fucomi",F_PAR,514,FUCOMI,0},
 /*235(0xEB)*/ {"fucomi",F_PAR,770,FUCOMI,0},
 /*236(0xEC)*/ {"fucomi",F_PAR,1026,FUCOMI,0},
 /*237(0xED)*/ {"fucomi",F_PAR,1282,FUCOMI,0},
 /*238(0xEE)*/ {"fucomi",F_PAR,1538,FUCOMI,0},
 /*239(0xEF)*/ {"fucomi",F_PAR,1794,FUCOMI,0},
 /*240(0xF0)*/ {"fcomi",F_PAR,2,FCOMI,0},
 /*241(0xF1)*/ {"fcomi",F_PAR,258,FCOMI,0},
 /*242(0xF2)*/ {"fcomi",F_PAR,514,FCOMI,0},
 /*243(0xF3)*/ {"fcomi",F_PAR,770,FCOMI,0},
 /*244(0xF4)*/ {"fcomi",F_PAR,1026,FCOMI,0},
 /*245(0xF5)*/ {"fcomi",F_PAR,1282,FCOMI,0},
 /*246(0xF6)*/ {"fcomi",F_PAR,1538,FCOMI,0},
 /*247(0xF7)*/ {"fcomi",F_PAR,1794,FCOMI,0},
 /*248(0xF8)*/ {0,0,0,0,0},
 /*249(0xF9)*/ {0,0,0,0,0},
 /*250(0xFA)*/ {0,0,0,0,0},
 /*251(0xFB)*/ {0,0,0,0,0},
 /*252(0xFC)*/ {0,0,0,0,0},
 /*253(0xFD)*/ {0,0,0,0,0},
 /*254(0xFE)*/ {0,0,0,0,0},
 /*255(0xFF)*/ {0,0,0,0,0}
};

//grupa DC - instrukcje koprocesora
op_t gDC[256]=
{
 /*0(0x00)*/ {"fadd",0,16,FADD,0},
 /*1(0x01)*/ {"fadd",0,272,FADD,0},
 /*2(0x02)*/ {"fadd",0,528,FADD,0},
 /*3(0x03)*/ {"fadd",0,784,FADD,0},
 /*4(0x04)*/ {"fadd",0,1040,FADD,0},
 /*5(0x05)*/ {"fadd",0,1296,FADD,0},
 /*6(0x06)*/ {"fadd",0,1552,FADD,0},
 /*7(0x07)*/ {"fadd",0,1808,FADD,0},
 /*8(0x08)*/ {"fmul",0,16,FMUL,0},
 /*9(0x09)*/ {"fmul",0,272,FMUL,0},
 /*10(0x0A)*/ {"fmul",0,528,FMUL,0},
 /*11(0x0B)*/ {"fmul",0,784,FMUL,0},
 /*12(0x0C)*/ {"fmul",0,1040,FMUL,0},
 /*13(0x0D)*/ {"fmul",0,1296,FMUL,0},
 /*14(0x0E)*/ {"fmul",0,1552,FMUL,0},
 /*15(0x0F)*/ {"fmul",0,1808,FMUL,0},
 /*16(0x10)*/ {"fcom",0,16,FCOM,0},
 /*17(0x11)*/ {"fcom",0,272,FCOM,0},
 /*18(0x12)*/ {"fcom",0,528,FCOM,0},
 /*19(0x13)*/ {"fcom",0,784,FCOM,0},
 /*20(0x14)*/ {"fcom",0,1040,FCOM,0},
 /*21(0x15)*/ {"fcom",0,1296,FCOM,0},
 /*22(0x16)*/ {"fcom",0,1552,FCOM,0},
 /*23(0x17)*/ {"fcom",0,1808,FCOM,0},
 /*24(0x18)*/ {"fcomp",0,16,FCOMP,0},
 /*25(0x19)*/ {"fcomp",0,272,FCOMP,0},
 /*26(0x1A)*/ {"fcomp",0,528,FCOMP,0},
 /*27(0x1B)*/ {"fcomp",0,784,FCOMP,0},
 /*28(0x1C)*/ {"fcomp",0,1040,FCOMP,0},
 /*29(0x1D)*/ {"fcomp",0,1296,FCOMP,0},
 /*30(0x1E)*/ {"fcomp",0,1552,FCOMP,0},
 /*31(0x1F)*/ {"fcomp",0,1808,FCOMP,0},
 /*32(0x20)*/ {"fsub",0,16,FSUB,0},
 /*33(0x21)*/ {"fsub",0,272,FSUB,0},
 /*34(0x22)*/ {"fsub",0,528,FSUB,0},
 /*35(0x23)*/ {"fsub",0,784,FSUB,0},
 /*36(0x24)*/ {"fsub",0,1040,FSUB,0},
 /*37(0x25)*/ {"fsub",0,1296,FSUB,0},
 /*38(0x26)*/ {"fsub",0,1552,FSUB,0},
 /*39(0x27)*/ {"fsub",0,1808,FSUB,0},
 /*40(0x28)*/ {"fsubr",0,16,FSUBR,0},
 /*41(0x29)*/ {"fsubr",0,272,FSUBR,0},
 /*42(0x2A)*/ {"fsubr",0,528,FSUBR,0},
 /*43(0x2B)*/ {"fsubr",0,784,FSUBR,0},
 /*44(0x2C)*/ {"fsubr",0,1040,FSUBR,0},
 /*45(0x2D)*/ {"fsubr",0,1296,FSUBR,0},
 /*46(0x2E)*/ {"fsubr",0,1552,FSUBR,0},
 /*47(0x2F)*/ {"fsubr",0,1808,FSUBR,0},
 /*48(0x30)*/ {"fdiv",0,16,FDIV,0},
 /*49(0x31)*/ {"fdiv",0,272,FDIV,0},
 /*50(0x32)*/ {"fdiv",0,528,FDIV,0},
 /*51(0x33)*/ {"fdiv",0,784,FDIV,0},
 /*52(0x34)*/ {"fdiv",0,1040,FDIV,0},
 /*53(0x35)*/ {"fdiv",0,1296,FDIV,0},
 /*54(0x36)*/ {"fdiv",0,1552,FDIV,0},
 /*55(0x37)*/ {"fdiv",0,1808,FDIV,0},
 /*56(0x38)*/ {"fdivr",0,16,FDIVR,0},
 /*57(0x39)*/ {"fdivr",0,272,FDIVR,0},
 /*58(0x3A)*/ {"fdivr",0,528,FDIVR,0},
 /*59(0x3B)*/ {"fdivr",0,784,FDIVR,0},
 /*60(0x3C)*/ {"fdivr",0,1040,FDIVR,0},
 /*61(0x3D)*/ {"fdivr",0,1296,FDIVR,0},
 /*62(0x3E)*/ {"fdivr",0,1552,FDIVR,0},
 /*63(0x3F)*/ {"fdivr",0,1808,FDIVR,0},
 /*64(0x40)*/ {"fadd",0,17,FADD,0},
 /*65(0x41)*/ {"fadd",0,273,FADD,0},
 /*66(0x42)*/ {"fadd",0,529,FADD,0},
 /*67(0x43)*/ {"fadd",0,785,FADD,0},
 /*68(0x44)*/ {"fadd",0,1041,FADD,0},
 /*69(0x45)*/ {"fadd",0,1297,FADD,0},
 /*70(0x46)*/ {"fadd",0,1553,FADD,0},
 /*71(0x47)*/ {"fadd",0,1809,FADD,0},
 /*72(0x48)*/ {"fmul",0,17,FMUL,0},
 /*73(0x49)*/ {"fmul",0,273,FMUL,0},
 /*74(0x4A)*/ {"fmul",0,529,FMUL,0},
 /*75(0x4B)*/ {"fmul",0,785,FMUL,0},
 /*76(0x4C)*/ {"fmul",0,1041,FMUL,0},
 /*77(0x4D)*/ {"fmul",0,1297,FMUL,0},
 /*78(0x4E)*/ {"fmul",0,1553,FMUL,0},
 /*79(0x4F)*/ {"fmul",0,1809,FMUL,0},
 /*80(0x50)*/ {"fcom",0,17,FCOM,0},
 /*81(0x51)*/ {"fcom",0,273,FCOM,0},
 /*82(0x52)*/ {"fcom",0,529,FCOM,0},
 /*83(0x53)*/ {"fcom",0,785,FCOM,0},
 /*84(0x54)*/ {"fcom",0,1041,FCOM,0},
 /*85(0x55)*/ {"fcom",0,1297,FCOM,0},
 /*86(0x56)*/ {"fcom",0,1553,FCOM,0},
 /*87(0x57)*/ {"fcom",0,1809,FCOM,0},
 /*88(0x58)*/ {"fcomp",0,17,FCOMP,0},
 /*89(0x59)*/ {"fcomp",0,273,FCOMP,0},
 /*90(0x5A)*/ {"fcomp",0,529,FCOMP,0},
 /*91(0x5B)*/ {"fcomp",0,785,FCOMP,0},
 /*92(0x5C)*/ {"fcomp",0,1041,FCOMP,0},
 /*93(0x5D)*/ {"fcomp",0,1297,FCOMP,0},
 /*94(0x5E)*/ {"fcomp",0,1553,FCOMP,0},
 /*95(0x5F)*/ {"fcomp",0,1809,FCOMP,0},
 /*96(0x60)*/ {"fsub",0,17,FSUB,0},
 /*97(0x61)*/ {"fsub",0,273,FSUB,0},
 /*98(0x62)*/ {"fsub",0,529,FSUB,0},
 /*99(0x63)*/ {"fsub",0,785,FSUB,0},
 /*100(0x64)*/ {"fsub",0,1041,FSUB,0},
 /*101(0x65)*/ {"fsub",0,1297,FSUB,0},
 /*102(0x66)*/ {"fsub",0,1553,FSUB,0},
 /*103(0x67)*/ {"fsub",0,1809,FSUB,0},
 /*104(0x68)*/ {"fsubr",0,17,FSUBR,0},
 /*105(0x69)*/ {"fsubr",0,273,FSUBR,0},
 /*106(0x6A)*/ {"fsubr",0,529,FSUBR,0},
 /*107(0x6B)*/ {"fsubr",0,785,FSUBR,0},
 /*108(0x6C)*/ {"fsubr",0,1041,FSUBR,0},
 /*109(0x6D)*/ {"fsubr",0,1297,FSUBR,0},
 /*110(0x6E)*/ {"fsubr",0,1553,FSUBR,0},
 /*111(0x6F)*/ {"fsubr",0,1809,FSUBR,0},
 /*112(0x70)*/ {"fdiv",0,17,FDIV,0},
 /*113(0x71)*/ {"fdiv",0,273,FDIV,0},
 /*114(0x72)*/ {"fdiv",0,529,FDIV,0},
 /*115(0x73)*/ {"fdiv",0,785,FDIV,0},
 /*116(0x74)*/ {"fdiv",0,1041,FDIV,0},
 /*117(0x75)*/ {"fdiv",0,1297,FDIV,0},
 /*118(0x76)*/ {"fdiv",0,1553,FDIV,0},
 /*119(0x77)*/ {"fdiv",0,1809,FDIV,0},
 /*120(0x78)*/ {"fdivr",0,17,FDIVR,0},
 /*121(0x79)*/ {"fdivr",0,273,FDIVR,0},
 /*122(0x7A)*/ {"fdivr",0,529,FDIVR,0},
 /*123(0x7B)*/ {"fdivr",0,785,FDIVR,0},
 /*124(0x7C)*/ {"fdivr",0,1041,FDIVR,0},
 /*125(0x7D)*/ {"fdivr",0,1297,FDIVR,0},
 /*126(0x7E)*/ {"fdivr",0,1553,FDIVR,0},
 /*127(0x7F)*/ {"fdivr",0,1809,FDIVR,0},
 /*128(0x80)*/ {"fadd",0,18,FADD,0},
 /*129(0x81)*/ {"fadd",0,274,FADD,0},
 /*130(0x82)*/ {"fadd",0,530,FADD,0},
 /*131(0x83)*/ {"fadd",0,786,FADD,0},
 /*132(0x84)*/ {"fadd",0,1042,FADD,0},
 /*133(0x85)*/ {"fadd",0,1298,FADD,0},
 /*134(0x86)*/ {"fadd",0,1554,FADD,0},
 /*135(0x87)*/ {"fadd",0,1810,FADD,0},
 /*136(0x88)*/ {"fmul",0,18,FMUL,0},
 /*137(0x89)*/ {"fmul",0,274,FMUL,0},
 /*138(0x8A)*/ {"fmul",0,530,FMUL,0},
 /*139(0x8B)*/ {"fmul",0,786,FMUL,0},
 /*140(0x8C)*/ {"fmul",0,1042,FMUL,0},
 /*141(0x8D)*/ {"fmul",0,1298,FMUL,0},
 /*142(0x8E)*/ {"fmul",0,1554,FMUL,0},
 /*143(0x8F)*/ {"fmul",0,1810,FMUL,0},
 /*144(0x90)*/ {"fcom",0,18,FCOM,0},
 /*145(0x91)*/ {"fcom",0,274,FCOM,0},
 /*146(0x92)*/ {"fcom",0,530,FCOM,0},
 /*147(0x93)*/ {"fcom",0,786,FCOM,0},
 /*148(0x94)*/ {"fcom",0,1042,FCOM,0},
 /*149(0x95)*/ {"fcom",0,1298,FCOM,0},
 /*150(0x96)*/ {"fcom",0,1554,FCOM,0},
 /*151(0x97)*/ {"fcom",0,1810,FCOM,0},
 /*152(0x98)*/ {"fcomp",0,18,FCOMP,0},
 /*153(0x99)*/ {"fcomp",0,274,FCOMP,0},
 /*154(0x9A)*/ {"fcomp",0,530,FCOMP,0},
 /*155(0x9B)*/ {"fcomp",0,786,FCOMP,0},
 /*156(0x9C)*/ {"fcomp",0,1042,FCOMP,0},
 /*157(0x9D)*/ {"fcomp",0,1298,FCOMP,0},
 /*158(0x9E)*/ {"fcomp",0,1554,FCOMP,0},
 /*159(0x9F)*/ {"fcomp",0,1810,FCOMP,0},
 /*160(0xA0)*/ {"fsub",0,18,FSUB,0},
 /*161(0xA1)*/ {"fsub",0,274,FSUB,0},
 /*162(0xA2)*/ {"fsub",0,530,FSUB,0},
 /*163(0xA3)*/ {"fsub",0,786,FSUB,0},
 /*164(0xA4)*/ {"fsub",0,1042,FSUB,0},
 /*165(0xA5)*/ {"fsub",0,1298,FSUB,0},
 /*166(0xA6)*/ {"fsub",0,1554,FSUB,0},
 /*167(0xA7)*/ {"fsub",0,1810,FSUB,0},
 /*168(0xA8)*/ {"fsubr",0,18,FSUBR,0},
 /*169(0xA9)*/ {"fsubr",0,274,FSUBR,0},
 /*170(0xAA)*/ {"fsubr",0,530,FSUBR,0},
 /*171(0xAB)*/ {"fsubr",0,786,FSUBR,0},
 /*172(0xAC)*/ {"fsubr",0,1042,FSUBR,0},
 /*173(0xAD)*/ {"fsubr",0,1298,FSUBR,0},
 /*174(0xAE)*/ {"fsubr",0,1554,FSUBR,0},
 /*175(0xAF)*/ {"fsubr",0,1810,FSUBR,0},
 /*176(0xB0)*/ {"fdiv",0,18,FDIV,0},
 /*177(0xB1)*/ {"fdiv",0,274,FDIV,0},
 /*178(0xB2)*/ {"fdiv",0,530,FDIV,0},
 /*179(0xB3)*/ {"fdiv",0,786,FDIV,0},
 /*180(0xB4)*/ {"fdiv",0,1042,FDIV,0},
 /*181(0xB5)*/ {"fdiv",0,1298,FDIV,0},
 /*182(0xB6)*/ {"fdiv",0,1554,FDIV,0},
 /*183(0xB7)*/ {"fdiv",0,1810,FDIV,0},
 /*184(0xB8)*/ {"fdivr",0,18,FDIVR,0},
 /*185(0xB9)*/ {"fdivr",0,274,FDIVR,0},
 /*186(0xBA)*/ {"fdivr",0,530,FDIVR,0},
 /*187(0xBB)*/ {"fdivr",0,786,FDIVR,0},
 /*188(0xBC)*/ {"fdivr",0,1042,FDIVR,0},
 /*189(0xBD)*/ {"fdivr",0,1298,FDIVR,0},
 /*190(0xBE)*/ {"fdivr",0,1554,FDIVR,0},
 /*191(0xBF)*/ {"fdivr",0,1810,FDIVR,0},
 /*192(0xC0)*/ {"fadd",F_CP,131074,FADD,0},
 /*193(0xC1)*/ {"fadd",F_CP,131330,FADD,0},
 /*194(0xC2)*/ {"fadd",F_CP,131586,FADD,0},
 /*195(0xC3)*/ {"fadd",F_CP,131842,FADD,0},
 /*196(0xC4)*/ {"fadd",F_CP,132098,FADD,0},
 /*197(0xC5)*/ {"fadd",F_CP,132354,FADD,0},
 /*198(0xC6)*/ {"fadd",F_CP,132610,FADD,0},
 /*199(0xC7)*/ {"fadd",F_CP,132866,FADD,0},
 /*200(0xC8)*/ {"fmul",F_CP,131074,FMUL,0},
 /*201(0xC9)*/ {"fmul",F_CP,131330,FMUL,0},
 /*202(0xCA)*/ {"fmul",F_CP,131586,FMUL,0},
 /*203(0xCB)*/ {"fmul",F_CP,131842,FMUL,0},
 /*204(0xCC)*/ {"fmul",F_CP,132098,FMUL,0},
 /*205(0xCD)*/ {"fmul",F_CP,132354,FMUL,0},
 /*206(0xCE)*/ {"fmul",F_CP,132610,FMUL,0},
 /*207(0xCF)*/ {"fmul",F_CP,132866,FMUL,0},
 /*208(0xD0)*/ {0,F_CP,131074,0,0},
 /*209(0xD1)*/ {0,F_CP,131330,0,0},
 /*210(0xD2)*/ {0,F_CP,131586,0,0},
 /*211(0xD3)*/ {0,F_CP,131842,0,0},
 /*212(0xD4)*/ {0,F_CP,132098,0,0},
 /*213(0xD5)*/ {0,F_CP,132354,0,0},
 /*214(0xD6)*/ {0,F_CP,132610,0,0},
 /*215(0xD7)*/ {0,F_CP,132866,0,0},
 /*216(0xD8)*/ {0,F_CP,131074,0,0},
 /*217(0xD9)*/ {0,F_CP,131330,0,0},
 /*218(0xDA)*/ {0,F_CP,131586,0,0},
 /*219(0xDB)*/ {0,F_CP,131842,0,0},
 /*220(0xDC)*/ {0,F_CP,132098,0,0},
 /*221(0xDD)*/ {0,F_CP,132354,0,0},
 /*222(0xDE)*/ {0,F_CP,132610,0,0},
 /*223(0xDF)*/ {0,F_CP,132866,0,0},
 /*224(0xE0)*/ {"fsubr",F_CP,131074,FSUBR,0},
 /*225(0xE1)*/ {"fsubr",F_CP,131330,FSUBR,0},
 /*226(0xE2)*/ {"fsubr",F_CP,131586,FSUBR,0},
 /*227(0xE3)*/ {"fsubr",F_CP,131842,FSUBR,0},
 /*228(0xE4)*/ {"fsubr",F_CP,132098,FSUBR,0},
 /*229(0xE5)*/ {"fsubr",F_CP,132354,FSUBR,0},
 /*230(0xE6)*/ {"fsubr",F_CP,132610,FSUBR,0},
 /*231(0xE7)*/ {"fsubr",F_CP,132866,FSUBR,0},
 /*232(0xE8)*/ {"fsub",F_CP,131074,FSUB,0},
 /*233(0xE9)*/ {"fsub",F_CP,131330,FSUB,0},
 /*234(0xEA)*/ {"fsub",F_CP,131586,FSUB,0},
 /*235(0xEB)*/ {"fsub",F_CP,131842,FSUB,0},
 /*236(0xEC)*/ {"fsub",F_CP,132098,FSUB,0},
 /*237(0xED)*/ {"fsub",F_CP,132354,FSUB,0},
 /*238(0xEE)*/ {"fsub",F_CP,132610,FSUB,0},
 /*239(0xEF)*/ {"fsub",F_CP,132866,FSUB,0},
 /*240(0xF0)*/ {"fdivr",F_CP,131074,FDIVR,0},
 /*241(0xF1)*/ {"fdivr",F_CP,131330,FDIVR,0},
 /*242(0xF2)*/ {"fdivr",F_CP,131586,FDIVR,0},
 /*243(0xF3)*/ {"fdivr",F_CP,131842,FDIVR,0},
 /*244(0xF4)*/ {"fdivr",F_CP,132098,FDIVR,0},
 /*245(0xF5)*/ {"fdivr",F_CP,132354,FDIVR,0},
 /*246(0xF6)*/ {"fdivr",F_CP,132610,FDIVR,0},
 /*247(0xF7)*/ {"fdivr",F_CP,132866,FDIVR,0},
 /*248(0xF8)*/ {"fdiv",F_CP,131074,FDIV,0},
 /*249(0xF9)*/ {"fdiv",F_CP,131330,FDIV,0},
 /*250(0xFA)*/ {"fdiv",F_CP,131586,FDIV,0},
 /*251(0xFB)*/ {"fdiv",F_CP,131842,FDIV,0},
 /*252(0xFC)*/ {"fdiv",F_CP,132098,FDIV,0},
 /*253(0xFD)*/ {"fdiv",F_CP,132354,FDIV,0},
 /*254(0xFE)*/ {"fdiv",F_CP,132610,FDIV,0},
 /*255(0xFF)*/ {"fdiv",F_CP,132866,FDIV,0}
};

//grupa DD - instrukcje koprocesora
op_t gDD[256]=
{
 /*0(0x00)*/ {"fld",F_PAR+F_QW+F_D,16,FLD,0},
 /*1(0x01)*/ {"fld",F_PAR+F_QW+F_D,272,FLD,0},
 /*2(0x02)*/ {"fld",F_PAR+F_QW+F_D,528,FLD,0},
 /*3(0x03)*/ {"fld",F_PAR+F_QW+F_D,784,FLD,0},
 /*4(0x04)*/ {"fld",F_PAR+F_QW+F_D,1040,FLD,0},
 /*5(0x05)*/ {"fld",F_PAR+F_QW+F_D,1296,FLD,0},
 /*6(0x06)*/ {"fld",F_PAR+F_QW+F_D,1552,FLD,0},
 /*7(0x07)*/ {"fld",F_PAR+F_QW+F_D,1808,FLD,0},
 /*8(0x08)*/ {0,F_PAR+F_QW+F_D,16,0,0},
 /*9(0x09)*/ {0,F_PAR+F_QW+F_D,272,0,0},
 /*10(0x0A)*/ {0,F_PAR+F_QW+F_D,528,0,0},
 /*11(0x0B)*/ {0,F_PAR+F_QW+F_D,784,0,0},
 /*12(0x0C)*/ {0,F_PAR+F_QW+F_D,1040,0,0},
 /*13(0x0D)*/ {0,F_PAR+F_QW+F_D,1296,0,0},
 /*14(0x0E)*/ {0,F_PAR+F_QW+F_D,1552,0,0},
 /*15(0x0F)*/ {0,F_PAR+F_QW+F_D,1808,0,0},
 /*16(0x10)*/ {"fst",F_PAR+F_QW+F_D,16,FST,0},
 /*17(0x11)*/ {"fst",F_PAR+F_QW+F_D,272,FST,0},
 /*18(0x12)*/ {"fst",F_PAR+F_QW+F_D,528,FST,0},
 /*19(0x13)*/ {"fst",F_PAR+F_QW+F_D,784,FST,0},
 /*20(0x14)*/ {"fst",F_PAR+F_QW+F_D,1040,FST,0},
 /*21(0x15)*/ {"fst",F_PAR+F_QW+F_D,1296,FST,0},
 /*22(0x16)*/ {"fst",F_PAR+F_QW+F_D,1552,FST,0},
 /*23(0x17)*/ {"fst",F_PAR+F_QW+F_D,1808,FST,0},
 /*24(0x18)*/ {"fstp",F_PAR+F_QW+F_D,16,FSTP,0},
 /*25(0x19)*/ {"fstp",F_PAR+F_QW+F_D,272,FSTP,0},
 /*26(0x1A)*/ {"fstp",F_PAR+F_QW+F_D,528,FSTP,0},
 /*27(0x1B)*/ {"fstp",F_PAR+F_QW+F_D,784,FSTP,0},
 /*28(0x1C)*/ {"fstp",F_PAR+F_QW+F_D,1040,FSTP,0},
 /*29(0x1D)*/ {"fstp",F_PAR+F_QW+F_D,1296,FSTP,0},
 /*30(0x1E)*/ {"fstp",F_PAR+F_QW+F_D,1552,FSTP,0},
 /*31(0x1F)*/ {"fstp",F_PAR+F_QW+F_D,1808,FSTP,0},
 /*32(0x20)*/ {"frstor",F_PAR,16,FRSTOR,0},
 /*33(0x21)*/ {"frstor",F_PAR,272,FRSTOR,0},
 /*34(0x22)*/ {"frstor",F_PAR,528,FRSTOR,0},
 /*35(0x23)*/ {"frstor",F_PAR,784,FRSTOR,0},
 /*36(0x24)*/ {"frstor",F_PAR,1040,FRSTOR,0},
 /*37(0x25)*/ {"frstor",F_PAR,1296,FRSTOR,0},
 /*38(0x26)*/ {"frstor",F_PAR,1552,FRSTOR,0},
 /*39(0x27)*/ {"frstor",F_PAR,1808,FRSTOR,0},
 /*40(0x28)*/ {0,F_PAR,16,0,0},
 /*41(0x29)*/ {0,F_PAR,272,0,0},
 /*42(0x2A)*/ {0,F_PAR,528,0,0},
 /*43(0x2B)*/ {0,F_PAR,784,0,0},
 /*44(0x2C)*/ {0,F_PAR,1040,0,0},
 /*45(0x2D)*/ {0,F_PAR,1296,0,0},
 /*46(0x2E)*/ {0,F_PAR,1552,0,0},
 /*47(0x2F)*/ {0,F_PAR,1808,0,0},
 /*48(0x30)*/ {"fsave",F_PAR,16,FSAVE,0},
 /*49(0x31)*/ {"fsave",F_PAR,272,FSAVE,0},
 /*50(0x32)*/ {"fsave",F_PAR,528,FSAVE,0},
 /*51(0x33)*/ {"fsave",F_PAR,784,FSAVE,0},
 /*52(0x34)*/ {"fsave",F_PAR,1040,FSAVE,0},
 /*53(0x35)*/ {"fsave",F_PAR,1296,FSAVE,0},
 /*54(0x36)*/ {"fsave",F_PAR,1552,FSAVE,0},
 /*55(0x37)*/ {"fsave",F_PAR,1808,FSAVE,0},
 /*56(0x38)*/ {"fstsw",F_PAR,16,FSTSW,0},
 /*57(0x39)*/ {"fstsw",F_PAR,272,FSTSW,0},
 /*58(0x3A)*/ {"fstsw",F_PAR,528,FSTSW,0},
 /*59(0x3B)*/ {"fstsw",F_PAR,784,FSTSW,0},
 /*60(0x3C)*/ {"fstsw",F_PAR,1040,FSTSW,0},
 /*61(0x3D)*/ {"fstsw",F_PAR,1296,FSTSW,0},
 /*62(0x3E)*/ {"fstsw",F_PAR,1552,FSTSW,0},
 /*63(0x3F)*/ {"fstsw",F_PAR,1808,FSTSW,0},
 /*64(0x40)*/ {"fld",F_PAR+F_QW+F_D,17,FLD,0},
 /*65(0x41)*/ {"fld",F_PAR+F_QW+F_D,273,FLD,0},
 /*66(0x42)*/ {"fld",F_PAR+F_QW+F_D,529,FLD,0},
 /*67(0x43)*/ {"fld",F_PAR+F_QW+F_D,785,FLD,0},
 /*68(0x44)*/ {"fld",F_PAR+F_QW+F_D,1041,FLD,0},
 /*69(0x45)*/ {"fld",F_PAR+F_QW+F_D,1297,FLD,0},
 /*70(0x46)*/ {"fld",F_PAR+F_QW+F_D,1553,FLD,0},
 /*71(0x47)*/ {"fld",F_PAR+F_QW+F_D,1809,FLD,0},
 /*72(0x48)*/ {0,F_PAR+F_QW+F_D,17,0,0},
 /*73(0x49)*/ {0,F_PAR+F_QW+F_D,273,0,0},
 /*74(0x4A)*/ {0,F_PAR+F_QW+F_D,529,0,0},
 /*75(0x4B)*/ {0,F_PAR+F_QW+F_D,785,0,0},
 /*76(0x4C)*/ {0,F_PAR+F_QW+F_D,1041,0,0},
 /*77(0x4D)*/ {0,F_PAR+F_QW+F_D,1297,0,0},
 /*78(0x4E)*/ {0,F_PAR+F_QW+F_D,1553,0,0},
 /*79(0x4F)*/ {0,F_PAR+F_QW+F_D,1809,0,0},
 /*80(0x50)*/ {"fst",F_PAR+F_QW+F_D,17,FST,0},
 /*81(0x51)*/ {"fst",F_PAR+F_QW+F_D,273,FST,0},
 /*82(0x52)*/ {"fst",F_PAR+F_QW+F_D,529,FST,0},
 /*83(0x53)*/ {"fst",F_PAR+F_QW+F_D,785,FST,0},
 /*84(0x54)*/ {"fst",F_PAR+F_QW+F_D,1041,FST,0},
 /*85(0x55)*/ {"fst",F_PAR+F_QW+F_D,1297,FST,0},
 /*86(0x56)*/ {"fst",F_PAR+F_QW+F_D,1553,FST,0},
 /*87(0x57)*/ {"fst",F_PAR+F_QW+F_D,1809,FST,0},
 /*88(0x58)*/ {"fstp",F_PAR+F_QW+F_D,17,FSTP,0},
 /*89(0x59)*/ {"fstp",F_PAR+F_QW+F_D,273,FSTP,0},
 /*90(0x5A)*/ {"fstp",F_PAR+F_QW+F_D,529,FSTP,0},
 /*91(0x5B)*/ {"fstp",F_PAR+F_QW+F_D,785,FSTP,0},
 /*92(0x5C)*/ {"fstp",F_PAR+F_QW+F_D,1041,FSTP,0},
 /*93(0x5D)*/ {"fstp",F_PAR+F_QW+F_D,1297,FSTP,0},
 /*94(0x5E)*/ {"fstp",F_PAR+F_QW+F_D,1553,FSTP,0},
 /*95(0x5F)*/ {"fstp",F_PAR+F_QW+F_D,1809,FSTP,0},
 /*96(0x60)*/ {"frstor",F_PAR,17,FRSTOR,0},
 /*97(0x61)*/ {"frstor",F_PAR,273,FRSTOR,0},
 /*98(0x62)*/ {"frstor",F_PAR,529,FRSTOR,0},
 /*99(0x63)*/ {"frstor",F_PAR,785,FRSTOR,0},
 /*100(0x64)*/ {"frstor",F_PAR,1041,FRSTOR,0},
 /*101(0x65)*/ {"frstor",F_PAR,1297,FRSTOR,0},
 /*102(0x66)*/ {"frstor",F_PAR,1553,FRSTOR,0},
 /*103(0x67)*/ {"frstor",F_PAR,1809,FRSTOR,0},
 /*104(0x68)*/ {0,F_PAR,17,0,0},
 /*105(0x69)*/ {0,F_PAR,273,0,0},
 /*106(0x6A)*/ {0,F_PAR,529,0,0},
 /*107(0x6B)*/ {0,F_PAR,785,0,0},
 /*108(0x6C)*/ {0,F_PAR,1041,0,0},
 /*109(0x6D)*/ {0,F_PAR,1297,0,0},
 /*110(0x6E)*/ {0,F_PAR,1553,0,0},
 /*111(0x6F)*/ {0,F_PAR,1809,0,0},
 /*112(0x70)*/ {"fsave",F_PAR,17,FSAVE,0},
 /*113(0x71)*/ {"fsave",F_PAR,273,FSAVE,0},
 /*114(0x72)*/ {"fsave",F_PAR,529,FSAVE,0},
 /*115(0x73)*/ {"fsave",F_PAR,785,FSAVE,0},
 /*116(0x74)*/ {"fsave",F_PAR,1041,FSAVE,0},
 /*117(0x75)*/ {"fsave",F_PAR,1297,FSAVE,0},
 /*118(0x76)*/ {"fsave",F_PAR,1553,FSAVE,0},
 /*119(0x77)*/ {"fsave",F_PAR,1809,FSAVE,0},
 /*120(0x78)*/ {"fstsw",F_PAR,17,FSTSW,0},
 /*121(0x79)*/ {"fstsw",F_PAR,273,FSTSW,0},
 /*122(0x7A)*/ {"fstsw",F_PAR,529,FSTSW,0},
 /*123(0x7B)*/ {"fstsw",F_PAR,785,FSTSW,0},
 /*124(0x7C)*/ {"fstsw",F_PAR,1041,FSTSW,0},
 /*125(0x7D)*/ {"fstsw",F_PAR,1297,FSTSW,0},
 /*126(0x7E)*/ {"fstsw",F_PAR,1553,FSTSW,0},
 /*127(0x7F)*/ {"fstsw",F_PAR,1809,FSTSW,0},
 /*128(0x80)*/ {"fld",F_PAR+F_QW+F_D,18,FLD,0},
 /*129(0x81)*/ {"fld",F_PAR+F_QW+F_D,274,FLD,0},
 /*130(0x82)*/ {"fld",F_PAR+F_QW+F_D,530,FLD,0},
 /*131(0x83)*/ {"fld",F_PAR+F_QW+F_D,786,FLD,0},
 /*132(0x84)*/ {"fld",F_PAR+F_QW+F_D,1042,FLD,0},
 /*133(0x85)*/ {"fld",F_PAR+F_QW+F_D,1298,FLD,0},
 /*134(0x86)*/ {"fld",F_PAR+F_QW+F_D,1554,FLD,0},
 /*135(0x87)*/ {"fld",F_PAR+F_QW+F_D,1810,FLD,0},
 /*136(0x88)*/ {0,F_PAR+F_QW+F_D,18,0,0},
 /*137(0x89)*/ {0,F_PAR+F_QW+F_D,274,0,0},
 /*138(0x8A)*/ {0,F_PAR+F_QW+F_D,530,0,0},
 /*139(0x8B)*/ {0,F_PAR+F_QW+F_D,786,0,0},
 /*140(0x8C)*/ {0,F_PAR+F_QW+F_D,1042,0,0},
 /*141(0x8D)*/ {0,F_PAR+F_QW+F_D,1298,0,0},
 /*142(0x8E)*/ {0,F_PAR+F_QW+F_D,1554,0,0},
 /*143(0x8F)*/ {0,F_PAR+F_QW+F_D,1810,0,0},
 /*144(0x90)*/ {"fst",F_PAR+F_QW+F_D,18,FST,0},
 /*145(0x91)*/ {"fst",F_PAR+F_QW+F_D,274,FST,0},
 /*146(0x92)*/ {"fst",F_PAR+F_QW+F_D,530,FST,0},
 /*147(0x93)*/ {"fst",F_PAR+F_QW+F_D,786,FST,0},
 /*148(0x94)*/ {"fst",F_PAR+F_QW+F_D,1042,FST,0},
 /*149(0x95)*/ {"fst",F_PAR+F_QW+F_D,1298,FST,0},
 /*150(0x96)*/ {"fst",F_PAR+F_QW+F_D,1554,FST,0},
 /*151(0x97)*/ {"fst",F_PAR+F_QW+F_D,1810,FST,0},
 /*152(0x98)*/ {"fstp",F_PAR+F_QW+F_D,18,FSTP,0},
 /*153(0x99)*/ {"fstp",F_PAR+F_QW+F_D,274,FSTP,0},
 /*154(0x9A)*/ {"fstp",F_PAR+F_QW+F_D,530,FSTP,0},
 /*155(0x9B)*/ {"fstp",F_PAR+F_QW+F_D,786,FSTP,0},
 /*156(0x9C)*/ {"fstp",F_PAR+F_QW+F_D,1042,FSTP,0},
 /*157(0x9D)*/ {"fstp",F_PAR+F_QW+F_D,1298,FSTP,0},
 /*158(0x9E)*/ {"fstp",F_PAR+F_QW+F_D,1554,FSTP,0},
 /*159(0x9F)*/ {"fstp",F_PAR+F_QW+F_D,1810,FSTP,0},
 /*160(0xA0)*/ {"frstor",F_PAR,18,FRSTOR,0},
 /*161(0xA1)*/ {"frstor",F_PAR,274,FRSTOR,0},
 /*162(0xA2)*/ {"frstor",F_PAR,530,FRSTOR,0},
 /*163(0xA3)*/ {"frstor",F_PAR,786,FRSTOR,0},
 /*164(0xA4)*/ {"frstor",F_PAR,1042,FRSTOR,0},
 /*165(0xA5)*/ {"frstor",F_PAR,1298,FRSTOR,0},
 /*166(0xA6)*/ {"frstor",F_PAR,1554,FRSTOR,0},
 /*167(0xA7)*/ {"frstor",F_PAR,1810,FRSTOR,0},
 /*168(0xA8)*/ {0,F_PAR,18,0,0},
 /*169(0xA9)*/ {0,F_PAR,274,0,0},
 /*170(0xAA)*/ {0,F_PAR,530,0,0},
 /*171(0xAB)*/ {0,F_PAR,786,0,0},
 /*172(0xAC)*/ {0,F_PAR,1042,0,0},
 /*173(0xAD)*/ {0,F_PAR,1298,0,0},
 /*174(0xAE)*/ {0,F_PAR,1554,0,0},
 /*175(0xAF)*/ {0,F_PAR,1810,0,0},
 /*176(0xB0)*/ {"fsave",F_PAR,18,FSAVE,0},
 /*177(0xB1)*/ {"fsave",F_PAR,274,FSAVE,0},
 /*178(0xB2)*/ {"fsave",F_PAR,530,FSAVE,0},
 /*179(0xB3)*/ {"fsave",F_PAR,786,FSAVE,0},
 /*180(0xB4)*/ {"fsave",F_PAR,1042,FSAVE,0},
 /*181(0xB5)*/ {"fsave",F_PAR,1298,FSAVE,0},
 /*182(0xB6)*/ {"fsave",F_PAR,1554,FSAVE,0},
 /*183(0xB7)*/ {"fsave",F_PAR,1810,FSAVE,0},
 /*184(0xB8)*/ {"fstsw",F_PAR,18,FSTSW,0},
 /*185(0xB9)*/ {"fstsw",F_PAR,274,FSTSW,0},
 /*186(0xBA)*/ {"fstsw",F_PAR,530,FSTSW,0},
 /*187(0xBB)*/ {"fstsw",F_PAR,786,FSTSW,0},
 /*188(0xBC)*/ {"fstsw",F_PAR,1042,FSTSW,0},
 /*189(0xBD)*/ {"fstsw",F_PAR,1298,FSTSW,0},
 /*190(0xBE)*/ {"fstsw",F_PAR,1554,FSTSW,0},
 /*191(0xBF)*/ {"fstsw",F_PAR,1810,FSTSW,0},
 /*192(0xC0)*/ {"ffree",F_PAR,2,FFREE,0},
 /*193(0xC1)*/ {"ffree",F_PAR,258,FFREE,0},
 /*194(0xC2)*/ {"ffree",F_PAR,514,FFREE,0},
 /*195(0xC3)*/ {"ffree",F_PAR,770,FFREE,0},
 /*196(0xC4)*/ {"ffree",F_PAR,1026,FFREE,0},
 /*197(0xC5)*/ {"ffree",F_PAR,1282,FFREE,0},
 /*198(0xC6)*/ {"ffree",F_PAR,1538,FFREE,0},
 /*199(0xC7)*/ {"ffree",F_PAR,1794,FFREE,0},
 /*200(0xC8)*/ {0,F_PAR,2,0,0},
 /*201(0xC9)*/ {0,F_PAR,258,0,0},
 /*202(0xCA)*/ {0,F_PAR,514,0,0},
 /*203(0xCB)*/ {0,F_PAR,770,0,0},
 /*204(0xCC)*/ {0,F_PAR,1026,0,0},
 /*205(0xCD)*/ {0,F_PAR,1282,0,0},
 /*206(0xCE)*/ {0,F_PAR,1538,0,0},
 /*207(0xCF)*/ {0,F_PAR,1794,0,0},
 /*208(0xD0)*/ {"fst",F_PAR,2,FST,0},
 /*209(0xD1)*/ {"fst",F_PAR,258,FST,0},
 /*210(0xD2)*/ {"fst",F_PAR,514,FST,0},
 /*211(0xD3)*/ {"fst",F_PAR,770,FST,0},
 /*212(0xD4)*/ {"fst",F_PAR,1026,FST,0},
 /*213(0xD5)*/ {"fst",F_PAR,1282,FST,0},
 /*214(0xD6)*/ {"fst",F_PAR,1538,FST,0},
 /*215(0xD7)*/ {"fst",F_PAR,1794,FST,0},
 /*216(0xD8)*/ {"fstp",F_PAR,2,FSTP,0},
 /*217(0xD9)*/ {"fstp",F_PAR,258,FSTP,0},
 /*218(0xDA)*/ {"fstp",F_PAR,514,FSTP,0},
 /*219(0xDB)*/ {"fstp",F_PAR,770,FSTP,0},
 /*220(0xDC)*/ {"fstp",F_PAR,1026,FSTP,0},
 /*221(0xDD)*/ {"fstp",F_PAR,1282,FSTP,0},
 /*222(0xDE)*/ {"fstp",F_PAR,1538,FSTP,0},
 /*223(0xDF)*/ {"fstp",F_PAR,1794,FSTP,0},
 /*224(0xE0)*/ {"fucom",F_PAR,2,FUCOM,0},
 /*225(0xE1)*/ {"fucom",F_PAR,258,FUCOM,0},
 /*226(0xE2)*/ {"fucom",F_PAR,514,FUCOM,0},
 /*227(0xE3)*/ {"fucom",F_PAR,770,FUCOM,0},
 /*228(0xE4)*/ {"fucom",F_PAR,1026,FUCOM,0},
 /*229(0xE5)*/ {"fucom",F_PAR,1282,FUCOM,0},
 /*230(0xE6)*/ {"fucom",F_PAR,1538,FUCOM,0},
 /*231(0xE7)*/ {"fucom",F_PAR,1794,FUCOM,0},
 /*232(0xE8)*/ {"fucomp",F_PAR,2,FUCOMP,0},
 /*233(0xE9)*/ {"fucomp",F_PAR,258,FUCOMP,0},
 /*234(0xEA)*/ {"fucomp",F_PAR,514,FUCOMP,0},
 /*235(0xEB)*/ {"fucomp",F_PAR,770,FUCOMP,0},
 /*236(0xEC)*/ {"fucomp",F_PAR,1026,FUCOMP,0},
 /*237(0xED)*/ {"fucomp",F_PAR,1282,FUCOMP,0},
 /*238(0xEE)*/ {"fucomp",F_PAR,1538,FUCOMP,0},
 /*239(0xEF)*/ {"fucomp",F_PAR,1794,FUCOMP,0},
 /*240(0xF0)*/ {0,F_PAR,2,0,0},
 /*241(0xF1)*/ {0,F_PAR,258,0,0},
 /*242(0xF2)*/ {0,F_PAR,514,0,0},
 /*243(0xF3)*/ {0,F_PAR,770,0,0},
 /*244(0xF4)*/ {0,F_PAR,1026,0,0},
 /*245(0xF5)*/ {0,F_PAR,1282,0,0},
 /*246(0xF6)*/ {0,F_PAR,1538,0,0},
 /*247(0xF7)*/ {0,F_PAR,1794,0,0},
 /*248(0xF8)*/ {0,F_PAR,2,0,0},
 /*249(0xF9)*/ {0,F_PAR,258,0,0},
 /*250(0xFA)*/ {0,F_PAR,514,0,0},
 /*251(0xFB)*/ {0,F_PAR,770,0,0},
 /*252(0xFC)*/ {0,F_PAR,1026,0,0},
 /*253(0xFD)*/ {0,F_PAR,1282,0,0},
 /*254(0xFE)*/ {0,F_PAR,1538,0,0},
 /*255(0xFF)*/ {0,F_PAR,1794,0,0}
};

//grupa DE - instrukcje koprocesora
op_t gDE[256]=
{
 /*0(0x00)*/ {"fiadd",F_PAR+F_DW,16,FIADD,0},
 /*1(0x01)*/ {"fiadd",F_PAR+F_DW,272,FIADD,0},
 /*2(0x02)*/ {"fiadd",F_PAR+F_DW,528,FIADD,0},
 /*3(0x03)*/ {"fiadd",F_PAR+F_DW,784,FIADD,0},
 /*4(0x04)*/ {"fiadd",F_PAR+F_DW,1040,FIADD,0},
 /*5(0x05)*/ {"fiadd",F_PAR+F_DW,1296,FIADD,0},
 /*6(0x06)*/ {"fiadd",F_PAR+F_DW,1552,FIADD,0},
 /*7(0x07)*/ {"fiadd",F_PAR+F_DW,1808,FIADD,0},
 /*8(0x08)*/ {"fimul",F_PAR+F_DW,16,FIMUL,0},
 /*9(0x09)*/ {"fimul",F_PAR+F_DW,272,FIMUL,0},
 /*10(0x0A)*/ {"fimul",F_PAR+F_DW,528,FIMUL,0},
 /*11(0x0B)*/ {"fimul",F_PAR+F_DW,784,FIMUL,0},
 /*12(0x0C)*/ {"fimul",F_PAR+F_DW,1040,FIMUL,0},
 /*13(0x0D)*/ {"fimul",F_PAR+F_DW,1296,FIMUL,0},
 /*14(0x0E)*/ {"fimul",F_PAR+F_DW,1552,FIMUL,0},
 /*15(0x0F)*/ {"fimul",F_PAR+F_DW,1808,FIMUL,0},
 /*16(0x10)*/ {"ficom",F_PAR+F_DW,16,FICOM,0},
 /*17(0x11)*/ {"ficom",F_PAR+F_DW,272,FICOM,0},
 /*18(0x12)*/ {"ficom",F_PAR+F_DW,528,FICOM,0},
 /*19(0x13)*/ {"ficom",F_PAR+F_DW,784,FICOM,0},
 /*20(0x14)*/ {"ficom",F_PAR+F_DW,1040,FICOM,0},
 /*21(0x15)*/ {"ficom",F_PAR+F_DW,1296,FICOM,0},
 /*22(0x16)*/ {"ficom",F_PAR+F_DW,1552,FICOM,0},
 /*23(0x17)*/ {"ficom",F_PAR+F_DW,1808,FICOM,0},
 /*24(0x18)*/ {"ficomp",F_PAR+F_DW,16,FICOMP,0},
 /*25(0x19)*/ {"ficomp",F_PAR+F_DW,272,FICOMP,0},
 /*26(0x1A)*/ {"ficomp",F_PAR+F_DW,528,FICOMP,0},
 /*27(0x1B)*/ {"ficomp",F_PAR+F_DW,784,FICOMP,0},
 /*28(0x1C)*/ {"ficomp",F_PAR+F_DW,1040,FICOMP,0},
 /*29(0x1D)*/ {"ficomp",F_PAR+F_DW,1296,FICOMP,0},
 /*30(0x1E)*/ {"ficomp",F_PAR+F_DW,1552,FICOMP,0},
 /*31(0x1F)*/ {"ficomp",F_PAR+F_DW,1808,FICOMP,0},
 /*32(0x20)*/ {"fisub",F_PAR+F_DW,16,FISUB,0},
 /*33(0x21)*/ {"fisub",F_PAR+F_DW,272,FISUB,0},
 /*34(0x22)*/ {"fisub",F_PAR+F_DW,528,FISUB,0},
 /*35(0x23)*/ {"fisub",F_PAR+F_DW,784,FISUB,0},
 /*36(0x24)*/ {"fisub",F_PAR+F_DW,1040,FISUB,0},
 /*37(0x25)*/ {"fisub",F_PAR+F_DW,1296,FISUB,0},
 /*38(0x26)*/ {"fisub",F_PAR+F_DW,1552,FISUB,0},
 /*39(0x27)*/ {"fisub",F_PAR+F_DW,1808,FISUB,0},
 /*40(0x28)*/ {"fisubr",F_PAR+F_DW,16,FISUBR,0},
 /*41(0x29)*/ {"fisubr",F_PAR+F_DW,272,FISUBR,0},
 /*42(0x2A)*/ {"fisubr",F_PAR+F_DW,528,FISUBR,0},
 /*43(0x2B)*/ {"fisubr",F_PAR+F_DW,784,FISUBR,0},
 /*44(0x2C)*/ {"fisubr",F_PAR+F_DW,1040,FISUBR,0},
 /*45(0x2D)*/ {"fisubr",F_PAR+F_DW,1296,FISUBR,0},
 /*46(0x2E)*/ {"fisubr",F_PAR+F_DW,1552,FISUBR,0},
 /*47(0x2F)*/ {"fisubr",F_PAR+F_DW,1808,FISUBR,0},
 /*48(0x30)*/ {"fidiv",F_PAR+F_DW,16,FIDIV,0},
 /*49(0x31)*/ {"fidiv",F_PAR+F_DW,272,FIDIV,0},
 /*50(0x32)*/ {"fidiv",F_PAR+F_DW,528,FIDIV,0},
 /*51(0x33)*/ {"fidiv",F_PAR+F_DW,784,FIDIV,0},
 /*52(0x34)*/ {"fidiv",F_PAR+F_DW,1040,FIDIV,0},
 /*53(0x35)*/ {"fidiv",F_PAR+F_DW,1296,FIDIV,0},
 /*54(0x36)*/ {"fidiv",F_PAR+F_DW,1552,FIDIV,0},
 /*55(0x37)*/ {"fidiv",F_PAR+F_DW,1808,FIDIV,0},
 /*56(0x38)*/ {"fidivr",F_PAR+F_DW,16,FIDIVR,0},
 /*57(0x39)*/ {"fidivr",F_PAR+F_DW,272,FIDIVR,0},
 /*58(0x3A)*/ {"fidivr",F_PAR+F_DW,528,FIDIVR,0},
 /*59(0x3B)*/ {"fidivr",F_PAR+F_DW,784,FIDIVR,0},
 /*60(0x3C)*/ {"fidivr",F_PAR+F_DW,1040,FIDIVR,0},
 /*61(0x3D)*/ {"fidivr",F_PAR+F_DW,1296,FIDIVR,0},
 /*62(0x3E)*/ {"fidivr",F_PAR+F_DW,1552,FIDIVR,0},
 /*63(0x3F)*/ {"fidivr",F_PAR+F_DW,1808,FIDIVR,0},
 /*64(0x40)*/ {"fiadd",F_PAR+F_DW,17,FIADD,0},
 /*65(0x41)*/ {"fiadd",F_PAR+F_DW,273,FIADD,0},
 /*66(0x42)*/ {"fiadd",F_PAR+F_DW,529,FIADD,0},
 /*67(0x43)*/ {"fiadd",F_PAR+F_DW,785,FIADD,0},
 /*68(0x44)*/ {"fiadd",F_PAR+F_DW,1041,FIADD,0},
 /*69(0x45)*/ {"fiadd",F_PAR+F_DW,1297,FIADD,0},
 /*70(0x46)*/ {"fiadd",F_PAR+F_DW,1553,FIADD,0},
 /*71(0x47)*/ {"fiadd",F_PAR+F_DW,1809,FIADD,0},
 /*72(0x48)*/ {"fimul",F_PAR+F_DW,17,FIMUL,0},
 /*73(0x49)*/ {"fimul",F_PAR+F_DW,273,FIMUL,0},
 /*74(0x4A)*/ {"fimul",F_PAR+F_DW,529,FIMUL,0},
 /*75(0x4B)*/ {"fimul",F_PAR+F_DW,785,FIMUL,0},
 /*76(0x4C)*/ {"fimul",F_PAR+F_DW,1041,FIMUL,0},
 /*77(0x4D)*/ {"fimul",F_PAR+F_DW,1297,FIMUL,0},
 /*78(0x4E)*/ {"fimul",F_PAR+F_DW,1553,FIMUL,0},
 /*79(0x4F)*/ {"fimul",F_PAR+F_DW,1809,FIMUL,0},
 /*80(0x50)*/ {"ficom",F_PAR+F_DW,17,FICOM,0},
 /*81(0x51)*/ {"ficom",F_PAR+F_DW,273,FICOM,0},
 /*82(0x52)*/ {"ficom",F_PAR+F_DW,529,FICOM,0},
 /*83(0x53)*/ {"ficom",F_PAR+F_DW,785,FICOM,0},
 /*84(0x54)*/ {"ficom",F_PAR+F_DW,1041,FICOM,0},
 /*85(0x55)*/ {"ficom",F_PAR+F_DW,1297,FICOM,0},
 /*86(0x56)*/ {"ficom",F_PAR+F_DW,1553,FICOM,0},
 /*87(0x57)*/ {"ficom",F_PAR+F_DW,1809,FICOM,0},
 /*88(0x58)*/ {"ficomp",F_PAR+F_DW,17,FICOMP,0},
 /*89(0x59)*/ {"ficomp",F_PAR+F_DW,273,FICOMP,0},
 /*90(0x5A)*/ {"ficomp",F_PAR+F_DW,529,FICOMP,0},
 /*91(0x5B)*/ {"ficomp",F_PAR+F_DW,785,FICOMP,0},
 /*92(0x5C)*/ {"ficomp",F_PAR+F_DW,1041,FICOMP,0},
 /*93(0x5D)*/ {"ficomp",F_PAR+F_DW,1297,FICOMP,0},
 /*94(0x5E)*/ {"ficomp",F_PAR+F_DW,1553,FICOMP,0},
 /*95(0x5F)*/ {"ficomp",F_PAR+F_DW,1809,FICOMP,0},
 /*96(0x60)*/ {"fisub",F_PAR+F_DW,17,FISUB,0},
 /*97(0x61)*/ {"fisub",F_PAR+F_DW,273,FISUB,0},
 /*98(0x62)*/ {"fisub",F_PAR+F_DW,529,FISUB,0},
 /*99(0x63)*/ {"fisub",F_PAR+F_DW,785,FISUB,0},
 /*100(0x64)*/ {"fisub",F_PAR+F_DW,1041,FISUB,0},
 /*101(0x65)*/ {"fisub",F_PAR+F_DW,1297,FISUB,0},
 /*102(0x66)*/ {"fisub",F_PAR+F_DW,1553,FISUB,0},
 /*103(0x67)*/ {"fisub",F_PAR+F_DW,1809,FISUB,0},
 /*104(0x68)*/ {"fisubr",F_PAR+F_DW,17,FISUBR,0},
 /*105(0x69)*/ {"fisubr",F_PAR+F_DW,273,FISUBR,0},
 /*106(0x6A)*/ {"fisubr",F_PAR+F_DW,529,FISUBR,0},
 /*107(0x6B)*/ {"fisubr",F_PAR+F_DW,785,FISUBR,0},
 /*108(0x6C)*/ {"fisubr",F_PAR+F_DW,1041,FISUBR,0},
 /*109(0x6D)*/ {"fisubr",F_PAR+F_DW,1297,FISUBR,0},
 /*110(0x6E)*/ {"fisubr",F_PAR+F_DW,1553,FISUBR,0},
 /*111(0x6F)*/ {"fisubr",F_PAR+F_DW,1809,FISUBR,0},
 /*112(0x70)*/ {"fidiv",F_PAR+F_DW,17,FIDIV,0},
 /*113(0x71)*/ {"fidiv",F_PAR+F_DW,273,FIDIV,0},
 /*114(0x72)*/ {"fidiv",F_PAR+F_DW,529,FIDIV,0},
 /*115(0x73)*/ {"fidiv",F_PAR+F_DW,785,FIDIV,0},
 /*116(0x74)*/ {"fidiv",F_PAR+F_DW,1041,FIDIV,0},
 /*117(0x75)*/ {"fidiv",F_PAR+F_DW,1297,FIDIV,0},
 /*118(0x76)*/ {"fidiv",F_PAR+F_DW,1553,FIDIV,0},
 /*119(0x77)*/ {"fidiv",F_PAR+F_DW,1809,FIDIV,0},
 /*120(0x78)*/ {"fidivr",F_PAR+F_DW,17,FIDIVR,0},
 /*121(0x79)*/ {"fidivr",F_PAR+F_DW,273,FIDIVR,0},
 /*122(0x7A)*/ {"fidivr",F_PAR+F_DW,529,FIDIVR,0},
 /*123(0x7B)*/ {"fidivr",F_PAR+F_DW,785,FIDIVR,0},
 /*124(0x7C)*/ {"fidivr",F_PAR+F_DW,1041,FIDIVR,0},
 /*125(0x7D)*/ {"fidivr",F_PAR+F_DW,1297,FIDIVR,0},
 /*126(0x7E)*/ {"fidivr",F_PAR+F_DW,1553,FIDIVR,0},
 /*127(0x7F)*/ {"fidivr",F_PAR+F_DW,1809,FIDIVR,0},
 /*128(0x80)*/ {"fiadd",F_PAR+F_DW,18,FIADD,0},
 /*129(0x81)*/ {"fiadd",F_PAR+F_DW,274,FIADD,0},
 /*130(0x82)*/ {"fiadd",F_PAR+F_DW,530,FIADD,0},
 /*131(0x83)*/ {"fiadd",F_PAR+F_DW,786,FIADD,0},
 /*132(0x84)*/ {"fiadd",F_PAR+F_DW,1042,FIADD,0},
 /*133(0x85)*/ {"fiadd",F_PAR+F_DW,1298,FIADD,0},
 /*134(0x86)*/ {"fiadd",F_PAR+F_DW,1554,FIADD,0},
 /*135(0x87)*/ {"fiadd",F_PAR+F_DW,1810,FIADD,0},
 /*136(0x88)*/ {"fimul",F_PAR+F_DW,18,FIMUL,0},
 /*137(0x89)*/ {"fimul",F_PAR+F_DW,274,FIMUL,0},
 /*138(0x8A)*/ {"fimul",F_PAR+F_DW,530,FIMUL,0},
 /*139(0x8B)*/ {"fimul",F_PAR+F_DW,786,FIMUL,0},
 /*140(0x8C)*/ {"fimul",F_PAR+F_DW,1042,FIMUL,0},
 /*141(0x8D)*/ {"fimul",F_PAR+F_DW,1298,FIMUL,0},
 /*142(0x8E)*/ {"fimul",F_PAR+F_DW,1554,FIMUL,0},
 /*143(0x8F)*/ {"fimul",F_PAR+F_DW,1810,FIMUL,0},
 /*144(0x90)*/ {"ficom",F_PAR+F_DW,18,FICOM,0},
 /*145(0x91)*/ {"ficom",F_PAR+F_DW,274,FICOM,0},
 /*146(0x92)*/ {"ficom",F_PAR+F_DW,530,FICOM,0},
 /*147(0x93)*/ {"ficom",F_PAR+F_DW,786,FICOM,0},
 /*148(0x94)*/ {"ficom",F_PAR+F_DW,1042,FICOM,0},
 /*149(0x95)*/ {"ficom",F_PAR+F_DW,1298,FICOM,0},
 /*150(0x96)*/ {"ficom",F_PAR+F_DW,1554,FICOM,0},
 /*151(0x97)*/ {"ficom",F_PAR+F_DW,1810,FICOM,0},
 /*152(0x98)*/ {"ficomp",F_PAR+F_DW,18,FICOMP,0},
 /*153(0x99)*/ {"ficomp",F_PAR+F_DW,274,FICOMP,0},
 /*154(0x9A)*/ {"ficomp",F_PAR+F_DW,530,FICOMP,0},
 /*155(0x9B)*/ {"ficomp",F_PAR+F_DW,786,FICOMP,0},
 /*156(0x9C)*/ {"ficomp",F_PAR+F_DW,1042,FICOMP,0},
 /*157(0x9D)*/ {"ficomp",F_PAR+F_DW,1298,FICOMP,0},
 /*158(0x9E)*/ {"ficomp",F_PAR+F_DW,1554,FICOMP,0},
 /*159(0x9F)*/ {"ficomp",F_PAR+F_DW,1810,FICOMP,0},
 /*160(0xA0)*/ {"fisub",F_PAR+F_DW,18,FISUB,0},
 /*161(0xA1)*/ {"fisub",F_PAR+F_DW,274,FISUB,0},
 /*162(0xA2)*/ {"fisub",F_PAR+F_DW,530,FISUB,0},
 /*163(0xA3)*/ {"fisub",F_PAR+F_DW,786,FISUB,0},
 /*164(0xA4)*/ {"fisub",F_PAR+F_DW,1042,FISUB,0},
 /*165(0xA5)*/ {"fisub",F_PAR+F_DW,1298,FISUB,0},
 /*166(0xA6)*/ {"fisub",F_PAR+F_DW,1554,FISUB,0},
 /*167(0xA7)*/ {"fisub",F_PAR+F_DW,1810,FISUB,0},
 /*168(0xA8)*/ {"fisubr",F_PAR+F_DW,18,FISUBR,0},
 /*169(0xA9)*/ {"fisubr",F_PAR+F_DW,274,FISUBR,0},
 /*170(0xAA)*/ {"fisubr",F_PAR+F_DW,530,FISUBR,0},
 /*171(0xAB)*/ {"fisubr",F_PAR+F_DW,786,FISUBR,0},
 /*172(0xAC)*/ {"fisubr",F_PAR+F_DW,1042,FISUBR,0},
 /*173(0xAD)*/ {"fisubr",F_PAR+F_DW,1298,FISUBR,0},
 /*174(0xAE)*/ {"fisubr",F_PAR+F_DW,1554,FISUBR,0},
 /*175(0xAF)*/ {"fisubr",F_PAR+F_DW,1810,FISUBR,0},
 /*176(0xB0)*/ {"fidiv",F_PAR+F_DW,18,FIDIV,0},
 /*177(0xB1)*/ {"fidiv",F_PAR+F_DW,274,FIDIV,0},
 /*178(0xB2)*/ {"fidiv",F_PAR+F_DW,530,FIDIV,0},
 /*179(0xB3)*/ {"fidiv",F_PAR+F_DW,786,FIDIV,0},
 /*180(0xB4)*/ {"fidiv",F_PAR+F_DW,1042,FIDIV,0},
 /*181(0xB5)*/ {"fidiv",F_PAR+F_DW,1298,FIDIV,0},
 /*182(0xB6)*/ {"fidiv",F_PAR+F_DW,1554,FIDIV,0},
 /*183(0xB7)*/ {"fidiv",F_PAR+F_DW,1810,FIDIV,0},
 /*184(0xB8)*/ {"fidivr",F_PAR+F_DW,18,FIDIVR,0},
 /*185(0xB9)*/ {"fidivr",F_PAR+F_DW,274,FIDIVR,0},
 /*186(0xBA)*/ {"fidivr",F_PAR+F_DW,530,FIDIVR,0},
 /*187(0xBB)*/ {"fidivr",F_PAR+F_DW,786,FIDIVR,0},
 /*188(0xBC)*/ {"fidivr",F_PAR+F_DW,1042,FIDIVR,0},
 /*189(0xBD)*/ {"fidivr",F_PAR+F_DW,1298,FIDIVR,0},
 /*190(0xBE)*/ {"fidivr",F_PAR+F_DW,1554,FIDIVR,0},
 /*191(0xBF)*/ {"fidivr",F_PAR+F_DW,1810,FIDIVR,0},
 /*192(0xC0)*/ {"faddp",F_PAR+F_CP,131074,FADDP,0},
 /*193(0xC1)*/ {"faddp",F_PAR+F_CP,131330,FADDP,0},
 /*194(0xC2)*/ {"faddp",F_PAR+F_CP,131586,FADDP,0},
 /*195(0xC3)*/ {"faddp",F_PAR+F_CP,131842,FADDP,0},
 /*196(0xC4)*/ {"faddp",F_PAR+F_CP,132098,FADDP,0},
 /*197(0xC5)*/ {"faddp",F_PAR+F_CP,132354,FADDP,0},
 /*198(0xC6)*/ {"faddp",F_PAR+F_CP,132610,FADDP,0},
 /*199(0xC7)*/ {"faddp",F_PAR+F_CP,132866,FADDP,0},
 /*200(0xC8)*/ {"fmulp",F_PAR+F_CP,131074,FMULP,0},
 /*201(0xC9)*/ {"fmulp",F_PAR+F_CP,131330,FMULP,0},
 /*202(0xCA)*/ {"fmulp",F_PAR+F_CP,131586,FMULP,0},
 /*203(0xCB)*/ {"fmulp",F_PAR+F_CP,131842,FMULP,0},
 /*204(0xCC)*/ {"fmulp",F_PAR+F_CP,132098,FMULP,0},
 /*205(0xCD)*/ {"fmulp",F_PAR+F_CP,132354,FMULP,0},
 /*206(0xCE)*/ {"fmulp",F_PAR+F_CP,132610,FMULP,0},
 /*207(0xCF)*/ {"fmulp",F_PAR+F_CP,132866,FMULP,0},
 /*208(0xD0)*/ {0,F_PAR+F_CP,131074,0,0},
 /*209(0xD1)*/ {0,F_PAR+F_CP,131330,0,0},
 /*210(0xD2)*/ {0,F_PAR+F_CP,131586,0,0},
 /*211(0xD3)*/ {0,F_PAR+F_CP,131842,0,0},
 /*212(0xD4)*/ {0,F_PAR+F_CP,132098,0,0},
 /*213(0xD5)*/ {0,F_PAR+F_CP,132354,0,0},
 /*214(0xD6)*/ {0,F_PAR+F_CP,132610,0,0},
 /*215(0xD7)*/ {0,F_PAR+F_CP,132866,0,0},
 /*216(0xD8)*/ {0,F_PAR+F_CP,131074,0,0},
 /*217(0xD9)*/ {"fcompp",0,0,FCOMPP,0},
 /*218(0xDA)*/ {0,F_PAR+F_CP,131586,0,0},
 /*219(0xDB)*/ {0,F_PAR+F_CP,131842,0,0},
 /*220(0xDC)*/ {0,F_PAR+F_CP,132098,0,0},
 /*221(0xDD)*/ {0,F_PAR+F_CP,132354,0,0},
 /*222(0xDE)*/ {0,F_PAR+F_CP,132610,0,0},
 /*223(0xDF)*/ {0,F_PAR+F_CP,132866,0,0},
 /*224(0xE0)*/ {"fsubrp",F_PAR+F_CP,131074,FSUBRP,0},
 /*225(0xE1)*/ {"fsubrp",F_PAR+F_CP,131330,FSUBRP,0},
 /*226(0xE2)*/ {"fsubrp",F_PAR+F_CP,131586,FSUBRP,0},
 /*227(0xE3)*/ {"fsubrp",F_PAR+F_CP,131842,FSUBRP,0},
 /*228(0xE4)*/ {"fsubrp",F_PAR+F_CP,132098,FSUBRP,0},
 /*229(0xE5)*/ {"fsubrp",F_PAR+F_CP,132354,FSUBRP,0},
 /*230(0xE6)*/ {"fsubrp",F_PAR+F_CP,132610,FSUBRP,0},
 /*231(0xE7)*/ {"fsubrp",F_PAR+F_CP,132866,FSUBRP,0},
 /*232(0xE8)*/ {"fsubp",F_PAR+F_CP,131074,FSUBP,0},
 /*233(0xE9)*/ {"fsubp",F_PAR+F_CP,131330,FSUBP,0},
 /*234(0xEA)*/ {"fsubp",F_PAR+F_CP,131586,FSUBP,0},
 /*235(0xEB)*/ {"fsubp",F_PAR+F_CP,131842,FSUBP,0},
 /*236(0xEC)*/ {"fsubp",F_PAR+F_CP,132098,FSUBP,0},
 /*237(0xED)*/ {"fsubp",F_PAR+F_CP,132354,FSUBP,0},
 /*238(0xEE)*/ {"fsubp",F_PAR+F_CP,132610,FSUBP,0},
 /*239(0xEF)*/ {"fsubp",F_PAR+F_CP,132866,FSUBP,0},
 /*240(0xF0)*/ {"fdivrp",F_PAR+F_CP,131074,FDIVRP,0},
 /*241(0xF1)*/ {"fdivrp",F_PAR+F_CP,131330,FDIVRP,0},
 /*242(0xF2)*/ {"fdivrp",F_PAR+F_CP,131586,FDIVRP,0},
 /*243(0xF3)*/ {"fdivrp",F_PAR+F_CP,131842,FDIVRP,0},
 /*244(0xF4)*/ {"fdivrp",F_PAR+F_CP,132098,FDIVRP,0},
 /*245(0xF5)*/ {"fdivrp",F_PAR+F_CP,132354,FDIVRP,0},
 /*246(0xF6)*/ {"fdivrp",F_PAR+F_CP,132610,FDIVRP,0},
 /*247(0xF7)*/ {"fdivrp",F_PAR+F_CP,132866,FDIVRP,0},
 /*248(0xF8)*/ {"fdivp",F_PAR+F_CP,131074,FDIVP,0},
 /*249(0xF9)*/ {"fdivp",F_PAR+F_CP,131330,FDIVP,0},
 /*250(0xFA)*/ {"fdivp",F_PAR+F_CP,131586,FDIVP,0},
 /*251(0xFB)*/ {"fdivp",F_PAR+F_CP,131842,FDIVP,0},
 /*252(0xFC)*/ {"fdivp",F_PAR+F_CP,132098,FDIVP,0},
 /*253(0xFD)*/ {"fdivp",F_PAR+F_CP,132354,FDIVP,0},
 /*254(0xFE)*/ {"fdivp",F_PAR+F_CP,132610,FDIVP,0},
 /*255(0xFF)*/ {"fdivp",F_PAR+F_CP,132866,FDIVP,0}
};

//grupa DF - instrukcje koprocesora
op_t gDF[256]=
{
 /*0(0x00)*/ {"fild",F_W,16,FILD,0},
 /*1(0x01)*/ {"fild",F_W,272,FILD,0},
 /*2(0x02)*/ {"fild",F_W,528,FILD,0},
 /*3(0x03)*/ {"fild",F_W,784,FILD,0},
 /*4(0x04)*/ {"fild",F_W,1040,FILD,0},
 /*5(0x05)*/ {"fild",F_W,1296,FILD,0},
 /*6(0x06)*/ {"fild",F_W,1552,FILD,0},
 /*7(0x07)*/ {"fild",F_W,1808,FILD,0},
 /*8(0x08)*/ {0,F_W,16,0,0},
 /*9(0x09)*/ {0,F_W,272,0,0},
 /*10(0x0A)*/ {0,F_W,528,0,0},
 /*11(0x0B)*/ {0,F_W,784,0,0},
 /*12(0x0C)*/ {0,F_W,1040,0,0},
 /*13(0x0D)*/ {0,F_W,1296,0,0},
 /*14(0x0E)*/ {0,F_W,1552,0,0},
 /*15(0x0F)*/ {0,F_W,1808,0,0},
 /*16(0x10)*/ {"fist",F_W,16,FIST,0},
 /*17(0x11)*/ {"fist",F_W,272,FIST,0},
 /*18(0x12)*/ {"fist",F_W,528,FIST,0},
 /*19(0x13)*/ {"fist",F_W,784,FIST,0},
 /*20(0x14)*/ {"fist",F_W,1040,FIST,0},
 /*21(0x15)*/ {"fist",F_W,1296,FIST,0},
 /*22(0x16)*/ {"fist",F_W,1552,FIST,0},
 /*23(0x17)*/ {"fist",F_W,1808,FIST,0},
 /*24(0x18)*/ {"fistp",F_W,16,FISTP,0},
 /*25(0x19)*/ {"fistp",F_W,272,FISTP,0},
 /*26(0x1A)*/ {"fistp",F_W,528,FISTP,0},
 /*27(0x1B)*/ {"fistp",F_W,784,FISTP,0},
 /*28(0x1C)*/ {"fistp",F_W,1040,FISTP,0},
 /*29(0x1D)*/ {"fistp",F_W,1296,FISTP,0},
 /*30(0x1E)*/ {"fistp",F_W,1552,FISTP,0},
 /*31(0x1F)*/ {"fistp",F_W,1808,FISTP,0},
 /*32(0x20)*/ {"fbld",F_TW,16,FBLD,0},
 /*33(0x21)*/ {"fbld",F_TW,272,FBLD,0},
 /*34(0x22)*/ {"fbld",F_TW,528,FBLD,0},
 /*35(0x23)*/ {"fbld",F_TW,784,FBLD,0},
 /*36(0x24)*/ {"fbld",F_TW,1040,FBLD,0},
 /*37(0x25)*/ {"fbld",F_TW,1296,FBLD,0},
 /*38(0x26)*/ {"fbld",F_TW,1552,FBLD,0},
 /*39(0x27)*/ {"fbld",F_TW,1808,FBLD,0},
 /*40(0x28)*/ {"fild",F_QW+F_D+F_DP,16,FILD,0},
 /*41(0x29)*/ {"fild",F_QW+F_D+F_DP,272,FILD,0},
 /*42(0x2A)*/ {"fild",F_QW+F_D+F_DP,528,FILD,0},
 /*43(0x2B)*/ {"fild",F_QW+F_D+F_DP,784,FILD,0},
 /*44(0x2C)*/ {"fild",F_QW+F_D+F_DP,1040,FILD,0},
 /*45(0x2D)*/ {"fild",F_QW+F_D+F_DP,1296,FILD,0},
 /*46(0x2E)*/ {"fild",F_QW+F_D+F_DP,1552,FILD,0},
 /*47(0x2F)*/ {"fild",F_QW+F_D+F_DP,1808,FILD,0},
 /*48(0x30)*/ {"fbstp",F_TW,16,FBSTP,0},
 /*49(0x31)*/ {"fbstp",F_TW,272,FBSTP,0},
 /*50(0x32)*/ {"fbstp",F_TW,528,FBSTP,0},
 /*51(0x33)*/ {"fbstp",F_TW,784,FBSTP,0},
 /*52(0x34)*/ {"fbstp",F_TW,1040,FBSTP,0},
 /*53(0x35)*/ {"fbstp",F_TW,1296,FBSTP,0},
 /*54(0x36)*/ {"fbstp",F_TW,1552,FBSTP,0},
 /*55(0x37)*/ {"fbstp",F_TW,1808,FBSTP,0},
 /*56(0x38)*/ {"fistp",F_QW+F_D+F_DP,16,FISTP,0},
 /*57(0x39)*/ {"fistp",F_QW+F_D+F_DP,272,FISTP,0},
 /*58(0x3A)*/ {"fistp",F_QW+F_D+F_DP,528,FISTP,0},
 /*59(0x3B)*/ {"fistp",F_QW+F_D+F_DP,784,FISTP,0},
 /*60(0x3C)*/ {"fistp",F_QW+F_D+F_DP,1040,FISTP,0},
 /*61(0x3D)*/ {"fistp",F_QW+F_D+F_DP,1296,FISTP,0},
 /*62(0x3E)*/ {"fistp",F_QW+F_D+F_DP,1552,FISTP,0},
 /*63(0x3F)*/ {"fistp",F_QW+F_D+F_DP,1808,FISTP,0},
 /*64(0x40)*/ {"fild",F_W,17,FILD,0},
 /*65(0x41)*/ {"fild",F_W,273,FILD,0},
 /*66(0x42)*/ {"fild",F_W,529,FILD,0},
 /*67(0x43)*/ {"fild",F_W,785,FILD,0},
 /*68(0x44)*/ {"fild",F_W,1041,FILD,0},
 /*69(0x45)*/ {"fild",F_W,1297,FILD,0},
 /*70(0x46)*/ {"fild",F_W,1553,FILD,0},
 /*71(0x47)*/ {"fild",F_W,1809,FILD,0},
 /*72(0x48)*/ {0,F_W,17,0,0},
 /*73(0x49)*/ {0,F_W,273,0,0},
 /*74(0x4A)*/ {0,F_W,529,0,0},
 /*75(0x4B)*/ {0,F_W,785,0,0},
 /*76(0x4C)*/ {0,F_W,1041,0,0},
 /*77(0x4D)*/ {0,F_W,1297,0,0},
 /*78(0x4E)*/ {0,F_W,1553,0,0},
 /*79(0x4F)*/ {0,F_W,1809,0,0},
 /*80(0x50)*/ {"fist",F_W,17,FIST,0},
 /*81(0x51)*/ {"fist",F_W,273,FIST,0},
 /*82(0x52)*/ {"fist",F_W,529,FIST,0},
 /*83(0x53)*/ {"fist",F_W,785,FIST,0},
 /*84(0x54)*/ {"fist",F_W,1041,FIST,0},
 /*85(0x55)*/ {"fist",F_W,1297,FIST,0},
 /*86(0x56)*/ {"fist",F_W,1553,FIST,0},
 /*87(0x57)*/ {"fist",F_W,1809,FIST,0},
 /*88(0x58)*/ {"fistp",F_W,17,FISTP,0},
 /*89(0x59)*/ {"fistp",F_W,273,FISTP,0},
 /*90(0x5A)*/ {"fistp",F_W,529,FISTP,0},
 /*91(0x5B)*/ {"fistp",F_W,785,FISTP,0},
 /*92(0x5C)*/ {"fistp",F_W,1041,FISTP,0},
 /*93(0x5D)*/ {"fistp",F_W,1297,FISTP,0},
 /*94(0x5E)*/ {"fistp",F_W,1553,FISTP,0},
 /*95(0x5F)*/ {"fistp",F_W,1809,FISTP,0},
 /*96(0x60)*/ {"fbld",F_TW,17,FBLD,0},
 /*97(0x61)*/ {"fbld",F_TW,273,FBLD,0},
 /*98(0x62)*/ {"fbld",F_TW,529,FBLD,0},
 /*99(0x63)*/ {"fbld",F_TW,785,FBLD,0},
 /*100(0x64)*/ {"fbld",F_TW,1041,FBLD,0},
 /*101(0x65)*/ {"fbld",F_TW,1297,FBLD,0},
 /*102(0x66)*/ {"fbld",F_TW,1553,FBLD,0},
 /*103(0x67)*/ {"fbld",F_TW,1809,FBLD,0},
 /*104(0x68)*/ {"fild",F_QW+F_D+F_DP,17,FILD,0},
 /*105(0x69)*/ {"fild",F_QW+F_D+F_DP,273,FILD,0},
 /*106(0x6A)*/ {"fild",F_QW+F_D+F_DP,529,FILD,0},
 /*107(0x6B)*/ {"fild",F_QW+F_D+F_DP,785,FILD,0},
 /*108(0x6C)*/ {"fild",F_QW+F_D+F_DP,1041,FILD,0},
 /*109(0x6D)*/ {"fild",F_QW+F_D+F_DP,1297,FILD,0},
 /*110(0x6E)*/ {"fild",F_QW+F_D+F_DP,1553,FILD,0},
 /*111(0x6F)*/ {"fild",F_QW+F_D+F_DP,1809,FILD,0},
 /*112(0x70)*/ {"fbstp",F_TW,17,FBSTP,0},
 /*113(0x71)*/ {"fbstp",F_TW,273,FBSTP,0},
 /*114(0x72)*/ {"fbstp",F_TW,529,FBSTP,0},
 /*115(0x73)*/ {"fbstp",F_TW,785,FBSTP,0},
 /*116(0x74)*/ {"fbstp",F_TW,1041,FBSTP,0},
 /*117(0x75)*/ {"fbstp",F_TW,1297,FBSTP,0},
 /*118(0x76)*/ {"fbstp",F_TW,1553,FBSTP,0},
 /*119(0x77)*/ {"fbstp",F_TW,1809,FBSTP,0},
 /*120(0x78)*/ {"fistp",F_QW+F_D+F_DP,17,FISTP,0},
 /*121(0x79)*/ {"fistp",F_QW+F_D+F_DP,273,FISTP,0},
 /*122(0x7A)*/ {"fistp",F_QW+F_D+F_DP,529,FISTP,0},
 /*123(0x7B)*/ {"fistp",F_QW+F_D+F_DP,785,FISTP,0},
 /*124(0x7C)*/ {"fistp",F_QW+F_D+F_DP,1041,FISTP,0},
 /*125(0x7D)*/ {"fistp",F_QW+F_D+F_DP,1297,FISTP,0},
 /*126(0x7E)*/ {"fistp",F_QW+F_D+F_DP,1553,FISTP,0},
 /*127(0x7F)*/ {"fistp",F_QW+F_D+F_DP,1809,FISTP,0},
 /*128(0x80)*/ {"fild",F_W,18,FILD,0},
 /*129(0x81)*/ {"fild",F_W,274,FILD,0},
 /*130(0x82)*/ {"fild",F_W,530,FILD,0},
 /*131(0x83)*/ {"fild",F_W,786,FILD,0},
 /*132(0x84)*/ {"fild",F_W,1042,FILD,0},
 /*133(0x85)*/ {"fild",F_W,1298,FILD,0},
 /*134(0x86)*/ {"fild",F_W,1554,FILD,0},
 /*135(0x87)*/ {"fild",F_W,1810,FILD,0},
 /*136(0x88)*/ {0,F_W,18,0,0},
 /*137(0x89)*/ {0,F_W,274,0,0},
 /*138(0x8A)*/ {0,F_W,530,0,0},
 /*139(0x8B)*/ {0,F_W,786,0,0},
 /*140(0x8C)*/ {0,F_W,1042,0,0},
 /*141(0x8D)*/ {0,F_W,1298,0,0},
 /*142(0x8E)*/ {0,F_W,1554,0,0},
 /*143(0x8F)*/ {0,F_W,1810,0,0},
 /*144(0x90)*/ {"fist",F_W,18,FIST,0},
 /*145(0x91)*/ {"fist",F_W,274,FIST,0},
 /*146(0x92)*/ {"fist",F_W,530,FIST,0},
 /*147(0x93)*/ {"fist",F_W,786,FIST,0},
 /*148(0x94)*/ {"fist",F_W,1042,FIST,0},
 /*149(0x95)*/ {"fist",F_W,1298,FIST,0},
 /*150(0x96)*/ {"fist",F_W,1554,FIST,0},
 /*151(0x97)*/ {"fist",F_W,1810,FIST,0},
 /*152(0x98)*/ {"fistp",F_W,18,FISTP,0},
 /*153(0x99)*/ {"fistp",F_W,274,FISTP,0},
 /*154(0x9A)*/ {"fistp",F_W,530,FISTP,0},
 /*155(0x9B)*/ {"fistp",F_W,786,FISTP,0},
 /*156(0x9C)*/ {"fistp",F_W,1042,FISTP,0},
 /*157(0x9D)*/ {"fistp",F_W,1298,FISTP,0},
 /*158(0x9E)*/ {"fistp",F_W,1554,FISTP,0},
 /*159(0x9F)*/ {"fistp",F_W,1810,FISTP,0},
 /*160(0xA0)*/ {"fbld",F_TW,18,FBLD,0},
 /*161(0xA1)*/ {"fbld",F_TW,274,FBLD,0},
 /*162(0xA2)*/ {"fbld",F_TW,530,FBLD,0},
 /*163(0xA3)*/ {"fbld",F_TW,786,FBLD,0},
 /*164(0xA4)*/ {"fbld",F_TW,1042,FBLD,0},
 /*165(0xA5)*/ {"fbld",F_TW,1298,FBLD,0},
 /*166(0xA6)*/ {"fbld",F_TW,1554,FBLD,0},
 /*167(0xA7)*/ {"fbld",F_TW,1810,FBLD,0},
 /*168(0xA8)*/ {"fild",F_QW+F_D+F_DP,18,FILD,0},
 /*169(0xA9)*/ {"fild",F_QW+F_D+F_DP,274,FILD,0},
 /*170(0xAA)*/ {"fild",F_QW+F_D+F_DP,530,FILD,0},
 /*171(0xAB)*/ {"fild",F_QW+F_D+F_DP,786,FILD,0},
 /*172(0xAC)*/ {"fild",F_QW+F_D+F_DP,1042,FILD,0},
 /*173(0xAD)*/ {"fild",F_QW+F_D+F_DP,1298,FILD,0},
 /*174(0xAE)*/ {"fild",F_QW+F_D+F_DP,1554,FILD,0},
 /*175(0xAF)*/ {"fild",F_QW+F_D+F_DP,1810,FILD,0},
 /*176(0xB0)*/ {"fbstp",F_TW,18,FBSTP,0},
 /*177(0xB1)*/ {"fbstp",F_TW,274,FBSTP,0},
 /*178(0xB2)*/ {"fbstp",F_TW,530,FBSTP,0},
 /*179(0xB3)*/ {"fbstp",F_TW,786,FBSTP,0},
 /*180(0xB4)*/ {"fbstp",F_TW,1042,FBSTP,0},
 /*181(0xB5)*/ {"fbstp",F_TW,1298,FBSTP,0},
 /*182(0xB6)*/ {"fbstp",F_TW,1554,FBSTP,0},
 /*183(0xB7)*/ {"fbstp",F_TW,1810,FBSTP,0},
 /*184(0xB8)*/ {"fistp",F_QW+F_D+F_DP,18,FISTP,0},
 /*185(0xB9)*/ {"fistp",F_QW+F_D+F_DP,274,FISTP,0},
 /*186(0xBA)*/ {"fistp",F_QW+F_D+F_DP,530,FISTP,0},
 /*187(0xBB)*/ {"fistp",F_QW+F_D+F_DP,786,FISTP,0},
 /*188(0xBC)*/ {"fistp",F_QW+F_D+F_DP,1042,FISTP,0},
 /*189(0xBD)*/ {"fistp",F_QW+F_D+F_DP,1298,FISTP,0},
 /*190(0xBE)*/ {"fistp",F_QW+F_D+F_DP,1554,FISTP,0},
 /*191(0xBF)*/ {"fistp",F_QW+F_D+F_DP,1810,FISTP,0},
 /*192(0xC0)*/ {0,0,0,0,0},
 /*193(0xC1)*/ {0,0,0,0,0},
 /*194(0xC2)*/ {0,0,0,0,0},
 /*195(0xC3)*/ {0,0,0,0,0},
 /*196(0xC4)*/ {0,0,0,0,0},
 /*197(0xC5)*/ {0,0,0,0,0},
 /*198(0xC6)*/ {0,0,0,0,0},
 /*199(0xC7)*/ {0,0,0,0,0},
 /*200(0xC8)*/ {0,0,0,0,0},
 /*201(0xC9)*/ {0,0,0,0,0},
 /*202(0xCA)*/ {0,0,0,0,0},
 /*203(0xCB)*/ {0,0,0,0,0},
 /*204(0xCC)*/ {0,0,0,0,0},
 /*205(0xCD)*/ {0,0,0,0,0},
 /*206(0xCE)*/ {0,0,0,0,0},
 /*207(0xCF)*/ {0,0,0,0,0},
 /*208(0xD0)*/ {0,0,0,0,0},
 /*209(0xD1)*/ {0,0,0,0,0},
 /*210(0xD2)*/ {0,0,0,0,0},
 /*211(0xD3)*/ {0,0,0,0,0},
 /*212(0xD4)*/ {0,0,0,0,0},
 /*213(0xD5)*/ {0,0,0,0,0},
 /*214(0xD6)*/ {0,0,0,0,0},
 /*215(0xD7)*/ {0,0,0,0,0},
 /*216(0xD8)*/ {0,0,0,0,0},
 /*217(0xD9)*/ {0,0,0,0,0},
 /*218(0xDA)*/ {0,0,0,0,0},
 /*219(0xDB)*/ {0,0,0,0,0},
 /*220(0xDC)*/ {0,0,0,0,0},
 /*221(0xDD)*/ {0,0,0,0,0},
 /*222(0xDE)*/ {0,0,0,0,0},
 /*223(0xDF)*/ {0,0,0,0,0},
 /*224(0xE0)*/ {"fstsw",F_PAR1D,0,FSTSW,0},
 /*225(0xE1)*/ {0,0,0,0,0},
 /*226(0xE2)*/ {0,0,0,0,0},
 /*227(0xE3)*/ {0,0,0,0,0},
 /*228(0xE4)*/ {0,0,0,0,0},
 /*229(0xE5)*/ {0,0,0,0,0},
 /*230(0xE6)*/ {0,0,0,0,0},
 /*231(0xE7)*/ {0,0,0,0,0},
 /*232(0xE8)*/ {"fucomip",F_PAR,2,FUCOMIP,0},
 /*233(0xE9)*/ {"fucomip",F_PAR,258,FUCOMIP,0},
 /*234(0xEA)*/ {"fucomip",F_PAR,514,FUCOMIP,0},
 /*235(0xEB)*/ {"fucomip",F_PAR,770,FUCOMIP,0},
 /*236(0xEC)*/ {"fucomip",F_PAR,1026,FUCOMIP,0},
 /*237(0xED)*/ {"fucomip",F_PAR,1282,FUCOMIP,0},
 /*238(0xEE)*/ {"fucomip",F_PAR,1538,FUCOMIP,0},
 /*239(0xEF)*/ {"fucomip",F_PAR,1794,FUCOMIP,0},
 /*240(0xF0)*/ {"fcomip",F_PAR,2,FCOMIP,0},
 /*241(0xF1)*/ {"fcomip",F_PAR,258,FCOMIP,0},
 /*242(0xF2)*/ {"fcomip",F_PAR,514,FCOMIP,0},
 /*243(0xF3)*/ {"fcomip",F_PAR,770,FCOMIP,0},
 /*244(0xF4)*/ {"fcomip",F_PAR,1026,FCOMIP,0},
 /*245(0xF5)*/ {"fcomip",F_PAR,1282,FCOMIP,0},
 /*246(0xF6)*/ {"fcomip",F_PAR,1538,FCOMIP,0},
 /*247(0xF7)*/ {"fcomip",F_PAR,1794,FCOMIP,0},
 /*248(0xF8)*/ {0,0,0,0,0},
 /*249(0xF9)*/ {0,0,0,0,0},
 /*250(0xFA)*/ {0,0,0,0,0},
 /*251(0xFB)*/ {0,0,0,0,0},
 /*252(0xFC)*/ {0,0,0,0,0},
 /*253(0xFD)*/ {0,0,0,0,0},
 /*254(0xFE)*/ {0,0,0,0,0},
 /*255(0xFF)*/ {0,0,0,0,0}
};

////grupa 0FBA - instrukcje bt/bts/btr/btc r/m,imm8
op_t g0FBA[256]=
{
 /*0(0x00)*/ {0,0,16,0,0},
 /*1(0x01)*/ {0,0,272,0,0},
 /*2(0x02)*/ {0,0,528,0,0},
 /*3(0x03)*/ {0,0,784,0,0},
 /*4(0x04)*/ {0,0,1040,0,0},
 /*5(0x05)*/ {0,0,1296,0,0},
 /*6(0x06)*/ {0,0,1552,0,0},
 /*7(0x07)*/ {0,0,1808,0,0},
 /*8(0x08)*/ {0,0,16,0,0},
 /*9(0x09)*/ {0,0,272,0,0},
 /*10(0x0A)*/ {0,0,528,0,0},
 /*11(0x0B)*/ {0,0,784,0,0},
 /*12(0x0C)*/ {0,0,1040,0,0},
 /*13(0x0D)*/ {0,0,1296,0,0},
 /*14(0x0E)*/ {0,0,1552,0,0},
 /*15(0x0F)*/ {0,0,1808,0,0},
 /*16(0x10)*/ {0,0,16,0,0},
 /*17(0x11)*/ {0,0,272,0,0},
 /*18(0x12)*/ {0,0,528,0,0},
 /*19(0x13)*/ {0,0,784,0,0},
 /*20(0x14)*/ {0,0,1040,0,0},
 /*21(0x15)*/ {0,0,1296,0,0},
 /*22(0x16)*/ {0,0,1552,0,0},
 /*23(0x17)*/ {0,0,1808,0,0},
 /*24(0x18)*/ {0,0,16,0,0},
 /*25(0x19)*/ {0,0,272,0,0},
 /*26(0x1A)*/ {0,0,528,0,0},
 /*27(0x1B)*/ {0,0,784,0,0},
 /*28(0x1C)*/ {0,0,1040,0,0},
 /*29(0x1D)*/ {0,0,1296,0,0},
 /*30(0x1E)*/ {0,0,1552,0,0},
 /*31(0x1F)*/ {0,0,1808,0,0},
 /*32(0x20)*/ {"bt",0,16,BT,0},
 /*33(0x21)*/ {"bt",0,272,BT,0},
 /*34(0x22)*/ {"bt",0,528,BT,0},
 /*35(0x23)*/ {"bt",0,784,BT,0},
 /*36(0x24)*/ {"bt",0,1040,BT,0},
 /*37(0x25)*/ {"bt",0,1296,BT,0},
 /*38(0x26)*/ {"bt",0,1552,BT,0},
 /*39(0x27)*/ {"bt",0,1808,BT,0},
 /*40(0x28)*/ {"bts",0,16,BTS,0},
 /*41(0x29)*/ {"bts",0,272,BTS,0},
 /*42(0x2A)*/ {"bts",0,528,BTS,0},
 /*43(0x2B)*/ {"bts",0,784,BTS,0},
 /*44(0x2C)*/ {"bts",0,1040,BTS,0},
 /*45(0x2D)*/ {"bts",0,1296,BTS,0},
 /*46(0x2E)*/ {"bts",0,1552,BTS,0},
 /*47(0x2F)*/ {"bts",0,1808,BTS,0},
 /*48(0x30)*/ {"btr",0,16,BTR,0},
 /*49(0x31)*/ {"btr",0,272,BTR,0},
 /*50(0x32)*/ {"btr",0,528,BTR,0},
 /*51(0x33)*/ {"btr",0,784,BTR,0},
 /*52(0x34)*/ {"btr",0,1040,BTR,0},
 /*53(0x35)*/ {"btr",0,1296,BTR,0},
 /*54(0x36)*/ {"btr",0,1552,BTR,0},
 /*55(0x37)*/ {"btr",0,1808,BTR,0},
 /*56(0x38)*/ {"btc",0,16,BTC,0},
 /*57(0x39)*/ {"btc",0,272,BTC,0},
 /*58(0x3A)*/ {"btc",0,528,BTC,0},
 /*59(0x3B)*/ {"btc",0,784,BTC,0},
 /*60(0x3C)*/ {"btc",0,1040,BTC,0},
 /*61(0x3D)*/ {"btc",0,1296,BTC,0},
 /*62(0x3E)*/ {"btc",0,1552,BTC,0},
 /*63(0x3F)*/ {"btc",0,1808,BTC,0},
 /*64(0x40)*/ {0,0,17,0,0},
 /*65(0x41)*/ {0,0,273,0,0},
 /*66(0x42)*/ {0,0,529,0,0},
 /*67(0x43)*/ {0,0,785,0,0},
 /*68(0x44)*/ {0,0,1041,0,0},
 /*69(0x45)*/ {0,0,1297,0,0},
 /*70(0x46)*/ {0,0,1553,0,0},
 /*71(0x47)*/ {0,0,1809,0,0},
 /*72(0x48)*/ {0,0,17,0,0},
 /*73(0x49)*/ {0,0,273,0,0},
 /*74(0x4A)*/ {0,0,529,0,0},
 /*75(0x4B)*/ {0,0,785,0,0},
 /*76(0x4C)*/ {0,0,1041,0,0},
 /*77(0x4D)*/ {0,0,1297,0,0},
 /*78(0x4E)*/ {0,0,1553,0,0},
 /*79(0x4F)*/ {0,0,1809,0,0},
 /*80(0x50)*/ {0,0,17,0,0},
 /*81(0x51)*/ {0,0,273,0,0},
 /*82(0x52)*/ {0,0,529,0,0},
 /*83(0x53)*/ {0,0,785,0,0},
 /*84(0x54)*/ {0,0,1041,0,0},
 /*85(0x55)*/ {0,0,1297,0,0},
 /*86(0x56)*/ {0,0,1553,0,0},
 /*87(0x57)*/ {0,0,1809,0,0},
 /*88(0x58)*/ {0,0,17,0,0},
 /*89(0x59)*/ {0,0,273,0,0},
 /*90(0x5A)*/ {0,0,529,0,0},
 /*91(0x5B)*/ {0,0,785,0,0},
 /*92(0x5C)*/ {0,0,1041,0,0},
 /*93(0x5D)*/ {0,0,1297,0,0},
 /*94(0x5E)*/ {0,0,1553,0,0},
 /*95(0x5F)*/ {0,0,1809,0,0},
 /*96(0x60)*/ {"bt",0,17,BT,0},
 /*97(0x61)*/ {"bt",0,273,BT,0},
 /*98(0x62)*/ {"bt",0,529,BT,0},
 /*99(0x63)*/ {"bt",0,785,BT,0},
 /*100(0x64)*/ {"bt",0,1041,BT,0},
 /*101(0x65)*/ {"bt",0,1297,BT,0},
 /*102(0x66)*/ {"bt",0,1553,BT,0},
 /*103(0x67)*/ {"bt",0,1809,BT,0},
 /*104(0x68)*/ {"bts",0,17,BTS,0},
 /*105(0x69)*/ {"bts",0,273,BTS,0},
 /*106(0x6A)*/ {"bts",0,529,BTS,0},
 /*107(0x6B)*/ {"bts",0,785,BTS,0},
 /*108(0x6C)*/ {"bts",0,1041,BTS,0},
 /*109(0x6D)*/ {"bts",0,1297,BTS,0},
 /*110(0x6E)*/ {"bts",0,1553,BTS,0},
 /*111(0x6F)*/ {"bts",0,1809,BTS,0},
 /*112(0x70)*/ {"btr",0,17,BTR,0},
 /*113(0x71)*/ {"btr",0,273,BTR,0},
 /*114(0x72)*/ {"btr",0,529,BTR,0},
 /*115(0x73)*/ {"btr",0,785,BTR,0},
 /*116(0x74)*/ {"btr",0,1041,BTR,0},
 /*117(0x75)*/ {"btr",0,1297,BTR,0},
 /*118(0x76)*/ {"btr",0,1553,BTR,0},
 /*119(0x77)*/ {"btr",0,1809,BTR,0},
 /*120(0x78)*/ {"btc",0,17,BTC,0},
 /*121(0x79)*/ {"btc",0,273,BTC,0},
 /*122(0x7A)*/ {"btc",0,529,BTC,0},
 /*123(0x7B)*/ {"btc",0,785,BTC,0},
 /*124(0x7C)*/ {"btc",0,1041,BTC,0},
 /*125(0x7D)*/ {"btc",0,1297,BTC,0},
 /*126(0x7E)*/ {"btc",0,1553,BTC,0},
 /*127(0x7F)*/ {"btc",0,1809,BTC,0},
 /*128(0x80)*/ {0,0,18,0,0},
 /*129(0x81)*/ {0,0,274,0,0},
 /*130(0x82)*/ {0,0,530,0,0},
 /*131(0x83)*/ {0,0,786,0,0},
 /*132(0x84)*/ {0,0,1042,0,0},
 /*133(0x85)*/ {0,0,1298,0,0},
 /*134(0x86)*/ {0,0,1554,0,0},
 /*135(0x87)*/ {0,0,1810,0,0},
 /*136(0x88)*/ {0,0,18,0,0},
 /*137(0x89)*/ {0,0,274,0,0},
 /*138(0x8A)*/ {0,0,530,0,0},
 /*139(0x8B)*/ {0,0,786,0,0},
 /*140(0x8C)*/ {0,0,1042,0,0},
 /*141(0x8D)*/ {0,0,1298,0,0},
 /*142(0x8E)*/ {0,0,1554,0,0},
 /*143(0x8F)*/ {0,0,1810,0,0},
 /*144(0x90)*/ {0,0,18,0,0},
 /*145(0x91)*/ {0,0,274,0,0},
 /*146(0x92)*/ {0,0,530,0,0},
 /*147(0x93)*/ {0,0,786,0,0},
 /*148(0x94)*/ {0,0,1042,0,0},
 /*149(0x95)*/ {0,0,1298,0,0},
 /*150(0x96)*/ {0,0,1554,0,0},
 /*151(0x97)*/ {0,0,1810,0,0},
 /*152(0x98)*/ {0,0,18,0,0},
 /*153(0x99)*/ {0,0,274,0,0},
 /*154(0x9A)*/ {0,0,530,0,0},
 /*155(0x9B)*/ {0,0,786,0,0},
 /*156(0x9C)*/ {0,0,1042,0,0},
 /*157(0x9D)*/ {0,0,1298,0,0},
 /*158(0x9E)*/ {0,0,1554,0,0},
 /*159(0x9F)*/ {0,0,1810,0,0},
 /*160(0xA0)*/ {"bt",0,18,BT,0},
 /*161(0xA1)*/ {"bt",0,274,BT,0},
 /*162(0xA2)*/ {"bt",0,530,BT,0},
 /*163(0xA3)*/ {"bt",0,786,BT,0},
 /*164(0xA4)*/ {"bt",0,1042,BT,0},
 /*165(0xA5)*/ {"bt",0,1298,BT,0},
 /*166(0xA6)*/ {"bt",0,1554,BT,0},
 /*167(0xA7)*/ {"bt",0,1810,BT,0},
 /*168(0xA8)*/ {"bts",0,18,BTS,0},
 /*169(0xA9)*/ {"bts",0,274,BTS,0},
 /*170(0xAA)*/ {"bts",0,530,BTS,0},
 /*171(0xAB)*/ {"bts",0,786,BTS,0},
 /*172(0xAC)*/ {"bts",0,1042,BTS,0},
 /*173(0xAD)*/ {"bts",0,1298,BTS,0},
 /*174(0xAE)*/ {"bts",0,1554,BTS,0},
 /*175(0xAF)*/ {"bts",0,1810,BTS,0},
 /*176(0xB0)*/ {"btr",0,18,BTR,0},
 /*177(0xB1)*/ {"btr",0,274,BTR,0},
 /*178(0xB2)*/ {"btr",0,530,BTR,0},
 /*179(0xB3)*/ {"btr",0,786,BTR,0},
 /*180(0xB4)*/ {"btr",0,1042,BTR,0},
 /*181(0xB5)*/ {"btr",0,1298,BTR,0},
 /*182(0xB6)*/ {"btr",0,1554,BTR,0},
 /*183(0xB7)*/ {"btr",0,1810,BTR,0},
 /*184(0xB8)*/ {"btc",0,18,BTC,0},
 /*185(0xB9)*/ {"btc",0,274,BTC,0},
 /*186(0xBA)*/ {"btc",0,530,BTC,0},
 /*187(0xBB)*/ {"btc",0,786,BTC,0},
 /*188(0xBC)*/ {"btc",0,1042,BTC,0},
 /*189(0xBD)*/ {"btc",0,1298,BTC,0},
 /*190(0xBE)*/ {"btc",0,1554,BTC,0},
 /*191(0xBF)*/ {"btc",0,1810,BTC,0},
 /*192(0xC0)*/ {0,0,0,0,0},
 /*193(0xC1)*/ {0,0,256,0,0},
 /*194(0xC2)*/ {0,0,512,0,0},
 /*195(0xC3)*/ {0,0,768,0,0},
 /*196(0xC4)*/ {0,0,1024,0,0},
 /*197(0xC5)*/ {0,0,1280,0,0},
 /*198(0xC6)*/ {0,0,1536,0,0},
 /*199(0xC7)*/ {0,0,1792,0,0},
 /*200(0xC8)*/ {0,0,0,0,0},
 /*201(0xC9)*/ {0,0,256,0,0},
 /*202(0xCA)*/ {0,0,512,0,0},
 /*203(0xCB)*/ {0,0,768,0,0},
 /*204(0xCC)*/ {0,0,1024,0,0},
 /*205(0xCD)*/ {0,0,1280,0,0},
 /*206(0xCE)*/ {0,0,1536,0,0},
 /*207(0xCF)*/ {0,0,1792,0,0},
 /*208(0xD0)*/ {0,0,0,0,0},
 /*209(0xD1)*/ {0,0,256,0,0},
 /*210(0xD2)*/ {0,0,512,0,0},
 /*211(0xD3)*/ {0,0,768,0,0},
 /*212(0xD4)*/ {0,0,1024,0,0},
 /*213(0xD5)*/ {0,0,1280,0,0},
 /*214(0xD6)*/ {0,0,1536,0,0},
 /*215(0xD7)*/ {0,0,1792,0,0},
 /*216(0xD8)*/ {0,0,0,0,0},
 /*217(0xD9)*/ {0,0,256,0,0},
 /*218(0xDA)*/ {0,0,512,0,0},
 /*219(0xDB)*/ {0,0,768,0,0},
 /*220(0xDC)*/ {0,0,1024,0,0},
 /*221(0xDD)*/ {0,0,1280,0,0},
 /*222(0xDE)*/ {0,0,1536,0,0},
 /*223(0xDF)*/ {0,0,1792,0,0},
 /*224(0xE0)*/ {"bt",0,0,BT,0},
 /*225(0xE1)*/ {"bt",0,256,BT,0},
 /*226(0xE2)*/ {"bt",0,512,BT,0},
 /*227(0xE3)*/ {"bt",0,768,BT,0},
 /*228(0xE4)*/ {"bt",0,1024,BT,0},
 /*229(0xE5)*/ {"bt",0,1280,BT,0},
 /*230(0xE6)*/ {"bt",0,1536,BT,0},
 /*231(0xE7)*/ {"bt",0,1792,BT,0},
 /*232(0xE8)*/ {"bts",0,0,BTS,0},
 /*233(0xE9)*/ {"bts",0,256,BTS,0},
 /*234(0xEA)*/ {"bts",0,512,BTS,0},
 /*235(0xEB)*/ {"bts",0,768,BTS,0},
 /*236(0xEC)*/ {"bts",0,1024,BTS,0},
 /*237(0xED)*/ {"bts",0,1280,BTS,0},
 /*238(0xEE)*/ {"bts",0,1536,BTS,0},
 /*239(0xEF)*/ {"bts",0,1792,BTS,0},
 /*240(0xF0)*/ {"btr",0,0,BTR,0},
 /*241(0xF1)*/ {"btr",0,256,BTR,0},
 /*242(0xF2)*/ {"btr",0,512,BTR,0},
 /*243(0xF3)*/ {"btr",0,768,BTR,0},
 /*244(0xF4)*/ {"btr",0,1024,BTR,0},
 /*245(0xF5)*/ {"btr",0,1280,BTR,0},
 /*246(0xF6)*/ {"btr",0,1536,BTR,0},
 /*247(0xF7)*/ {"btr",0,1792,BTR,0},
 /*248(0xF8)*/ {"btc",0,0,BTC,0},
 /*249(0xF9)*/ {"btc",0,256,BTC,0},
 /*250(0xFA)*/ {"btc",0,512,BTC,0},
 /*251(0xFB)*/ {"btc",0,768,BTC,0},
 /*252(0xFC)*/ {"btc",0,1024,BTC,0},
 /*253(0xFD)*/ {"btc",0,1280,BTC,0},
 /*254(0xFE)*/ {"btc",0,1536,BTC,0},
 /*255(0xFF)*/ {"btc",0,1792,BTC,0}
};
